#ifndef _CONVOLVE_H
#define _CONVOLVE_H
/////////////////////////////////////////////////////////////////////////////
#include  "windowing_fun.h" 
/////////////////////////////////////////////////////////////////////////////
namespace Vtl 
{
namespace ImAlg
{
/*!
 * \brief  :  , , ,     
 *      ().     
 *    ,     .      ( 
 *   Plf)       (   ).
 *  
 *   .  
 * \example test_fast_gauss.cpp 
 *
 ***************************************************************************************************/
namespace Conv
{
/////////////////////////////////////////////////////////////////////////////
	///   ,     
	template<typename T,int stride=0> class SliceIter
	{
		T* d;
	public:
		SliceIter<T,stride>(T* data) : d(data) {}
		SliceIter<T,stride>(const SliceIter<T,stride>& si) : d(si.d) {}
		SliceIter<T,stride>& operator=(const SliceIter<T,stride>& si) { (*this)(si); return *this; }
		SliceIter& operator++() { d+=stride; return *this; }
		SliceIter  operator++(int) { SliceIter<T,stride> si(*this); this->operator++(); return si; }
		SliceIter& operator--() { d-=stride; return *this; }
		SliceIter  operator--(int) { SliceIter<T,stride> si(*this); this->operator--(); return si; }
		SliceIter& operator+=(int n) { d+=stride*n; return *this; }
		SliceIter& operator-=(int n) { return this->operator+=(-n); }
		SliceIter operator+(int n) const { return SliceIter<T,stride>(d+stride*n,stride); }
		SliceIter operator-(int n) const { return this->operator+(-n); }
		bool operator==(const SliceIter& si) const { return d==si.d; }
		bool operator!=(const SliceIter& si) const { return !this->operator==(si); }
		T  operator*() const { return *d; }
		T& operator*() { return *d; }
	};
	///   ,            runtime
	template<typename T> class SliceIter<T,0>
	{
		T* d;
		const int stride_var;
	public:
		SliceIter<T,0>(T* data,int step) : d(data), stride_var(step) {}
		SliceIter<T,0>(const SliceIter<T,0>& si) : d(si.d), stride_var(si.stride_var) {}
		//    ,      stride_var.
		SliceIter<T,0>& operator=(const SliceIter<T,0>& si) { this->SliceIter<T,0>::SliceIter(si); return *this; }
		SliceIter& operator++() { d+=stride_var; return *this; }
		SliceIter  operator++(int) { SliceIter<T,0> si(*this); this->operator++(); return si; }
		SliceIter& operator--() { d-=stride_var; return *this; }
		SliceIter  operator--(int) { SliceIter<T,0> si(*this); this->operator--(); return si; }
		SliceIter& operator+=(int n) { d+=stride_var*n; return *this; }
		SliceIter&  operator-=(int n) { return this->operator+=(-n); }
		SliceIter operator+(int n) const { return SliceIter<T,0>(d+stride_var*n,stride_var); }
		SliceIter operator-(int n) const { return this->operator+(-n); }
		bool operator==(const SliceIter& si) const { return d==si.d; }
		bool operator!=(const SliceIter& si) const { return !this->operator==(si); }
		T  operator*() const { return *d; }
		T& operator*() { return *d; }
	};
	///   ,     
	template<typename T,int stride=0> class SliceIterConst
	{
		const T* d;
	public:
		SliceIterConst<T,stride>(const T* data) : d(data) {}
		SliceIterConst<T,stride>(const SliceIterConst<T,stride>& si) : d(si.d) {}
		SliceIterConst<T,stride>& operator=(const SliceIterConst<T,stride>& si) { (*this)(si); return *this; }
		SliceIterConst& operator++() { d+=stride; return *this; }
		SliceIterConst  operator++(int) { SliceIterConst<T,stride> si(*this); this->operator++(); return si; }
		SliceIterConst& operator--() { d-=stride; return *this; }
		SliceIterConst  operator--(int) { SliceIterConst<T,stride> si(*this); this->operator--(); return si; }
		SliceIterConst& operator+=(int n) { d+=stride*n; return *this; }
		SliceIterConst& operator-=(int n) { return this->operator+=(-n); }
		SliceIterConst operator+(int n) const { return SliceIterConst<T,stride>(d+stride*n,stride); }
		SliceIterConst operator-(int n) const { return this->operator+(-n); }
		bool operator==(const SliceIterConst& si) const { return d==si.d; }
		bool operator!=(const SliceIterConst& si) const { return !this->operator==(si); }
		T  operator*() const { return *d; }
	};
	///   ,            runtime
	template<typename T> class SliceIterConst<T,0>
	{
		const T* d;
		const int stride_var;
	public:
		SliceIterConst<T,0>(const T* data,int step) : d(data), stride_var(step) {}
		SliceIterConst<T,0>(const SliceIterConst<T,0>& si) : d(si.d), stride_var(si.stride_var) {}
		//    ,      stride_var.
		SliceIterConst<T,0>& operator=(const SliceIterConst<T,0>& si) { this->SliceIterConst<T,0>::SliceIterConst(si); return *this; }
		SliceIterConst& operator++() { d+=stride_var; return *this; }
		SliceIterConst  operator++(int) { SliceIterConst<T,0> si(*this); this->operator++(); return si; }
		SliceIterConst& operator--() { d-=stride_var; return *this; }
		SliceIterConst  operator--(int) { SliceIterConst<T,0> si(*this); this->operator--(); return si; }
		SliceIterConst& operator+=(int n) { d+=stride_var*n; return *this; }
		SliceIterConst&  operator-=(int n) { return this->operator+=(-n); }
		SliceIterConst operator+(int n) const { return SliceIterConst<T,0>(d+stride_var*n,stride_var); }
		SliceIterConst operator-(int n) const { return this->operator+(-n); }
		bool operator==(const SliceIterConst& si) const { return d==si.d; }
		bool operator!=(const SliceIterConst& si) const { return !this->operator==(si); }
		T  operator*() const { return *d; }
	};

	///     
	template<typename T,typename iterator_in,typename iterator_ker>
	T Conv1pt(iterator_in in,iterator_ker begin,iterator_ker end)
	{
		T s=0;
		for(; begin!=end; ++in, ++begin)
			s+=(*in)*(*begin);
		return s;
	}
	/*! \brief  1D   2D ,   .    ,  \n
	 *  \brief <b> </b>  <b></b>,  <b>   </b>   , 
	 *      ()    . \n
	 *  \brief   <b> ,    </b> value_type. \n
	 *  <b> :</b>   ,       , 
	 *         <b>  </b>  .
	 *      ,     , 
	 *   -  Kernel.left()     Kernel.right() . \see Base
	 ************************************************************************************/
	template<typename iterator_in,typename iterator_out,typename Kernel>
	void ConvolveNoExt(iterator_in begin,iterator_in end,iterator_out out,const Kernel& ker)
	{
		for( ; begin!=end ; begin++ )
		{
			*out++=Conv1pt< iterator_out::value_type >(*begin,ker.begin(),ker.end());
		}
	}
	/*! \brief  1D   2D ,   .    ,  \n
	 *  \brief <b> </b>  <b></b>,  <b>   </b>   , 
	 *      ()    . \n
	 *  \brief    <b> </b> . \n
	 *  \brief <b> :</b>   ,       , 
	 *         <b>  </b>  . \n
	 *  \brief     ,     , 
	 *   -  Kernel.left()     Kernel.right() . \see Base
	 ************************************************************************************/
	template<typename iterator_in,typename iterator_out,typename Kernel>
	void ConvolveNoExt(iterator_in begin,iterator_in end,iterator_out *out,const Kernel& ker)
	{
		for( ; begin!=end ; ++begin )
		{
			*out++=Conv1pt< iterator_out >(*begin,ker.begin(),ker.end());
		}
	}
	/*! \brief  1D ,   .    ,  \n
	 *  \brief <b> </b>  <b> </b>, begin <b>     </b>.\n
	 *  \brief  ()      . \n
	 *  \brief   <b> ,    </b> value_type. \n
	 *  \brief <b> :</b>  ()   
	 *     . \n
	 *  \brief     ,     , 
	 *   -  Kernel.left()     Kernel.right() . \see Base
	 ************************************************************************************/
	template<typename iterator_in,typename iterator_out,typename Kernel>
	void ConvolveNoExt(iterator_in* begin,iterator_in* end,iterator_out out,const Kernel& ker)
	{
		for( ; begin!=end ; begin++ )
		{
			*out++=Conv1pt< iterator_out::value_type  >(begin,ker.begin(),ker.end());
		}
	}
	/*! \brief  1D ,   .    , 
	 *  <b> </b>  <b> </b>, begin <b>     </b>.\n
	 *  \brief    <b> </b> .\n
	 *  \brief <b> :</b>  ()   
	 *     . \n
	 *  \brief     ,     , 
	 *   -  Kernel.left()     Kernel.right() . \see Base
	 ************************************************************************************/
	template<typename iterator_in,typename iterator_out,typename Kernel>
	void ConvolveNoExt(iterator_in* begin,iterator_in* end,iterator_out *out,const Kernel& ker)
	{
		for( ; begin!=end ; begin++ )
		{
			*out++=Conv1pt< iterator_out >(begin,ker.begin(),ker.end());
		}
	}
	/// ,   !
	namespace Impl
	{
		/*!
		 * \brief    
		 *
		 *  :\n
		 * T -  ,    \n
		 * K -   ()     Kernel::Base   
		 *  ,    kernel_type_=KT_Conv  KT_Corr (    
		 *     ) ,   Base double_kernel_   true.
		 ************************************************************************************/
		template<typename T,class K>
		class Buffer
		{
		protected:
			std::vector<T> data;
			int pos;
		public:
			const K& ker;
		public:
			Buffer(const K& kernel) : ker(kernel)
			{
				LOKI_STATIC_CHECK((ker.double_kernel==1),Kernel_must_have_template_parameter_double_kernel_is_true);
				data.resize(ker.size());
				rewind();
			}
			void rewind() 
			{
				pos=0;
			}
			void push(T data_elem)
			{
				data[pos++]=data_elem;
				if(pos==ker.size())
					rewind();
			}
			T pop()
			{
				const T* k=ker.Right()+1-pos;
				return Conv1pt<T>(&(data[0]),k,k+ker.size());
			}
		};
	}
	/*!
	 * \brief   
	 *
	 *  :\n
	 * T -  ,    \n
	 * K -   ()     Kernel::Base   
	 *  ,    kernel_type_=KT_Conv  KT_Corr (    
	 *     ) ,   Base double_kernel_   true.
	 *
	 *           
	 *   ,..   Run()   end-begin >= kernel.size() .
	 *   ,          
	 *          ,   
	 *   .
	 * 
	 ************************************************************************************/
	template<typename T,class K>
	class CyclicConv
	{
		Impl::Buffer<T,K> buf; 
	public:
		CyclicConv(const K& kernel) :buf(kernel) 
		{
		}
		/*!
		 * \brief  IterIn  IterOut    forward_iterator,  operator++(),
		 *     (IterIn) /   (IterOut)    T. 
		 *     kernrl.left()+kernel.right()==kernle.size()-1  
		 * ,      ,     IterIn   
		 *  .      ,
		 *    kernrl.left() .    ,      
		 * ,    STL .       
		 *  STL . IterIn        .
		 * 
		 ************************************************************************************/
		template<typename IterIn,typename IterOut>
		void operator()(IterIn begin,IterIn end,IterOut out)
		{
			buf.rewind();
			//push only
			IterIn cur=_start(begin);
			//central
			IterOut out_cur=_central(cur,end,out);
			//right and than left process
			_right_end_left(begin,out_cur,out);
		}
	private:
		template<typename IterIn>
		IterIn _start(IterIn begin)
		{
			int c=buf.ker.size()-1;
			while(c--)
				buf.push(*begin++);
			return begin;
		}
		template<typename IterIn,typename IterOut>
		IterOut _central(IterIn begin,IterIn end,IterOut out)
		{
			int r=buf.ker.right();
			while(r--)
				out++;
//			out+=r;
			while(begin!=end)
			{
				buf.push(*begin++);
				*out++=buf.pop();
			}
			return out;
		}
		template<typename IterIn,typename IterOut>
		void _right_end_left(IterIn begin,IterOut out,IterOut out0)
		{
			int r=buf.ker.right();
			while(r--)
			{
				buf.push(*begin++);
				*out++=buf.pop();
			}
			int l=buf.ker.left();
			while(l--)
			{
				buf.push(*begin++);
				*out0++=buf.pop();
			}
		}
	};
	/*!
	 * \brief    inplace
	 *
	 *  :\n
	 * T -  ,   
	 * T -  ,    \n
	 * K -   ()     Kernel::Base   
	 *  ,    kernel_type_=KT_Conv  KT_Corr (    
	 *     ) ,   Base double_kernel_   true.
	 *
	 *           
	 *   ,..   Run()   end-begin >= kernel.size() .
	 *   ,          
	 *          ,   
	 *   .
	 ************************************************************************************/
	template<typename T,class K>
	class CyclicConvInplace
	{
		Impl::Buffer<T,K> buf; 
		std::vector<T> save;
	public:
		CyclicConvInplace(const K& kernel) :buf(kernel) 
		{
			save.resize(buf.ker.size());
		}
		/*!
		 * \brief  IterIn    forward_iterator,  operator++(),
		 *         T. 
		 *     ,   .
		 *      ,    kernrl.right() , 
		 *         kernel.size()-1   
		 *     .
		 *    ,      
		 * ,    STL .       
		 *  STL . 
		 * 
		 ************************************************************************************/
		template<typename Iter>
		void operator()(Iter begin,Iter end)
		{
			buf.rewind();
			//left
			Iter cur=_left(begin,end);
			//central
			Iter out=_central(cur,end,begin);
			//right
			_right(begin,end,out);
		}
	private:
		template<typename Iter>
		Iter _left(Iter begin,Iter end)
		{
			int c=buf.ker.size()-1;
			T* s=&(save[0]);
			while(c--)
			{
				T v=*begin++;
				buf.push(v);
				*s++=v;
			}
			return begin;
		}
		template<typename Iter>
		Iter _central(Iter begin,Iter end,Iter out)
		{
			int r=buf.ker.right();
			while(r--)
				out++;
			while(begin!=end)
			{
				buf.push(*begin++);
				*out++=buf.pop();
			}
			return out;
		}
		template<typename Iter>
		void _right(Iter begin,Iter end,Iter out)
		{
			const T* s=&(save[0]);
			while(out!=end)
			{
				buf.push(*s++);
				*out++=buf.pop();
			}
			int r=buf.ker.right();
			while(r--)
			{
				buf.push(*s++);
				*begin++=buf.pop();
			}
		}
	};
	/*!
	 * \brief  ,    .
	 *
	 *  :\n
	 * T -  ,   
	 * T -  ,    \n
	 * K -   ()     Kernel::Base   
	 *  ,    kernel_type_=KT_Conv  KT_Corr (    
	 *     ) ,   Base double_kernel_   true.
	 *
	 *           
	 *   ,..   Run()   end-begin >= kernel.size() .
	 *   ,          
	 *          ,   
	 *   .
	 ************************************************************************************/
	template<typename T,class K>
	class Conv
	{
	protected:
		Impl::Buffer<T,K> buf; 
	public:
		Conv(const K& kernel) :buf(kernel) 
		{
		}
		/*!
		 * \brief  IterIn    random_access_iterator,   , 
		 *  operator++(), operator--(), operator+(int step),    
		 *  T   ()  .  ( kernel.left())   ( kernel.right())
		 *    .
		 *     IterOut,     forward_iterator, 
		 *  operator++()         T.    
		 *      .
		 *    ,      
		 * ,    STL .       
		 *  STL . 
		 * 
		 ************************************************************************************/
		template<typename IterIn,typename IterOut>
		void operator()(IterIn begin,IterIn end,IterOut out)
		{
			buf.rewind();
			//left, push only
			IterIn cur=_left(begin,end);
			//central
			out=_central(cur,end,out);
			//right
			_right(--end,out);
		}
//	private:
	protected:
		template<typename IterIn>
		IterIn _left(IterIn begin,IterIn end)
		{
			IterIn it=begin+buf.ker.left();
			while(it!=begin)
				buf.push(*it--);
			int r=buf.ker.right();
			while(r--)
				buf.push(*begin++);
			return begin;
		}
		template<typename IterIn,typename IterOut>
		IterOut _central(IterIn begin,IterIn end,IterOut out)
		{
			while(begin!=end)
			{
				buf.push(*begin++);
				*out++=buf.pop();
			}
			return out;
		}
		template<typename IterIn,typename IterOut>
		void _right(IterIn end,IterOut out)
		{
			int r=buf.ker.right();
			while(r--)
			{
				buf.push(*(--end));
				*out++=buf.pop();
			}
		}
	};
	/*!
	 * \brief  ,    .
	 *
	 *  :\n
	 * T -  ,   
	 * T -  ,    \n
	 * K -   ()     Kernel::Base   
	 *  ,    kernel_type_=KT_Conv  KT_Corr (    
	 *     ) ,   Base double_kernel_   true.
	 *
	 *           
	 *   ,..   Run()   end-begin >= kernel.size() .
	 *   ,          
	 *          ,   
	 *   .
	 ************************************************************************************/
	template<typename T,class K>
	class ConvInplace
		: private Conv<T,K>
	{
		std::vector<T> save;
	public:
		ConvInplace(const K& kernel) :Conv(kernel) 
		{
			save.resize(buf.ker.right());
		}
		/*!
		 * \brief  Iter    random_access_iterator,  
		 *  operator++(), operator--(), operator+(int step), operator-(int step),   
		 *    T   ()  .  ( kernel.left())  
		 *   .
		 *          .
		 *    ,      
		 * ,    STL .       
		 *  STL . 
		 * 
		 ************************************************************************************/
		template<typename Iter>
		void operator()(Iter begin,Iter end)
		{
			buf.rewind();
			Iter out=begin;
			//left
			begin=_left_2(begin,end,out);
			//central
			out=_central_2(begin,end,out);
			//right
			_right_2(out);
		}
		/*!
		 * No inplace version. In 2D, 3D, etc. sometimes it is required no inplace (with copy) 
		 * operation to process first dimension simultaneously moving data from source to target.
		 * Then, second, etc. dimensions should be processed inplaced, because data is already 
		 * trasferred fro source to target and now processing performed at the target container only.
		 *
		 * Identical to Conv::operator()(IterIn begin,IterIn end,IterOut out)
		 **************************************************************************************/
		template<typename IterIn,typename IterOut>
		void operator()(IterIn begin,IterIn end,IterOut out)
		{
			Conv<T,K>::operator()(begin,end,out);
		}
	private:
		template<typename Iter>
		Iter _left_2(Iter begin,Iter end,Iter& out)
		{
			Iter it=begin+buf.ker.left();
			while(it!=begin)
				buf.push(*it--);
			int r=buf.ker.right();
			while(r--)
				buf.push(*begin++);
			return begin;
		}
		template<typename Iter>
		Iter _central_2(Iter begin,Iter end,Iter out)
		{
			int r=buf.ker.right();
			Iter end1=end-(r+1);
			while(begin!=end1)
			{
				buf.push(*begin++);
				*out++=buf.pop();
			}
			T* s=&(save[0]);
			if(r>0)
			{
				while(r--)
				{
					T v=*begin++;
					s[r]=v;
					buf.push(v);
					*out++=buf.pop();
				}
			}
			buf.push(*begin++);
			*out++=buf.pop();
			return out;
		}
		template<typename Iter>
		void _right_2(Iter out)
		{
			const int r0=buf.ker.right();
			const T* s=&(save[0]);
			for(int r=0; r< r0; ++r)
			{
				buf.push(*s++);
				*out++=buf.pop();
			}
		}
	};

	/*!
	 * \brief  /        
	 *   (),       
	 *  data,width,height, nls 
	 * (Next line shift, ,     ,      
	 *     )    ,   /.
	 * K, K1, K2 -   ()     Kernel::Base   
	 *  ,    kernel_type_=KT_Conv  KT_Corr (    
	 *     ) ,   Base double_kernel_   true.
	 *
	 **************************************************************************************/
	class Conv2D
	{
	public:
		template<typename T, class 1, class 2> 
		static void conv2D(const T* data_in,int width,int height,int nls_in,T* data_out,int nls_out,1& cx,2& cy)
		{
			const T* di=data_in;
			T* dout=data_out;
			for(int y=0; y<height; ++y, di+=nls_in, dout+=nls_out)
				cx(di,di+width,dout);
			//
			for(int x=0; x<width; ++x,++data_out)
			{
				SliceIter<T,0> si(data_out,nls_out);
				cy(si,si+height);
			}
		}
		template<typename T, class 1, class 2> 
		static void conv2DInplace(T* data,int width,int height,int nls,1& cx,2& cy)
		{
			T* d=data;
			for(int y=0; y<height; ++y, d+=nls)
				cx(d,d+width);
			d=data;
			for(int x=0; x<width; ++x, ++d)
			{
				SliceIter<T,0> si(d,nls);
				cy(si,si+height);
			}
		}
	public:
		/*!
		 * \brief  <b></b> /   (),  
		 *       data_in,width,height, nls_in .
		 *       data_out,width,height, nls_out.
		 **************************************************************************************/
		template<typename T, class K1, class K2> 
		static void Cyclic(const T* data_in,int width,int height,int nls_in,T* data_out,int nls_out,const K1& ker_x,const K2& ker_y)
		{
			CyclicConv<T,K1> cx(ker_x);
			CyclicConv<T,K2> cy(ker_y);
			conv2D(data_in,width,height,nls_in,data_out,nls_out,cx,cy);
		}
		/*!
		 * \brief  <b></b> /   (),  
		 *       data_in,width,height, nls_in 
		 *    ,    ker.
		 *       data_out,width,height, nls_out.
		 **************************************************************************************/
		template<typename T, class K> 
		static void Cyclic(const T* data_in,int width,int height,int nls_in,T* data_out,int nls_out,const K& ker)
		{
			CyclicConv<T,K> cc(ker);
			conv2D(data_in,width,height,nls_in,data_out,nls_out,cc,cc);
		}
		/*!
		 * \brief  <b></b> /   (),  
		 *       data,width,height, nls    , 
		 *    ker_x,ker_y .
		 *     .
		 **************************************************************************************/
		template<typename T, class K1, class K2> 
		static void CyclicInplace(T* data,int width,int height,int nls,const K1& ker_x,const K2& ker_y)
		{
			CyclicConvInplace<T,K1> cx(ker_x);
			CyclicConvInplace<T,K2> cy(ker_y);
			conv2DInplace(data,width,height,nls,cx,cy);
		}
		/*!
		 * \brief  <b></b> /   (),  
		 *       data,width,height, nls 
		 *    ,    ker.
		 *     .
		 **************************************************************************************/
		template<typename T, class K> 
		static void CyclicInplace(T* data,int width,int height,int nls,const K& ker)
		{
			CyclicConvInplace<T,K> cc(ker);
			conv2DInplace(data,width,height,nls,cc,cc);
		}
		/*!
		 * \brief  /   (),  
		 *       data_in,width,height, nls_in 
		 *    ,   ker_x,ker_y.
		 *       data_out,width,height, nls_out.
		 *       (),  
		 *     (, , ) .
		 **************************************************************************************/
		template<typename T, class K1, class K2> 
		static void Reflect(const T* data_in,int width,int height,int nls_in,T* data_out,int nls_out,const K1& ker_x,const K2& ker_y)
		{
			Conv<T,K1> cx(ker_x);
			Conv<T,K2> cy(ker_y);
			conv2D(data_in,width,height,nls_in,data_out,nls_out,cx,cy);
		}
		/*!
		 * \brief  /   (),  
		 *       data_in,width,height, nls_in 
		 *    ,    ker.
		 *       data_out,width,height, nls_out.
		 *       (),  
		 *     (, , ) .
		 **************************************************************************************/
		template<typename T, class K> 
		static void Reflect(const T* data_in,int width,int height,int nls_in,T* data_out,int nls_out,const K& ker)
		{
			Conv<T,K> cc(ker);
			conv2D(data_in,width,height,nls_in,data_out,nls_out,cc,cc);
		}
		/*!
		 * \brief  /   (),  
		 *       data,width,height, nls 
		 *    ,   ker_x,ker_y .
		 *     .
		 *       (),  
		 *     (, , ) .
		 **************************************************************************************/
		template<typename T, class K1, class K2> 
		static void ReflectInplace(T* data,int width,int height,int nls,const K1& ker_x,const K2& ker_y)
		{
			ConvInplace<T,K1> cx(ker_x);
			ConvInplace<T,K2> cy(ker_y);
			conv2DInplace(data,width,height,nls,cx,cy);
		}
		/*!
		 * \brief  /   (),  
		 *       data,width,height, nls 
		 *    ,    ker.
		 *     .
		 *       (),  
		 *     (, , ) .
		 **************************************************************************************/
		template<typename T, class K> 
		static void ReflectInplace(T* data,int width,int height,int nls,const K& ker)
		{
			ConvInplace<T,K> cc(ker);
			conv2DInplace(data,width,height,nls,cc,cc);
		}
	};
	namespace Impl
	{
		namespace FastGaussConvolve
		{
			template<typename T,unsigned D> class HeadF;
			template<typename T,unsigned D> class HeadR;
			template<typename T> class HeadF<T,0>
			{
				const T B;
			public:
				HeadF(T B_,T prev_in) :B(B_) {}
				T operator()(T cur,T next) const { return  B*cur;}
			};
			template<typename T> class HeadR<T,0>
			{
				const T B;
			public:
				HeadR(T B_,T prev_in) :B(B_) {}
				T operator()(T cur,T next)  const {return  B*cur;}
			};
			template<typename T> class HeadF<T,1>
			{
				const T B2;
				T prev;
			public:
				HeadF(T B_,T prev_in) :B2(B_/2.0f) { prev=prev_in; }
				T operator()(T cur,T next) { T r=B2*(next-prev) ;  prev=cur; return  r; }
			};
			template<typename T> class HeadR<T,1>
			{
				const T B;
			public:
				HeadR(T B_,T prev_in) :B(B_) {}
				T operator()(T cur,T next)  const { return  B * cur;}
			};
			template<typename T> class HeadF<T,2>
			{
				const T B;
				T prev;
			public:
				HeadF(T B_,T prev_in) :B(B_) {prev=prev_in;}
				T operator()(T cur,T next)  { T r=B*(cur-prev);  prev=cur; return  r; }
			};
			template<typename T> class HeadR<T,2>
			{
				const T B;
				T prev;
			public:
				HeadR(T B_,T prev_in) :B(B_) {prev=prev_in;}
				T operator()(T cur,T next)  {  T r=B*(prev-cur); prev=cur; return r;}
			};
			template<typename T> class HeadF<T,3>
			{
				const T B;
				T prev;
			public:
				HeadF(T B_,T prev_in) :B(B_) {prev=prev_in;}
				T operator()(T cur,T next) { T r=B*(next-2*cur+prev); prev=cur; return  r;}
			};
			template<typename T> class HeadR<T,3>
			{
				const T B2;
				T prev;
			public:
				HeadR(T B_,T prev_in) : B2(B_/2) { prev=prev_in; }
				T operator()(T cur,T next)  { T r=B2*(prev-next); prev=cur; return  r;}
			};

			template<typename T> class Tail
			{
				const T a1;
				const T a2;
				const T a3;
				T s1;
				T s2;
				T s3;
			public:
				Tail(T a1_,T a2_,T a3_) :a1(a1_),a2(a2_),a3(a3_) 
				{
					s1=s2=s3=0;
				}
				///      (   - )
				///   .    ,    
				///      .  -     .
				/// /      ,    
				/// .        .  :
				///T fr(T* w)	{return a1 * w[1] + a2 * w[-2] + a3 * w[-3];}
				///T rr(T* out) {	return a1 * out[1] + a2 * out[2] + a3 * out[3];	}
				T operator()() 
				{ 
					return s1;
				}
				void operator()(T v) 
				{ 
					s1=s2+v*a1;
					s2=s3+v*a2;
					s3=v*a3;
				}
			};
			template<typename T,unsigned D> class Forward
			{
				HeadF<T,D> h;
				Tail<T> t;
			public:
				Forward(T B,T a1,T a2,T a3,T prev_in) : h(B,prev_in),t(a1,a2,a3){}
				void operator()(T& out,T in_cur,T in_next)
				{
					out=h(in_cur,in_next)+t();
					t(out);
				}
				T extent(T out,T in_cur,T in_next)
				{
					T out2=h(in_cur,in_next)+t();
					t(out2);
					return out2;
				}
				void operator()(T out)
				{
					t(out);
				}
			};
			template<typename T,unsigned D> class Reverse
			{
				HeadR<T,D> h;
				Tail<T> t;
			public:
				Reverse(T B,T a1,T a2,T a3,T prev_in) : h(B,prev_in),t(a1,a2,a3) {}
				void operator()(T& out,T in_cur,T in_next)
				{
					out=h(in_cur,in_next)+t();
					t(out);
				}
				void operator()(T out)
				{
					t(out);
				}
			};

			template<int n> class Triggs;
			template<>
			class Triggs<3>
			{
				double   M[9];
				double   den;
				double sum;
			public:
				Triggs(double B,double b1,double b2,double b3)
					:den(1.0-b1-b2-b3)
				{
					//sumsq = B;
					sum = B*B;
					TriggsM(b1,b2,b3,M);
				}
				void operator()(double iplus,double& p1,double& p2,double& p3)
				{
					double uplus = iplus/den;
					double vplus = uplus/den;

					double unp  = p1-uplus;
					double unp1 = p2-uplus;
					double unp2 = p3-uplus;


					p1 = M[0]*unp + M[1]*unp1 + M[2]*unp2 + vplus;
					p2  = M[3]*unp + M[4]*unp1 + M[5]*unp2 + vplus;
					p3  = M[6]*unp + M[7]*unp1 + M[8]*unp2 + vplus;

					p1 *= sum; p2 *= sum; p3 *= sum;
				}
				static void TriggsM(double a1,double a2,double a3,double *M)
				{
					double scale = 1.0/((1.0+a1-a2+a3)*(1.0-a1-a2-a3)*(1.0+a2+(a1-a3)*a3));
					M[0] =  scale * (-a3*a1+1.0-a3*a3-a2);
					M[1] =  scale * (a3+a1)*(a2+a3*a1);
					M[2] =  scale * a3*(a1+a3*a2);
					M[3] =  scale * (a1+a3*a2);
					M[4] = -scale * (a2-1.0)*(a2+a3*a1);
					M[5] = -scale * (a3*a1+a3*a3+a2-1.0)*a3;
					M[6] =  scale * (a3*a1+a2+a1*a1-a2*a2);
					M[7] =  scale * (a1*a2+a3*a2*a2-(a1+a3)*a3*a3-a3*a2+a3);
					M[8] =  scale * a3*(a1+a3*a2);
				}
			};
			template<typename T>
			void border(const T *src,T *dest,int sx)
			{
				T   B, b1, b2, b3;
				T   pix, p1, p2, p3;
				T   sum, sumsq;
				T   iplus;


				Triggs<3> triggs(B,b1,b2,b3);

				sumsq = B;
				sum = B*B;

				/* causal filter */
				p1 = *src/sumsq; p2 = p1; p3 = p1;
				iplus = src[sx-1];

				for (int j = 0; j < sx; j++) {
					pix = *src++ + b1*p1 + b2*p2 + b3*p3;
					*dest++ = pix;
					p3 = p2; p2 = p1; p1 = pix; /* update history */
				}

				// anti-causal filter 

				// apply Triggs border condition 
				triggs(iplus,p1,p2,p3);
				*(--dest) = p1;
					
			   for (int j = sx-2; j >= 0; j--) {
					pix = sum * *(--dest) + b1*p1 + b2*p2 + b3*p3;
					*dest = pix;
					p3 = p2; p2 = p1; p1 = pix;
				}
				dest += sx;
			}
		}
	}
	/*! \brief          3 .   \ref Lit_FastGauss
	 *
	 *  :
	 *
	 *  fr(w,n)   = a1 * w[n-1] + a2 * w[n-2] + a3 * w[n-3)        \n
	 *  rr(out,n) = a1 * out[n+1] + a2 * out[n+2] + a3 * out[n+3)  \n
	 *
	 * forward iteration:                                 \n
	 *   w[n]   = fh(in,n) + fr(w,n)                      \n
	 * reverse iteration:                                 \n
	 *   out[n] = rh(w,n) + rr(out,n) 
	 *
     *     :             \n
	 * Gauss:                                             \n
	 *   fh(in,n)= B*in[n]                                \n
	 *   rh(w,n) = B*w[n]                                 \n
	 * Gauss 1-st derivative:                             \n
	 *   fh(in,n)= B/2 * ( in[n+1] - in[n-1] )            \n
	 *   rh(w,n) = B*w[n]                                 \n
	 * Gauss 2-st derivative:                             \n
	 *   fh(in,n)= B * ( in[n] - in[n-1] )                \n
	 *   rh(w,n) = B * ( w[n+1] - w[n] )                  \n
	 * Gauss 3-st derivative:                             \n
	 *   fh(in,n)= B * ( in[n+1] - 2*in[n] + in[n-1] )    \n
	 *   rh(w,n) = B/2 * ( w[n+1] - w[n-1] )                     
	 *
	 * : \n
	 * a1=b1/b0     a2=b2/b0     a3=b3/b0                        \n
	 *
	 * b0 = 1.57825 + 2.44413*q + 1.4281*q^2 + 0.422205*q^3      \n
	 * b1 = 2.44413*q + 2.85619*q^2 + 1.26661*q^3                \n
	 * b2 = -1.4281*q^2 + 1.26661*q^3                            \n
	 * b3 = 0.422205*q^3                                         \n
	 *
	 * B = 1-(b1+b2+b3)/b0;                                      \n
	 *
	 * q = 0.98711*sigma - 0.96330        if sigma >= 2.5        \n
	 * q = 3.97156 - 4.14554*sqrt(1-0.26891*sigma)        if 0.5 <= sigma <= 2.5            \n
     *
	 *
	 */
	template<typename T,unsigned D> class FastGaussConvolve
	{
		T a1;
		T a2;
		T a3;
		T B;
	private:
		static void check()
		{
			LOKI_STATIC_CHECK(( (D>=0) && (D<=3) ),Fast_gauss_convolution_implemented_only_for_function_and_first_3_derivatives);
		}
	public:
		enum { Deriv=D };
		template<typename TT>
		static T Mult(TT sigma) 
		{
			switch(D)
			{
				case 0:
					return T(1.0f);
				case 1:
					return T(0.6f*sqrt(2.0f*3.1415926f)*sigma);
				case 2:
					return T(1.2f*sqrt(2.0f*3.1415926f)*sigma);
				case 3:
					return T(2.0f*sqrt(2.0f*3.1415926f)*sigma);
				default:
					return T(1.0f);
			};
		}
		FastGaussConvolve()
		{
			check();
		}
		template<typename TT>
		FastGaussConvolve(TT sigma)
		{
			check();
			Init(sigma);
		}
		template<typename TT>
		void Init(TT sigma)
		{
			double q = (sigma<2.5) ? (3.97156 - 4.14554*sqrt(1.0-0.26891*sigma)) : (0.98711*sigma - 0.96330);
			double b0 = 1.57825 + (2.44413 + (1.4281 + 0.422205*q)*q)*q;
			double b1 = (2.44413 + (2.85619 + 1.26661*q)*q)*q;
			double b2 = -(1.4281 + 1.26661*q)*q*q;
			double b3 = 0.422205*q*q*q;
			B = T(1-(b1+b2+b3)/b0);
			a1=T(b1/b0);
			a2=T(b2/b0);
			a3=T(b3/b0); 
//			std::cout << "FastGaussConvolve: "<< std::endl;
//			std::cout << "q= " << q << "  B="<<B<<"   a1="<<a1<<"   a2="<<a2<<"   a3="<<a3<< std::endl;
		}

		///  -  -     inplace
		template<typename IterIn,typename IterOut>
		void operator()(IterIn begin,IterIn end,IterOut out)
		{
			IterOut o=out;
			{
				
			}
		}

		///   inplace - OK!
		template<typename Iter>
		void operator()(Iter begin,Iter end)
		{
			Iter d=begin;
			Impl::FastGaussConvolve::Forward<T,D> f(B,a1,a2,a3,*d);
			if(D==0)
			{
				Iter i=d;
				T v1=*(++i);
				T v2=*(++i);
				T v3=*(++i);
				f(v3); 
				f(v2); 
				f(v1);
			}
			--end;
			--end;
			--end;
			--end;
			Iter in_cur=d;
			for( ; ++d != end ; in_cur=d )
				f(*in_cur,*in_cur,*d);
			T v4=*in_cur;
			f(*in_cur,*in_cur,*d); in_cur=d; ++d;
			T v3=*in_cur;
			f(*in_cur,*in_cur,*d); in_cur=d; ++d;
			T v2=*in_cur;
			f(*in_cur,*in_cur,*d); in_cur=d; ++d;
			T v1=*in_cur;
			f(*in_cur,*in_cur,*d); in_cur=d; 

			*in_cur=f.extent(*in_cur,*in_cur,v1); 
			v1=f.extent(v1,v1,v2); 
			v2=f.extent(v2,v2,v3); 
			v3=f.extent(v3,v3,v4); 

			Impl::FastGaussConvolve::Reverse<T,D> r(B,a1,a2,a3,v1);
			r(v3); 
			r(v2); 
			r(v1);
			--begin;//before first element 
			for( Iter out_cur=d ; --d != begin ; out_cur=d )
				r(*out_cur,*out_cur,*d);
			++d;
			r(*d,*d,0);

		}
	};
/////////////////////////////////////////////////////////////////////////////
} //namespace Conv
} //namespace ImAlg
} //namespace Vtl 
/////////////////////////////////////////////////////////////////////////////
#endif //_CONVOLVE_H 
