
#include "AbsoluteDualQuadric.h"

namespace Basics3D
{
	void AbsoluteDualQuadric::CreateSystem
		                      (Vtl::NumAlg::Lin::ublas::c_matrix<double, 8, 10> &out_system,
			                     const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> &in_P,
			                     const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> &in_P1)
	{
		// First view
		// aspect ratio = 0  =>  w11 = w22
		out_system(0, 0) =      in_P(0,0) * in_P(0,0) - in_P(1,0) * in_P(1,0);	// (0, 0)
		out_system(0, 1) = 2 * (in_P(0,0) * in_P(0,1) - in_P(1,0) * in_P(1,1));	// (0, 1)
		out_system(0, 2) = 2 * (in_P(0,0) * in_P(0,2) - in_P(1,0) * in_P(1,2));	// (0, 2)
		out_system(0, 3) = 2 * (in_P(0,0) * in_P(0,3) - in_P(1,0) * in_P(1,3));	// (0, 3)
		out_system(0, 4) =      in_P(0,1) * in_P(0,1) - in_P(1,1) * in_P(1,1);  // (1, 1)
		out_system(0, 5) = 2 * (in_P(0,1) * in_P(0,2) - in_P(1,1) * in_P(1,2));	// (1, 2)
		out_system(0, 6) = 2 * (in_P(0,1) * in_P(0,3) - in_P(1,1) * in_P(1,3));	// (1, 3)
		out_system(0, 7) =      in_P(0,2) * in_P(0,2) - in_P(1,2) * in_P(1,2);	// (2, 2)
		out_system(0, 8) = 2 * (in_P(0,2) * in_P(0,3) - in_P(1,2) * in_P(1,3));	// (2, 3)
		out_system(0, 9) =      in_P(0,3) * in_P(0,3) - in_P(1,3) * in_P(1,3);	// (3, 3)

  	// skew = 0  =>  w12 = 0
		out_system(1, 0) = in_P(0,0) * in_P(1,0);	                        // (0, 0)
		out_system(1, 1) = in_P(0,0) * in_P(1,1) + in_P(0,1) * in_P(1,0); // (0, 1)
		out_system(1, 2) = in_P(0,0) * in_P(1,2) + in_P(0,2) * in_P(1,0); // (0, 2)
		out_system(1, 3) = in_P(0,0) * in_P(1,3) + in_P(0,3) * in_P(1,0); // (0, 3)
		out_system(1, 4) = in_P(0,1) * in_P(1,1);	                        // (1, 1)
		out_system(1, 5) = in_P(0,1) * in_P(1,2) + in_P(0,2) * in_P(1,1); // (1, 2)
		out_system(1, 6) = in_P(0,1) * in_P(1,3) + in_P(0,3) * in_P(1,1); // (1, 3)
		out_system(1, 7) = in_P(0,2) * in_P(1,2);                       	// (2, 2)
		out_system(1, 8) = in_P(0,2) * in_P(1,3) + in_P(0,3) * in_P(1,2); // (2, 3)
		out_system(1, 9) = in_P(0,3) * in_P(1,3);		                      // (3, 3)

		// principal point = (0, 0)  =>  w13 = 0
		out_system(2, 0) = in_P(0,0) * in_P(2,0);		                      // (0, 0)
		out_system(2, 1) = in_P(0,0) * in_P(2,1) + in_P(0,1) * in_P(2,0); // (0, 1)
		out_system(2, 2) = in_P(0,0) * in_P(2,2) + in_P(0,2) * in_P(2,0); // (0, 2)
		out_system(2, 3) = in_P(0,0) * in_P(2,3) + in_P(0,3) * in_P(2,0); // (0, 3)
		out_system(2, 4) = in_P(0,1) * in_P(2,1);		                      // (1, 1)
		out_system(2, 5) = in_P(0,1) * in_P(2,2) + in_P(0,2) * in_P(2,1); // (1, 2)
		out_system(2, 6) = in_P(0,1) * in_P(2,3) + in_P(0,3) * in_P(2,1); // (1, 3)
		out_system(2, 7) = in_P(0,2) * in_P(2,2);		                      // (2, 2)
		out_system(2, 8) = in_P(0,2) * in_P(2,3) + in_P(0,3) * in_P(2,2); // (2, 3)
		out_system(2, 9) = in_P(0,3) * in_P(2,3);                      	  // (3, 3)

		// principal point = (0, 0)  =>  w23 = 0
		out_system(3, 0) = in_P(1,0) * in_P(2,0);		                      // (0, 0)
		out_system(3, 1) = in_P(1,0) * in_P(2,1) + in_P(1,1) * in_P(2,0); // (0, 1)
		out_system(3, 2) = in_P(1,0) * in_P(2,2) + in_P(1,2) * in_P(2,0); // (0, 2)
		out_system(3, 3) = in_P(1,0) * in_P(2,3) + in_P(1,3) * in_P(2,0); // (0, 3)
		out_system(3, 4) = in_P(1,1) * in_P(2,1);		                      // (1, 1)
		out_system(3, 5) = in_P(1,1) * in_P(2,2) + in_P(1,2) * in_P(2,1);	// (1, 2)
		out_system(3, 6) = in_P(1,1) * in_P(2,3) + in_P(1,3) * in_P(2,1);	// (1, 3)
		out_system(3, 7) = in_P(1,2) * in_P(2,2);		                      // (2, 2)
		out_system(3, 8) = in_P(1,2) * in_P(2,3) + in_P(1,3) * in_P(2,2);	// (2, 3)
		out_system(3, 9) = in_P(1,3) * in_P(2,3);		                      // (3, 3)

		
		//_____________________________________________________
		// Second view
		// aspect ratio = 0  =>  w11 = w22
		out_system(4, 0) =      in_P1(0,0) * in_P1(0,0) - in_P1(1,0) * in_P1(1,0);	// (0, 0)
		out_system(4, 1) = 2 * (in_P1(0,0) * in_P1(0,1) - in_P1(1,0) * in_P1(1,1));	// (0, 1)
		out_system(4, 2) = 2 * (in_P1(0,0) * in_P1(0,2) - in_P1(1,0) * in_P1(1,2));	// (0, 2)
		out_system(4, 3) = 2 * (in_P1(0,0) * in_P1(0,3) - in_P1(1,0) * in_P1(1,3));	// (0, 3)
		out_system(4, 4) =      in_P1(0,1) * in_P1(0,1) - in_P1(1,1) * in_P1(1,1);	// (1, 1)
		out_system(4, 5) = 2 * (in_P1(0,1) * in_P1(0,2) - in_P1(1,1) * in_P1(1,2));	// (1, 2)
		out_system(4, 6) = 2 * (in_P1(0,1) * in_P1(0,3) - in_P1(1,1) * in_P1(1,3));	// (1, 3)
		out_system(4, 7) =      in_P1(0,2) * in_P1(0,2) - in_P1(1,2) * in_P1(1,2);	// (2, 2)
		out_system(4, 8) = 2 * (in_P1(0,2) * in_P1(0,3) - in_P1(1,2) * in_P1(1,3));	// (2, 3)
		out_system(4, 9) =      in_P1(0,3) * in_P1(0,3) - in_P1(1,3) * in_P1(1,3);	// (3, 3)

  	// skew = 0  =>  w12 = 0
		out_system(5, 0) = in_P1(0,0) * in_P1(1,0);	                          // (0, 0)
		out_system(5, 1) = in_P1(0,0) * in_P1(1,1) + in_P1(0,1) * in_P1(1,0); // (0, 1)
		out_system(5, 2) = in_P1(0,0) * in_P1(1,2) + in_P1(0,2) * in_P1(1,0); // (0, 2)
		out_system(5, 3) = in_P1(0,0) * in_P1(1,3) + in_P1(0,3) * in_P1(1,0); // (0, 3)
		out_system(5, 4) = in_P1(0,1) * in_P1(1,1);	                          // (1, 1)
		out_system(5, 5) = in_P1(0,1) * in_P1(1,2) + in_P1(0,2) * in_P1(1,1); // (1, 2)
		out_system(5, 6) = in_P1(0,1) * in_P1(1,3) + in_P1(0,3) * in_P1(1,1); // (1, 3)
		out_system(5, 7) = in_P1(0,2) * in_P1(1,2);                       	  // (2, 2)
		out_system(5, 8) = in_P1(0,2) * in_P1(1,3) + in_P1(0,3) * in_P1(1,2); // (2, 3)
		out_system(5, 9) = in_P1(0,3) * in_P1(1,3);		                        // (3, 3)

		// principal point = (0, 0)  =>  w13 = 0
		out_system(6, 0) = in_P1(0,0) * in_P1(2,0);		                        // (0, 0)
		out_system(6, 1) = in_P1(0,0) * in_P1(2,1) + in_P1(0,1) * in_P1(2,0); // (0, 1)
		out_system(6, 2) = in_P1(0,0) * in_P1(2,2) + in_P1(0,2) * in_P1(2,0); // (0, 2)
		out_system(6, 3) = in_P1(0,0) * in_P1(2,3) + in_P1(0,3) * in_P1(2,0); // (0, 3)
		out_system(6, 4) = in_P1(0,1) * in_P1(2,1);		                        // (1, 1)
		out_system(6, 5) = in_P1(0,1) * in_P1(2,2) + in_P1(0,2) * in_P1(2,1); // (1, 2)
		out_system(6, 6) = in_P1(0,1) * in_P1(2,3) + in_P1(0,3) * in_P1(2,1); // (1, 3)
		out_system(6, 7) = in_P1(0,2) * in_P1(2,2);		                        // (2, 2)
		out_system(6, 8) = in_P1(0,2) * in_P1(2,3) + in_P1(0,3) * in_P1(2,2); // (2, 3)
		out_system(6, 9) = in_P1(0,3) * in_P1(2,3);                      	    // (3, 3)

		// principal point = (0, 0)  =>  w23 = 0
		out_system(7, 0) = in_P1(1,0) * in_P1(2,0);		                        // (0, 0)
		out_system(7, 1) = in_P1(1,0) * in_P1(2,1) + in_P1(1,1) * in_P1(2,0); // (0, 1)
		out_system(7, 2) = in_P1(1,0) * in_P1(2,2) + in_P1(1,2) * in_P1(2,0); // (0, 2)
		out_system(7, 3) = in_P1(1,0) * in_P1(2,3) + in_P1(1,3) * in_P1(2,0); // (0, 3)
		out_system(7, 4) = in_P1(1,1) * in_P1(2,1);		                        // (1, 1)
		out_system(7, 5) = in_P1(1,1) * in_P1(2,2) + in_P1(1,2) * in_P1(2,1);	// (1, 2)
		out_system(7, 6) = in_P1(1,1) * in_P1(2,3) + in_P1(1,3) * in_P1(2,1);	// (1, 3)
		out_system(7, 7) = in_P1(1,2) * in_P1(2,2);		                        // (2, 2)
		out_system(7, 8) = in_P1(1,2) * in_P1(2,3) + in_P1(1,3) * in_P1(2,2);	// (2, 3)
		out_system(7, 9) = in_P1(1,3) * in_P1(2,3);		                        // (3, 3)
	}
	
	///////////////////////////////////////////////////////////////////////////////////////
	int AbsoluteDualQuadric::SolveDet0
		                  (Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 16> &out_Q,
			                 const Vtl::NumAlg::Lin::ublas::c_matrix<double, 8, 10> &in_system)
	{
		/*std::cout << std::endl << std::endl << "System" << std::endl;
		T_MatrixOperations::Print(in_system);

		double ev = 0;
		for (int i = 4; i < 8; i++)
		{
			for (int j = 0; j < 16; j++)
			{
				ev += in_system(i, j);
			}
		}
		std::cout << std::endl << std::endl << "ev = "
			<< ev / (4 * 16);*/

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 8, 10> U;
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 10, 10> V;
		Vtl::NumAlg::Lin::ublas::c_vector<double, 10> W;
		Vtl::NumAlg::Lin::ublas::c_vector<double, 10> RV1;

		U = in_system;

		Vtl::NumAlg::Lin::svd::decomp(U, W, V, RV1);

		std::cout << std::endl << std::endl << "U" << std::endl;
		T_MatrixOperations::Print(U);
		std::cout << std::endl << std::endl << "V" << std::endl;
		T_MatrixOperations::Print(V);
		std::cout << std::endl << std::endl << "W" << std::endl;
		for (int i = 0; i < 8; i++)
		{
			std::cout << W(i) << "  ";
		}

		double min1 = fabs(W(0));
		double min2 = fabs(W(0));
		int k1 = 0;
		int k2 = 0;
		
		for (int j = 1; j < 8; j++)
		{
			if (fabs(W(j)) <= min1)
			{
				min2 = min1;
				k2 = k1;
				min1 = fabs(W(j));
				k1 = j;
			}
			else if (fabs(W(j)) <= min2)
			{
				min2 = fabs(W(j));
				k2 = j;
			}
		}

		std::cout << std::endl << std::endl << "k1 = " << k1 << "  k2 = " << k2;

		std::vector<double> a(16);
		a[0]  = V(0, k1); a[1]  = V(1, k1); a[2]  = V(2, k1); a[3]  = V(3, k1);
		a[4]  = V(1, k1); a[5]  = V(4, k1); a[6]  = V(5, k1); a[7]  = V(6, k1);
		a[8]  = V(2, k1); a[9]  = V(5, k1); a[10] = V(7, k1); a[11] = V(8, k1);
		a[12] = V(3, k1); a[13] = V(6, k1); a[14] = V(8, k1); a[15] = V(9, k1);

		std::vector<double> b(16);
		b[0]  = V(0, k2); b[1]  = V(1, k2); b[2]  = V(2, k2); b[3]  = V(3, k2);
		b[4]  = V(1, k2); b[5]  = V(4, k2); b[6]  = V(5, k2); b[7]  = V(6, k2);
		b[8]  = V(2, k2); b[9]  = V(5, k2); b[10] = V(7, k2); b[11] = V(8, k2);
		b[12] = V(3, k2); b[13] = V(6, k2); b[14] = V(8, k2); a[15] = V(9, k2);

		std::cout << std::endl << "a   ";
		for (int i = 0; i < 16; i++)
		{
			std::cout << a[i] << "  ";
		}
		std::cout << std::endl << "b   ";
		for (int i = 0; i < 16; i++)
		{
			std::cout << b[i] << "  ";
		}

		// a * k^4 + b * k^3 + c * k^2 + d * k + e = 0
		double a1 = -b[12]*b[9]*b[2]*b[7]-a[4]*b[9]*b[14]*a[3]
		-a[4]*b[9]*a[2]*a[15]+a[4]*b[9]*a[2]*b[15]+a[4]*b[9]*b[2]*a[15]
		-b[8]*a[5]*b[2]*a[15]+b[8]*a[5]*b[2]*b[15]-b[8]*a[1]*a[6]*a[15]
		+b[8]*a[1]*a[6]*b[15]-b[0]*a[13]*a[6]*a[11]-a[8]*b[1]*a[6]*a[15]
		+a[8]*b[1]*a[6]*b[15]+a[8]*b[1]*b[6]*a[15]-b[12]*b[5]*b[10]*b[3]
		+b[0]*a[5]*a[10]*b[15]+b[0]*b[9]*b[6]*a[15]+b[0]*b[9]*a[6]*b[15]
		-b[0]*b[9]*a[6]*a[15]-b[0]*a[9]*b[14]*b[7]-b[0]*a[13]*b[10]*a[7]
		+b[12]*b[1]*b[10]*b[7]+b[12]*b[5]*b[2]*b[11]-a[8]*b[5]*a[14]*a[3]
		+a[8]*b[5]*a[14]*b[3]-b[0]*a[9]*a[14]*a[7]+b[0]*a[9]*a[14]*b[7]
		+b[0]*a[9]*b[14]*a[7]-a[4]*b[13]*a[10]*a[3]+a[4]*b[13]*b[2]*b[11]
		-a[4]*b[13]*b[10]*b[3]+a[4]*b[13]*a[10]*b[3]-b[4]*a[13]*b[2]*a[11]
		+b[4]*a[13]*b[2]*b[11]+b[12]*a[1]*b[10]*a[7]-a[12]*a[9]*b[6]*a[3]
		+a[12]*b[9]*b[2]*b[7]-b[12]*a[1]*b[6]*a[11]+b[12]*a[1]*b[6]*b[11]
		+a[0]*a[5]*a[10]*a[15]-a[0]*b[5]*b[11]*a[14]+a[0]*a[9]*a[14]*a[7]
		-a[0]*a[9]*a[14]*b[7]-b[12]*b[1]*b[6]*b[11]-b[12]*a[9]*a[2]*b[7]
		-b[12]*a[9]*b[6]*b[3]+b[12]*b[9]*b[6]*b[3]+a[8]*a[13]*a[2]*a[7]
		-a[8]*a[13]*a[2]*b[7]-b[0]*b[13]*a[10]*a[7]-b[0]*b[9]*a[14]*b[7]
		+b[0]*b[9]*a[14]*a[7]-b[0]*b[9]*b[14]*a[7]+a[8]*a[13]*a[6]*b[3]
		+a[8]*a[13]*b[6]*a[3]-a[8]*a[13]*b[6]*b[3]-a[12]*a[9]*a[2]*a[7]
		+a[12]*a[9]*a[2]*b[7]+a[12]*a[9]*b[2]*a[7]+a[12]*a[9]*b[6]*b[3]
		-b[12]*b[9]*a[6]*b[3]-b[12]*b[9]*b[6]*a[3]+a[12]*b[1]*a[6]*a[11]
		-a[12]*b[1]*a[6]*b[11]-a[12]*b[1]*b[6]*a[11]-b[8]*b[1]*b[14]*b[7]
		+b[8]*b[1]*b[6]*b[15]-a[12]*b[9]*a[6]*a[3]+a[12]*b[9]*a[6]*b[3]
		+b[4]*a[13]*a[2]*a[11]+b[4]*a[13]*a[10]*b[3]+b[4]*a[13]*b[10]*a[3]
		-b[4]*a[13]*b[10]*b[3]-b[4]*a[13]*a[10]*a[3]+b[0]*b[13]*a[6]*a[11]
		-b[0]*b[13]*a[6]*b[11]-b[0]*b[5]*b[11]*b[14]+a[12]*a[1]*b[6]*a[11]
		-a[12]*a[1]*b[6]*b[11]+b[8]*a[13]*a[6]*a[3]-b[8]*a[13]*a[6]*b[3]
		-b[8]*a[13]*b[6]*a[3]+b[8]*a[13]*b[6]*b[3]-a[4]*a[1]*a[10]*a[15]
		+a[4]*a[1]*a[10]*b[15]+a[4]*b[13]*a[2]*a[11]-a[4]*b[13]*a[2]*b[11]
		-a[4]*b[13]*b[2]*a[11]+a[12]*a[1]*a[10]*a[7]-a[12]*a[1]*a[10]*b[7]
		-b[12]*b[9]*a[2]*a[7]+b[12]*b[9]*a[2]*b[7]+b[12]*b[9]*b[2]*a[7]
		+b[12]*a[9]*b[6]*a[3]+b[12]*b[9]*a[6]*a[3]+b[0]*a[9]*a[6]*a[15]
		+b[0]*a[13]*b[10]*b[7]-b[4]*b[1]*b[10]*b[15]-b[4]*b[13]*a[2]*a[11]
		+b[4]*b[13]*a[2]*b[11]+b[4]*b[13]*b[2]*a[11]-a[8]*a[5]*b[14]*a[3]
		+a[8]*a[5]*b[14]*b[3]-b[4]*b[9]*a[14]*a[3]+b[4]*b[9]*a[14]*b[3]
		+b[4]*a[1]*b[11]*a[14]-b[4]*a[1]*b[11]*b[14]-b[0]*b[5]*b[10]*a[15]
		+b[4]*b[13]*a[10]*a[3]-b[4]*b[13]*a[10]*b[3]-b[4]*b[13]*b[10]*a[3]
		-a[0]*a[9]*b[14]*a[7]+a[0]*a[9]*b[14]*b[7]+a[0]*a[13]*b[10]*a[7]
		-b[0]*b[5]*a[11]*a[14]+b[0]*b[5]*a[11]*b[14]-b[0]*a[5]*a[10]*a[15]
		+b[0]*b[5]*b[11]*a[14]-a[4]*b[1]*b[10]*a[15]-b[12]*a[9]*a[6]*a[3]
		-a[0]*b[9]*b[6]*a[15]-a[0]*b[9]*a[6]*b[15]+a[0]*b[9]*a[6]*a[15]
		-b[12]*a[1]*a[6]*b[11]-a[0]*b[13]*a[10]*b[7]-a[0]*b[13]*b[10]*a[7]
		+a[0]*a[13]*b[6]*b[11]+a[4]*b[9]*a[14]*a[3]-a[4]*b[9]*a[14]*b[3]
		-a[4]*a[1]*b[11]*a[14]+a[4]*a[1]*b[11]*b[14]+b[4]*a[9]*a[14]*a[3]
		-b[4]*b[1]*a[10]*a[15]+b[4]*b[1]*a[10]*b[15]+b[4]*b[1]*b[10]*a[15]
		-a[0]*a[13]*b[6]*a[11]-b[8]*b[13]*b[6]*b[3]+a[4]*a[1]*b[10]*a[15]
		-a[4]*a[1]*b[10]*b[15]+a[0]*b[9]*a[14]*b[7]-a[0]*b[9]*a[14]*a[7]
		+a[0]*b[9]*b[14]*a[7]-a[0]*a[5]*a[10]*b[15]+a[4]*a[1]*a[11]*a[14]
		-a[4]*a[1]*a[11]*b[14]-b[8]*b[5]*b[14]*a[3]-b[8]*b[5]*a[2]*a[15]
		+b[8]*b[5]*a[2]*b[15]+b[8]*b[5]*b[2]*a[15]-a[8]*a[1]*a[14]*a[7]
		+a[8]*a[1]*a[14]*b[7]+a[8]*a[1]*b[14]*a[7]-a[8]*a[1]*b[14]*b[7]
		+b[8]*b[13]*b[2]*b[7]-a[0]*a[13]*a[6]*b[11]+a[0]*b[9]*b[6]*b[15]
		-a[0]*a[13]*a[10]*a[7]+a[0]*a[13]*a[10]*b[7]-a[0]*a[5]*b[10]*a[15]
		+b[0]*b[9]*b[14]*b[7]-b[0]*b[13]*b[10]*b[7]+b[0]*b[13]*b[6]*b[11]
		-b[0]*b[9]*b[6]*b[15]+b[4]*b[9]*b[2]*b[15]-b[0]*a[9]*b[6]*a[15]
		+b[0]*a[9]*b[6]*b[15]-a[0]*a[9]*b[6]*b[15]+a[0]*a[13]*a[6]*a[11]+
			b[4]*b[1]*a[11]*a[14]-b[4]*b[1]*a[11]*b[14]-b[4]*b[1]*b[11]*a[14]
		-a[0]*b[13]*a[6]*a[11]-b[12]*b[1]*a[6]*a[11]+b[12]*b[1]*a[6]*b[11]
		+b[12]*b[1]*a[10]*a[7]-b[12]*b[1]*a[10]*b[7]-b[12]*b[1]*b[10]*a[7]
		+b[12]*a[5]*a[10]*a[3]-a[8]*a[1]*b[6]*a[15]+a[8]*a[1]*b[6]*b[15]
		-b[12]*a[5]*a[10]*b[3]+b[12]*a[9]*b[2]*b[7]-b[12]*a[5]*b[10]*a[3]
		+b[12]*a[5]*b[10]*b[3]-b[12]*a[1]*a[10]*a[7]+b[12]*a[1]*a[10]*b[7]
		-a[8]*a[13]*b[2]*a[7]+a[8]*a[13]*b[2]*b[7]-a[8]*b[13]*b[2]*b[7]
		+a[8]*b[13]*b[6]*b[3]-a[8]*a[13]*a[6]*a[3]-a[0]*a[13]*b[10]*b[7]
		+a[0]*b[13]*b[10]*b[7]-a[0]*a[9]*a[6]*a[15]-a[4]*a[9]*a[2]*b[15]
		-b[12]*a[9]*b[2]*a[7]+b[12]*a[1]*a[6]*a[11]+a[8]*b[13]*a[6]*a[3]
		-a[8]*b[13]*a[6]*b[3]-a[8]*b[13]*b[6]*a[3]-a[8]*b[13]*a[2]*a[7]
		+a[8]*b[13]*a[2]*b[7]+a[8]*b[13]*b[2]*a[7]-a[4]*a[9]*b[2]*a[15]
		+a[4]*a[9]*b[2]*b[15]-b[4]*a[1]*b[10]*a[15]+b[4]*a[1]*b[10]*b[15]
		-b[4]*a[1]*a[11]*a[14]+a[4]*b[1]*b[10]*b[15]-a[4]*b[1]*b[11]*b[14]
		-a[4]*a[13]*a[2]*a[11]-a[4]*a[13]*a[10]*b[3]-b[8]*b[5]*b[2]*b[15]
		+b[8]*a[1]*b[6]*a[15]-b[8]*a[1]*b[6]*b[15]-a[8]*b[5]*b[14]*b[3]
		+a[8]*b[5]*b[2]*b[15]+b[8]*b[5]*b[14]*b[3]+b[4]*b[1]*b[11]*b[14]
		-b[4]*b[13]*b[2]*b[11]+b[4]*b[13]*b[10]*b[3]-b[4]*b[9]*b[14]*b[3]
		+b[8]*a[5]*b[14]*a[3]-b[8]*a[5]*b[14]*b[3]+b[8]*a[5]*a[2]*a[15]
		-b[8]*a[5]*a[2]*b[15]-a[4]*a[9]*b[14]*b[3]+a[4]*a[9]*a[2]*a[15]
		+a[12]*a[5]*a[2]*a[11]-a[12]*a[5]*a[2]*b[11]-a[12]*a[5]*b[2]*a[11]
		+a[12]*a[5]*b[2]*b[11]+a[12]*a[9]*a[6]*a[3]-a[12]*a[9]*a[6]*b[3]
		-b[4]*a[9]*a[2]*a[15]+b[4]*a[9]*a[2]*b[15]+b[4]*a[9]*b[2]*a[15]
		-b[4]*a[9]*b[2]*b[15]-b[4]*a[9]*b[14]*a[3]+b[4]*a[9]*b[14]*b[3]
		-b[8]*a[13]*a[2]*a[7]+b[8]*a[13]*a[2]*b[7]+b[8]*a[13]*b[2]*a[7]
		-b[8]*a[13]*b[2]*b[7]+b[0]*b[13]*a[10]*b[7]+b[0]*b[13]*b[10]*a[7]
		-b[12]*a[5]*a[2]*a[11]+b[12]*a[5]*a[2]*b[11]+b[12]*b[1]*b[6]*a[11]
		+a[0]*a[9]*a[6]*b[15]-a[0]*b[13]*b[6]*b[11]+a[0]*a[9]*b[6]*a[15]
		+a[0]*b[13]*a[6]*b[11]+a[0]*b[13]*b[6]*a[11]+a[0]*b[13]*a[10]*a[7]
		+b[0]*a[13]*a[10]*a[7]-b[0]*a[13]*a[10]*b[7]+b[0]*a[5]*b[10]*a[15]
		+a[4]*a[9]*a[14]*b[3]+a[4]*a[9]*b[14]*a[3]-b[8]*a[5]*a[14]*a[3]
		+b[8]*a[5]*a[14]*b[3]+b[12]*a[5]*b[2]*a[11]-b[12]*a[5]*b[2]*b[11]
		-b[0]*a[13]*b[6]*b[11]+b[0]*a[13]*b[6]*a[11]+b[0]*a[13]*a[6]*b[11]
		-b[0]*a[9]*a[6]*b[15]+b[8]*a[1]*b[14]*b[7]-a[4]*b[1]*a[11]*a[14]
		+a[4]*b[1]*a[11]*b[14]+a[4]*b[1]*b[11]*a[14]-a[4]*a[9]*a[14]*a[3]
		+a[4]*b[1]*a[10]*a[15]-a[4]*b[1]*a[10]*b[15]-a[8]*a[5]*a[2]*a[15]
		+a[8]*a[5]*a[2]*b[15]+a[8]*a[5]*b[2]*a[15]-a[8]*a[5]*b[2]*b[15]
		+b[0]*b[5]*a[10]*a[15]-b[0]*a[5]*b[10]*b[15]+b[0]*a[5]*a[11]*a[14]
		-b[0]*a[5]*a[11]*b[14]-b[0]*b[5]*a[10]*b[15]+b[12]*a[9]*a[2]*a[7]
		+b[0]*b[5]*b[10]*b[15]+a[4]*b[13]*b[10]*a[3]+a[4]*b[9]*b[14]*b[3]
		-a[4]*b[9]*b[2]*b[15]-b[8]*b[1]*a[14]*a[7]+b[8]*b[1]*a[14]*b[7]
		+b[8]*b[1]*b[14]*a[7]-a[4]*a[13]*b[10]*a[3]+b[8]*b[13]*a[2]*a[7]
		-b[8]*b[13]*a[2]*b[7]-b[8]*b[13]*b[2]*a[7]-b[8]*b[13]*a[6]*a[3]
		+b[8]*b[13]*a[6]*b[3]+b[4]*b[9]*b[14]*a[3]-a[12]*b[5]*a[10]*b[3]
		-a[12]*b[5]*b[10]*a[3]-a[12]*b[5]*a[2]*a[11]+a[12]*b[5]*a[2]*b[11]
		+a[12]*b[5]*b[2]*a[11]-a[8]*b[1]*b[6]*b[15]+a[8]*b[1]*b[14]*b[7]
		+a[8]*a[1]*a[6]*a[15]-a[8]*a[1]*a[6]*b[15]+b[4]*b[9]*a[2]*a[15]
		-b[4]*b[9]*a[2]*b[15]-b[4]*b[9]*b[2]*a[15]-b[4]*a[13]*a[2]*b[11]
		-b[0]*a[5]*b[11]*a[14]+b[0]*a[5]*b[11]*b[14]-b[12]*b[5]*a[10]*a[3]
		+b[12]*b[5]*a[10]*b[3]+b[12]*b[5]*b[10]*a[3]-b[8]*b[5]*a[14]*b[3]
		-a[12]*b[9]*b[6]*b[3]+a[12]*b[1]*b[6]*b[11]-a[12]*b[1]*b[10]*b[7]
		-a[12]*b[5]*b[2]*b[11]-a[12]*a[1]*a[6]*a[11]+a[12]*a[1]*a[6]*b[11]
		-a[12]*a[1]*b[10]*a[7]+a[12]*b[5]*a[10]*a[3]+a[8]*a[5]*a[14]*a[3]
		-a[8]*a[5]*a[14]*b[3]-a[0]*b[5]*a[10]*a[15]-a[0]*b[5]*b[10]*b[15]
		+a[0]*a[5]*b[10]*b[15]-a[12]*b[1]*a[10]*a[7]+a[12]*b[1]*a[10]*b[7]
		+a[12]*b[1]*b[10]*a[7]-a[12]*a[5]*a[10]*a[3]+a[0]*a[5]*b[11]*a[14]
		-a[0]*a[5]*b[11]*b[14]+a[0]*b[5]*b[11]*b[14]+a[8]*b[5]*b[14]*a[3]
		+a[8]*b[5]*a[2]*a[15]-a[8]*b[5]*a[2]*b[15]-a[8]*b[5]*b[2]*a[15]
		-a[12]*b[9]*b[2]*a[7]-b[12]*a[1]*b[10]*b[7]+b[4]*a[1]*a[11]*b[14]
		-a[0]*a[5]*a[11]*a[14]+a[0]*a[5]*a[11]*b[14]+a[0]*b[5]*a[10]*b[15]
		+a[0]*b[5]*b[10]*a[15]+a[4]*a[13]*b[10]*b[3]+a[4]*a[13]*a[10]*a[3]
		+b[8]*b[13]*b[6]*a[3]+b[8]*a[1]*a[14]*a[7]-b[8]*a[1]*a[14]*b[7]
		-b[8]*a[1]*b[14]*a[7]-b[4]*a[9]*a[14]*b[3]+b[12]*a[9]*a[6]*b[3]
		+b[12]*b[5]*a[2]*a[11]-b[12]*b[5]*a[2]*b[11]-b[12]*b[5]*b[2]*a[11]
		-b[0]*b[13]*b[6]*a[11]-a[0]*b[9]*b[14]*b[7]+a[0]*b[5]*a[11]*a[14]
		-a[0]*b[5]*a[11]*b[14]+b[8]*b[1]*a[6]*a[15]-b[8]*b[1]*a[6]*b[15]
		-b[8]*b[1]*b[6]*a[15]+a[12]*b[9]*b[6]*a[3]+a[12]*b[9]*a[2]*a[7]
		-a[12]*b[9]*a[2]*b[7]+a[12]*a[1]*b[10]*b[7]+b[4]*a[1]*a[10]*a[15]
		-b[4]*a[1]*a[10]*b[15]+a[12]*b[5]*b[10]*b[3]+a[12]*a[5]*a[10]*b[3]
		-a[12]*a[9]*b[2]*b[7]+a[12]*a[5]*b[10]*a[3]-a[12]*a[5]*b[10]*b[3]
		+a[4]*a[13]*a[2]*b[11]+a[4]*a[13]*b[2]*a[11]-a[4]*a[13]*b[2]*b[11]
		+a[8]*b[1]*a[14]*a[7]-a[8]*b[1]*a[14]*b[7]-a[8]*b[1]*b[14]*a[7]+b[8]*b[5]*a[14]*a[3];

		double b1 = 4*b[12]*b[9]*b[2]*b[7]+2*a[4]*b[9]*b[14]*a[3]+a[4]*b[9]*a[2]*a[15]
		-2*a[4]*b[9]*a[2]*b[15]-2*a[4]*b[9]*b[2]*a[15]+2*b[8]*a[5]*b[2]*a[15]
		-3*b[8]*a[5]*b[2]*b[15]+b[8]*a[1]*a[6]*a[15]-2*b[8]*a[1]*a[6]*b[15]
		+b[0]*a[13]*a[6]*a[11]+a[8]*b[1]*a[6]*a[15]-2*a[8]*b[1]*a[6]*b[15]
		-2*a[8]*b[1]*b[6]*a[15]+4*b[12]*b[5]*b[10]*b[3]-2*b[0]*a[5]*a[10]*b[15]
		-3*b[0]*b[9]*b[6]*a[15]-3*b[0]*b[9]*a[6]*b[15]+2*b[0]*b[9]*a[6]*a[15]
		+3*b[0]*a[9]*b[14]*b[7]+2*b[0]*a[13]*b[10]*a[7]-4*b[12]*b[1]*b[10]*b[7]
		-4*b[12]*b[5]*b[2]*b[11]+a[8]*b[5]*a[14]*a[3]-2*a[8]*b[5]*a[14]*b[3]
		+b[0]*a[9]*a[14]*a[7]-2*b[0]*a[9]*a[14]*b[7]-2*b[0]*a[9]*b[14]*a[7]
		+a[4]*b[13]*a[10]*a[3]-3*a[4]*b[13]*b[2]*b[11]+3*a[4]*b[13]*b[10]*b[3]
		-2*a[4]*b[13]*a[10]*b[3]+2*b[4]*a[13]*b[2]*a[11]-3*b[4]*a[13]*b[2]*b[11]
		-2*b[12]*a[1]*b[10]*a[7]+a[12]*a[9]*b[6]*a[3]-3*a[12]*b[9]*b[2]*b[7]
		+2*b[12]*a[1]*b[6]*a[11]-3*b[12]*a[1]*b[6]*b[11]+2*a[0]*b[5]*b[11]*a[14]
		+a[0]*a[9]*a[14]*b[7]+4*b[12]*b[1]*b[6]*b[11]+2*b[12]*a[9]*a[2]*b[7]
		+3*b[12]*a[9]*b[6]*b[3]-4*b[12]*b[9]*b[6]*b[3]+a[8]*a[13]*a[2]*b[7]
		+2*b[0]*b[13]*a[10]*a[7]+3*b[0]*b[9]*a[14]*b[7]-2*b[0]*b[9]*a[14]*a[7]
		+3*b[0]*b[9]*b[14]*a[7]-a[8]*a[13]*a[6]*b[3]-a[8]*a[13]*b[6]*a[3]
		+2*a[8]*a[13]*b[6]*b[3]-a[12]*a[9]*a[2]*b[7]-a[12]*a[9]*b[2]*a[7]
		-2*a[12]*a[9]*b[6]*b[3]+3*b[12]*b[9]*a[6]*b[3]+3*b[12]*b[9]*b[6]*a[3]
		-a[12]*b[1]*a[6]*a[11]+2*a[12]*b[1]*a[6]*b[11]+2*a[12]*b[1]*b[6]*a[11]
		+4*b[8]*b[1]*b[14]*b[7]-4*b[8]*b[1]*b[6]*b[15]+a[12]*b[9]*a[6]*a[3]
		-2*a[12]*b[9]*a[6]*b[3]-b[4]*a[13]*a[2]*a[11]-2*b[4]*a[13]*a[10]*b[3]
		-2*b[4]*a[13]*b[10]*a[3]+3*b[4]*a[13]*b[10]*b[3]+b[4]*a[13]*a[10]*a[3]
		-2*b[0]*b[13]*a[6]*a[11]+3*b[0]*b[13]*a[6]*b[11]+4*b[0]*b[5]*b[11]*b[14]
		-a[12]*a[1]*b[6]*a[11]+2*a[12]*a[1]*b[6]*b[11]-b[8]*a[13]*a[6]*a[3]
		+2*b[8]*a[13]*a[6]*b[3]+2*b[8]*a[13]*b[6]*a[3]-3*b[8]*a[13]*b[6]*b[3]
		-a[4]*a[1]*a[10]*b[15]-a[4]*b[13]*a[2]*a[11]+2*a[4]*b[13]*a[2]*b[11]
		+2*a[4]*b[13]*b[2]*a[11]+a[12]*a[1]*a[10]*b[7]+2*b[12]*b[9]*a[2]*a[7]
		-3*b[12]*b[9]*a[2]*b[7]-3*b[12]*b[9]*b[2]*a[7]-2*b[12]*a[9]*b[6]*a[3]
		-2*b[12]*b[9]*a[6]*a[3]-b[0]*a[9]*a[6]*a[15]-3*b[0]*a[13]*b[10]*b[7]
		+4*b[4]*b[1]*b[10]*b[15]+2*b[4]*b[13]*a[2]*a[11]-3*b[4]*b[13]*a[2]*b[11]
		-3*b[4]*b[13]*b[2]*a[11]+a[8]*a[5]*b[14]*a[3]-2*a[8]*a[5]*b[14]*b[3]
		+2*b[4]*b[9]*a[14]*a[3]-3*b[4]*b[9]*a[14]*b[3]-2*b[4]*a[1]*b[11]*a[14]
		+3*b[4]*a[1]*b[11]*b[14]+3*b[0]*b[5]*b[10]*a[15]-2*b[4]*b[13]*a[10]*a[3]
		+3*b[4]*b[13]*a[10]*b[3]+3*b[4]*b[13]*b[10]*a[3]+a[0]*a[9]*b[14]*a[7]
		-2*a[0]*a[9]*b[14]*b[7]-a[0]*a[13]*b[10]*a[7]+2*b[0]*b[5]*a[11]*a[14]
		-3*b[0]*b[5]*a[11]*b[14]+b[0]*a[5]*a[10]*a[15]-3*b[0]*b[5]*b[11]*a[14]
		+2*a[4]*b[1]*b[10]*a[15]+b[12]*a[9]*a[6]*a[3]+2*a[0]*b[9]*b[6]*a[15]
		+2*a[0]*b[9]*a[6]*b[15]-a[0]*b[9]*a[6]*a[15]+2*b[12]*a[1]*a[6]*b[11]
		+2*a[0]*b[13]*a[10]*b[7]+2*a[0]*b[13]*b[10]*a[7]-2*a[0]*a[13]*b[6]*b[11]
		-a[4]*b[9]*a[14]*a[3]+2*a[4]*b[9]*a[14]*b[3]+a[4]*a[1]*b[11]*a[14]
		-2*a[4]*a[1]*b[11]*b[14]-b[4]*a[9]*a[14]*a[3]+2*b[4]*b[1]*a[10]*a[15]
		-3*b[4]*b[1]*a[10]*b[15]-3*b[4]*b[1]*b[10]*a[15]+a[0]*a[13]*b[6]*a[11]
		+4*b[8]*b[13]*b[6]*b[3]-a[4]*a[1]*b[10]*a[15]+2*a[4]*a[1]*b[10]*b[15]
		-2*a[0]*b[9]*a[14]*b[7]+a[0]*b[9]*a[14]*a[7]-2*a[0]*b[9]*b[14]*a[7]
		+a[0]*a[5]*a[10]*b[15]+a[4]*a[1]*a[11]*b[14]+3*b[8]*b[5]*b[14]*a[3]
		+2*b[8]*b[5]*a[2]*a[15]-3*b[8]*b[5]*a[2]*b[15]-3*b[8]*b[5]*b[2]*a[15]
		-a[8]*a[1]*a[14]*b[7]-a[8]*a[1]*b[14]*a[7]+2*a[8]*a[1]*b[14]*b[7]
		-4*b[8]*b[13]*b[2]*b[7]+a[0]*a[13]*a[6]*b[11]-3*a[0]*b[9]*b[6]*b[15]
		-a[0]*a[13]*a[10]*b[7]+a[0]*a[5]*b[10]*a[15]-4*b[0]*b[9]*b[14]*b[7]
		+4*b[0]*b[13]*b[10]*b[7]-4*b[0]*b[13]*b[6]*b[11]+4*b[0]*b[9]*b[6]*b[15]
		-4*b[4]*b[9]*b[2]*b[15]+2*b[0]*a[9]*b[6]*a[15]-3*b[0]*a[9]*b[6]*b[15]
		+2*a[0]*a[9]*b[6]*b[15]-2*b[4]*b[1]*a[11]*a[14]+3*b[4]*b[1]*a[11]*b[14]
		+3*b[4]*b[1]*b[11]*a[14]+a[0]*b[13]*a[6]*a[11]+2*b[12]*b[1]*a[6]*a[11]
		-3*b[12]*b[1]*a[6]*b[11]-2*b[12]*b[1]*a[10]*a[7]+3*b[12]*b[1]*a[10]*b[7]
		+3*b[12]*b[1]*b[10]*a[7]-b[12]*a[5]*a[10]*a[3]+a[8]*a[1]*b[6]*a[15]
		-2*a[8]*a[1]*b[6]*b[15]+2*b[12]*a[5]*a[10]*b[3]-3*b[12]*a[9]*b[2]*b[7]
		+2*b[12]*a[5]*b[10]*a[3]-3*b[12]*a[5]*b[10]*b[3]+b[12]*a[1]*a[10]*a[7]
		-2*b[12]*a[1]*a[10]*b[7]+a[8]*a[13]*b[2]*a[7]-2*a[8]*a[13]*b[2]*b[7]
		+3*a[8]*b[13]*b[2]*b[7]-3*a[8]*b[13]*b[6]*b[3]+2*a[0]*a[13]*b[10]*b[7]
		-3*a[0]*b[13]*b[10]*b[7]+a[4]*a[9]*a[2]*b[15]+2*b[12]*a[9]*b[2]*a[7]
		-b[12]*a[1]*a[6]*a[11]-a[8]*b[13]*a[6]*a[3]+2*a[8]*b[13]*a[6]*b[3]
		+2*a[8]*b[13]*b[6]*a[3]+a[8]*b[13]*a[2]*a[7]-2*a[8]*b[13]*a[2]*b[7]
		-2*a[8]*b[13]*b[2]*a[7]+a[4]*a[9]*b[2]*a[15]-2*a[4]*a[9]*b[2]*b[15]
		+2*b[4]*a[1]*b[10]*a[15]-3*b[4]*a[1]*b[10]*b[15]+b[4]*a[1]*a[11]*a[14]
		-3*a[4]*b[1]*b[10]*b[15]+3*a[4]*b[1]*b[11]*b[14]+a[4]*a[13]*a[10]*b[3]
		+4*b[8]*b[5]*b[2]*b[15]-2*b[8]*a[1]*b[6]*a[15]+3*b[8]*a[1]*b[6]*b[15]
		+3*a[8]*b[5]*b[14]*b[3]-3*a[8]*b[5]*b[2]*b[15]-4*b[8]*b[5]*b[14]*b[3]
		-4*b[4]*b[1]*b[11]*b[14]+4*b[4]*b[13]*b[2]*b[11]-4*b[4]*b[13]*b[10]*b[3]
		+4*b[4]*b[9]*b[14]*b[3]-2*b[8]*a[5]*b[14]*a[3]+3*b[8]*a[5]*b[14]*b[3]
		-b[8]*a[5]*a[2]*a[15]+2*b[8]*a[5]*a[2]*b[15]+2*a[4]*a[9]*b[14]*b[3]
		+a[12]*a[5]*a[2]*b[11]+a[12]*a[5]*b[2]*a[11]-2*a[12]*a[5]*b[2]*b[11]
		+a[12]*a[9]*a[6]*b[3]+b[4]*a[9]*a[2]*a[15]-2*b[4]*a[9]*a[2]*b[15]
		-2*b[4]*a[9]*b[2]*a[15]+3*b[4]*a[9]*b[2]*b[15]+2*b[4]*a[9]*b[14]*a[3]
		-3*b[4]*a[9]*b[14]*b[3]+b[8]*a[13]*a[2]*a[7]-2*b[8]*a[13]*a[2]*b[7]
		-2*b[8]*a[13]*b[2]*a[7]+3*b[8]*a[13]*b[2]*b[7]-3*b[0]*b[13]*a[10]*b[7]
		-3*b[0]*b[13]*b[10]*a[7]+b[12]*a[5]*a[2]*a[11]-2*b[12]*a[5]*a[2]*b[11]
		-3*b[12]*b[1]*b[6]*a[11]-a[0]*a[9]*a[6]*b[15]+3*a[0]*b[13]*b[6]*b[11]
		-a[0]*a[9]*b[6]*a[15]-2*a[0]*b[13]*a[6]*b[11]-2*a[0]*b[13]*b[6]*a[11]
		-a[0]*b[13]*a[10]*a[7]-b[0]*a[13]*a[10]*a[7]+2*b[0]*a[13]*a[10]*b[7]
		-2*b[0]*a[5]*b[10]*a[15]-a[4]*a[9]*a[14]*b[3]-a[4]*a[9]*b[14]*a[3]
		+b[8]*a[5]*a[14]*a[3]-2*b[8]*a[5]*a[14]*b[3]-2*b[12]*a[5]*b[2]*a[11]
		+3*b[12]*a[5]*b[2]*b[11]+3*b[0]*a[13]*b[6]*b[11]-2*b[0]*a[13]*b[6]*a[11]
		-2*b[0]*a[13]*a[6]*b[11]+2*b[0]*a[9]*a[6]*b[15]-3*b[8]*a[1]*b[14]*b[7]
		+a[4]*b[1]*a[11]*a[14]-2*a[4]*b[1]*a[11]*b[14]-2*a[4]*b[1]*b[11]*a[14]
		-a[4]*b[1]*a[10]*a[15]+2*a[4]*b[1]*a[10]*b[15]-a[8]*a[5]*a[2]*b[15]
		-a[8]*a[5]*b[2]*a[15]+2*a[8]*a[5]*b[2]*b[15]-2*b[0]*b[5]*a[10]*a[15]
		+3*b[0]*a[5]*b[10]*b[15]-b[0]*a[5]*a[11]*a[14]+2*b[0]*a[5]*a[11]*b[14]
		+3*b[0]*b[5]*a[10]*b[15]-b[12]*a[9]*a[2]*a[7]-4*b[0]*b[5]*b[10]*b[15]
		-2*a[4]*b[13]*b[10]*a[3]-3*a[4]*b[9]*b[14]*b[3]+3*a[4]*b[9]*b[2]*b[15]
		+2*b[8]*b[1]*a[14]*a[7]-3*b[8]*b[1]*a[14]*b[7]-3*b[8]*b[1]*b[14]*a[7]
		+a[4]*a[13]*b[10]*a[3]-2*b[8]*b[13]*a[2]*a[7]+3*b[8]*b[13]*a[2]*b[7]
		+3*b[8]*b[13]*b[2]*a[7]+2*b[8]*b[13]*a[6]*a[3]-3*b[8]*b[13]*a[6]*b[3]
		-3*b[4]*b[9]*b[14]*a[3]+2*a[12]*b[5]*a[10]*b[3]+2*a[12]*b[5]*b[10]*a[3]
		+a[12]*b[5]*a[2]*a[11]-2*a[12]*b[5]*a[2]*b[11]-2*a[12]*b[5]*b[2]*a[11]
		+3*a[8]*b[1]*b[6]*b[15]-3*a[8]*b[1]*b[14]*b[7]+a[8]*a[1]*a[6]*b[15]
		-2*b[4]*b[9]*a[2]*a[15]+3*b[4]*b[9]*a[2]*b[15]+3*b[4]*b[9]*b[2]*a[15]
		+2*b[4]*a[13]*a[2]*b[11]+2*b[0]*a[5]*b[11]*a[14]-3*b[0]*a[5]*b[11]*b[14]
		+2*b[12]*b[5]*a[10]*a[3]-3*b[12]*b[5]*a[10]*b[3]-3*b[12]*b[5]*b[10]*a[3]
		+3*b[8]*b[5]*a[14]*b[3]+3*a[12]*b[9]*b[6]*b[3]-3*a[12]*b[1]*b[6]*b[11]
		+3*a[12]*b[1]*b[10]*b[7]+3*a[12]*b[5]*b[2]*b[11]-a[12]*a[1]*a[6]*b[11]
		+a[12]*a[1]*b[10]*a[7]-a[12]*b[5]*a[10]*a[3]+a[8]*a[5]*a[14]*b[3]
		+a[0]*b[5]*a[10]*a[15]+3*a[0]*b[5]*b[10]*b[15]-2*a[0]*a[5]*b[10]*b[15]
		+a[12]*b[1]*a[10]*a[7]-2*a[12]*b[1]*a[10]*b[7]-2*a[12]*b[1]*b[10]*a[7]
		-a[0]*a[5]*b[11]*a[14]+2*a[0]*a[5]*b[11]*b[14]-3*a[0]*b[5]*b[11]*b[14]
		-2*a[8]*b[5]*b[14]*a[3]-a[8]*b[5]*a[2]*a[15]+2*a[8]*b[5]*a[2]*b[15]
		+2*a[8]*b[5]*b[2]*a[15]+2*a[12]*b[9]*b[2]*a[7]+3*b[12]*a[1]*b[10]*b[7]
		-2*b[4]*a[1]*a[11]*b[14]-a[0]*a[5]*a[11]*b[14]-2*a[0]*b[5]*a[10]*b[15]
		-2*a[0]*b[5]*b[10]*a[15]-2*a[4]*a[13]*b[10]*b[3]-3*b[8]*b[13]*b[6]*a[3]
		-b[8]*a[1]*a[14]*a[7]+2*b[8]*a[1]*a[14]*b[7]+2*b[8]*a[1]*b[14]*a[7]
		+2*b[4]*a[9]*a[14]*b[3]-2*b[12]*a[9]*a[6]*b[3]-2*b[12]*b[5]*a[2]*a[11]
		+3*b[12]*b[5]*a[2]*b[11]+3*b[12]*b[5]*b[2]*a[11]+3*b[0]*b[13]*b[6]*a[11]
		+3*a[0]*b[9]*b[14]*b[7]-a[0]*b[5]*a[11]*a[14]+2*a[0]*b[5]*a[11]*b[14]
		-2*b[8]*b[1]*a[6]*a[15]+3*b[8]*b[1]*a[6]*b[15]+3*b[8]*b[1]*b[6]*a[15]
		-2*a[12]*b[9]*b[6]*a[3]-a[12]*b[9]*a[2]*a[7]+2*a[12]*b[9]*a[2]*b[7]
		-2*a[12]*a[1]*b[10]*b[7]-b[4]*a[1]*a[10]*a[15]+2*b[4]*a[1]*a[10]*b[15]
		-3*a[12]*b[5]*b[10]*b[3]-a[12]*a[5]*a[10]*b[3]+2*a[12]*a[9]*b[2]*b[7]
		-a[12]*a[5]*b[10]*a[3]+2*a[12]*a[5]*b[10]*b[3]-a[4]*a[13]*a[2]*b[11]
		-a[4]*a[13]*b[2]*a[11]+2*a[4]*a[13]*b[2]*b[11]-a[8]*b[1]*a[14]*a[7]
		+2*a[8]*b[1]*a[14]*b[7]+2*a[8]*b[1]*b[14]*a[7]-2*b[8]*b[5]*a[14]*a[3];

		double c = -6*b[12]*b[9]*b[2]*b[7]-a[4]*b[9]*b[14]*a[3]+a[4]*b[9]*a[2]*b[15]
		+a[4]*b[9]*b[2]*a[15]-b[8]*a[5]*b[2]*a[15]+3*b[8]*a[5]*b[2]*b[15]+b[8]*a[1]*a[6]*b[15]
		+a[8]*b[1]*a[6]*b[15]+a[8]*b[1]*b[6]*a[15]-6*b[12]*b[5]*b[10]*b[3]
		+b[0]*a[5]*a[10]*b[15]+3*b[0]*b[9]*b[6]*a[15]+3*b[0]*b[9]*a[6]*b[15]
		-b[0]*b[9]*a[6]*a[15]-3*b[0]*a[9]*b[14]*b[7]-b[0]*a[13]*b[10]*a[7]
		+6*b[12]*b[1]*b[10]*b[7]+6*b[12]*b[5]*b[2]*b[11]+a[8]*b[5]*a[14]*b[3]
		+b[0]*a[9]*a[14]*b[7]+b[0]*a[9]*b[14]*a[7]+3*a[4]*b[13]*b[2]*b[11]
		-3*a[4]*b[13]*b[10]*b[3]+a[4]*b[13]*a[10]*b[3]-b[4]*a[13]*b[2]*a[11]
		+3*b[4]*a[13]*b[2]*b[11]+b[12]*a[1]*b[10]*a[7]+3*a[12]*b[9]*b[2]*b[7]
		-b[12]*a[1]*b[6]*a[11]+3*b[12]*a[1]*b[6]*b[11]-a[0]*b[5]*b[11]*a[14]
		-6*b[12]*b[1]*b[6]*b[11]-b[12]*a[9]*a[2]*b[7]-3*b[12]*a[9]*b[6]*b[3]
		+6*b[12]*b[9]*b[6]*b[3]-b[0]*b[13]*a[10]*a[7]-3*b[0]*b[9]*a[14]*b[7]
		+b[0]*b[9]*a[14]*a[7]-3*b[0]*b[9]*b[14]*a[7]-a[8]*a[13]*b[6]*b[3]
		+a[12]*a[9]*b[6]*b[3]-3*b[12]*b[9]*a[6]*b[3]-3*b[12]*b[9]*b[6]*a[3]
		-a[12]*b[1]*a[6]*b[11]-a[12]*b[1]*b[6]*a[11]-6*b[8]*b[1]*b[14]*b[7]
		+6*b[8]*b[1]*b[6]*b[15]+a[12]*b[9]*a[6]*b[3]+b[4]*a[13]*a[10]*b[3]
		+b[4]*a[13]*b[10]*a[3]-3*b[4]*a[13]*b[10]*b[3]+b[0]*b[13]*a[6]*a[11]
		-3*b[0]*b[13]*a[6]*b[11]-6*b[0]*b[5]*b[11]*b[14]-a[12]*a[1]*b[6]*b[11]
		-b[8]*a[13]*a[6]*b[3]-b[8]*a[13]*b[6]*a[3]+3*b[8]*a[13]*b[6]*b[3]
		-a[4]*b[13]*a[2]*b[11]-a[4]*b[13]*b[2]*a[11]-b[12]*b[9]*a[2]*a[7]
		+3*b[12]*b[9]*a[2]*b[7]+3*b[12]*b[9]*b[2]*a[7]+b[12]*a[9]*b[6]*a[3]
		+b[12]*b[9]*a[6]*a[3]+3*b[0]*a[13]*b[10]*b[7]-6*b[4]*b[1]*b[10]*b[15]
		-b[4]*b[13]*a[2]*a[11]+3*b[4]*b[13]*a[2]*b[11]+3*b[4]*b[13]*b[2]*a[11]
		+a[8]*a[5]*b[14]*b[3]-b[4]*b[9]*a[14]*a[3]+3*b[4]*b[9]*a[14]*b[3]
		+b[4]*a[1]*b[11]*a[14]-3*b[4]*a[1]*b[11]*b[14]-3*b[0]*b[5]*b[10]*a[15]
		+b[4]*b[13]*a[10]*a[3]-3*b[4]*b[13]*a[10]*b[3]-3*b[4]*b[13]*b[10]*a[3]
		+a[0]*a[9]*b[14]*b[7]-b[0]*b[5]*a[11]*a[14]+3*b[0]*b[5]*a[11]*b[14]
		+3*b[0]*b[5]*b[11]*a[14]-a[4]*b[1]*b[10]*a[15]-a[0]*b[9]*b[6]*a[15]
		-a[0]*b[9]*a[6]*b[15]-b[12]*a[1]*a[6]*b[11]-a[0]*b[13]*a[10]*b[7]
		-a[0]*b[13]*b[10]*a[7]+a[0]*a[13]*b[6]*b[11]-a[4]*b[9]*a[14]*b[3]
		+a[4]*a[1]*b[11]*b[14]-b[4]*b[1]*a[10]*a[15]+3*b[4]*b[1]*a[10]*b[15]
		+3*b[4]*b[1]*b[10]*a[15]-6*b[8]*b[13]*b[6]*b[3]-a[4]*a[1]*b[10]*b[15]
		+a[0]*b[9]*a[14]*b[7]+a[0]*b[9]*b[14]*a[7]-3*b[8]*b[5]*b[14]*a[3]
		-b[8]*b[5]*a[2]*a[15]+3*b[8]*b[5]*a[2]*b[15]+3*b[8]*b[5]*b[2]*a[15]
		-a[8]*a[1]*b[14]*b[7]+6*b[8]*b[13]*b[2]*b[7]+3*a[0]*b[9]*b[6]*b[15]
		+6*b[0]*b[9]*b[14]*b[7]-6*b[0]*b[13]*b[10]*b[7]+6*b[0]*b[13]*b[6]*b[11]
		-6*b[0]*b[9]*b[6]*b[15]+6*b[4]*b[9]*b[2]*b[15]-b[0]*a[9]*b[6]*a[15]
		+3*b[0]*a[9]*b[6]*b[15]-a[0]*a[9]*b[6]*b[15]+b[4]*b[1]*a[11]*a[14]
		-3*b[4]*b[1]*a[11]*b[14]-3*b[4]*b[1]*b[11]*a[14]-b[12]*b[1]*a[6]*a[11]
		+3*b[12]*b[1]*a[6]*b[11]+b[12]*b[1]*a[10]*a[7]-3*b[12]*b[1]*a[10]*b[7]
		-3*b[12]*b[1]*b[10]*a[7]+a[8]*a[1]*b[6]*b[15]-b[12]*a[5]*a[10]*b[3]
		+3*b[12]*a[9]*b[2]*b[7]-b[12]*a[5]*b[10]*a[3]+3*b[12]*a[5]*b[10]*b[3]
		+b[12]*a[1]*a[10]*b[7]+a[8]*a[13]*b[2]*b[7]-3*a[8]*b[13]*b[2]*b[7]
		+3*a[8]*b[13]*b[6]*b[3]-a[0]*a[13]*b[10]*b[7]+3*a[0]*b[13]*b[10]*b[7]
		-b[12]*a[9]*b[2]*a[7]-a[8]*b[13]*a[6]*b[3]-a[8]*b[13]*b[6]*a[3]+a[8]*b[13]*a[2]*b[7]
		+a[8]*b[13]*b[2]*a[7]+a[4]*a[9]*b[2]*b[15]-b[4]*a[1]*b[10]*a[15]
		+3*b[4]*a[1]*b[10]*b[15]+3*a[4]*b[1]*b[10]*b[15]-3*a[4]*b[1]*b[11]*b[14]
		-6*b[8]*b[5]*b[2]*b[15]+b[8]*a[1]*b[6]*a[15]-3*b[8]*a[1]*b[6]*b[15]
		-3*a[8]*b[5]*b[14]*b[3]+3*a[8]*b[5]*b[2]*b[15]+6*b[8]*b[5]*b[14]*b[3]
		+6*b[4]*b[1]*b[11]*b[14]-6*b[4]*b[13]*b[2]*b[11]+6*b[4]*b[13]*b[10]*b[3]
		-6*b[4]*b[9]*b[14]*b[3]+b[8]*a[5]*b[14]*a[3]-3*b[8]*a[5]*b[14]*b[3]
		-b[8]*a[5]*a[2]*b[15]-a[4]*a[9]*b[14]*b[3]+a[12]*a[5]*b[2]*b[11]+b[4]*a[9]*a[2]*b[15]
		+b[4]*a[9]*b[2]*a[15]-3*b[4]*a[9]*b[2]*b[15]-b[4]*a[9]*b[14]*a[3]
		+3*b[4]*a[9]*b[14]*b[3]+b[8]*a[13]*a[2]*b[7]+b[8]*a[13]*b[2]*a[7]
		-3*b[8]*a[13]*b[2]*b[7]+3*b[0]*b[13]*a[10]*b[7]+3*b[0]*b[13]*b[10]*a[7]
		+b[12]*a[5]*a[2]*b[11]+3*b[12]*b[1]*b[6]*a[11]-3*a[0]*b[13]*b[6]*b[11]
		+a[0]*b[13]*a[6]*b[11]+a[0]*b[13]*b[6]*a[11]-b[0]*a[13]*a[10]*b[7]
		+b[0]*a[5]*b[10]*a[15]+b[8]*a[5]*a[14]*b[3]+b[12]*a[5]*b[2]*a[11]
		-3*b[12]*a[5]*b[2]*b[11]-3*b[0]*a[13]*b[6]*b[11]+b[0]*a[13]*b[6]*a[11]
		+b[0]*a[13]*a[6]*b[11]-b[0]*a[9]*a[6]*b[15]+3*b[8]*a[1]*b[14]*b[7]
		+a[4]*b[1]*a[11]*b[14]+a[4]*b[1]*b[11]*a[14]-a[4]*b[1]*a[10]*b[15]
		-a[8]*a[5]*b[2]*b[15]+b[0]*b[5]*a[10]*a[15]-3*b[0]*a[5]*b[10]*b[15]
		-b[0]*a[5]*a[11]*b[14]-3*b[0]*b[5]*a[10]*b[15]+6*b[0]*b[5]*b[10]*b[15]
		+a[4]*b[13]*b[10]*a[3]+3*a[4]*b[9]*b[14]*b[3]-3*a[4]*b[9]*b[2]*b[15]
		-b[8]*b[1]*a[14]*a[7]+3*b[8]*b[1]*a[14]*b[7]+3*b[8]*b[1]*b[14]*a[7]
		+b[8]*b[13]*a[2]*a[7]-3*b[8]*b[13]*a[2]*b[7]-3*b[8]*b[13]*b[2]*a[7]
		-b[8]*b[13]*a[6]*a[3]+3*b[8]*b[13]*a[6]*b[3]+3*b[4]*b[9]*b[14]*a[3]
		-a[12]*b[5]*a[10]*b[3]-a[12]*b[5]*b[10]*a[3]+a[12]*b[5]*a[2]*b[11]
		+a[12]*b[5]*b[2]*a[11]-3*a[8]*b[1]*b[6]*b[15]+3*a[8]*b[1]*b[14]*b[7]
		+b[4]*b[9]*a[2]*a[15]-3*b[4]*b[9]*a[2]*b[15]-3*b[4]*b[9]*b[2]*a[15]
		-b[4]*a[13]*a[2]*b[11]-b[0]*a[5]*b[11]*a[14]+3*b[0]*a[5]*b[11]*b[14]
		-b[12]*b[5]*a[10]*a[3]+3*b[12]*b[5]*a[10]*b[3]+3*b[12]*b[5]*b[10]*a[3]
		-3*b[8]*b[5]*a[14]*b[3]-3*a[12]*b[9]*b[6]*b[3]+3*a[12]*b[1]*b[6]*b[11]
		-3*a[12]*b[1]*b[10]*b[7]-3*a[12]*b[5]*b[2]*b[11]-3*a[0]*b[5]*b[10]*b[15]
		+a[0]*a[5]*b[10]*b[15]+a[12]*b[1]*a[10]*b[7]+a[12]*b[1]*b[10]*a[7]
		-a[0]*a[5]*b[11]*b[14]+3*a[0]*b[5]*b[11]*b[14]+a[8]*b[5]*b[14]*a[3]
		-a[8]*b[5]*a[2]*b[15]-a[8]*b[5]*b[2]*a[15]-a[12]*b[9]*b[2]*a[7]
		-3*b[12]*a[1]*b[10]*b[7]+b[4]*a[1]*a[11]*b[14]+a[0]*b[5]*a[10]*b[15]
		+a[0]*b[5]*b[10]*a[15]+a[4]*a[13]*b[10]*b[3]+3*b[8]*b[13]*b[6]*a[3]
		-b[8]*a[1]*a[14]*b[7]-b[8]*a[1]*b[14]*a[7]-b[4]*a[9]*a[14]*b[3]+b[12]*a[9]*a[6]*b[3]
		+b[12]*b[5]*a[2]*a[11]-3*b[12]*b[5]*a[2]*b[11]-3*b[12]*b[5]*b[2]*a[11]
		-3*b[0]*b[13]*b[6]*a[11]-3*a[0]*b[9]*b[14]*b[7]-a[0]*b[5]*a[11]*b[14]
		+b[8]*b[1]*a[6]*a[15]-3*b[8]*b[1]*a[6]*b[15]-3*b[8]*b[1]*b[6]*a[15]
		+a[12]*b[9]*b[6]*a[3]-a[12]*b[9]*a[2]*b[7]+a[12]*a[1]*b[10]*b[7]-b[4]*a[1]*a[10]*b[15]
		+3*a[12]*b[5]*b[10]*b[3]-a[12]*a[9]*b[2]*b[7]-a[12]*a[5]*b[10]*b[3]
		-a[4]*a[13]*b[2]*b[11]-a[8]*b[1]*a[14]*b[7]-a[8]*b[1]*b[14]*a[7]+b[8]*b[5]*a[14]*a[3];

		double d = 4*b[12]*b[9]*b[2]*b[7]-b[8]*a[5]*b[2]*b[15]+4*b[12]*b[5]*b[10]*b[3]
		-b[0]*b[9]*b[6]*a[15]-b[0]*b[9]*a[6]*b[15]+b[0]*a[9]*b[14]*b[7]-4*b[12]*b[1]*b[10]*b[7]
		-4*b[12]*b[5]*b[2]*b[11]-a[4]*b[13]*b[2]*b[11]+a[4]*b[13]*b[10]*b[3]
		-b[4]*a[13]*b[2]*b[11]-a[12]*b[9]*b[2]*b[7]-b[12]*a[1]*b[6]*b[11]
		+4*b[12]*b[1]*b[6]*b[11]+b[12]*a[9]*b[6]*b[3]-4*b[12]*b[9]*b[6]*b[3]
		+b[0]*b[9]*a[14]*b[7]+b[0]*b[9]*b[14]*a[7]+b[12]*b[9]*a[6]*b[3]+b[12]*b[9]*b[6]*a[3]
		+4*b[8]*b[1]*b[14]*b[7]-4*b[8]*b[1]*b[6]*b[15]+b[4]*a[13]*b[10]*b[3]
		+b[0]*b[13]*a[6]*b[11]+4*b[0]*b[5]*b[11]*b[14]-b[8]*a[13]*b[6]*b[3]
		-b[12]*b[9]*a[2]*b[7]-b[12]*b[9]*b[2]*a[7]-b[0]*a[13]*b[10]*b[7]
		+4*b[4]*b[1]*b[10]*b[15]-b[4]*b[13]*a[2]*b[11]-b[4]*b[13]*b[2]*a[11]
		-b[4]*b[9]*a[14]*b[3]+b[4]*a[1]*b[11]*b[14]+b[0]*b[5]*b[10]*a[15]
		+b[4]*b[13]*a[10]*b[3]+b[4]*b[13]*b[10]*a[3]-b[0]*b[5]*a[11]*b[14]
		-b[0]*b[5]*b[11]*a[14]-b[4]*b[1]*a[10]*b[15]-b[4]*b[1]*b[10]*a[15]
		+4*b[8]*b[13]*b[6]*b[3]+b[8]*b[5]*b[14]*a[3]-b[8]*b[5]*a[2]*b[15]
		-b[8]*b[5]*b[2]*a[15]-4*b[8]*b[13]*b[2]*b[7]-a[0]*b[9]*b[6]*b[15]
		-4*b[0]*b[9]*b[14]*b[7]+4*b[0]*b[13]*b[10]*b[7]-4*b[0]*b[13]*b[6]*b[11]
		+4*b[0]*b[9]*b[6]*b[15]-4*b[4]*b[9]*b[2]*b[15]-b[0]*a[9]*b[6]*b[15]
		+b[4]*b[1]*a[11]*b[14]+b[4]*b[1]*b[11]*a[14]-b[12]*b[1]*a[6]*b[11]
		+b[12]*b[1]*a[10]*b[7]+b[12]*b[1]*b[10]*a[7]-b[12]*a[9]*b[2]*b[7]
		-b[12]*a[5]*b[10]*b[3]+a[8]*b[13]*b[2]*b[7]-a[8]*b[13]*b[6]*b[3]-a[0]*b[13]*b[10]*b[7]
		-b[4]*a[1]*b[10]*b[15]-a[4]*b[1]*b[10]*b[15]+a[4]*b[1]*b[11]*b[14]
		+4*b[8]*b[5]*b[2]*b[15]+b[8]*a[1]*b[6]*b[15]+a[8]*b[5]*b[14]*b[3]
		-a[8]*b[5]*b[2]*b[15]-4*b[8]*b[5]*b[14]*b[3]-4*b[4]*b[1]*b[11]*b[14]
		+4*b[4]*b[13]*b[2]*b[11]-4*b[4]*b[13]*b[10]*b[3]+4*b[4]*b[9]*b[14]*b[3]
		+b[8]*a[5]*b[14]*b[3]+b[4]*a[9]*b[2]*b[15]-b[4]*a[9]*b[14]*b[3]+b[8]*a[13]*b[2]*b[7]
		-b[0]*b[13]*a[10]*b[7]-b[0]*b[13]*b[10]*a[7]-b[12]*b[1]*b[6]*a[11]
		+a[0]*b[13]*b[6]*b[11]+b[12]*a[5]*b[2]*b[11]+b[0]*a[13]*b[6]*b[11]-b[8]*a[1]*b[14]*b[7]
		+b[0]*a[5]*b[10]*b[15]+b[0]*b[5]*a[10]*b[15]-4*b[0]*b[5]*b[10]*b[15]
		-a[4]*b[9]*b[14]*b[3]+a[4]*b[9]*b[2]*b[15]-b[8]*b[1]*a[14]*b[7]-b[8]*b[1]*b[14]*a[7]
		+b[8]*b[13]*a[2]*b[7]+b[8]*b[13]*b[2]*a[7]-b[8]*b[13]*a[6]*b[3]-b[4]*b[9]*b[14]*a[3]
		+a[8]*b[1]*b[6]*b[15]-a[8]*b[1]*b[14]*b[7]+b[4]*b[9]*a[2]*b[15]+b[4]*b[9]*b[2]*a[15]
		-b[0]*a[5]*b[11]*b[14]-b[12]*b[5]*a[10]*b[3]-b[12]*b[5]*b[10]*a[3]+b[8]*b[5]*a[14]*b[3]
		+a[12]*b[9]*b[6]*b[3]-a[12]*b[1]*b[6]*b[11]+a[12]*b[1]*b[10]*b[7]+a[12]*b[5]*b[2]*b[11]
		+a[0]*b[5]*b[10]*b[15]-a[0]*b[5]*b[11]*b[14]+b[12]*a[1]*b[10]*b[7]-b[8]*b[13]*b[6]*a[3]
		+b[12]*b[5]*a[2]*b[11]+b[12]*b[5]*b[2]*a[11]+b[0]*b[13]*b[6]*a[11]+a[0]*b[9]*b[14]*b[7]
		+b[8]*b[1]*a[6]*b[15]+b[8]*b[1]*b[6]*a[15]-a[12]*b[5]*b[10]*b[3];

		double e = -b[12]*b[9]*b[2]*b[7]-b[12]*b[5]*b[10]*b[3]+b[12]*b[1]*b[10]*b[7]
		+b[12]*b[5]*b[2]*b[11]-b[12]*b[1]*b[6]*b[11]+b[12]*b[9]*b[6]*b[3]-b[8]*b[1]*b[14]*b[7]
		+b[8]*b[1]*b[6]*b[15]-b[0]*b[5]*b[11]*b[14]-b[4]*b[1]*b[10]*b[15]-b[8]*b[13]*b[6]*b[3]
		+b[8]*b[13]*b[2]*b[7]+b[0]*b[9]*b[14]*b[7]-b[0]*b[13]*b[10]*b[7]+b[0]*b[13]*b[6]*b[11]
		-b[0]*b[9]*b[6]*b[15]+b[4]*b[9]*b[2]*b[15]-b[8]*b[5]*b[2]*b[15]+b[8]*b[5]*b[14]*b[3]
		+b[4]*b[1]*b[11]*b[14]-b[4]*b[13]*b[2]*b[11]+b[4]*b[13]*b[10]*b[3]-b[4]*b[9]*b[14]*b[3]
		+b[0]*b[5]*b[10]*b[15];

		a1 *= 100000;
		b1 *= 100000;
		c *= 100000;
		d *= 100000;
		e *= 100000;
		std::cout << std::endl << std::endl << "a = " << a1 << "  b = " << b1 << "  c = "
			<< c << "  d = " << d << "  e = " << e;

		const Eq4Degree eq(a1, b1, c, d, e, 4);
		std::vector<double> sol(4);
		SolveEq solveEq;
		int nRoots = solveEq.Apply(sol, &eq, 0.000000001);

		std::cout << std::endl << std::endl << "nRoots = " << nRoots << std::endl;

		for (int i = 0; i < nRoots; i++)
		{
			std::cout << sol[i] << "  ";
			double t2 = sol[i] * sol[i];
			double t3 = t2 * sol[i];
			double t4 = t3 * sol[i];
			std::cout << (a1 * t4 + b1 * t3 + c * t2 + d * sol[i] + e) << std::endl;
		}

		// save nRoots solutions for Q
		for (int i = 0; i < nRoots; i++)
		{
			for (int j = 0; j < 16; j++)
			{
				out_Q(i, j) = sol[i] * a[j] + (1 - sol[i]) * b[j];
			}
		}
		
		// check if det == 0
		std::cout << std::endl << std::endl << "Check if det == 0";
		for (int i = 0; i < nRoots; i++)
		{
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 4> U_;
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 4> V_;
			Vtl::NumAlg::Lin::ublas::c_vector<double, 4> W_;
			Vtl::NumAlg::Lin::ublas::c_vector<double, 4> RV1_;

			for (int l = 0; l < 4; l++)
			{
				for (int m = 0; m < 4; m++)
				{
					U_(l, m) = out_Q(i, l * 4 + m);
				}
			}
			
			Vtl::NumAlg::Lin::svd::decomp(U_, W_, V_, RV1_);

			std::cout  << std::endl << "W   ";
			for (int i = 0; i < 4; i++)
			{
				std::cout << W_(i) << "  ";
			}
		}

		// check if obtained Q is solution of the system
		std::cout << std::endl << std::endl
			<< "check if obtained Q is solution of the system";
		for (int i = 0; i < nRoots; i++)
		{
			std::cout << std::endl << "i = " << i << std::endl;
			for (int k = 0; k < 8; k++)
			{
				double c = 0;
				for (int l = 0; l < 16; l++)
				{
					c += in_system(k, l) * out_Q(i, l);
				}
				std::cout << c << "  ";
			}
		}
		return nRoots;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	int AbsoluteDualQuadric::Get
		                      (Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 16> &out_Q,
			                     const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> &in_P,
				                   const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> &in_P1)
	{
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 8, 10> system;
		CreateSystem(system, in_P, in_P1);

		std::cout << std::endl << std::endl << "system" << std::endl;
		T_MatrixOperations::Print(system);

		int n = SolveDet0(out_Q, system);
		return n;
	}
}