
#include "DrawEpipolarLines.h"

namespace Basics3D
{
	///////////////////////////////////////////////////////////////////////////////////////
	C_DrawEpipolarLines::C_DrawEpipolarLines(const AImageMem8 *in_src0,
			                                     const AImageMem8 *in_src1)
	{
		m_src0 = in_src0;
		m_src1 = in_src1;
		m_dst0 = dynamic_cast<AImageMemRgb8 *>(::AImageCreate(0, AImage::IMGC_RGB, 
																           m_src0->Width(), m_src0->Height(), 8));
		m_dst1 = dynamic_cast<AImageMemRgb8 *>(::AImageCreate(0, AImage::IMGC_RGB, 
																           m_src1->Width(), m_src1->Height(), 8));

		if (m_dst0 == 0)
		{
			throw "m_dst0 == 0";
			return;
		}
		if (m_dst1 == 0)
		{
			throw "m_dst1 == 0";
			return;
		}

		AImageMem8::Pixel *p_s0 = (*m_src0)[0];
		AImageMem8::Pixel *p_d00 = (*m_dst0->GetPlane8(0))[0];
		AImageMem8::Pixel *p_d01 = (*m_dst0->GetPlane8(1))[0];
		AImageMem8::Pixel *p_d02 = (*m_dst0->GetPlane8(2))[0];
		
		AImageMem8::Pixel *p_s1 = (*m_src1)[0];
		AImageMem8::Pixel *p_d10 = (*m_dst1->GetPlane8(0))[0];
		AImageMem8::Pixel *p_d11 = (*m_dst1->GetPlane8(1))[0];
		AImageMem8::Pixel *p_d12 = (*m_dst1->GetPlane8(2))[0];

		// copy src image to dst
		for (int y = 0; y < m_src0->Height(); y++)
		{
			for (int x = 0; x < m_src0->Width(); x++)
			{
				*p_d00 = *p_d01 = *p_d02 = (*p_s0);
				p_s0++;
				p_d00++;
				p_d01++;
				p_d02++;

				*p_d10 = *p_d11 = *p_d12 = (*p_s1);
				p_s1++;
				p_d10++;
				p_d11++;
				p_d12++;
			}
		}
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_DrawEpipolarLines::GetRandColor(int *r,
			                                   int *g,
												                 int *b)
	{
		double coef = 255.0 / (pow(2.0, 15) - 1);
		*r = int(rand() * coef);
		*g = int(rand() * coef);
		*b = int(rand() * coef);
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_DrawEpipolarLines::Ends(const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> 
		                                                                            &in_Line,
		                             const int in_h,
		                             const int in_w,
			                           int *out_x1,
					                       int *out_y1,
                                 int *out_x2,
						                     int *out_y2)
	{
	  if (abs(in_Line(0, 0)) > 0.00000001 && abs(in_Line(1, 0)) > 0.00000001)
		{
			// y = kx + b
			double k = - in_Line(0, 0) / in_Line(1, 0);
			double b = - in_Line(2, 0) / in_Line(1, 0);

			/*std::cout << std::endl << " b = " << b << "   k = " << k << std::endl;*/

			(*out_x1) = 0;
			(*out_y1) = int( -in_Line(2, 0) / in_Line(1, 0));
			(*out_x2) = in_w;
			(*out_y2) = int(( -in_Line(0, 0) * in_w - in_Line(2, 0)) / in_Line(1, 0));

			/*std::cout << std::endl << "" << "   x1 = " << (*out_x1) << "  y1 = " << (*out_y1)
				<< " | x2 = " << (*out_x2) << "  y2 = " << (*out_y2) << "\n";*/
			
		  if (/**/in_h - (*out_y1) > in_h)
			{
				//std::cout << std::endl << "V";
				(*out_y1) = in_h;
				(*out_x1) = int(- (in_Line(1, 0) * in_h + in_Line(2, 0)) / in_Line(0, 0));
			}

			if (/**/in_h - (*out_y1) < 0)
			{
				//std::cout << std::endl << "VI";
				(*out_y1) = 0;
				(*out_x1) = int(- in_Line(2, 0) / in_Line(0, 0));
			}

			if (/**/in_h - (*out_y2) > in_h)
			{
				//std::cout << std::endl << "VII";
				(*out_y2) = in_h;
				(*out_x2) = int(- (in_Line(1, 0) * in_h + in_Line(2, 0)) / in_Line(0, 0));
			}

			if (/**/in_h - (*out_y2) < 0)
			{
				//std::cout << std::endl << "VIII";
				(*out_y2) = 0;
				(*out_x2) = int(- in_Line(2, 0) / in_Line(0, 0));
			}

			if ((*out_x1) > in_w)
			{
				//std::cout << std::endl << "I";
				(*out_x1) = in_w;
				(*out_y1) = int(( -in_Line(0, 0) * in_w - in_Line(2, 0)) / in_Line(1, 0));
			}

			if ((*out_x1) < 0)
			{
				//std::cout << std::endl << "II";
				(*out_x1) = 0;
				(*out_y1) = int( -in_Line(2, 0) / in_Line(1, 0));
			}

			if ((*out_x2) > in_w)
			{
				//std::cout << std::endl << "III";
				(*out_x2) = in_w;
				(*out_y2) = int(( -in_Line(0, 0) * in_w - in_Line(2, 0)) / in_Line(1, 0));
			}

			if ((*out_x2) < 0)
			{
				//std::cout << std::endl << "IV";
				(*out_x2) = 0;
				(*out_y2) = int( -in_Line(2, 0) / in_Line(1, 0));
			}
		}
		else
		{
			if (abs(in_Line(0, 0)) <= 0.00000001)
			{
				//std::cout << std::endl << "IX";

				int q = int(-in_Line(2, 0) / in_Line(1, 0));

				if (q >= 0 && q <= in_h)
				{
					(*out_x1) = 0;
					(*out_y1) = q;
					(*out_x2) = in_w;
					(*out_y2) = q;
				}
			}
			if (abs(in_Line(1, 0)) <= 0.00000001)
			{
				//std::cout << std::endl << "X";

				int q = int(-in_Line(2, 0)/in_Line(0, 0));

				if (q >= 0 && q <= in_w)
				{
					(*out_x1) = q;
					(*out_y1) = 0;
					(*out_x2) = q;
					(*out_y2) = in_h;
				}
			}
		}

	/*	std::cout << std::endl << "res" << "   x1 = " << (*out_x1) << "  y1 = " << (*out_y1)
				<< " | x2 = " << (*out_x2) << "  y2 = " << (*out_y2) << "\n";*/
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_DrawEpipolarLines::DrawLine(AImageMem8 *out_img,
		                                 const int in_nls,
			                               const int in_col,
									                   const int in_x1,
									                   const int in_y1,
									                   const int in_x2,
									                   const int in_y2)
	{
		typedef Vtl::ImAlg::Bres::Draw<AImageMem8::Pixel> DrawFunctor;
		typedef Vtl::ImAlg::Bres::SetPixelGray<AImageMem8::Pixel, DrawFunctor> SetPixelFunctor;

		AImageMem8::Pixel color_pen = AImageMem8::Pixel(in_col);
		DrawFunctor draw_functor(color_pen);
		SetPixelFunctor set_pix((*out_img)[0], in_nls, draw_functor);
		Vtl::ImAlg::Bres::Line<SetPixelFunctor> line(set_pix);

		set_pix.Set0(in_x1, in_y1);
		line.Draw(in_x2 - in_x1, in_y2 - in_y1);
	}
	
	///////////////////////////////////////////////////////////////////////////////////////
	void C_DrawEpipolarLines::Apply(const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> 
		                                                                          &in_matrix,
																  const T_Points2DSet<C_Point2D> &in_points,
																	int i)
	{
		int n = in_points.Size();

		if (i > n || i == -1)
		{
			i = n;
		}

		int shift = n / i;
		double tg = (double)m_src0->Height() / m_src0->Width();

		int nls0 = (m_dst0->GetPlane8(0))->NextLineShift();
		int nls1 = (m_dst1->GetPlane8(0))->NextLineShift();
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> MT =
			                                              boost::numeric::ublas::trans(in_matrix);
		int h = m_src0->Height() - 1;
		int w = m_src0->Width() - 1;

		for (int j = 1; j < n; j += shift)
		{
			int r, g, b;
			GetRandColor(&r, &g, &b);
			
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> Point0(3, 1);
			in_points[2 * j].Matrix(&Point0);
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> Line0(3, 1);
			Line0 = boost::numeric::ublas::prod(in_matrix, Point0);
					
			int x1, y1, x2, y2;
			Ends(Line0, h, w, &x1, &y1, &x2, &y2);

			if ( x1 >= 0 && x1 <= w && x2 >= 0 && x2 <= w &&
				   /**/h - y1 >= 0 && /**/h - y1 <= h && /**/h - y2 >= 0 && /**/h - y2 <= h)
			{
				DrawLine(m_dst1->GetPlane8(0), nls1, r, x1, /**/h - y1, x2, /**/h - y2);
				DrawLine(m_dst1->GetPlane8(1), nls1, g, x1, /**/h - y1, x2, /**/h - y2);
				DrawLine(m_dst1->GetPlane8(2), nls1, b, x1, /**/h - y1, x2, /**/h - y2);
			}
						
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> Point1(3, 1);
			in_points[2 * j + 1].Matrix(&Point1);
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> Line1(3, 1);
			Line1 = boost::numeric::ublas::prod(MT, Point1);
						
			Ends(Line1, h, w, &x1, &y1, &x2, &y2);
			if ( x1 >= 0 && x1 <= w && x2 >= 0 && x2 <= w &&
				   /**/h - y1 >= 0 && /**/h - y1 <= h && /**/h - y2 >= 0 && /**/h - y2 <= h)
			{
				DrawLine(m_dst0->GetPlane8(0), nls0, r, x1, /**/h - y1, x2, /**/h - y2);
				DrawLine(m_dst0->GetPlane8(1), nls0, g, x1, /**/h - y1, x2, /**/h - y2);
				DrawLine(m_dst0->GetPlane8(2), nls0, b, x1, /**/h - y1, x2, /**/h - y2);
			}
		}
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	void C_DrawEpipolarLines::Save(const char* in_outputImageName,
		                             const int in_i)
	{
		if (m_dst0 == 0)
		{
			std::cout << "Image was not created\n";
			return;
		}

		std::ostringstream os0;
		os0 << in_outputImageName << in_i << ".bmp";
		std::string filename0 = os0.str();

		::AImageSave(m_dst0, filename0.c_str());

		if (m_dst1 == 0)
		{
			std::cout << "Image was not created\n";
			return;
		}

		std::ostringstream os1;
		os1 << in_outputImageName << in_i + 1 << ".bmp";
		std::string filename1 = os1.str();

		::AImageSave(m_dst1, filename1.c_str());
	}
}