
#include "Func.h"

namespace Basics3D
{
	Eq4Degree::Eq4Degree(const double in_a,
			                 const double in_b,
							         const double in_c,
							         const double in_d,
							         const double in_e,
											 const int in_n)
											 :m_a(in_a), m_b(in_b), m_c(in_c), m_d(in_d), m_e(in_e), m_n(in_n)
	{}

	///////////////////////////////////////////////////////////////////////////////////////
	int Eq4Degree::N() const
	{
		return m_n;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	double Eq4Degree::Func(const double in_t) const
	{
		double t2 = in_t * in_t;
		double t3 = t2 * in_t;
		double t4 = t3 * in_t;
		double res = m_a * t4 + m_b * t3 + m_c * t2 + m_d * in_t + m_e;
		return res;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	double Eq4Degree::Deriv(const double in_t) const
	{
		double t2 = in_t * in_t;
		double t3 = t2 * in_t;
		double res = 4.0 * m_a * t3 + 3.0 * m_b * t2 + 2.0 * m_c * in_t + m_d;
		return res;
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	g::g(const double in_a,
			 const double in_b,
			 const double in_c,
			 const double in_d,
			 const double in_f,
			 const double in_f1,
			 const int in_n)
			 :a(in_a), b(in_b), c(in_c), d(in_d), f(in_f), f1(in_f1), n(in_n){}

	/////////////////////////////////////////////////////////////////////////////////////////
	int g::N()
	{
		return n;
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	double g::Func(const double in_t) const
	{
		double A = (a * in_t + b) * (a * in_t + b) + f1 * f1 * (c * in_t + d) * (c * in_t + d);
		double B = 1 + f * f * in_t * in_t;
		double res = in_t * A * A - (a * d - b * c) * B * B * (a * in_t + b) * (c * in_t + d);
		return res;
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	double g::Deriv(const double in_t) const
	{
		double atb = a * in_t + b;
		double ctd = c * in_t + d;
		double A = atb * atb + f1 * f1 * ctd * ctd;
		double B = 1 + f * f * in_t * in_t;

		double res = A * A + 2 * in_t * A * (2 * atb * a + 2 * f1 * f1 * ctd * c)
							 - 4 * (a * d - b * c) * (1 + f * f * in_t * in_t) * atb * ctd * f * f * in_t
							 - (a * d - b * c) * B * B * a * ctd - (a * d - b * c) * B * B * atb * c;
		
		return res;
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	test::test(const double in_a,
						 const double in_b)
						 : a(in_a), b(in_b) {}

	/////////////////////////////////////////////////////////////////////////////////////////
	double test::Func(const double in_t) const
	{
		return ((in_t - a) * (in_t - b));
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	double test::Deriv(const double in_t) const
	{
		return (2 * in_t - (a + b));
	}
}