
#include "Basics3D/Points2DParamSet.h"

namespace Basics3D
{
	/////////////////////////////////////////////////////////////////////////////////////////
	C_Points2DParamSet::C_Points2DParamSet(int in_groupSize)
	{
		this->m_groupSize = in_groupSize;

		m_MeanX.resize(m_groupSize);
		m_MeanY.resize(m_groupSize);
		m_MeanDev.resize(m_groupSize);

		ClearMeans();
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	int C_Points2DParamSet::Size() const
	{
		return m_Points.size();
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	void C_Points2DParamSet::ClearMeans()
	{
		for (int i = 0; i < m_groupSize; i++)
		{
			m_MeanX[i] = m_MeanY[i] = m_MeanDev[i] = -1;
		}
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	double C_Points2DParamSet::MeanX(int i)
	{
			if (i >= m_groupSize)
			{
				throw "\nIndex is to be less than group size";
				return -1;
			}

			if (m_MeanX[i] == -1)
			{
				m_MeanX[i] = 0;

				for (int j = 0; j < Size(); j++)
				{
					m_MeanX[i] += m_Points[m_groupSize * j + i].X();
				}

				m_MeanX[i] /= Size();
			}

			return m_MeanX[i];
		}

	/////////////////////////////////////////////////////////////////////////////////////////
	double C_Points2DParamSet::MeanY(int i)
	{
			if (i >= m_groupSize)
			{
				throw "\nIndex is to be less than group size";
				return -1;
			}

			if (m_MeanY[i] == -1)
			{
				m_MeanY[i] = 0;

				for (int j = 0; j < Size(); j++)
				{
					m_MeanY[i] += m_Points[m_groupSize * j + i].Y();
				}

				m_MeanY[i] /= Size();
			}

			return m_MeanY[i];
		}

	/////////////////////////////////////////////////////////////////////////////////////////
	double C_Points2DParamSet::MeanDev(int i)
	{
		if (i >= m_groupSize)
		{
			throw "\nIndex is to be less than group size";
			return -1;
		}

		if (m_MeanDev[i] < 0)
		{
			m_MeanDev[i] = 0;

			for (int j = 0; j < Size(); j++)
			{
				C_Point2D P = m_Points[m_groupSize * j + i];
				m_MeanDev[i] += sqrt((P.X() - MeanX(i)) * (P.X() - MeanX(i)) +
														 (P.Y() - MeanY(i)) * (P.Y() - MeanY(i)));
			}

			m_MeanDev[i] /= Size();
		}
		return m_MeanDev[i];
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	void C_Points2DParamSet::Load(const char* in_filename,
		                            const int in_imHeight)
	{
		std::ifstream inFile;
		inFile.open(in_filename, std::ios::in);

		if (!inFile)
		{
			std::cerr << "Can't open " << in_filename << "\n";
		}
		
		std::vector<float> param;
		param.resize(C_Point2DParam::ParamNumb());
		double x, y, sigma, angle;
		int c;
		int i = -1;
		
		while(1)
		{
			i++;

			inFile >> x;
			inFile >> y;
			inFile >> sigma;
			inFile >> angle;

			for (int j = 0; j < C_Point2DParam::ParamNumb(); j++)
			{
				inFile >> param[j];
			}

			m_Points.push_back(C_Point2DParam(x, in_imHeight - y, i, sigma, angle, param));

			c = inFile.peek();

			if (c == EOF)
			{
				break;
			}
		}

		inFile.close();

		//m_MeanX /= m_Points.size();
		//m_MeanY /= m_Points.size();
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	const C_Point2DParam *C_Points2DParamSet::operator[](int i) const
	{
		return &(m_Points[i]);
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	void C_Points2DParamSet::Disable(T_Points2DSet<C_Point2D> &in_Set,
		                               int j)
	{
		for (int i = 0; i < in_Set.Size(); i++)
		{
			int numb = in_Set[2 * i + j].Numb();
			if (numb < 0)
			{
				continue;
			}
			 m_Points[numb].SetNumb(-1);
		}
	}

	////////////////////////////////////////////////////////////////////////////////////////
	void C_Points2DParamSet::EnableAll()
	{
		for (int i = 0; i < Size(); i++)
		{
			m_Points[i].SetNumb(i);
		}
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	// Shifts and scales points to achieve MeanX = 0, MeanY = 0, MeanDev = sqrt(2)
	void C_Points2DParamSet::Normalize()
	{
		std::vector<double> coef(m_groupSize);
			for (int i = 0; i < m_groupSize; i++)
			{
				coef[i] = sqrt(2.0) / MeanDev(i);
			}

			for (int i = 0; i < Size(); i++)
			{
				for (int j = 0; j < m_groupSize; j++)
				{
					m_Points[m_groupSize * i + j].Shift(MeanX(j), MeanY(j));
					m_Points[m_groupSize * i + j].Scale(coef[j]);
				}
			}
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	//C_Points2DParamSet::~C_Points2DParamSet()
	//{
	//	//m_Points.~vector();
	//}
}