#ifndef PROJECTIVE_MATRIX_H_
#define PROJECTIVE_MATRIX_H_

// library includings
#include "num_linalg.h"

// own includings
#include "Epipole.h"
#include "TransformMatrices.h"
#include "MatrixOperations.h"
#include "Sort.h"

namespace Basics3D
{
	// Computes projective (camera) matrices in canonical form from
	// fundamental matrix according with Result 9.14 (Hartley, p.274)
  class C_ProjectiveMatrix
	{
	private:
		static void GetW(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_W);

		static void SortColumns(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_M,
			                      const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_M,
														const int k[3]);

		static void SortSV(int k[3]);

		// Hartley res. 9.19 (p.277)
		// computes one of four variants of camera matrix out_P retrieved from ess.m.
		// in_UWV - U * W * V^T or U * W^T * V^T
		// k -  number of the smallest of singular values of ess.m.
		// sgn = 1 / -1
		static void GetP(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> &out_P,
			               const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_UWV,
									   const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_U,
									   const int in_k,
									   const int in_sgn);

	public:
	  static void Left(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> &out);

		static void Right(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> &out,
							        const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_F);

		static void Right(std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4>> &out_P,
							        const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_E);
	};
}

#endif //PROJECTIVE_MATRIX_H_
