#include "sfm_structure_from_2_views.h" 
#include "StructureComputation.h"

namespace Basics3D
{
	C_StructureComputation::C_StructureComputation
		                        (const T_Points2DSet<C_Point2D> &in_points2D,
			                       const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_F,
														 const int in_width,
														 const int in_height,
														 const double in_sigma)
	{
		m_texCoord = in_points2D;
		m_normPoints2D = in_points2D;

		//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		// HERE NORMALIZATION OF 2D POINTS AND FUND M IS MADE
		// origin moves from the up left corner to the center of image
		// sqrt(MaxX^2 + MaxY^2) -> sqrt(2)
		//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		// normalize input 2D points
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> Norm;
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> Norm1;
		double err1 = m_normPoints2D.Normalize(Norm, Norm1, in_width, in_height, in_sigma);
		// replace prev. line with
		// double err1 = m_normPoints2D.Normalize(Norm, Norm1, in_sigma);
		// in order to move origin not to the center of image but to center of mass of points

		double d = sqrt(double(in_width * in_width + in_height * in_height)) / 4.0;
		std::cout << std::endl << "1/d = " << 1.0/d;
		std::cout << std::endl << "MeanX[0] = " << m_normPoints2D.MeanX(0) * d
			<< "  MeanY[0] = " << m_normPoints2D.MeanY(0) * d << "  MeanX[1] = " 
			<< m_normPoints2D.MeanX(1) * d << "  MeanY[1] = "
			<< m_normPoints2D.MeanY(1) * d;

		// transform input fundamental matrix into normalized coordinates
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> Norm_1;
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> Norm1_1T;
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> Norm1_T;
		Norm_1 = Vtl::NumAlg::Lin::svd::inverse_save(Norm, err1);
		Norm1_T = boost::numeric::ublas::trans(Norm1);
		Norm1_1T = Vtl::NumAlg::Lin::svd::inverse_save(Norm1_T, err1);

		T_MatrixOperations::Normalize(m_fundM, in_F, Norm1_1T, Norm_1);

		// count inliers for normalized fund. m.
		/*temp*/T_Model<C_FundamentalMatrix> model0(m_normPoints2D);
		/*temp*/m_normPoints2D.EnableAll();
		/*temp*/int numbInl = 0;
		/*temp*/for (int m = 0; m < m_normPoints2D.Size(); m++)
		/*temp*/{
		/*temp*/	if (model0.IsInlier(m_fundM, m, in_sigma / 
			                  sqrt(double(in_width * in_width + in_height * in_height)) * 4.0))
		/*temp*/	{
		/*temp*/		numbInl ++;
		/*temp*/	}
		/*temp*/}
		/*temp*/std::cout << std::endl << std::endl << "number of inliers for FM:  "
		/*temp*/	<< numbInl;
		
		// create array for 3D points
		m_points3D.resize(in_points2D.Size(), 3);

		// @@Yurin@@:
		Structure::FromFM(m_normPoints2D,m_fundM,in_sigma);
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_StructureComputation::Reconstruction(const double in_g)
	{
		// compute projective matrices
		//Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> P;
		//Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> P1;
		//C_ProjectiveMatrix::Left(P);
		//C_ProjectiveMatrix::Right(P1, m_fundM);

		//// print projective matrices
		//std::cout << std::endl << std::endl << "P";
		//T_MatrixOperations::Print(P);
		//std::cout << std::endl << std::endl << "P1";
		//T_MatrixOperations::Print(P1);
		
		// compute projective reconstruction
		/*C_Triangulation triang(P, P1);

		int size = m_normPoints2D.Size();
		Vtl::NumAlg::Lin::ublas::matrix<double> points3D(size, 3);

		for (int i = 0; i < size; i++)
		{
			Vtl::NumAlg::Lin::ublas::c_vector<double, 3> point;
			triang.Get3DPoint(point, m_normPoints2D[2 * i], m_normPoints2D[2 * i + 1]);

			for (int j = 0; j < 3; j++)
			{ 
				points3D(i, j) = point(j);
			}
		}*/

		m_texCoord.Save("X.txt", "Y.txt", 0);
		std::cout << "";

		// compute metric reconstruction

		// compute projective matrices
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> P;
		C_ProjectiveMatrix::Left(P);
		
		C_AutoCalibration calibr;
		//std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 4>> H(4);
		//int nH = calibr.GetH(H, P, P1);
		std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4>> P1(4);
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> K;
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> K1;
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> E;
		calibr.GetRightP(P1, K, K1, E, m_fundM);

		int size = m_normPoints2D.Size();
		 
		// x^ = K^-1 * x; x^' = K'^-1 * x';
		for (int i = 0; i < size; i++)
		{
			m_normPoints2D[2 * i].Scale(1.0 / K(0, 0));
			m_normPoints2D[2 * i + 1].Scale(1.0 / K1(0, 0));
		}

		// count number of inliers for ess. matrix
		const double SIGMA = 0.5;
		/*temp*/T_Points2DSet<C_Point2D> inl(2);
		/*temp*/T_Model<C_FundamentalMatrix> model0(m_normPoints2D);
		/*temp*/m_normPoints2D.EnableAll();
		/*temp*/int numbInl = 0;
		/*temp*/for (int m = 0; m < m_normPoints2D.Size(); m++)
		/*temp*/{
		/*temp*/	if (model0.IsInlier(E, m, SIGMA))
		/*temp*/	{
		/*temp*/		numbInl ++;
		/*temp*/	}
		/*temp*/}
		/*temp*/std::cout << std::endl << std::endl << "number of inliers for EsM:  "
		/*temp*/	<< numbInl;

		// 
		for (int i = 0; i < 4; i++)
		{
			/*Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> corP;
		  corP = boost::numeric::ublas::prod(P, H[i]);
		  Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> corP1;
		  corP1 = boost::numeric::ublas::prod(P1, H[i]);*/
			
			C_Triangulation triang/*(corP, corP1)*/(P, P1[i]);

			for (int j = 0; j < size; j++)
		  {
			  Vtl::NumAlg::Lin::ublas::c_vector<double, 3> point;
			  triang.Get3DPoint(point, m_normPoints2D[2 * j], m_normPoints2D[2 * j + 1]);
		  }
			
			std::ostringstream os1;
			os1 << "ModelPoints" << i << ".wrl";
			std::string filename1 = os1.str();
			BuildModelPoints(filename1.c_str());

			std::ostringstream os;
			os << "ModelFaces" << i << ".wrl";
			std::string filename = os.str();
			std::cout << std::endl << filename.c_str();
			BuildModelFaces(filename.c_str(), "triang.txt");
		}
		
		std::cout << "";
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_StructureComputation::Correct
		                        (const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> in_P,
														 const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> in_P1,
														 const Vtl::NumAlg::Lin::ublas::matrix<double> in_points,
			                       const Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 4> &in_H)
	{
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> P;
		P = boost::numeric::ublas::prod(in_P, in_H);
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> P1;
		P1 = boost::numeric::ublas::prod(in_P1, in_H);

		std::cout << std::endl << std::endl << "P" << std::endl;
		T_MatrixOperations::Print(P);
		std::cout << std::endl << std::endl << "P1" << std::endl;
		T_MatrixOperations::Print(P1);

		// Decompose projective matrices
		C_AutoCalibration calib;
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> K;
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> R;
		Vtl::NumAlg::Lin::ublas::c_vector<double, 3> t;
		calib.DecomposeProjMatr(K, R, t, P);

		std::cout << std::endl << std::endl << "K" << std::endl;
		T_MatrixOperations::Print(K);
		std::cout << std::endl << std::endl << "R" << std::endl;
		T_MatrixOperations::Print(R);
		std::cout << std::endl << std::endl << "t" << std::endl;

		calib.DecomposeProjMatr(K, R, t, P1);

		std::cout << std::endl << std::endl << "K1" << std::endl;
		T_MatrixOperations::Print(K);
		std::cout << std::endl << std::endl << "R1" << std::endl;
		T_MatrixOperations::Print(R);
		std::cout << std::endl << std::endl << "t1" << std::endl;
		

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 4> H_1;
		H_1 = Vtl::NumAlg::Lin::svd::inverse_save(in_H, 0.000001);
		//std::cout << std::endl << std::endl << "H * H^-1" << std::endl;
		//T_MatrixOperations::Print(boost::numeric::ublas::prod(in_H, H_1));

		int size = in_points.size1();
		for (int i = 0; i < size; i++)
		{
			double x = in_points(i, 0);
			double y = in_points(i, 1);
			double z = in_points(i, 2);

			double newX = in_H(0,0) * x + in_H(0,1) * y + in_H(0,2) * z + in_H(0,3);
			double newY = in_H(1,0) * x + in_H(1,1) * y + in_H(1,2) * z + in_H(1,3);
			double newZ = in_H(2,0) * x + in_H(2,1) * y + in_H(2,2) * z + in_H(2,3);
			double newW = in_H(3,0) * x + in_H(3,1) * y + in_H(3,2) * z + in_H(3,3);
			
			m_points3D(i, 0) = newX / newW;
			m_points3D(i, 1) = newY / newW;
			m_points3D(i, 2) = newZ / newW;
		}
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_StructureComputation::BuildModelPoints(const char* in_filename)
	{
		std::ofstream outFile;
		outFile.open(in_filename);

		outFile 
			<< "#VRML V2.0 utf8" << std::endl << std::endl
			<< "Viewpoint" << std::endl
			<< "{" << std::endl
			<< "  position  0 8.0  -20.0" << std::endl
			<< "  orientation 1 0 0 3.1415926" << std::endl
			<< "  description \"1\"" << std::endl
			<< "}"  << std::endl << std::endl
			/*<< "DEF PointRed Shape" << std::endl
			<< "          {" << std::endl
			<< "            geometry Sphere {radius 2.0}" << std::endl
			<< "            appearance Appearance" << std::endl
			<< "            {" << std::endl
			<< "              material Material" << std::endl
			<< "              {" << std::endl
			<< "                diffuseColor     1.0 0.0 0.0" << std::endl
			<< "                shininess        0.3" << std::endl
			<< "                specularColor    0 0 0.2" << std::endl
			<< "                emissiveColor    0.1 0.1 0.1" << std::endl
			<< "                ambientIntensity 0.5" << std::endl
			<< "                transparency     0.0" << std::endl
			<< "              }" << std::endl
			<< "            }" << std::endl
			<< "          }" << std::endl << std::endl*/
			<< "DEF	Point Shape" << std::endl
			<< "          {" << std::endl
			<< "            geometry Sphere {radius 0.1}" << std::endl
			<< "            appearance Appearance" << std::endl
			<< "            {" << std::endl
			<< "              material Material" << std::endl
			<< "              {" << std::endl
			<< "                diffuseColor     0.8 0.55 0.4" << std::endl
			<< "                shininess        0.3" << std::endl
			<< "                specularColor    0 0 0.2" << std::endl
			<< "                emissiveColor    0.1 0.1 0.1" << std::endl
			<< "                ambientIntensity 0.5" << std::endl
			<< "                transparency     0.0" << std::endl
			<< "              }" << std::endl
			<< "            }" << std::endl
			<< "          }" << std::endl << std::endl;

		/*outFile 
				<< "Transform { translation " << m_center(0) << " "
				<< m_center(1) << " " << m_center(2) << std::endl
				<< "            children [ USE PointRed ]}" << std::endl;*/
			
		for (int i = 0; i < (int)m_points3D.size1(); i++)
		{
			/*if ((m_points3D(i, 0) == m_maxX) || (m_points3D(i, 0) == m_minX) ||
			  (m_points3D(i, 1) == m_maxY) || (m_points3D(i, 1) == m_minY) ||
					(m_points3D(i, 2) == m_maxY) || (m_points3D(i, 2) == m_minY) )
			{
				outFile 
				<< "Transform { translation " << m_points3D(i, 0) << " "
				<< m_points3D(i, 1) << " " << m_points3D(i, 2) << std::endl
				<< "            children [ USE PointRed ]}" << std::endl;
			}
			else
			{*/
			outFile 
				<< "Transform { translation " << m_points3D(i, 0) << " "
				<< m_points3D(i, 1) << " " << m_points3D(i, 2) << std::endl
				<< "            children [ USE Point ]}" << std::endl;
			/*}*/
		}
	
		outFile.close();
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_StructureComputation::BuildModelFaces(const char* in_filename,
		                                           const char* in_triang)
	{
		std::ofstream outFile;
		outFile.open(in_filename);

		outFile 
			<< "#VRML V2.0 utf8" << std::endl
			<< std::endl
			<< "Shape" << std::endl
			<< "{" << std::endl
			<< "  appearance Appearance" << std::endl
			<< "  {" << std::endl
			<< "    material Material" << std::endl
			<< "    {" << std::endl
			<< "      diffuseColor      0.8 0.55 0.4" << std::endl
			<< "      shininess         0.3" << std::endl
			<< "      specularColor     0 0 0.2" << std::endl
			<< "      emissiveColor     0.1 0.1 0.1" << std::endl
			<< "      ambientIntensity  0.5" << std::endl
			<< "      transparency      0.0" << std::endl
			<< "    }" << std::endl
			<< "    texture ImageTexture" << std::endl
			<< "    {" << std::endl
			<< "      url  \"src.jpg\"" << std::endl
			<< "    }" << std::endl
			<< "  }" << std::endl
			<< "  geometry IndexedFaceSet" << std::endl
			<< "  {" << std::endl
			<< "    coord Coordinate" << std::endl
			<< "    {" << std::endl
			<< "      point" << std::endl
			<< "      [" << std::endl;

		for (int i = 0; i < (int)m_points3D.size1(); i++)
		{
			outFile << "        " << m_points3D(i, 0) << " " 
				<< m_points3D(i, 1) << " " << m_points3D(i, 2) << std::endl;
		}

		outFile
			<< "      ]" << std::endl
			<< "    }" << std::endl
			<< "    coordIndex" << std::endl
			<< "    [" << std::endl;

		std::ifstream inFile;
		inFile.open(in_triang, std::ios::in);

		double p1, p2, p3;
		int c;
		int k = 0;

		while (1)
		{
			k++;

			inFile >> p1;
			inFile >> p2;
			inFile >> p3;

			outFile << "      " << int(p1-1) << " " << int(p2-1) << " "
				<< int(p3-1) << " -1" << std::endl;
			outFile << "      " << int(p3-1) << " " << int(p2-1) << " "
				<< int(p1-1) << " -1" << std::endl;

			/*std::cout << "      " << int(p1-1) << " " << int(p2-1) << " "
				<< int(p3-1) << " -1" << std::endl;
			std::cout << "      " << int(p3-1) << " " << int(p2-1) << " "
				<< int(p1-1) << " -1" << std::endl;*/

			c = inFile.peek();

			if (c == EOF)
			{
				break;
			}
		}

		//std::cout << std::endl << "k = " << k;

		inFile.close();

		outFile
			<< "    ]" << std::endl
			<< "    texCoord TextureCoordinate" << std::endl
			<< "    {" << std::endl
			<< "      point" << std::endl
			<< "      [" << std::endl;

		for (int i = 0; i < (int)m_points3D.size1(); i++)
		{
			/*if (m_texCoord[2 * i].Y() > 345)
			{
				std::cout << std::endl << m_texCoord[2 * i].X() / 551.0 << " "
					<< m_texCoord[2 * i].Y() / 750.0;
			}*/

			outFile << "        " << /*(551.0 - */m_texCoord[2 * i].X()/*)*/ / 551.0 << " " << 
				(750.0 - m_texCoord[2 * i].Y()) / 750.0 << std::endl;
		}

		outFile
			<< "      ]" << std::endl
			<< "    }" << std::endl
			<< "  }" << std::endl
			<< "}" << std::endl;

		outFile.close();
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_StructureComputation::OptimalSolution()
	{
		//// compute right epipole
		//Vtl::NumAlg::Lin::ublas::c_vector<double, 3> er;
		//C_Epipole::Right(er, m_fundM);

		// compute projective matrices
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> P;
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> P1;
		C_ProjectiveMatrix::Left(P);
		C_ProjectiveMatrix::Right(P1, m_fundM);

		C_Triangulation triang(P, P1);
			
		int size = m_normPoints2D.Size();

		for (int i = 0; i < size; i++)
		{
			/*std::cout << std::endl << "i = " << i;
			if (i == 6)
				std::cout << "";*/

			// Build translational matrices
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> T_1;
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> T1_1;
			C_TransformMatrices::TranslationalInverse(T_1, m_normPoints2D[2 * i].X(),
				                                        m_normPoints2D[2 * i].Y());
			C_TransformMatrices::TranslationalInverse(T1_1, m_normPoints2D[2 * i + 1].X(),
				                                        m_normPoints2D[2 * i + 1].Y());

			// Replace F by T'_T*F*T_1
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> F;
			T_MatrixOperations::Normalize(F, m_fundM, T1_1, T_1);

			// Compute epipoles and normalize them
			// new epipoles are need as the fund matrix was transformed
			Vtl::NumAlg::Lin::ublas::c_vector<double, 3> e;
			Vtl::NumAlg::Lin::ublas::c_vector<double, 3> e1;
			C_Epipole::Left(e, m_fundM);
			C_Epipole::Right(e1, m_fundM);
			C_Epipole::Normalize(e);
			C_Epipole::Normalize(e1);

			// Build rotation matrices
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> R;
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> R_T;
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> R1;
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> R1_T;
			C_TransformMatrices::Rotation(R, e);
			C_TransformMatrices::Rotation(R1, e1);
			R_T = boost::numeric::ublas::trans(R);
			R1_T = boost::numeric::ublas::trans(R1);

			/*temp*/std::cout << std::endl << std::endl << "e" << std::endl;
			/*temp*/std::cout << e(0) << "  " << e(1) << "  " << e(2);
			/*temp*/std::cout << std::endl << std::endl << "e1" << std::endl;
			/*temp*/std::cout << e1(0) << "  " << e1(1) << "  " << e1(2);
			/*temp*/Vtl::NumAlg::Lin::ublas::c_vector<double, 3> Re = boost::numeric::ublas::prod(R, e);
			/*temp*/std::cout << std::endl << std::endl << "R * e" << std::endl;
			/*temp*/std::cout << Re(0) << "  " << Re(1) << "  " << Re(2);
			/*temp*/Vtl::NumAlg::Lin::ublas::c_vector<double, 3> R1e1 = boost::numeric::ublas::prod(R1, e1);
			/*temp*/std::cout << std::endl << std::endl << "R1 * e1" << std::endl;
			/*temp*/std::cout << R1e1(0) << "  " << R1e1(1) << "  " << R1e1(2);

			// Replace F by R'*F*R^T
			T_MatrixOperations::Normalize(F, F, R1, R_T);

			/*temp*/std::cout << std::endl << std::endl << "R' * F * R^T" << std::endl;
			/*temp*/T_MatrixOperations::Print(F);

			// Form the polynomial g(t)
			double f = e(2);
			double f1 = e1(2);
			double a = F(1, 1);
			double b = F(1, 2);
			double c = F(2, 1);
			double d = F(2, 2);

			/*temp*/std::cout << std::endl << "a = " << a << "  b = " << b << "  c = " << c << "  d = "
			/*temp*/	<< d << "  f = " << f << "  f1 = " << f1;
			/*temp*/std::cout << "";

			g func(a, b, c, d, f, f1, 6);

			// Solve g(t)
			std::vector<double> t(6);
			SolveEq solve;
			int n = solve.Apply(t, &func, 0.00000001);

			double t_min = t[0];
			double min = CostFunction(a, b, c, d, f, f1, t[0]);

			for (int j = 1; j < n; j++)
			{
				double cost = CostFunction(a, b, c, d, f, f1, t[j]);
				if (cost < min)
				{
					t_min = t[j];
					min = cost;
				}
			}

			double asimpt = AsimptValue(a, c, f, f1);

			// Evaluate lines
			Vtl::NumAlg::Lin::ublas::c_vector<double, 3> l;
			l(0) = t_min * f;
			l(1) = 1;
			l(2) = -t_min;

			Vtl::NumAlg::Lin::ublas::c_vector<double, 3> l1;
			l1(0) = -f1 * (c * t_min + d);
			l1(1) = a * t_min + b;
			l1(2) = c * t_min + d;
			
			Vtl::NumAlg::Lin::ublas::c_vector<double, 3> x_new;
			Vtl::NumAlg::Lin::ublas::c_vector<double, 3> x1_new;

			ClosestToOriginPoint(x_new, l);
			ClosestToOriginPoint(x1_new, l1);
			std::cout << std::endl << "1  " << x_new;
			std::cout << std::endl << "   " << x1_new;

			x_new = boost::numeric::ublas::prod(R_T, x_new);
			x_new = boost::numeric::ublas::prod(T_1, x_new);
			x1_new = boost::numeric::ublas::prod(R1_T, x1_new);
			x1_new = boost::numeric::ublas::prod(T1_1, x1_new);
			std::cout << std::endl << "2  " << x_new;
			std::cout << std::endl << "   " << x1_new;

			for (int j = 0; j < 2; j++)
			{
				x_new(j) /= x_new(2);
				x1_new(j) /= x1_new(2);
			}
			std::cout << std::endl << "3  " << x_new;
			std::cout << std::endl << "   " << x1_new;

			//LinearTriangulation
			Vtl::NumAlg::Lin::ublas::c_vector<double, 3> point;
			triang.Get3DPoint(point, C_Point2D(x_new(0), x_new(1), 1),
				                C_Point2D(x1_new(0), x1_new(1), 1));

			for (int j = 0; j < 3; j++)
			{
			  m_points3D(i, j) = point(j);
			}
		}
	}

	///////////////////////////////////////////////////////////////////////////////////////
	double C_StructureComputation::CostFunction(const double a,
			                                        const double b,
												                      const double c,
												                      const double d,
												                      const double f,
												                      const double f1,
												                      const double t)
	{
		double res = t * t / (1 + f * f * t * t);
		double k = c * t + d;
		double m = a * t + b;
		k = k * k;
		m = m * m;
		res += k / (m + f1 * f1 * k);
		return res;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	double C_StructureComputation::AsimptValue(const double a,
		                                         const double c,
										                         const double f,
										                         const double f1)
	{
		double res = 1 / (f * f) + c * c / (a * a + f1 * f1 * c * c);
		return res;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_StructureComputation::ClosestToOriginPoint
		                         (Vtl::NumAlg::Lin::ublas::c_vector<double, 3> out_p,
			                        const Vtl::NumAlg::Lin::ublas::c_vector<double, 3> in_line)
	{
		out_p(0) = - in_line(0) * in_line(2);
		out_p(1) = - in_line(1) * in_line(2);
		out_p(2) = in_line(0) * in_line(0) + in_line(1) * in_line(1);
	}

}