#ifndef STRUCTURE_COMPUTATION_H_
#define STRUCTURE_COMPUTATION_H_

#include <fstream>
#include <math.h>

//library includings
#include "num_linalg.h"
//own includings
#include "Point2D.h"
#include "Points2DSet.h"
#include "Epipole.h"
#include "ProjectiveMatrix.h"
#include "Triangulation.h"
#include "TransformMatrices.h"
#include "Func.h"
#include "SolveEq.h"
#include "FundamentalMatrix.h"
#include "Model.h"
#include "AutoCalibration.h"

namespace Basics3D
{
	class C_StructureComputation
	{
	private:
	  Vtl::NumAlg::Lin::ublas::matrix<double> m_points3D;
		T_Points2DSet<C_Point2D> m_normPoints2D;
		T_Points2DSet<C_Point2D> m_texCoord;
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> m_fundM;
		double m_maxX;
		double m_minX;
		double m_maxY;
		double m_minY;
		double m_maxZ;
		double m_minZ;

	private:
		void OptimalSolution();

		double CostFunction(const double a,
			                  const double b,
												const double c,
												const double d,
												const double f,
												const double f1,
												const double t);

		double AsimptValue(const double a,
		                   const double c,
										   const double f,
										   const double f1);

		void ClosestToOriginPoint(Vtl::NumAlg::Lin::ublas::c_vector<double, 3> out_p,
			                        const Vtl::NumAlg::Lin::ublas::c_vector<double, 3> in_line);
		
		void BuildModelPoints(const char* in_filename);

		void BuildModelFaces(const char* in_filename,
		                     const char* in_triang);

		void Correct(const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> in_P,
								 const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> in_P1,
								 const Vtl::NumAlg::Lin::ublas::matrix<double> in_points,
			           const Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 4> &in_H);
		
	public:
		C_StructureComputation(const T_Points2DSet<C_Point2D> &in_points2D,
		                       const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_F,
													 const int in_width,
													 const int in_height,
													 const double in_sigma);

		void Reconstruction(const double in_g);
	};
}

#endif // STRUCTURE_COMPUTATION_H_
