#ifndef _SFM_CALIBRATION_H_
#define _SFM_CALIBRATION_H_ 
////////////////////////////////////////////////////////////////////////////////////////////
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <functional>      // For greater<int>( )
#include "loki_typeManip.h"
//#include "loki_Typelist.h"
#include "loki_static_check.h"
#include "num_linalg.h"
#include "num_optimization.h"
#include "alexandresku.h" 
#include "vtl_meta_base.h" 

#include "sfm_types.h"
#include "sfm_dbg.h" 
////////////////////////////////////////////////////////////////////////////////////////////
//template<typename T> struct Fmt;

namespace Basics3D
{
	///      . ,    ,   
	///     (float, double)      (Matrix).   .
	template<typename T>
	struct FmTools 
	{
		typedef Vtl::NumAlg::Lin::ublas::c_matrix<T, 3, 3> Matrix;
		typedef Vtl::NumAlg::Lin::ublas::c_vector<T, 3> Vector3;
		typedef Vtl::NumAlg::Lin::ublas::c_matrix<T, 3, 4> Matrix_3x4;
		typedef typename Vtl::NumAlg::Lin::ublas::c_matrix<T, 3, 3>::value_type value_type;

		//     
		static Matrix Build_K(T focus_x,T focus_y,T u_0=T(0), T v_0=T(0), T skew=T(0))
		{
			Matrix K;
			Build_K(K,focus_x,focus_y,u_0,v_0,skew);
			return K;
		}
		//     
		static void Build_K(Matrix& K,T focus_x,T focus_y,T u_0=T(0), T v_0=T(0), T skew=T(0))
		{
			K(0,0)=focus_x;  K(0,1)=skew;     K(0,2)=u_0;
			K(1,0)=0;        K(1,1)=focus_y;  K(1,2)=v_0;
			K(2,0)=0;        K(2,1)=0;        K(2,2)=1;
		}
		//     ,  , principal points and skew is zero, aspect ratio is 1. 
		static Matrix Build_K(T focus)
		{
			Matrix K;
			Build_K(K,focus);
			return K;
		}
		//     ,  , principal points and skew is zero, aspect ratio is 1. 
		static void Build_K(Matrix& K,T focus)
		{
			K(0,0)=focus;    K(0,1)=0;      K(0,2)=0;
			K(1,0)=0;        K(1,1)=focus;  K(1,2)=0;
			K(2,0)=0;        K(2,1)=0;      K(2,2)=1;
		}
		/// E = K'^T * F * K , here K2==K' and K1==K 
		static Matrix fm2essential(const Matrix& K2,const Matrix& F,const Matrix& K1)
		{
			Matrix E;
			fm2essential(E,K2,F,K1);
			return E;
		}
		/// E = K'^T * F * K , here K2==K' and K1==K
		static void fm2essential(Matrix& E,const Matrix& K2,const Matrix& F,const Matrix& K1)
		{
			using namespace Vtl::NumAlg::Lin::ublas;
			E= prod(trans(K2),F);
			E= prod(F,K1);
		}
		/// F= (K'^T)^(-1) * E  * K^(-1)  , here K2==K' and K1==K
		static Matrix essential2fm(const Matrix& K2,const Matrix& E,const Matrix& K1)
		{
			Matrix F;
			fm2essential(F,K2,E,K1);
			return F;
		}
		/// F= (K'^T)^(-1) * E  * K^(-1)  , here K2==K' and K1==K
		static void essential2fm(Matrix& F,const Matrix& K2,const Matrix& E,const Matrix& K1)
		{
			using namespace Vtl::NumAlg::Lin::ublas;
			using namespace Vtl::NumAlg::Lin;
			Matrix Kinv;
			svd::inverse_save(Kinv,K2,1.0e-06);
			F= prod(trans(Kinv),F);
			svd::inverse_save(Kinv,K1,1.0e-06);
			F= prod(F,Kinv);
		}
		//    P=K*(R|t)
		static void Build_P(Matrix_3x4& P,const Matrix& K,const Matrix& R,const Vector3& t)
		{
			using namespace Vtl::NumAlg::Lin::ublas;
			matrix_slice<Matrix_3x4> RP(P, slice(0, 1, 3), slice(0, 1, 3) );
			RP=R;
			matrix_column<Matrix_3x4> tP( P, 3 );
			tP=t;
			P=prod(K,P);
		}
	};

	/// :       .
	namespace Fm2Essential
	{
		//     ublas::c_matrix<T, N, N>.      
		//       enum (   ,   c_matrix -  
		//     ),        
		//  DirectMinimization.
		template<class Matrix> 
		class DirectMinimization;
		//   DirectMinimization     ublas::c_matrix<T, 3, 3>.
		template<typename T,int N> 
		struct DirectMinimization<Vtl::NumAlg::Lin::ublas::c_matrix<T, N, N> >
		{
			enum {n=N};
			typedef Vtl::NumAlg::Lin::ublas::c_matrix<T, N, N> Matrix;
			typedef Vtl::NumAlg::Lin::ublas::c_vector<T, N>  Vector;
			typedef typename Matrix::value_type value_type;
			typedef typename Matrix::size_type size_type;
			const Matrix F;//!< Input fundamental matrix
			Matrix E;//!< Output essential matrix
			Matrix K1;//!< Output calibration matrix K
			Matrix K2;//!< Output calibration matrix K'
		public:
			DirectMinimization(const Matrix& Fm)
				: F(Fm)
			{
				LOKI_STATIC_CHECK(N==3,Fundamental_matrix_should_be_3x3_in_size);
				using namespace Vtl::NumAlg::Lin::ublas;

				std::cout << "Fm2Essential" << std::endl;

			}
			void operator()()
			{// Some experiments can be maked here - call propere function

				std::cout << "========== Direct search ============" <<std::endl;
//   ,   .         .
#if 0
				value_type focus = full_search();

				K1=FmTools<value_type>::Build_K(focus);
				K2=K1;
				E=FmTools<value_type>::fm2essential(K2,F,K1);
				{
					std::cout << "====== Optimal E via direct search =====" <<std::endl;
					std::cout << " focus= " << focus <<std::endl ;
					std::cout << " E= " << std::endl ;
					std::cout << Dbg::Fmt<Matrix,Dbg::Fmt<Matrix>::RowsAndCols>(E) <<std::endl; 
					Matrix U=E;
					value_type eps = find_sing_value_dev(U);
					std::cout << " |s1-s2|/(s1+s2)= " << eps << std::endl;
				}
#else
				value_type focus = optimal_search();

				K1=FmTools<value_type>::Build_K(focus);
				K2=K1;
				E=FmTools<value_type>::fm2essential(K2,F,K1);

				{
					std::cout << "====== Optimal E via fast direct search =====" <<std::endl;
					std::cout << " focus= " << focus <<std::endl ;
					std::cout << " E= " << std::endl ;
					std::cout << Dbg::Fmt<Matrix,Dbg::Fmt<Matrix>::RowsAndCols>(E) <<std::endl; 
					Matrix U=E;
					value_type eps = find_sing_value_dev(U);
					std::cout << " |s1-s2|/(s1+s2)= " << eps << std::endl;
				}
#endif
			}
		private:
			value_type optimal_search()
			{
				struct F
				{
					DirectMinimization& dm;
					F(DirectMinimization& obj) : dm(obj) {}
					value_type operator()(value_type focus) {return dm.build_E(focus);}
				};
				const value_type f_min=0.1;
				const value_type f_max=10.0;
				value_type f_opt= Vtl::NumAlg::Opt::Fmin( F(*this) , f_min , f_max );
				return f_opt;
			}
			value_type full_search()
			{
				const int count=30;
				value_type f_min=0.1;
				value_type f_max=10.0;
				value_type f_step=(f_max-f_min)/count;
				value_type f_opt=f_min;
				value_type eps;
				for(int i=0; i<5 ; ++i)
				{
					f_opt=full_search_subproc(f_min,f_max,f_step,eps);
					f_min=f_opt-f_step;
					f_max=f_opt+f_step;
					f_step=(f_max-f_min)/count;
				}
				std::cout << " |s1-s2|/(s1+s2)= " << eps << std::endl;
				return f_opt;
			}
			value_type full_search_subproc(value_type f_min,value_type f_max,value_type f_step,value_type& eps)
			{
				value_type f_opt=f_min;
				eps=build_E(f_min);
				for(value_type f=f_min+f_step ; f<f_max; f+=f_step)
				{
					value_type eps2=build_E(f);
					if(eps2<eps)
					{
						eps=eps2;
						f_opt=f;
						std::cout << " |s1-s2|/(s1+s2)= " << eps << std::endl;
					}
				}
				return f_opt;
			}
			value_type build_E(value_type focus)
			{
				using namespace Vtl::NumAlg::Lin::ublas;
				K1=FmTools<value_type>::Build_K(focus);
				// K1 == K2 - both cameras supposed identical
				Matrix U=FmTools<value_type>::fm2essential(K1,F,K1);
				return find_sing_value_dev(U);
			}
			///    
			value_type find_sing_value_dev(Matrix& E)
			{
				using namespace Vtl::NumAlg::Lin::ublas;
				Matrix V;
				Vector W;
				Vector rv1;
				Vtl::NumAlg::Lin::svd::decomp(E,W,V,rv1,false,false);

				Vtl::NumAlg::Implicit::Sort<Vector,Vtl::NumAlg::Implicit::Pred::gt<Vector> > s(W);
				//    :
				//Vtl::NumAlg::Implicit::Pred::gt<Vector> decrement_order(W);
				//Vtl::NumAlg::Implicit::Sort<Vector> s(W,decrement_order);
				value_type   sigma_1( W(s[0]) );
				value_type   sigma_2( W(s[1]) );
				return std::abs(sigma_1-sigma_2)/(sigma_1+sigma_2);
			}
			
		};
		///     ublas::c_matrix<T, N, N>.      
		///       enum (   ,   c_matrix -  
		///     ),        
		///  KruppaEq.
		template<class Matrix> 
		class KruppaEq;
		///   KruppaEq     ublas::c_matrix<T, 3, 3>.
		/// !!!  !!!    ,      .   
		///       ( )     
		///   ,    6       .
		///      ?            ?
		template<typename T,int N> 
		class KruppaEq<Vtl::NumAlg::Lin::ublas::c_matrix<T, N, N> >
		{
			enum {n=N};
			typedef Vtl::NumAlg::Lin::ublas::c_matrix<T, N, N> Matrix;
			typedef Vtl::NumAlg::Lin::ublas::c_vector<T, N>  Vector;
			typedef typename Matrix::value_type value_type;
			typedef typename Matrix::size_type size_type;
			const Matrix F;//!< Input fundamental matrix
			Matrix E[6];//!< Output essential matrix
			Matrix K1[6];//!< Output calibration matrix K
			Matrix K2[6];//!< Output calibration matrix K'
		public:
			// Now Kruppa equations in form:
			//
			//  (0) = (1) = (2)
			//
			// (i) = (j) form equation
			//
			//  k0 = i*2;
			//  k1 = k1+1;
			//  k2 = j*2;
			//  k3 = j*2+1;
			//
			//   a[k0]*y + b[k0]   a[k2]*y + b[k2]  
			//   --------------- = ---------------- 
			//   a[k1]*x + b[k1]   a[k3]*x + b[k3]  
			//
			//  (a[k0]*y + b[k0])*(a[k3]*x + b[k3])-(a[k1]*x + b[k1])*(a[k2]*y + b[k2]) =0 
			//
			// Or, rewriting:
			//
			//    a[k0]*a[k3]*x*y + a[k0]*b[k3]*y + b[k0]*a[k3]*x + b[k0]*b[k3]
			//   - 
			//    a[k1]*a[k2]*x*y + b[k1]*a[k2]*y + a[k1]*b[k2]*x + b[k1]*b[k2] =0 
			// 
			// Or, rewriting:
			//
			//           | a[k0]  a[k1] |          | a[k0]  b[k1] |          | b[k0]  a[k1] |      | b[k0]  b[k1] |
			//  xy * det |              | + y * det|              | + x * det|              | + det|              | = 0
			//           | a[k2]  a[k3] |          | a[k2]  b[k3] |          | b[k2]  a[k3] |      | b[k2]  b[k3] |
			//
			// i_coef           0                         1                         2                     3
			// i_col        0       1                 0       1                 0       1             0       1
			// i_coef(hex):    00                        01                        10                    11      
			// desired:     0       0                 0       1                 1       0             1       1
			// -------------------------------------------------------------------------------------------------------
			class Km
			{
				value_type a[6];
				value_type b[6];
			public:
				Km() {}
				template<int i>
				void build_expression(value_type  multiplier,const Vtl::NumAlg::Lin::ublas::matrix_column<Matrix>& v_1,const Vtl::NumAlg::Lin::ublas::matrix_column<Matrix>& v_2)
				{
					mul(a[i],b[i],multiplier, v_1, v_2 );
				}
				template<int ratio_1,int ratio_2>
				void build_equation_coef(value_type* coef)
				{
					coef[0]=det<ratio_1,ratio_2,0>();
					coef[1]=det<ratio_1,ratio_2,1>();
					coef[2]=det<ratio_1,ratio_2,2>();
					coef[3]=det<ratio_1,ratio_2,3>();
				}
				void print_ab()
				{
					std::cout << std::resetiosflags(std::ios::showpos);
					Vtl::Meta::print_array<6>(a,"a",false);
					Vtl::Meta::print_array<6>(b,"b",false);
				}
				void to_single_focus(value_type* coef)
				{
					coef[1]+=coef[2];
					coef[2]=coef[3];
					coef[1]/=coef[0];
					coef[2]/=coef[0];
					coef[0]/=coef[0];
				}
				void solve_qe(const value_type* coef,value_type& x1,value_type& x2)
				{
					value_type d=Vtl::sqr(coef[1])-4*coef[0]*coef[2];
					std::cout << "Kruppa Discriminant = " << d <<std::endl;
					if( coef[1] > 0 )
					{
						x1=(-coef[1]-sqrt(d))/(2*coef[0]);
						x2=coef[2]/x1;
					}
					else
					{
						x2=(-coef[1]+sqrt(d))/(2*coef[0]);
						x1=coef[2]/x2;
					}
					x1=Vtl::NumAlg::sign(sqrt(std::abs(x1)),x1);
					x2=Vtl::NumAlg::sign(sqrt(std::abs(x2)),x2);
				}
			private:
				void mul(value_type& a,value_type& b,value_type coef,const Vtl::NumAlg::Lin::ublas::matrix_column<Matrix> v1,const Vtl::NumAlg::Lin::ublas::matrix_column<Matrix> v2)
				{
					a=coef*(v1[0]*v2[0]+v1[1]*v2[1]);
					b=coef*(v1[2]*v2[2]);
				}
				template<int i_coef,int i_col,int k>
				value_type col() const
				{
					return ab_selector<k>(Loki::Int2Type< ( (i_coef/(2-i_col)) & 1 ) >() );
				}
				template<int k>
				value_type ab_selector(Loki::Int2Type<0> ) const {return a[k];}
				template<int k>
				value_type ab_selector(Loki::Int2Type<1> ) const {return b[k];}
				template<int ratio_1,int ratio_2,int i_coef>
				value_type det()
				{
					enum {k0=ratio_1*2};
					enum {k1=k0+1};
					enum {k2=ratio_2*2};
					enum {k3=k2+1};
					return col<i_coef,0,k0>()*col<i_coef,1,k3>()-col<i_coef,0,k2>()*col<i_coef,1,k1>();
				}
			};
		public:
			KruppaEq(const Matrix& Fm)
				: F(Fm)
			{
				LOKI_STATIC_CHECK(N==3,Fundamental_matrix_should_be_3x3_in_size);
				using namespace Vtl::NumAlg::Lin::ublas;

				std::cout << std::endl << "===== Kruppa equations =============" << std::endl;

				Matrix U=F;
				Matrix V;
				Vector W;
				Vector rv1;
				Vtl::NumAlg::Lin::svd::decomp(U,W,V,rv1);

				Vtl::NumAlg::Implicit::Sort<Vector,Vtl::NumAlg::Implicit::Pred::gt<Vector> > s(W);

				value_type        sigma_1( W[ s[0] ] );
				matrix_column<Matrix> u_1( U, s[0]   );
				matrix_column<Matrix> v_1( V, s[0]   );
				value_type        sigma_2( W[ s[1] ] );
				matrix_column<Matrix> u_2( U, s[1]   );
				matrix_column<Matrix> v_2( V, s[1]   );
	#if 1
				for(int i=0; i<3; ++i)
				{
					std::cout << "    W[" << i << "]=" << W[i] << std::endl;
					std::cout << "    s[" << i << "].index=" << s[i] << std::endl;
					std::cout << "    s[" << i << "].sigma=" << s[i] << std::endl;
				}
				std::cout << "U" << std::endl;
				std::cout << Dbg::Fmt<Matrix,Dbg::Fmt<Matrix>::RowsAndCols>(U) <<std::endl; 
				std::cout << "V" << std::endl;
				std::cout << Dbg::Fmt<Matrix,Dbg::Fmt<Matrix>::RowsAndCols>(V) <<std::endl; 

				std::cout << "  sigma_1=" << sigma_1 << "  sigma_2=" << sigma_2 << std::endl;
				std::cout << "u_1 :  " << u_1 << std::endl;
				std::cout << "v_1 :  " << v_1 << std::endl;
				std::cout << "u_2 :  " << u_2 << std::endl;
				std::cout << "v_2 :  " << v_2 << std::endl;
	#endif
				// Now Kruppa equations in form:
				//
				//   a[0]*y + b[0]   a[2]*y + b[2]   a[4]*y + b[4]   
				//   ------------- = ------------- = -------------
				//   a[1]*x + b[1]   a[3]*x + b[3]   a[5]*x + b[5]   
				//
				//  Where y = focus_2^2 and x == focus_1^2
				//
				//  (0) = (1) = (2)
				//

				value_type a[4];
				Km km;
				km.build_expression<0>( 1                 , u_2, u_2 );
				km.build_expression<1>( Vtl::sqr(sigma_1) , v_1, v_1 );
				km.build_expression<2>( -1                , u_1, u_2 );
				km.build_expression<3>( sigma_1*sigma_2   , v_1, v_2 );
				km.build_expression<4>( 1                 , u_1, u_1 );
				km.build_expression<5>( Vtl::sqr(sigma_2) , v_2, v_2 );
				km.print_ab();
				value_type x1;
				value_type x2;

				std::cout << "--------------------------" <<std::endl;
				km.build_equation_coef<0,1>(a);
				Vtl::Meta::print_array<4>(a,"eq_0_1",false);
				km.to_single_focus(a);
				Vtl::Meta::print_array<3>(a,"eq_0_1",false);
				km.solve_qe(a,x1,x2);
				std::cout << "x1= " << x1 << "      x2= " << x2 <<std::endl;
				build_E(x1,0);
				build_E(x2,1);

				std::cout << "--------------------------" <<std::endl;
				km.build_equation_coef<0,2>(a);
				Vtl::Meta::print_array<4>(a,"eq_0_2",false);
				km.to_single_focus(a);
				Vtl::Meta::print_array<3>(a,"eq_0_2",false);
				km.solve_qe(a,x1,x2);
				std::cout << "x1= " << x1 << "      x2= " << x2 <<std::endl;
				build_E(x1,2);
				build_E(x2,3);

				std::cout << "--------------------------" <<std::endl;
				km.build_equation_coef<1,2>(a);
				Vtl::Meta::print_array<4>(a,"eq_1_2",false);
				km.to_single_focus(a);
				Vtl::Meta::print_array<3>(a,"eq_1_2",false);
				km.solve_qe(a,x1,x2);
				std::cout << "x1= " << x1 << "      x2= " << x2 <<std::endl;
				build_E(x1,4);
				build_E(x2,5);
 			}
		private:
			value_type build_E(value_type focus,int hypothesis)
			{
				using namespace Vtl::NumAlg::Lin::ublas;

				std::cout << std::endl << "Hypothesis[  "  << hypothesis << " ]:  focus= "<< focus << std::endl;

				K1[hypothesis]=FmTools<value_type>::Build_K(focus);
				K2[hypothesis]=K1[hypothesis];
				// K1 == K2 - both cameras supposed identical
				E[hypothesis]=FmTools<value_type>::fm2essential(K2[hypothesis],F,K1[hypothesis]);
				Matrix U=E[hypothesis];
				value_type   eps = find_sing_value_dev(U);
				std::cout << " |s1-s2|/(s1+s2)= " << eps << std::endl;
				return eps;
			}
			///    
			value_type find_sing_value_dev(Matrix& E)
			{
				using namespace Vtl::NumAlg::Lin::ublas;
				Matrix V;
				Vector W;
				Vector rv1;
				Vtl::NumAlg::Lin::svd::decomp(E,W,V,rv1,false,false);

				Vtl::NumAlg::Implicit::Sort<Vector,Vtl::NumAlg::Implicit::Pred::gt<Vector> > s(W);
				value_type   sigma_1( W(s[0]) );
				value_type   sigma_2( W(s[1]) );
				std::cout << " Sigmas = " << sigma_1 << "  "<< sigma_2 << "  "<< W(s[2]) << std::endl;
				return std::abs(sigma_1-sigma_2)/(sigma_1+sigma_2);
			}
		};
		///         . 
		/// \ref Lit_HartleyZissermanBook, Result 9.19 page 259.
		template<typename T> 
		class CameraPoseFromEssential
		{
			typedef T value_type;
			enum { N3=3};
			typedef Vtl::NumAlg::Lin::ublas::c_matrix<value_type, N3, N3> Matrix_3x3;
			typedef Vtl::NumAlg::Lin::ublas::c_vector<value_type, N3> Vector_3;
			enum { N4=4};
			typedef Vtl::NumAlg::Lin::ublas::c_matrix<value_type, N3, N4> Matrix_4x3;
			typedef Vtl::NumAlg::Lin::ublas::c_vector<value_type, N4> Vector_4;
			const Matrix_3x3& E;
			const Matrix_3x3  W;
			const Matrix_3x3  WT;
		public:
			Matrix_3x3  R0;
			Vector_3  t0;
			Matrix_3x3  R;
			Vector_3  t;
			CameraPoseFromEssential(const Matrix_3x3 E_)
				: E(E_) , W(build_W()), WT(Vtl::NumAlg::Lin::ublas::trans(W))
			{
				using namespace Vtl::NumAlg::Lin::ublas;
				using namespace Vtl::NumAlg::Lin;
				using namespace Vtl::NumAlg; //sign

				Matrix_3x3 U=E;
				Matrix_3x3 V;
				Vector_3 sigma;
				Vector_3 rv1;
				Vtl::NumAlg::Lin::svd::decomp(U,sigma,V,rv1);
				
				Vtl::NumAlg::Lin::svd::sort(U,sigma,V);

				matrix_column<Matrix_3x3> u( U, 2 );
				Vector_3 u3(u);

				V=trans(V);
				Matrix_3x3 R1=mult(U,W,V);
				Matrix_3x3 R2=mult(U,WT,V);

				Vector_3 d1;
				d1(0)=0; d1(1)=0; d1(2)=1; 
				Vector_3   d2=prod(R1,d1);
				value_type d2_norm = boost::numeric::ublas::blas_1::dot(d2,d2);
				value_type deg1 = boost::numeric::ublas::blas_1::dot(d1,d2)/sqrt(d2_norm);
				std::cout << "Camera rotation: deg = " << acos(deg1)*180/M_PI << "      d2_norm = " << d2_norm << std::endl;;
				std::cout << "with z direction: " << d2 << std::endl;;
						   d2=prod(R2,d1);
						   d2_norm = boost::numeric::ublas::blas_1::dot(d2,d2);
				value_type deg2 = boost::numeric::ublas::blas_1::dot(d1,d2)/sqrt(d2_norm);
				std::cout << "Camera rotation: deg = " << acos(deg2)*180/M_PI << "      d2_norm = " << d2_norm << std::endl;;
				std::cout << "with z direction: " << d2 << std::endl;;

				if( sign(1,deg1) == sign(1,deg2))
					std::cout << "BOTH CAMERAROTATION IS APPROPRIATE !!!! NOT PROCESSED! "  << std::endl;;
				//Set R|t values
				Vector_3 d;
				if(deg1>0)
				{
					R=R1;
					d=d1;
				}
				else
				{
					R=R2;
					d=d2;
				}
				value_type ofs = boost::numeric::ublas::blas_1::dot(u3,d);
				t;
				if(ofs > 0)
					t=u3;
				else
					t=-u3;
				std::cout << "translation selected: " << t << "      ofs = " << ofs << std::endl;;
				R0=ublas::identity_matrix<value_type>(N3,N3);
				t0=boost::numeric::ublas::zero_vector<value_type>(N3);
				
			}
		private:
			static Matrix_3x3 mult(const Matrix_3x3& a,const Matrix_3x3& b,const Matrix_3x3& c)
			{
				using namespace Vtl::NumAlg::Lin::ublas;
				Matrix_3x3 r=prod(a,b);
				return prod(r,c);
			}
			static Matrix_3x3 build_W()
			{
				Matrix_3x3 w;
				build_W(w);
				return w;
			}
			static void build_W(Matrix_3x3& w )
			{
				w(0,0) = 0;    w(0,1) = -1;    w(0,2) = 0;
				w(1,0) = 1;    w(1,1) =  0;    w(1,2) = 0;
				w(2,0) = 0;    w(2,1) =  0;    w(2,2) = 1;
			}
		};
	}//namespace Fm2Essential;
}//namespace Basics3D
////////////////////////////////////////////////////////////////////////////////////////////
#endif //_SFM_CALIBRATION_H_
