//file #include "sfm_export_3D_model.h" 
#ifndef _SFM_EXPORT_3D_MODEL_H_
#define _SFM_EXPORT_3D_MODEL_H_
///////////////////////////////////////////////////////////////////////////////////////////////
#include "num_linalg.h"
#include "sfm_quaternion.h" 
////////////////////////////////////////////////////////////////////////////////////////
namespace Basics3D
{
namespace Structure
{
	///      3D .  VRML       ,
	///            .
	template<typename T>
	static typename Types<T>::Vector_3 FindTranslationToObjectCenterOfMass(const std::vector<Pt<T> >& points)
	{
		typedef typename Types<T>::Vector_3 Vector_3;
		typedef typename Types<T>::value_type value_type;
		using namespace boost::numeric::ublas;
		Vector_3 cm; //object center of mass
		cm=zero_vector<value_type>(3);
		{
			size_t count=0;
			for (size_t i = 0; i < points.size() ; ++i)
			{
				if(points[i].flag == PointStatus::Ok)
				{
					cm+=points[i].p3D;
					++count;
				}
			}
			cm/=count;
		}
		return cm;
	}
	///      3D .  VRML       ,
	///            .
	template<typename T>
	static T FindCameraQuaternion(const typename Types<T>::Matrix_3x3& R,typename Types<T>::Vector_3& direction)
	{
		typedef typename Types<T>::Vector_3 Vector_3;
		typedef typename Types<T>::value_type value_type;
		typedef typename Types<value_type>::Matrix_3x3 Matrix_3x3;
		value_type angle=3.14;
		{
			Quaternion<value_type> qR(R);
			{//Temporary test - enlarge finded angle
				Vector_3 v;
				Quaternion<value_type> tmp(qR);
				v=tmp.toRot(angle);
				std::cout << "Quaternion angle=" << angle*180/M_PI <<std::endl;
				//angle*=4;
				//angle=-angle;
				v(0)=-v(0); v(1)=-v(1); v(2)=v(2);
				qR=Quaternion<value_type>(v,angle);
			}
			Vector_3 v;
			v(0)=1;v(1)=0;v(2)=0;
			Quaternion<value_type> qV(v,M_PI);
			//    3D  qR ,      VRML qV , .. qV * qR . 
			Quaternion<value_type> qRes(qV*qR);
			//Quaternion<value_type> qRes=qV*qR;
			direction=qRes.toRot(angle);
			direction=v;
			return angle;
		}
	}


	///
	template<typename T>
	static void SaveVRML_AsPointsSet(const char* filename,const std::vector<Pt<T> >& points,const std::vector<CameraPos<T> >& cam)
	{
		typedef typename Types<T>::Vector_3 Vector_3;
		typedef typename Types<T>::value_type value_type;
		typedef typename Types<value_type>::Matrix_3x3 Matrix_3x3;

		std::cout << "VRML:" << std::endl;
		for(size_t i=0; i<cam.size(); ++i)
		{
			std::cout << "cam # " << i << std::endl;
			std::cout << "R = " << cam[i].R << std::endl;
			std::cout << "t = " << cam[i].t << std::endl;
			std::cout << "focus= " << cam[i].focus << std::endl;
		}

		//object center of mass
		Vector_3 cm=FindTranslationToObjectCenterOfMass<value_type>(points); 
		std::cout << "Center of mass =" << cm << std::endl;
		Matrix_3x3 R2=cam[1].R;
		Vector_3 direction;
		value_type angle=FindCameraQuaternion<value_type>(R2,direction);
		std::cout << "angle =" << angle << std::endl;
		std::cout << "direction =" << direction << std::endl;

		std::ofstream out(filename);

		//std::string navig="EXAMINE";
		std::string navig="FLY";
		out 
			<< "#VRML V2.0 utf8" << std::endl << std::endl
			<< "NavigationInfo {" << std::endl
//			<< " headlight FALSE" << std::endl
			<< " type \"" << navig << "\"" << std::endl
			<< "}" << std::endl
			<< "Viewpoint" << std::endl
			<< "{" << std::endl
			<< "  position  " << cam[0].t(0)-cm(0) << " " << cam[0].t(1)-cm(1) << " " << cam[0].t(2)-cm(2) << " "  << std::endl
			<< "  orientation 1 0 0 3.1415926" << std::endl
			<< "  fieldOfView  0.985398 " << std::endl
			<< "  description \"1\"" << std::endl
			<< "}"  << std::endl << std::endl
			<< "Viewpoint" << std::endl
			<< "{" << std::endl
			<< "  position  " << cam[1].t(0)-cm(0) << " " << cam[1].t(1)-cm(1) << " " << cam[1].t(2)-cm(2) << " "  << std::endl
			<< "  orientation " << direction(0) << " "<< direction(1) << " "<< direction(2) << " "<< angle  << std::endl
			<< "  fieldOfView  0.985398 " << std::endl
			<< "  description \"2\"" << std::endl
			<< "}"  << std::endl << std::endl
			////////////////////////////////////////
			<< "DEF	Point Shape" << std::endl
			<< "          {" << std::endl
			<< "            geometry Sphere {radius 0.01}" << std::endl
			<< "            appearance Appearance" << std::endl
			<< "            {" << std::endl
			<< "              material Material" << std::endl
			<< "              {" << std::endl
			<< "                diffuseColor     0.8 0.55 0.4" << std::endl
			<< "                shininess        0.3" << std::endl
			<< "                specularColor    0 0 0.2" << std::endl
			<< "                emissiveColor    0.1 0.1 0.1" << std::endl
			<< "                ambientIntensity 0.5" << std::endl
			<< "                transparency     0.0" << std::endl
			<< "              }" << std::endl
			<< "            }" << std::endl
			<< "          }" << std::endl << std::endl;

		for (size_t i = 0; i < points.size() ; ++i)
		{
			if(points[i].flag == PointStatus::Ok)
			{
				out 
					<< "Transform { translation " << -(points[i].p3D(0)-cm(0)) << " "
					<< -(points[i].p3D(1)-cm(0)) << " " << (points[i].p3D(2)-cm(2)) << std::endl
					<< "            children [ USE Point ]}" << std::endl;
			}
		}
	
		out.close();
		out.flush();
	}
}//namespace Structure
}//namespace Basics3D
///////////////////////////////////////////////////////////////////////////////////////////////
#endif //_SFM_EXPORT_3D_MODEL_H_
