// file: #include "sfm_quaternion.h" 
#ifndef _SFM_QUATERNION_H 
#define _SFM_QUATERNION_H 
////////////////////////////////////////////////////////////////////////
// TEMPORARY! DO NOT USE! ONLY FOR TESTS WITH VRML  ////////////////////
////////////////////////////////////////////////////////////////////////
#include "vtl_meta_base.h"
#include "num_linalg.h"
////////////////////////////////////////////////////////////////////////
template<typename T>
struct Quaternion
{
	typedef typename boost::numeric::ublas::c_matrix<T,3,3> RotationMatrix;
	typedef typename boost::numeric::ublas::c_vector<T,3>  Vector3;
	T q[4];
public:
	Quaternion() {}
	Quaternion(const Quaternion & p) { (*this)=p;}
	explicit Quaternion(T q0,T q1,T q2,T q3) { q[0]=q0; q[1]=q1; q[2]=q2; q[3]=q3;}
	Quaternion& operator=(const Quaternion & p) { q[0]=p.q[0]; q[1]=p.q[1]; q[2]=p.q[2]; q[3]=p.q[3]; return *this;}
	/// Construct from ROTATION MATRIX
	explicit Quaternion(const RotationMatrix  R)
	{
		T trace4 = 1 + R(0,0) + R(1,1) + R(2,2);
		if(trace4 > 0.00000001 )
		{
			T s=2*sqrt(trace4);
			q[0]=s/4;
			q[1]=(R(1,2)-R(2,1))/s;
			q[2]=(R(2,0)-R(0,2))/s;
			q[3]=(R(0,1)-R(1,0))/s;
		}
		else if( R(0,0) > R(1,1) && R(0,0) > R(2,2) )  
		{	// Column 0: 
			T s= sqrt( 1.0 + R(0,0) - R(1,1) - R(2,2) ) * 2;
			q[1] = 0.25 * s;
			q[2] = (R(0,1) + R(1,0) ) / s;
			q[3] = (R(2,0) + R(0,2) ) / s;
			q[0] = (R(1,2) - R(2,1) ) / s;
		} 
		else if( R(1,1) > R(2,2) ) 
		{			// Column 1: 
			T s= sqrt( 1.0 + R(1,1) - R(0,0) - R(2,2) ) * 2;
			q[1] = (R(0,1) + R(1,0) ) / s;
			q[2] = 0.25 * s;
			q[3] = (R(1,2) + R(2,1) ) / s;
			q[0] = (R(2,0) - R(0,2) ) / s;
		}
		else 
		{						// Column 2:
			T s= sqrt( 1.0 + R(2,2) - R(0,0) - R(1,1) ) * 2;
			q[1] = (R(2,0) + R(0,2) ) / s;
			q[2] = (R(1,2) + R(2,1) ) / s;
			q[3] = 0.25 * s;
			q[0] = (R(0,1) - R(1,0) ) / s;
		}

	}
	T Modul() const
	{
		return sqrt(q[0]*q[0] + q[1]*q[1] + q[2]*q[2] + q[3]*q[3]);
	}
	void Normalize()
	{
		T s= Modul();
		q[0]/=s;
		q[1]/=s;
		q[2]/=s;
		q[3]/=s;
	}
	/// Construct from ROTATION MATRIX
	operator RotationMatrix()
	{
		Normalize();
		RotationMatrix R;
		R(0,0)  = 1 - 2 * ( q[2]*q[2] + q[3]*q[3] );
		R(1,0)  =     2 * ( q[1]*q[2] - q[3]*q[0] );
		R(2,0)  =     2 * ( q[1]*q[3] + q[2]*q[0] );
		R(0,1)  =     2 * ( q[1]*q[2] + q[3]*q[0] );
		R(1,1)  = 1 - 2 * ( q[1]*q[1] + q[3]*q[3] );
		R(2,1)  =     2 * ( q[2]*q[3] - q[1]*q[0] );
		R(0,2)  =     2 * ( q[1]*q[3] - q[2]*q[0] );
		R(1,2)  =     2 * ( q[2]*q[3] + q[1]*q[0] );
		R(2,2)  = 1 - 2 * ( q[1]*q[1] + q[2]*q[2] );
		return R;
	}
	Quaternion operator*(const Quaternion& p) const 
	{
		Quaternion qp;
		qp.q[0] = q[0]*p.q[0] - q[1]*p.q[1] - q[2]*p.q[2] - q[3]*p.q[3];
		qp.q[1] = q[0]*p.q[1] + q[1]*p.q[0] + q[2]*p.q[3] - q[3]*p.q[2];
		qp.q[2] = q[0]*p.q[2] + q[2]*p.q[0] + q[3]*p.q[1] - q[1]*p.q[3];
		qp.q[3] = q[0]*p.q[3] + q[3]*p.q[0] + q[1]*p.q[2] - q[2]*p.q[1];
		return qp;
	}
	Quaternion operator*(T cof) const 
	{
		Quaternion p(*this);
		p.q[0] *= cof ;
		p.q[1] *= cof ;
		p.q[2] *= cof ;
		p.q[3] *= cof ;
		return p;
	}
	Quaternion operator/(T cof) const 
	{
		Quaternion p(*this);
		return p * (1/cof) ;
	}
	Quaternion inverse() const 
	{
		Quaternion qi(*this);
		qi/=Modul();
		qi.q[1]=-qi.q[1];
		qi.q[2]=-qi.q[2];
		qi.q[3]=-qi.q[3];
		return qi;
	}
	Vector3 toRot(T& angle)
	{
		using namespace boost::numeric::ublas;
		Vector3 axis;
		Normalize();
		T cos_a = q[0];
		angle = acos( cos_a ) * 2;
		T sin_a = sqrt( 1.0 - cos_a * cos_a );
		if ( abs( sin_a ) < 0.0005 ) sin_a = 1;
		axis(0) = q[1] / sin_a;
		axis(1) = q[2] / sin_a;
		axis(2) = q[3] / sin_a;
		return axis;
	}
	explicit Quaternion(const Vector3& axis_,T angle)
	{
		using namespace boost::numeric::ublas;
		Vector3 axis=axis_;
		axis /= sqrt(blas_1::dot(axis,axis));
		T sin_a = sin( angle / 2 );
		T cos_a = cos( angle / 2 );
		axis *= sin_a;
		q[1]  = axis(0);
		q[2]  = axis(1);
		q[3]  = axis(2);
		q[0]  = cos_a;
	}
};
////////////////////////////////////////////////////////////////////////
#endif _SFM_QUATERNION_H 
