// file: #include "sfm_structure_from_2_views.h"
#ifndef _SFM_RECONSRUCTION_H 
#define _SFM_RECONSRUCTION_H
////////////////////////////////////////////////////////////////////////////////////////
#include "sfm_types.h"
#include "sfm_calibration.h"
#include "sfm_reconsruction_dlt.h"
#include "num_optimization.h"

#include "Point2D.h"
#include "Points2DSet.h"
////////////////////////////////////////////////////////////////////////////////////////
namespace Basics3D
{
namespace Structure
{
	/// 
	template<typename T>
	struct Pt
	{
		typename Types<T>::Vector_3 p3D;
		typename Types<T>::Vector_2 p1;
		typename Types<T>::Vector_2 p2;
		PointStatus::Flags flag;
		int id;
	};
	template<typename T>
	struct CameraPos
	{
		typename Types<T>::Vector_3  t;
		typename Types<T>::Matrix_3x3 R;
		typename Types<T>::value_type focus;
		int id;
	};
	///
	template<typename T>
	class Comp3D
	{
	public:
		typedef typename Types<T>::value_type value_type;
		typedef typename Types<T>::Vector_2 Vector_2;
		typedef typename Types<T>::Vector_3 Vector_3;
		typedef typename Types<T>::Vector_4 Vector_4;
	private:
		const T_Points2DSet<C_Point2D> &points2D;
		Structure::Dlt<value_type>& dlt;
	public:
		std::vector<Pt<value_type> > points;
		Comp3D(const T_Points2DSet<C_Point2D> &in_points2D,Structure::Dlt<value_type>& dlt_)
			: points2D(in_points2D), dlt(dlt_)
		{
			int sz = points2D.Size();
			points.resize(sz);
			// x^ = K^-1 * x; x^' = K'^-1 * x';
			for (int i = 0; i < sz; i++)
			{
				points[i].p1(0)=points2D[2*i].X();
				points[i].p1(1)=points2D[2*i].Y();
				points[i].p2(0)=points2D[2*i+1].X();
				points[i].p2(1)=points2D[2*i+1].Y();
				points[i].id=i;
			}
		}
		void operator()()
		{
			size_t count=0;
			Vector_4 X;
			
			size_t rep_cnt=0;
			size_t rep_sum=0;
			const value_type rep_mult=10000;


			for (size_t i = 0; i < points.size(); i++)
			{
				Pt<value_type> & pt=points[i];
				dlt(pt.p1,pt.p2,X);
				homogeneous2euclidean(X,pt.p3D);

				value_type dev = dlt.reProjectionStdDev(pt.p1,pt.p2,X);
				if(pt.p3D(2)<0 )
					pt.flag = PointStatus::Behind;
				else if( pt.p3D(2) > 50)
					pt.flag = PointStatus::Uncertain;
				else
				{
					pt.flag = PointStatus::Ok;
					++count;
				}
				if(pt.flag==PointStatus::Ok)
				{
					++rep_cnt;
					rep_sum+=size_t(dev*rep_mult);
				}

				std::cout << "3D[ " << i << " ] =  " << pt.p3D << "    flag=" << pt.flag << "  reproj_err= " << dev << std::endl;
			}
			std::cout << " OK points " << count << " from " << points.size() << "    %=" << double(count)/points.size()*100 << std::endl;

			value_type dev_mean=(rep_sum/rep_mult)/rep_cnt;
			std::cout << " dev_mean " << dev_mean  << std::endl;
			{//Filter points with great reprojection error
				value_type stddev_thresh=2*dev_mean;
				rep_cnt=0;
				Vector_4 X;
				for (size_t i = 0; i < points.size(); i++)
				{
					Pt<value_type> & pt=points[i];
					if(pt.flag==PointStatus::Ok)
					{
						euclidean2homogeneous(pt.p3D,X);
						value_type dev = dlt.reProjectionStdDev(pt.p1,pt.p2,X);
						if(dev > stddev_thresh)
							pt.flag=PointStatus::BadReproj;
						else
							++rep_cnt;
					}
				}
				std::cout << " Filter by reproj error:  Points remains = " << rep_cnt << "    %=" << double(rep_cnt)/points.size()*100 << std::endl;
			}

		}
		void homogeneous2euclidean(const Vector_4& v,Vector_3& result)
		{
			using namespace boost::numeric::ublas;
//			result=boost::numeric::ublas::vector_slice<Vector_4>(v, slice(0,1,3));
			//result=vs;
			result(0)=v(0);
			result(1)=v(1);
			result(2)=v(2);
			result/=v(3);
		}
		void euclidean2homogeneous(const Vector_3& v,Vector_4& result)
		{
			using namespace boost::numeric::ublas;
//			result=boost::numeric::ublas::vector_slice<Vector_4>(v, slice(0,1,3));
			//result=vs;
			result(0)=v(0);
			result(1)=v(1);
			result(2)=v(2);
			result(3)=1;
		}
	};
	///      \ref  Lit_HartleyZissermanBook   
	///  (19.16) .467.     Pi    Ki  ,
	///      , AspectRatio =1      .
	///    p=(p1,p2,p3)^T.
	template<typename T>
	class ProjectiveToMetric
	{
	public:
		typedef typename Types<T>::value_type value_type;
//		typedef typename Types<T>::Vector_2 Vector_2;
		typedef typename Types<T>::Vector_3 Vector_3;
		typedef typename Types<T>::Vector_4 Vector_4;
		typedef typename Types<T>::Matrix_3x3 Matrix_3x3;
		typedef typename Types<T>::Matrix_3x4 Matrix_3x4;
		typedef typename Types<T>::Matrix_4x4 Matrix_4x4;
	private:
		const value_type focus_sqr;
		const Matrix_3x4& P1;
		const Matrix_3x4& P2;
//		Matrix_3x3 KKT;
	public:
		ProjectiveToMetric(value_type focus, Matrix_3x4& P1_,Matrix_3x4& P2_)
			: focus_sqr(Vtl::sqr(focus)),P1(P1_),P2(P2_)
		{
		}
/// 1 -  p, 0 -  p  f .
#if 1
		void operator()(Matrix_4x4& H)
		{
			std::cout << "-------------------------"  << std::endl;
			std::cout << "class ProjectiveToMetric:"  << std::endl;
			std::cout << std::setiosflags(std::ios::scientific);
			//|std::ios::showpos

			class Func
			{
				const ProjectiveToMetric& pm;
				const value_type focus_sqr;
			public:
				Func(const ProjectiveToMetric& pm_,value_type focus): pm(pm_), focus_sqr(Vtl::sqr(focus)) {}
				value_type  operator()(const Vector_3& p) const
				{
					return 1.0e+08*pm.build_Functional(p,focus_sqr);
				}
				Vector_3 operator()(const Vector_3& p,Loki::Int2Type<1> ) const
				{
					Vector_3 grad;
					grad(0)=1.0e+08*pm.build_Functional_derivative<0>(p,focus_sqr);
					grad(1)=1.0e+08*pm.build_Functional_derivative<1>(p,focus_sqr);
					grad(2)=1.0e+08*pm.build_Functional_derivative<2>(p,focus_sqr);
					return grad;
				}
			};

			// Set initial point p

			Vector_3 p;
			p(0)=0;  p(1)=0;  p(2)=0;
			Func func(*this,sqrt(focus_sqr));
			Vector_3 grad;

			{// debug print function value and gradient at 
				std::cout << "p= " << p << std::endl;
				value_type f_value;
				f_value=func(p);
				std::cout << "f_value= " << f_value << std::endl;
				std::cout << "focus= " << sqrt(focus_sqr) << std::endl;
				grad=func(p,Loki::Int2Type<1>() );
				std::cout << "grad= " << grad << std::endl;
			}
			

			value_type focus=sqrt(focus_sqr);
			value_type gtol=1.0e-20;
			int iter;
			value_type result = Vtl::NumAlg::Opt::nrc_dfpmin(p,gtol,iter,func);
			
			{
				std::cout << "After dfpmin :" << std::endl;
				std::cout << "p = " << p << std::endl;
				std::cout << "f_value = " << result << std::endl;
				std::cout << "focus= " << focus << std::endl;
				grad=func(p,Loki::Int2Type<1>() );
				std::cout << "grad= " << grad << std::endl;
				std::cout << "iter = " << iter << std::endl;
			}
			build_H(H,p,focus);
		}
#else
		void operator()(Matrix_4x4& H)
		{
			std::cout << "-------------------------"  << std::endl;
			std::cout << "class ProjectiveToMetric:"  << std::endl;
			std::cout << std::setiosflags(std::ios::scientific);
			//|std::ios::showpos

			class Func4D
			{
				const ProjectiveToMetric& pm;
				const value_type focus_sqr;
			public:
				Func4D(const ProjectiveToMetric& pm_,value_type focus): pm(pm_), focus_sqr(Vtl::sqr(focus)) {}
				value_type  operator()(const Vector_4& p_f) const
				{
					using namespace boost::numeric::ublas;
					Vector_3 p = subrange(p_f, 0, 3);
					return pm.build_Functional(p,p_f(3));
				}
				Vector_4 operator()(const Vector_4& p_f,Loki::Int2Type<1> ) const
				{
					Vector_4 grad;
					Vector_3 p = subrange(p_f, 0, 3);
					grad(0)=pm.build_Functional_derivative<0>(p,focus_sqr);
					grad(1)=pm.build_Functional_derivative<1>(p,focus_sqr);
					grad(2)=pm.build_Functional_derivative<2>(p,focus_sqr);
					grad(3)=pm.build_Functional_derivative_d_focus_sqr(p,focus_sqr);
					return grad;
				}
			};

			// Set initial point p


			Vector_4 p;
			p(0)=0;  p(1)=0;  p(2)=0;  p(3)=focus_sqr;
			Func4D func(*this,sqrt(focus_sqr));
			Vector_4 grad;
			
			{// debug print function value and gradient at 
				std::cout << "p= " << p << std::endl;
				value_type f_value;
				f_value=func(p);
				std::cout << "f_value= " << f_value << std::endl;
				std::cout << "focus= " << sqrt(focus_sqr) << std::endl;
				grad=func(p,Loki::Int2Type<1>() );
				std::cout << "grad= " << grad << std::endl;
			}
			

			value_type focus=sqrt(focus_sqr);
			value_type gtol=1.0e-20;
			int iter;
			value_type result = Vtl::NumAlg::Opt::nrc_dfpmin(p,gtol,iter,func);
			focus=sqrt(p(3));
			
			{
				std::cout << "After dfpmin :" << std::endl;
				std::cout << "p = " << p << std::endl;
				std::cout << "f_value = " << result << std::endl;
				std::cout << "focus= " << focus << std::endl;
				grad=func(p,Loki::Int2Type<1>() );
				std::cout << "grad= " << grad << std::endl;
				std::cout << "iter = " << iter << std::endl;
			}
			Vector_3 pp = subrange(p, 0, 3);
			build_H(H,pp,focus);
		}
#endif
	private:
		value_type build_Functional(const Vector_3& p,value_type focus_sqr) const
		{
			//std::cout << " build_Functional= " << std::endl ;

			Matrix_4x4 Q = build_Qoo(p,focus_sqr);

			Matrix_3x3 KKT;
			build_KKT(KKT,focus_sqr);
			
			//std::cout << " Q= " << std::endl ;
			//std::cout << Dbg::Fmt<Matrix_4x4,Dbg::Fmt<Matrix_4x4>::RowsAndCols>(Q) <<std::endl; 
			
			// P1 * Qoo * P1^T - KKT
			Matrix_3x3 b = build_PQPT_KKT(Q,P1,KKT);
			value_type result = frobenius_norm_sqr(b);
			
			//std::cout << " P1 * Q * P1^T - KKT = " << std::endl ;
			//std::cout << Dbg::Fmt<Matrix_3x3,Dbg::Fmt<Matrix_3x3>::RowsAndCols>(b) <<std::endl; 
			//std::cout << "frob = " << result <<std::endl; 
			
			// P2 * Qoo * P2^T - KKT
			b = build_PQPT_KKT(Q,P2,KKT);
			result += frobenius_norm_sqr(b);
			
			//std::cout << " P2 * Q * P2^T - KKT = " << std::endl ;
			//std::cout << Dbg::Fmt<Matrix_3x3,Dbg::Fmt<Matrix_3x3>::RowsAndCols>(b) <<std::endl; 
			//std::cout << "frob = " << result <<std::endl; 
			
			return result;
		}
		//    p(i)   p. i=0..2
		template<int i>
		value_type build_Functional_derivative(const Vector_3& p,value_type focus_sqr)  const
		{
			//std::cout << " d build_Functional / d p" << i << std::endl ;
			Matrix_4x4 Q = build_Qoo(p,focus_sqr);
			Matrix_4x4 dQ = build_Qoo_d_p<i>(p,focus_sqr);
			Matrix_3x3 KKT;
			build_KKT(KKT,focus_sqr);

			// P1 * Qoo * P1^T - KKT
			Matrix_3x3 a = build_PQPT_KKT(Q,P1,KKT);
			// d/d(p_i) { P1 * Qoo * P1^T - KKT ) = P1 * d(Qoo)/d(p_i) * P1^T 
			Matrix_3x3 b = build_PQPT(dQ,P1);
			value_type result = frobenius_norm_sqr_derivative(a,b);
			// P2 * Qoo * P2^T - KKT
			a = build_PQPT_KKT(Q,P2,KKT);
			// d/d(p_i) { P2 * Qoo * P2^T - KKT ) = P2 * d(Qoo)/d(p_i) * P2^T 
			b = build_PQPT(dQ,P2);
			result += frobenius_norm_sqr_derivative(a,b);
			return result;
		}
		//    focus_sqr .
		value_type build_Functional_derivative_d_focus_sqr(const Vector_3& p,value_type focus_sqr)  const
		{
			//std::cout << " d build_Functional / d p" << i << std::endl ;
			Matrix_4x4 Q = build_Qoo(p,focus_sqr);
			Matrix_4x4 dQ = build_Qoo_d_focus_sqr(p);
			Matrix_3x3 KKT;
			build_KKT(KKT,focus_sqr);
			Matrix_3x3 dKKT;
			build_KKT_d_focus_sqr(dKKT);

			// P1 * Qoo * P1^T - KKT
			Matrix_3x3 a = build_PQPT_KKT(Q,P1,KKT);
			// d/d(f^2) { P1 * Qoo * P1^T - KKT ) = P1 * d(Qoo)/d(f^2) * P1^T  - d(KKT)/d(f^2)
			Matrix_3x3 b = build_PQPT_KKT(dQ,P1,dKKT); 
			value_type result = frobenius_norm_sqr_derivative(a,b);
			// P2 * Qoo * P2^T - KKT
			a = build_PQPT_KKT(Q,P2,KKT);
			// d/d(f^2) { P2 * Qoo * P2^T - KKT ) = P2 * d(Qoo)/d(f^2) * P2^T  - d(KKT)/d(f^2)
			b = build_PQPT_KKT(dQ,P2,dKKT); 
			result += frobenius_norm_sqr_derivative(a,b);
			return result;
		}
		// build P * Qoo * P^T - KKT
		Matrix_3x3 build_PQPT_KKT(const Matrix_4x4& Q,const Matrix_3x4& P,const Matrix_3x3& KKT) const
		{
			Matrix_3x3 b = build_PQPT(Q,P);
			b -= KKT;
			return b;
		}
		// build P * Qoo * P^T 
		static Matrix_3x3 build_PQPT(const Matrix_4x4& Q,const Matrix_3x4& P)
		{
			Matrix_3x4 a = prod(P,Q);
			Matrix_3x3 b = prod(a,trans(P));
			return b;
		}
		static Matrix_4x4 build_Qoo(const Vector_3& p,value_type focus_sqr) 
		{
			Matrix_4x4 Q;
			build_Qoo(Q,p,focus_sqr);
			return Q;
		}
		static void build_Qoo(Matrix_4x4& Q,const Vector_3& p,value_type focus_sqr) 
		{
			using namespace boost::numeric::ublas;
			Q=identity_matrix<value_type>(4,4);
			Q(0,0) = Q(1,1) = focus_sqr;
			Q(0,3) = Q(3,0) = -focus_sqr*p(0);
			Q(1,3) = Q(3,1) = -focus_sqr*p(1);
			Q(2,3) = Q(3,2) = -p(2);
			Q(3,3) = focus_sqr * ( Vtl::sqr(p(0)) + Vtl::sqr(p(1)) ) + Vtl::sqr(p(2));
		}
		//    Qoo  p(i)   p. i=0..2
		template<int i>
		static Matrix_4x4 build_Qoo_d_p(const Vector_3& p,value_type focus_sqr)
		{
			Matrix_4x4 Qd;
			build_Qoo_d_p<i>(Qd,p,focus_sqr);
			return Qd;
		}
		//    Qoo  p(i)   p. i=0..2
		template<int i>
		static void build_Qoo_d_p(Matrix_4x4& Qd,const Vector_3& p,value_type focus_sqr)
		{
			using namespace boost::numeric::ublas;
			Qd=zero_matrix<value_type>(4,4);
			value_type t=((i<2)?focus_sqr:1);
			Qd(i,3)=Qd(3,i)=-t;
			Qd(3,3)=2*t*p(i);
		}
		//    Qoo  focus_sqr 
		static Matrix_4x4 build_Qoo_d_focus_sqr(const Vector_3& p)
		{
			Matrix_4x4 Qd;
			build_Qoo_d_focus_sqr(Qd,p);
			return Qd;
		}
		//    Qoo  focus_sqr 
		static void build_Qoo_d_focus_sqr(Matrix_4x4& Qd,const Vector_3& p)
		{
			using namespace boost::numeric::ublas;
			Qd=zero_matrix<value_type>(4,4);
			Qd(0,0) = Qd(1,1) = 1;
			Qd(0,3) = Qd(3,0) = -p(0);
			Qd(1,3) = Qd(3,1) = -p(1);
			Qd(3,3) = Vtl::sqr(p(0)) + Vtl::sqr(p(1));
		}
		static void build_KKT(Matrix_3x3& KKT,value_type focus_sqr)
		{
			using namespace boost::numeric::ublas;
			KKT = identity_matrix<value_type>(3,3);
			KKT(0,0)=KKT(1,1)=focus_sqr;
		}
		//    KKT  focus_sqr 
		static void build_KKT_d_focus_sqr(Matrix_3x3& KKT)
		{
			using namespace boost::numeric::ublas;
			KKT = identity_matrix<value_type>(3,3);
			KKT(2,2)=0;
		}
		static value_type frobenius_norm_sqr(const Matrix_3x3& A)
		{
			value_type r=0;
			for(int i=0; i<3; ++i)
			for(int j=0; j<3; ++j)
				r+=Vtl::sqr(A(i,j));
			return r;
		}
		static value_type frobenius_norm_sqr_derivative(const Matrix_3x3& A,const Matrix_3x3& A_deriv)
		{
			value_type r=0;
			for(int i=0; i<3; ++i)
			for(int j=0; j<3; ++j)
				r+=A(i,j)*A_deriv(i,j);
			return 2*r;
		}
		static void build_H(Matrix_4x4& H,const Vector_3& p,value_type focus)
		{
			using namespace boost::numeric::ublas;
			Matrix_3x3 K(identity_matrix<value_type>(3,3));
			K(0,0)=K(1,1)=focus;
			Vector_3 v = prod(p,K);
			H=identity_matrix<value_type>(4,4);
			H(0,0)=H(1,1)=focus;
			H(3,0)=-v(0);
			H(3,1)=-v(1);
			H(3,2)=-v(2);
		}
	};
}//namespace Structure
}//namespace Basics3D
////////////////////////////////////////////////////////////////////////////////////////
#endif //_SFM_RECONSRUCTION_H
