#ifndef _SFM_RECONSRUCTION_DLT_H
#define _SFM_RECONSRUCTION_DLT_H 
////////////////////////////////////////////////////////////////////////////////////////////
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <functional>      // For greater<int>( )
#include "loki_typeManip.h"
//#include "loki_Typelist.h"
#include "loki_static_check.h"
#include "num_linalg.h"
#include "alexandresku.h" 
#include "vtl_meta_base.h" 
#include "sfm_dbg.h" 
////////////////////////////////////////////////////////////////////////////////////////////
//template<typename T> struct Fmt;

namespace Basics3D
{
namespace Structure
{
	template<typename T> 
	struct Dlt
	{
		typedef T value_type;
		typedef boost::numeric::ublas::c_matrix<value_type,3,4> MatrixP;
		typedef boost::numeric::ublas::c_vector<value_type,4> Vector4;
		typedef boost::numeric::ublas::c_vector<value_type,3> Vector3;
		typedef boost::numeric::ublas::c_vector<value_type,2> Vector2;
	private:
		const MatrixP& P1;
		const MatrixP& P2;
		const Vector4 p1_1;
		const Vector4 p1_2;
		const Vector4 p1_3;
		const Vector4 p2_1;
		const Vector4 p2_2;
		const Vector4 p2_3;
		enum {M=4};
		enum {N=4};
		typedef boost::numeric::ublas::c_matrix<T,M,N> MatrixA;
		MatrixA A;
	public:
		Dlt(const MatrixP& P1_,const MatrixP& P2_)
			: P1(P1_)
			 ,P2(P2_)
			 ,p1_1(boost::numeric::ublas::row(P1_,0))
			 ,p1_2(boost::numeric::ublas::row(P1_,1))
			 ,p1_3(boost::numeric::ublas::row(P1_,2))
			 ,p2_1(boost::numeric::ublas::row(P2_,0))
			 ,p2_2(boost::numeric::ublas::row(P2_,1))
			 ,p2_3(boost::numeric::ublas::row(P2_,2))
		{
			using namespace boost::numeric::ublas;
			//std::cout << "p1_1 = "<<p1_1 << std::endl;
			//std::cout << "p1_2 = "<<p1_2 << std::endl;
			//std::cout << "p1_3 = "<<p1_3 << std::endl;
			//std::cout << "p2_1 = "<<p2_1 << std::endl;
			//std::cout << "p2_2 = "<<p2_2 << std::endl;
			//std::cout << "p2_3 = "<<p2_3 << std::endl;
		}
		void operator()(const Vector2& p1,const Vector2& p2,Vector4& X)
		{
			using namespace boost::numeric::ublas;
			using namespace Vtl::NumAlg::Lin;
			build_A(p1(0),p1(1),p2(0),p2(1));
			//std::cout << "A = " << A <<std::endl;
			//Matrix U=FmTools<value_type>::fm2essential(K1,F,K1);
			c_matrix<value_type,N,N> V;
			Vector4 W;
			Vector4 rv1;
			svd::decomp(A,W,V,rv1,true,true);
			
			Vtl::NumAlg::Implicit::Sort<Vector4,Vtl::NumAlg::Implicit::Pred::gt<Vector4> > s(W);
			//Vtl::NumAlg::Implicit::Pred::gt<Vector4> decrement_order(W);
			//Vtl::NumAlg::Implicit::Sort<Vector4> s(W,decrement_order);
			X=column( V, s[N-1] );
		}
		value_type  reProjectionStdDev(const Vector2& p1,const Vector2& p2,const Vector4& X) const
		{
			using namespace boost::numeric::ublas;
			Vector2 pt_rep;
			pt_rep = toEuclidean(prod(P1,X));
			value_type d1=distEuclidean(p1,pt_rep); 
			pt_rep = toEuclidean(prod(P2,X));
			value_type d2=distEuclidean(p2,pt_rep); 
			return std::max(d1,d2);
		}
	private:
		static Vector2 toEuclidean(const Vector3& pt) 
		{
			Vector2 euc;
			euc(0)=pt(0)/pt(2);
			euc(1)=pt(1)/pt(2);
			return euc;
		}
		static value_type distEuclidean(const Vector2& pt1,const Vector2& pt2) 
		{
			return std::sqrt(Vtl::sqr(pt1(0)-pt2(0))+Vtl::sqr(pt1(1)-pt2(1)));
		}
		void build_A(value_type x1,value_type y1,value_type x2,value_type y2)
		{
			using namespace boost::numeric::ublas;
			matrix_row<MatrixA> a0( A, 0 );
			a0= x1 * p1_3-p1_1;
			matrix_row<MatrixA> a1( A, 1 );
			a1= y1 * p1_3-p1_2;
			matrix_row<MatrixA> a2( A, 2 );
			a2= x2 * p2_3-p2_1;
			matrix_row<MatrixA> a3( A, 3 );
			a3= y2 * p2_3-p2_2;
		}
	};//class Dlt
}//namespace Structure
}//namespace Basics3D
////////////////////////////////////////////////////////////////////////////////////////////
#endif //_SFM_RECONSRUCTION_DLT_H
