#ifndef _EDGE_DET_H
#define _EDGE_DET_H
////////////////////////////////////////////////////////////////////////////////////////////////////
#include "convolve.h"
#include "imalg_tools.h"
#include "plf_functors.h"
#include "plf_kernels_1d.h"
////////////////////////////////////////////////////////////////////////////////////////////////////
namespace Students
{
namespace Curve
{
	template<typename T>
	class EdgeDetect
	{
		//   (     )
		typedef Vtl::ImAlg::WndFun::Gauss<T,0, true> FF0; 
		typedef Vtl::Plf::Kernel::FKern<T,Vtl::Plf::Kernel::KT_Conv,true> K0;
		//1 
		typedef Vtl::ImAlg::WndFun::Gauss<T,1,true> FF1; 
		typedef Vtl::Plf::Kernel::FKern<T,Vtl::Plf::Kernel::KT_Conv,true> K1;

		const float sigma;
		const float n_sigmas_in_eviron;
		//- 
		FF0 f0;
		K0 ker0;
		Vtl::ImAlg::Conv::ConvInplace<T,K0> g0;
		// 1 
		FF1 f1;
		K1 ker1;
		Vtl::ImAlg::Conv::ConvInplace<T,K1> g1;

	public:
		EdgeDetect(float sigma_,float n_sigmas_in_eviron_=4.0f) 
			: sigma(sigma_), n_sigmas_in_eviron(n_sigmas_in_eviron_)
			, f0(sigma,n_sigmas_in_eviron),ker0(f0),g0(ker0),f1(sigma,n_sigmas_in_eviron),ker1(f1),g1(ker1)
		{
		}
		void Canny(int w,int h,int nls,T* edge,T* gx,T* gy,T* fri)
		{
			const int sz=w*h;

			// 1 ,        gx,gy, 
			//     edge,    inplace   -   .
			// Gx :
			//std::copy(edge,edge+sz,gx);
			//Vtl::ImAlg::Conv::Conv2D::conv2DInplace(gx,w,h,w,g1,g0);
			Vtl::ImAlg::Conv::Conv2D::conv2D(edge,w,h,nls,gx,nls,g1,g0);

			// Gy :
			//std::copy(edge,edge+sz,gy);
			//Vtl::ImAlg::Conv::Conv2D::conv2DInplace(gy,w,h,w,g0,g1);
			Vtl::ImAlg::Conv::Conv2D::conv2D(edge,w,h,nls,gy,nls,g0,g1);

			//  :
			Vtl::Plf::F21::SqrtOfSumSqr<T> sssq;
			Vtl::ImAlg::Tools::copy_2_1(gx,w,h,nls,gy,nls,fri,nls,sssq);
			Vtl::Plf::F11::Mult<T> mul(T(sqrt(2.0f*M_PI)*sigma));
			Vtl::ImAlg::Tools::copy_1_1(fri,w,h,nls,mul);
			// Non Maxima Supression:
			Vtl::Plf::Fpie::CannyNMS<float> nms(0.1f);
			Vtl::Plf::Extent ext=nms.ext();
			std::fill(edge,edge+sz,0.0f);
			Vtl::ImAlg::Tools::envTransf_3e1_1(ext,fri,w,h,nls,gx,nls,gy,nls,edge,nls,nms);
		}
		void DiZenzo(int w,int h,int nls,T* edge,T* gx,T* gy,T* fri,T* tmp,T* tmp2)
		{
			const int sz=w*h;
			//Vtl::ImAlg::Conv::Conv2D::conv2DInplace(gy,w,h,w,g0,g1);

			// Rx :  in fri
			Vtl::ImAlg::Conv::Conv2D::conv2D(edge,w,h,w,fri,w,g1,g0);
			// Ry :  in edge
			Vtl::ImAlg::Conv::Conv2D::conv2DInplace(edge,w,h,w,g0,g1);

			Vtl::Plf::F23::CorrMatrix<T> cm;
			Vtl::ImAlg::Tools::copy_2_3(fri,w,h,nls,edge,nls,edge,nls,fri,nls,tmp,nls,cm);
			// Now:  edge==Rx^2 fri==RxRy  tmp Ry^2

			// Gx :  in tmp2
			Vtl::ImAlg::Conv::Conv2D::conv2D(gx,w,h,w,tmp2,w,g1,g0);
			// Gy :  in gx
			Vtl::ImAlg::Conv::Conv2D::conv2DInplace(gx,w,h,w,g0,g1);
			Vtl::Plf::F23::CorrMatrixAdd<T> cma;
			Vtl::ImAlg::Tools::copy_2_3(tmp2,w,h,nls,gx,nls,edge,nls,fri,nls,tmp,nls,cma);

			// Bx :  in tmp2
			Vtl::ImAlg::Conv::Conv2D::conv2D(gy,w,h,w,tmp2,w,g1,g0);
			// By :  in gy
			Vtl::ImAlg::Conv::Conv2D::conv2DInplace(gy,w,h,w,g0,g1);
			Vtl::ImAlg::Tools::copy_2_3(tmp2,w,h,nls,gy,nls,edge,nls,fri,nls,tmp,nls,cma);

			// ..  ..:
			Vtl::Plf::F33::DiZenzo<T> dz;
			Vtl::ImAlg::Tools::copy_3_3(edge,w,h,nls,fri,nls,tmp,nls,fri,nls,gx,nls,gy,nls,dz);
			Vtl::Plf::F11::Mult<T> mul3(T(sqrt(2.0f*M_PI)*sigma));
			Vtl::ImAlg::Tools::copy_1_1(fri,w,h,nls,mul3);
			// Non Maxima Supression:
			Vtl::Plf::Fpie::CannyNMS<float> nms(0.1f);
			Vtl::Plf::Extent ext=nms.ext();
			std::fill(edge,edge+sz,0.0f);
			Vtl::ImAlg::Tools::envTransf_3e1_1(ext,fri,w,h,w,gx,w,gy,w,edge,w,nms);
		}
	};
}//namespace Curve
}//namespace Students
////////////////////////////////////////////////////////////////////////////////////////////////////
#endif //_EDGE_DET_H
