#ifndef _SS_RIDGE_DET_H 
#define _SS_RIDGE_DET_H 
///////////////////////////////////////////////////////////////////////////////////////
#include "num_linalg.h"
#include "AImage.h"
#include "plf_functors.h"
#include "plf_kernels_1d.h"
#include "Hanina/ss_ridge_functors.h" 
///////////////////////////////////////////////////////////////////////////////////////
namespace Hanina
{
	/////////////////////////////////////////////////////////////////////////////////////
	///
	///         4 
	///
	///
	/////////////////////////////////////////////////////////////////////////////////////
	template<typename T> 
	class Hessian3D
	{
		private:
			const int w;
			const int h;
			const int sz;
			std::vector<T> _data;
			std::vector<T> _Gxx;
			std::vector<T> _Gxy;
			std::vector<T> _Gyy;
			std::vector<T> _Gxxx;
			std::vector<T> _Gyyy;
			std::vector<T> _Gxyy;
			std::vector<T> _Gyxx;
			std::vector<T> _Gxxxx;
			std::vector<T> _Gxxyy;
			std::vector<T> _Gyyyy;

			T *data;
			T *Gxx;
			T *Gxy;
			T *Gyy;
			T *Gxxx;
			T *Gyyy;
			T *Gxyy;
			T *Gyxx;
			T *Gxxxx;
			T *Gxxyy;
			T *Gyyyy;
		
		public:
			// ,        
			template<typename T1>
			Hessian3D(const T1* img,int width,int height,int nls)
			: w(width), h(height), sz(width*height), _data(sz)
			{
				//   1-    ,   , ,  
				 _data.resize(sz);
				 _Gxx.resize(sz);
				 _Gxy.resize(sz);
				 _Gyy.resize(sz);
				 _Gxxx.resize(sz);
				 _Gyyy.resize(sz);
				 _Gxyy.resize(sz);
				 _Gyxx.resize(sz);
				 _Gxxxx.resize(sz);
				 _Gxxyy.resize(sz);
				 _Gyyyy.resize(sz);
				
				//        .
				data=&(_data[0]);
				Gxx=&(_Gxx[0]);
				Gxy=&(_Gxy[0]);
				Gyy=&(_Gyy[0]);
				Gxxx=&(_Gxxx[0]);
				Gyyy=&(_Gyyy[0]);
				Gxyy=&(_Gxyy[0]);
				Gyxx=&(_Gyxx[0]);
				Gxxxx=&(_Gxxxx[0]);
				Gxxyy=&(_Gxxyy[0]);
				Gyyyy=&(_Gyyyy[0]);
				
				//          float
				Vtl::Plf::F11::Convert<T> c_to_float;
				Vtl::ImAlg::Tools::copy_1_1(img,w,h,nls,data,w,c_to_float);	
				//     ()
				//    
			}
			//         
			// 3    ,   
			template<typename T1> 
			void operator()(float sigma, T1* max_eig,T1* vx1, T1* vy1, T1* vt1, T1* vx2, T1*vy2, T1* vt2)
			{
				Hanina::FillingHessianMatrix3x3 Filling_Hessian_Matrix_3x3(sigma);

				calc_deriv_from(sigma,data);
				// 3d      ,   t
				Vtl::ImAlg::Tools::copy_10_7(Gxx,w,h,w, Gyy, w, Gxy, w,Gxxx,w, Gyyy,w,Gxyy,w,Gyxx,w,
											  Gxxxx, w, Gyyyy, w,Gxxyy, w, max_eig,w,vx1,w, vy1,w, vt1,w,
											  vx2,w,vy2,w,vt2,w,Filling_Hessian_Matrix_3x3);

#if 0  //double eigen (hessian) approach
				calc_deriv_from(sigma,max_eig);
				// 3d      ,   t
				Vtl::ImAlg::Tools::copy_10_7(Gxx,w,h,w, Gyy, w, Gxy, w,Gxxx,w, Gyyy,w,Gxyy,w,Gyxx,w,
											  Gxxxx, w, Gyyyy, w,Gxxyy, w, max_eig,w,vx1,w, vy1,w, vt1,w,
											  vx2,w,vy2,w,vt2,w,Filling_Hessian_Matrix_3x3);
#endif
				
			}
			template<typename T1> 
			void operator()(float sigma, T1* max_eig,T1* vx1, T1* vy1, T1* vt1, T1* vx2, T1*vy2, T1* vt2,T *grad_x,T *grad_y)
			{
				this->operator ()(sigma,max_eig,vx1, vy1, vt1, vx2, vy2, vt2);
				calc_grad_from(sigma,data,grad_x,grad_y);
			}
			void calc_grad_from(float sigma,const T *data,T *grad_x,T *grad_y)
			{
				// ,         4 
				// 
				typedef Vtl::ImAlg::WndFun::Gauss<float,0, false> FF0; 
				FF0 f0(sigma,4);
				typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K0;
				K0 ker0(f0);
				Vtl::ImAlg::Conv::ConvInplace<float,K0> g0(ker0);
				//1 
				typedef Vtl::ImAlg::WndFun::Gauss<float,1,false> FF1; 
				FF1 f1(sigma,4);
				typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K1;
				K1 ker1(f1);
				Vtl::ImAlg::Conv::ConvInplace<float,K1> g1(ker1);

				memcpy(grad_x,data,sz*sizeof(float));
				memcpy(grad_y,data,sz*sizeof(float));
				// Gx :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(grad_x,w,h,w,g1,g0);
				// Gy :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(grad_y,w,h,w,g0,g1);
			}
			void calc_deriv_from(float sigma,const T *data)
			{
				memcpy(Gxx,data,sz*sizeof(float));
				memcpy(Gxy,data,sz*sizeof(float));
				memcpy(Gyy,data,sz*sizeof(float));
				memcpy(Gxxx,data,sz*sizeof(float));
				memcpy(Gyyy,data,sz*sizeof(float));
				memcpy(Gxyy,data,sz*sizeof(float));
				memcpy(Gyxx,data,sz*sizeof(float));
				memcpy(Gxxxx,data,sz*sizeof(float));
				memcpy(Gxxyy,data,sz*sizeof(float));
				memcpy(Gyyyy,data,sz*sizeof(float));
				calc_deriv(sigma);
			}
			//,          
			void calc_deriv(float sigma)
			{
				// ,         4 
				// 
				typedef Vtl::ImAlg::WndFun::Gauss<float,0, false> FF0; 
				FF0 f0(sigma,4);
				typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K0;
				K0 ker0(f0);
				Vtl::ImAlg::Conv::ConvInplace<float,K0> g0(ker0);

				//1 
				typedef Vtl::ImAlg::WndFun::Gauss<float,1,false> FF1; 
				FF1 f1(sigma,4);
				typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K1;
				K1 ker1(f1);
				Vtl::ImAlg::Conv::ConvInplace<float,K1> g1(ker1);
			
				//2 
				typedef Vtl::ImAlg::WndFun::Gauss<float,2,false> FF2; 
				FF2 f2(sigma,4);
				typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K2;
				K2 ker2(f2);
				Vtl::ImAlg::Conv::ConvInplace<float,K2> g2(ker2);
				
				//3 
				typedef Vtl::ImAlg::WndFun::Gauss<float,3,false> FF3; 
				FF3 f3(sigma,4);
				typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K3;
				K3 ker3(f3);
				Vtl::ImAlg::Conv::ConvInplace<float,K3> g3(ker3);

				//4 
				typedef Vtl::ImAlg::WndFun::Gauss<float,4,false> FF4; 
				FF4 f4(sigma,4);
				typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K4;
				K4 ker4(f4);
				Vtl::ImAlg::Conv::ConvInplace<float,K4> g4(ker4);
								
				// 2 
				// Gxx :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Gxx,w,h,w,g2,g0);
				// Gyy :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Gyy,w,h,w,g0,g2);
				// Gxy :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Gxy,w,h,w,g1,g1);
				
				// 3 
				// Gxxx :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Gxxx,w,h,w,g3,g0);
				// Gyyy :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Gyyy,w,h,w,g0,g3);
				// Gxyy :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Gxyy,w,h,w,g1,g2);
				// Gyxx :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Gyxx,w,h,w,g2,g1);

				// 4 
				// Gxxxx :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Gxxxx,w,h,w,g4,g0);
				// Gxxyy :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Gxxyy,w,h,w,g2,g2);
				// Gyyyy :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Gyyyy,w,h,w,g0,g4);
			}
	};
	//////////////////////////////////////////////////////////////////////////////////////////
	///
	///   3   3  
	///
	//////////////////////////////////////////////////////////////////////////////////////////
	template<typename T> 
	class Planes3
	{
		private:
			int w;
			int h;
			int sz;
			Hessian3D<T> hess3d;
			std::vector<T> _eig_prev;
			std::vector<T> _eig_curr;
			std::vector<T> _eig_next;

			std::vector<T> _grad_x_curr;
			std::vector<T> _grad_y_curr;
			std::vector<T> _grad_x_next;
			std::vector<T> _grad_y_next;

			std::vector<T> _h2_x_curr;
			std::vector<T> _h2_y_curr;
			std::vector<T> _h2_z_curr;
			
			std::vector<T> _h3_x_curr;
			std::vector<T> _h3_y_curr;
			std::vector<T> _h3_z_curr;
			
			std::vector<T> _h2_x_next;
			std::vector<T> _h2_y_next;
			std::vector<T> _h2_z_next;
			
			std::vector<T> _h3_x_next;
			std::vector<T> _h3_y_next;
			std::vector<T> _h3_z_next;
			std::vector<T> _image;
				
			T *eig_prev;
			T *eig_curr;
			T *eig_next;

			T *grad_x_curr;
			T *grad_y_curr;
			T *grad_x_next;
			T *grad_y_next;


			T *h2_x_curr;
			T *h2_y_curr;
			T *h2_z_curr;
			T *h3_x_curr;
			T *h3_y_curr;
			T *h3_z_curr;
			T *h2_x_next;
			T *h2_y_next;
			T *h2_z_next;
			T *h3_x_next;
			T *h3_y_next;
			T *h3_z_next;
			T *image;

		public:
			//   Planes3 
			template<typename T1>
			Planes3(const T1* img,int width,int height,int nls)
			: w(width),h(height), sz(width*height), hess3d(img,width,height,nls)
			{
				_eig_prev.resize(sz);
				_eig_curr.resize(sz);
				_eig_next.resize(sz);

				_grad_x_curr.resize(sz);
				_grad_y_curr.resize(sz);
				_grad_x_next.resize(sz);
				_grad_y_next.resize(sz);

				_h2_x_curr.resize(sz);
				_h2_y_curr.resize(sz);
				_h2_z_curr.resize(sz);
				_h3_x_curr.resize(sz);
				_h3_y_curr.resize(sz);
				_h3_z_curr.resize(sz);
				_h2_x_next.resize(sz);
				_h2_y_next.resize(sz);
				_h2_z_next.resize(sz);
				_h3_x_next.resize(sz);
				_h3_y_next.resize(sz);
				_h3_z_next.resize(sz);
				_image.resize(sz);

				eig_prev=&(_eig_prev[0]);
				eig_curr=&(_eig_curr[0]);
				eig_next=&(_eig_next[0]);

				grad_x_curr=&(_grad_x_curr[0]);
				grad_y_curr=&(_grad_y_curr[0]);
				grad_x_next=&(_grad_x_next[0]);
				grad_y_next=&(_grad_y_next[0]);

				h2_x_curr=&(_h2_x_curr[0]);
				h2_y_curr=&(_h2_y_curr[0]);
				h2_z_curr=&(_h2_z_curr[0]);
				h3_x_curr=&(_h3_x_curr[0]);
				h3_y_curr=&(_h3_y_curr[0]);
				h3_z_curr=&(_h3_z_curr[0]);
				h2_x_next=&(_h2_x_next[0]);
				h2_y_next=&(_h2_y_next[0]);
				h2_z_next=&(_h2_z_next[0]);
				h3_x_next=&(_h3_x_next[0]);
				h3_y_next=&(_h3_y_next[0]);
				h3_z_next=&(_h3_z_next[0]);
				image=&(_image[0]);
			}
			void operator()()
			{
				std::auto_ptr<AImage> _img(::AImageCreate(0,AImage::IMGC_GRAYSCALE,w,h,8));
				std::auto_ptr<AImage> _img_2(::AImageCreate(0,AImage::IMGC_GRAYSCALE,w,h,8));

				//     ,    8 ,   
				//     -         1.
				AImageMem8* img=dynamic_cast<AImageMem8*>(_img.get());
				AImageMem8* img_total_ridge=dynamic_cast<AImageMem8*>(_img_2.get());
				if(!img || !img_total_ridge)
				{
					throw "error";
				}
				img_total_ridge->SetTo(0);

				const float sigma_mult=float(pow(2.,1/4.0));
//				const float sigma_start=0.5f;
				const float sigma_start=0.5f*float(pow(2.,1/2.0));
				const float sigma_max=0.1f*std::min(w,h);
				//indicator->start("Generate set of ridge edge map for different scales",int(100*log(sigma_start)),int(100*log(sigma_max)));
				////for( float sigma=sigma_start ; sigma<= sigma_max ; sigma*=sigma_mult)
				////{
				////}
				float sigma=sigma_start;
				float sigma_count=1;
//				hess3d(sigma,eig_curr,h2_x_curr,h2_y_curr,h2_z_curr,h3_x_curr,h3_y_curr,h3_z_curr);
				hess3d(sigma,eig_curr,h2_x_curr,h2_y_curr,h2_z_curr,h3_x_curr,h3_y_curr,h3_z_curr,grad_x_curr,grad_y_curr);
				sigma*=sigma_mult; ++sigma_count;
//				hess3d(sigma,eig_next,h2_x_next,h2_y_next,h2_z_next,h3_x_next,h3_y_next,h3_z_next);
				hess3d(sigma,eig_next,h2_x_next,h2_y_next,h2_z_next,h3_x_next,h3_y_next,h3_z_next,grad_x_next,grad_y_next);
				////std::copy(eig_next,eig_next+sz,eig_prev);

				Check_4_grad_sign_change<T> supress_edge(w);
				ScaleScan3x3<T> func(w);

				////T* planes[]={eig_prev,eig_curr,eig_next};
				////Vtl::ImAlg::Tools::envTransf_3D(planes, w,h,w,h2_x_curr,h2_y_curr,h2_z_curr,h3_x_curr,h3_y_curr,h3_z_curr, image, func);
				////Vtl::ImAlg::Tools::envTransf_2Ddircheck(image, w,h,w,h2_x_curr,h2_y_curr,grad_x_curr,grad_y_curr,supress_edge);
				//////image save
				////SaveImage(img,image,w,h,w,sigma,"aaa.tif","eig3Dss");
				//SaveImage(img,eig_curr,w,h,w,sigma,"aaa.tif","eig3Dss");

				using namespace Vtl::ImAlg::Tools; 
				for( ; sigma< sigma_max ; )
				{
					//    sigma,     333     sigma_prev. 
					//            .
					float sigma_prev=sigma;
					std::cout << "sigma=" << sigma <<std::endl;
					sigma*=sigma_mult; ++sigma_count;

					xchg(eig_prev,eig_curr);
					xchg(eig_curr,eig_next);

					xchg(h2_x_curr,h2_x_next);
					xchg(h2_y_curr,h2_y_next);
					xchg(h2_z_curr,h2_z_next);

					xchg(h3_x_curr,h3_x_next);
					xchg(h3_y_curr,h3_y_next);
					xchg(h3_z_curr,h3_z_next);

					xchg(grad_x_curr,grad_x_next);
					xchg(grad_y_curr,grad_y_next);

					////planes[0]=eig_prev;
					////planes[1]=eig_curr;
					////planes[2]=eig_next;
					T* planes[]={eig_prev,eig_curr,eig_next};

//					hess3d(sigma,eig_next,h2_x_next,h2_y_next,h2_z_next,h3_x_next,h3_y_next,h3_z_next);
					hess3d(sigma,eig_next,h2_x_next,h2_y_next,h2_z_next,h3_x_next,h3_y_next,h3_z_next,grad_x_next,grad_y_next);
					SaveImage(img,planes[1],w,h,w,sigma_prev,"aab.tif","eig3Dss");

//-,     //	Vtl::ImAlg::Tools::envTransf_3D(planes, w,h,w,h2_x_next,h2_y_next,h2_z_next,h3_x_next,h3_y_next,h3_z_next, image, func);
					Vtl::ImAlg::Tools::envTransf_3D(planes, w,h,w,h2_x_curr,h2_y_curr,h2_z_curr,h3_x_curr,h3_y_curr,h3_z_curr, image, func);

					Vtl::ImAlg::Tools::envTransf_2Ddircheck(image, w,h,w,h2_x_curr,h2_y_curr,grad_x_curr,grad_y_curr,supress_edge);
					
					Comb_ridge<T,uint8_t> crid(uint8_t(unsigned(sigma_count*10)));
					Vtl::ImAlg::Tools::copy_2_1(image,w,h,w,(*img_total_ridge)[0],img_total_ridge->NextLineShift(),(*img_total_ridge)[0],img_total_ridge->NextLineShift(),crid);

					
					//image save
					SaveImage(img,image,w,h,w,sigma_prev,"aaa.tif","eig3Dss");
					//SaveImage(img,eig_curr,w,h,w,sigma,"aaa.tif","eig3Dss");
				}
				AImageSave(img_total_ridge,"img_total_ridge.tif",0);


			}
			void SaveImage(AImageMem8* img,const T *data,int w,int h,int nls,float sigma,const char *base_name,const char *postfix)
			{
				//    8-       .   , 
			//    .
			
			//    ()
			Vtl::Plf::F11::GammaCorMult<float> lin_3mod(255,2.2,1.0);
			//     float    byte. ,   float  
			//     0-255 -       
			// Hess_Eig_diff  HessianDetSqrt  -         .
			//   Hess_Eig_diff
			Vtl::Plf::F11::Convert<uint8_t> c_to_Byte;
			//  ,      
			Vtl::Plf::F11::Comb<uint8_t,Vtl::Plf::F11::GammaCorMult<float>,Vtl::Plf::F11::Convert<uint8_t> > adj_3mod(lin_3mod,c_to_Byte);
				//          ()
				Vtl::ImAlg::Tools::copy_1_1(data,w,h,w,(*img)[0],img->NextLineShift(),adj_3mod);
				//    
				{
					//  :
					std::ostringstream os;
					os <<	YFileSys::getNameWithoutExt(base_name)
						<< std::setprecision(2) << std::setiosflags(std::ios::showpoint|std::ios::fixed)
						<< "_"<<postfix<<"_sigma=" ;
					if(sigma<10)
						os << "0";
					os << sigma 
						<< YFileSys::getExt(base_name);
					std::string outfname=os.str();
					// 
					::AImageSave(img,outfname.c_str());
				}
			}
	};

	

}//namespace Hanina


///////////////////////////////////////////////////////////////////////////////////////
#endif //_SS_RIDGE_DET_H 