#ifndef _SS_RIDGE_FUNCTORS_2_H 
#define _SS_RIDGE_FUNCTORS_2_H
//
//
//     
//
//
///////////////////////////////////////////////////////////////////////////////////////
//#include "num_linalg.h"
//#include "AImage.h"
//#include "plf_functors.h"
//#include "plf_kernels_1d.h"
///////////////////////////////////////////////////////////////////////////////////////
namespace Hanina
{
	// 
	template<typename T> 
	struct HessianDetSqrt
	{
		template<typename T1> 
		T operator()(T1 Dxx,T1 Dxy,T1 Dyy) 
		{
			T1 d=Dxx*Dyy-Dxy*Dxy; 
			return T(d>=0?sqrt(d):-sqrt(d));
		}
	};
	// ..  
	template<typename T> 
	struct HessianEigenDiff
	{
		template<typename T1> 
		T operator()(T1 Dxx,T1 Dxy,T1 Dyy) 
		{
			/// Calc eigen values: (Dxx - Lambda)(Dyy-Lambda)-Dxy^2=0=Lambda^2 - Lambda * (Dxx+Dyy) +  Dxx*Dyy-Dxy^2
			/// discriminant = b^2 - 4ac = (Dxx+Dyy)^2 - 4*(Dxx*Dyy-Dxy^2)= Dxx^2+Dyy^2 + 2*Dxx*Dyy - 4*Dxx*Dyy+4*Dxy*Dxy
			/// = Dxx^2+Dyy^2 - 2*Dxx*Dyy+4*Dxy*Dxy = (Dxx - Dyy)^2 +4+Dxy^2 = 4* ( ((Dxx - Dyy)/2)^2 +Dxy^2)
			T1 d = sqrt(Vtl::sqr(Dxx-Dyy)+4*Vtl::sqr(Dxy));
			// b2=Dxx+Dyy; //       1    .
			// Lambda1 = b2 + d/2;
			// Lambda2 = b2 - d/2;
			// Lambda1 -Lambda2 = d;
			T1 diff = d;
			return diff;
		}
	};
	// ..     . 
	struct HessianEigenDiffAndMaxCurvatureDirection
	{
		template<typename T1> 
		void operator()(T1 Dxx,T1 Dxy,T1 Dyy, T1 & eigen_diff, T1 & vx, T1 & vy  ) 
		{
			/// Calc eigen values: (Dxx - Lambda)(Dyy-Lambda)-Dxy^2=0=Lambda^2 - Lambda * (Dxx+Dyy) +  Dxx*Dyy-Dxy^2
			/// discriminant = b^2 - 4ac = (Dxx+Dyy)^2 - 4*(Dxx*Dyy-Dxy^2)= Dxx^2+Dyy^2 + 2*Dxx*Dyy - 4*Dxx*Dyy+4*Dxy*Dxy
			/// = Dxx^2+Dyy^2 - 2*Dxx*Dyy+4*Dxy*Dxy = (Dxx - Dyy)^2 +4+Dxy^2 = 4* ( ((Dxx - Dyy)/2)^2 +Dxy^2)
			 T1 d = sqrt(Vtl::sqr(Dxx-Dyy)+4*Vtl::sqr(Dxy));
			 T1 b2 = Dxx+Dyy;  //       1    .
//			 T1 Lambda1 = (b2 + d)/2;
//			 T1 Lambda2 = (b2 - d)/2;
//			 T1 vect1Y = 1;
//			 T1 vect1X = (Dxy*vect1Y)/Dxx-Lambda1; 
//			 T1 vect2X = 1;
//			 T1 vect2Y = (Dxx-Lambda2*vect2X)/Dxy; 
//   1 .. -    . 
//    , . ..     ..
// ..  b2 > 0  Lambda1,  Lambda2
			 T1 Lambda = ( b2 >= 0 ) ? (b2 + d/2) : (b2 - d/2);
			 vx = -Dxy;
			 vy = Dxx-Lambda; 
			 eigen_diff=d;
		}
	};
	///  \f$ L_{uv}*(L_x^2+L_y^2) \f$ .    \f$ (L_x^2+L_y^2) \f$   
	///          -  ,   ,
	///  Luv  ,    ,     Luv. 
	template<typename T> 
	struct Luv_from_xy_derivatives
	{
		template<typename T1> 
		T operator()(T1 Lx,T1 Ly,T1 Lxx,T1 Lxy,T1 Lyy) 
		{
			return Lx*Ly*(Lxx-Lyy)+(Vtl::sqr(Lx)-Vtl::sqr(Ly))*Lxy;
		}
	};
	
	template<typename T>
	class ZeroCrossVisualization
	{
		T maxval;
		T k;
	public:
		ZeroCrossVisualization(T max_value=255,T slope=1):maxval(T(max_value/M_PI*2)),k(slope) {}
		T operator()(T val) const {return maxval*atan(k*abs(val));}
	};
		template<typename T> 
		class ZeroCross
		{
			T thr;
			int nls;
			int pos;
			int neg;

		public:
			/// !!! MUST be called by filter BEFORE using 'operator()', by default
			/// next line shist parameter set to value which probably most quick lead 
			/// a 'Invalid memory access' error (for bug search convenience).
			void set_IterNextLineShift(int modul_nls) { nls=modul_nls;}
			Vtl::Plf::Extent ext() const {return Vtl::Plf::Extent(1,1,1,1); }
			ZeroCross(T threshold=0):thr(threshold) {nls=0x5FFFFFFF;}
			T operator()(const T* value)
			{
				//if(abs(value[0]) <= thr)
				//	return 0;
				pos=neg=0;
				check(value[0],pos,neg);
				check(value[-1],pos,neg);
				check(value[+1],pos,neg);
				check(value[nls],pos,neg);
				check(value[nls-1],pos,neg);
				check(value[nls+1],pos,neg);
				check(value[-nls],pos,neg);
				check(value[-nls-1],pos,neg);
				check(value[-nls+1],pos,neg);
				if(pos>2 && neg >2 )
					return 0;
				return 255;
			}
		private:
			static void check(T value,int& pos,int& neg)
			{
				if(value>0)
					++pos;
				else if(value<0)
					++neg;
			}
		};



}//namespace Hanina


///////////////////////////////////////////////////////////////////////////////////////
#endif //_SS_RIDGE_FUNCTORS_2_H