#ifndef POINTS_CONTAINER_H
#define POINTS_CONTAINER_H

// system includings
#include <vector>

// own includings
#include "KMeans.h"
#include "CurveDetection/Point.h"

namespace Vtl
{
namespace CurveDetection
{
	// select points with high absolute value (>thresh) from image iter
	// and save them to the conteiner
  template <typename T, typename IterType>
	void SelectPoints(std::vector<Point> &container,
		                IterType iter,
	                  int width,
			              int height,
			              T thresh)
	{
		container.clear();

	  for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (fabs(*iter) > thresh)
				{
					container.push_back(Point(x, y));
				}
        iter++;
			}
		}
	}

	//-------------------------------------------------------------------------
	template <typename T, typename IterType>
	void SelectPoints(std::vector< DatumR<T> > &container,
		                IterType iter,
	                  int width,
			              int height,
			              T thresh = T(0))
	{
		container.clear();

	  for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (fabs(*iter) > thresh)
				{
					container.push_back(DatumR<T>(x, y, *iter));
				}
        iter++;
			}
		}
	}
  //-------------------------------------------------------------------------
	// make selection of random points which allow finding of model parameters
	template <typename ModelType>
	int GetSample(const std::vector<Point> &container,
		            std::vector<const Point *> &sample)
	{
		double coef = int(container.size() - 1) / (pow(2.0, 15) - 1);
		//std::cout << std::endl << "Get random sample";
		//std::cout << std::endl << "coef = " << coef;
		
		int sampleSize =  ModelType::sampleSize;
		sample.resize(sampleSize);
		
		// we don't check more than container.size() samples
		int maxC = container.size();
		//std::cout << "   sample size = " << sampleSize << "  maxC = " << maxC;
		for (int c = 0; c < maxC; c++)
		{ 
			//std::cout << std::endl << "______________" <<  std::endl << "c = " << c;
			for (int i = 0; i < sampleSize; i++)
			{
				int n = int(rand() * coef);
				sample[i] = &(container[n]);
				//std::cout << std::endl << n << "  " << sample[i]->x << "  " << sample[i]->y; 
			}
       
			// check distanses between points in sample
			bool distIsOK = true;

			double minDist = ModelType::minDist;
			double maxDist = ModelType::maxDist;
	
			for (int i = 0; i < sampleSize - 1; i++)
			{
				for (int j = i + 1; j < sampleSize; j++)
				{
					const Point *p1 = sample[i];
					const Point *p2 = sample[j];

					double dist = sqrt(double((p1->x - p2->x) * (p1->x - p2->x) +
														 (p1->y - p2->y) * (p1->y - p2->y)));
					//std::cout << std::endl << i << "  " << j << "  " << dist;
					if (dist < minDist || dist > maxDist)
					{
						distIsOK = false;
					}
				}
			}
			
			if (distIsOK)
			{
        return 1;
			}
		}
		return 0;
	}
}
}

#endif // POINTS_CONTAINER_H