#ifndef DRAW_ELLIPSE_H
#define DRAW_ELLIPSE_H

// std includes
#define _USE_MATH_DEFINES
#include <cmath>

// library includes
#include "alg_bresenham.h"

template <typename T>
void DrawEllipse(T* io_dstR,        // destination
                 T* io_dstG,        //             image
                 T* io_dstB,        //                   components
                 int w,             // image
                 int h,             //       size
                 int nls,           // image line length
                 int cx,            // center
                 int cy,            //        coordinates    
                 double Ax,         // large semi-axis
                 double Ay,         // small semi-axis
                 double fi,         // slope angle
                 int pointsNumb,    // number of lines approximating ellipse
                 T colR = T(0),     // brush
                 T colG = T(0),     //      
                 T colB = T(0))     //        colour
{
  // pen is black
	AImageMem8::Pixel color_pen = AImageMem8::Pixel(0); 

  typedef Vtl::ImAlg::Bres::Draw<AImageMem8::Pixel> DrawFunctor;
	DrawFunctor draw_functor(color_pen);

	typedef Vtl::ImAlg::Bres::SetPixelGray<AImageMem8::Pixel, DrawFunctor>
																														   SetPixelFunctor;
  SetPixelFunctor set_pixR(io_dstR, nls, draw_functor);
  SetPixelFunctor set_pixG(io_dstG, nls, draw_functor);
  SetPixelFunctor set_pixB(io_dstB, nls, draw_functor);

	Vtl::ImAlg::Bres::Line<SetPixelFunctor> lineR(set_pixR);
  Vtl::ImAlg::Bres::Line<SetPixelFunctor> lineG(set_pixG);
  Vtl::ImAlg::Bres::Line<SetPixelFunctor> lineB(set_pixB);

  std::vector<int> x(pointsNumb + 1);
	std::vector<int> y(pointsNumb + 1);

  double sinFi = sin(fi);
  double cosFi = cos(fi);

  for(int i = 0; i <= pointsNumb ; i ++)
	{
		double x_ = Ax * cos(2 * M_PI / pointsNumb * i);
		double y_ = Ay * sin(2 * M_PI / pointsNumb * i);

    x[i] = cx + int(x_ * cosFi - y_ * sinFi + 0.5);
    y[i] = cy + int(x_ * sinFi + y_ * cosFi + 0.5);

    x[i] = std::min(w, x[i]);
    x[i] = std::max(0, x[i]);
    y[i] = std::min(h, y[i]);
    y[i] = std::max(0, y[i]);
	}

  set_pixR.Set0(x[0], y[0]);
  set_pixG.Set0(x[0], y[0]);
  set_pixB.Set0(x[0], y[0]);

  for(int i = 1; i <= pointsNumb ; i ++)
	{
		lineR.Draw(x[i] - x[i - 1], y[i] - y[i - 1]);
    lineG.Draw(x[i] - x[i - 1], y[i] - y[i - 1]);
    lineB.Draw(x[i] - x[i - 1], y[i] - y[i - 1]);
  }
}

#endif // DRAW_ELLIPSE_H