#ifndef CANNY_EDGE_DETECTOR_H
#define CANNY_EDGE_DETECTOR_H

// std includings
#include <vector>

// own includings
#include "EdgeDetection/GradientOperations.h"
#include "EdgeDetection/Hysteresis.h"

// Gradient module + nonmaxima supression + hysteresis
template< typename T >
class CannyEdgeDetector
{
public:
  std::vector<T> edges;
  std::vector<T> grMod;
  std::vector<T> Lx;
  std::vector<T> Ly;

private:
  T* src;                 // image to be processed
  int w;                  // image
  int h;                  //       size

  //===========================================================================
public:
  CannyEdgeDetector(T* in_src,
                    int in_w,
                    int in_h)
  {
    src = in_src;
    w = in_w;
    h = in_h;

    int size = w * h;

    Lx.resize(size);
    Ly.resize(size);
    grMod.resize(size);
    edges.resize(size);
  }

  //---------------------------------------------------------------------------
  void Apply(float sigma,   
             float nSigmasEnv,
             T hisThStart,         // histeresis
             T hisThEnd = T(0))    //            thresholds       
  {
    // calculate gradient components and absolute value
	  Gradient(&(grMod[0]), &(Lx[0]), &(Ly[0]), src, w, h, sigma, nSigmasEnv);
  	  
    // non maxima supression
    Vtl::Plf::Fpie::CannyNMS<T> nms(T(0));
    Vtl::Plf::Extent ext = nms.ext();
    Vtl::ImAlg::Tools::envTransf_3e1_1(ext, &(grMod[0]), w, h, w, &(Lx[0]), w,  
                                             &(Ly[0]), w, &(edges[0]), w, nms);
    // Hysteresis
    Hysteresis<T> hyst(&(edges[0]), w, h);
    hyst.Apply(hisThStart, hisThEnd, &(edges[0]));
  }
};

#endif CANNY_EDGE_DETECTOR_H