#ifndef EDT_H
#define EDT_H

namespace Vtl 
{ 
namespace ImAlg 
{ 
namespace Edt 
{
  class EdtEdgeCoord
  {
 public:
	  template<typename T1, typename T2, class Functor>
    static void edt(T1* src,
                    int w,
                    int h,
                    int nls_s,
                    T2* res,
                    int* edgeX,
                    int* edgeY,
                    int nls_d,
                    Functor isEdgeFunc)
    {
      T2 MaxValue = T2(std::max(w, h) + 1); // instead INT32_MAX

      //First Step - vertical
      std::vector<T2> t1(h); // instead t1(w)
      std::vector<T2> t2(h);

      std::vector<int> y1(h);
      std::vector<int> y2(h);

      // for each column
      for (int i = 0; i < w; ++i)
      {
        T1* col = src + i;

        // up -> down
        int j = 0;

        // before we meet first edge pixel fill colomn with imposible value
        for( ; j < h && !isEdgeFunc(*col); ++j, col += nls_s)
        {
            t1[j] = MaxValue;
            y1[j] = -1;
        }

        // fill colomn with distances from the up closest edge point
        for( ; j < h; ++j, col += nls_s)
        {
            if (isEdgeFunc(*col))
            {
              t1[j] = 0;
              y1[j] = j;
            }
            else
            {
              t1[j] = t1[j - 1] + 1;
              y1[j] = y1[j - 1];
            }
        }

        // down -> up
        --j; // j = h - 1;
				col -= nls_s;  //col = src + j * nls_s;
        
        // before we meet first edge pixel fill colomn with imposible value
        for( ; j >= 0 && !isEdgeFunc(*col); --j, col -= nls_s)
        {
          t2[j] = MaxValue;
          y2[j] = -1;
        }

        // fill colomn with distances from the down closest edge point
        for( ; j >= 0; --j, col -= nls_s)
        {
          if (isEdgeFunc(*col))
          {
            t2[j] = 0;
            y2[j] = j;
          }
          else
          {
            t2[j] = t2[j + 1] + 1;
            y2[j] = y2[j + 1];
          }
        }

        // choose between up and down edge point the closest and save distance 
        T2* r = res + i;
        int* y = edgeY + i;

        for(j = 0; j < h; ++j, r += nls_d, y += nls_d)
        {
          if (t1[j] < t2[j])
          {
            *r = t1[j];
            *y = y1[j];
          }
          else
          {
            *r = t2[j];
            *y = y2[j];
          }
	        //*r = std::min(t1[j], t2[j]);
        }
      }


      //Second Step - horizoltal
       for (int i = 0; i < h ; ++i)
      {
        edt1D(&(res[i * nls_d]), &(edgeX[i * nls_d]), &(edgeY[i * nls_d]), w,
                                                                     MaxValue);
      }
    }

    //=========================================================================
private:
    template< typename T2 >
	  static void edt1D(T2* dist,    // vertical distances
                      int* edgeXrow,
                      int* edgeYrow,
                      int N,       // row length  
                      T2 MaxValue)
    {
      // save numbers of previous and next elements because some elements
      // should be skiped
      std::vector<int> prev(N);
      std::vector<int> next(N);

      // initialize without skipping any elements
      for (int i = 0; i < N; ++i)
      {
          prev[i] = i - 1;
          next[i] = i + 1;
      }

      // Non-existent Points Removal
      int first = 0;
      int last = N - 1;

      // search for number of the first colomn containing edge pixel
      while(dist[first] == MaxValue && first < last)
      {
        ++first;
      }

      // search for number of the last colomn containing edge pixel
      while (dist[last] == MaxValue && first < last)
      {
        --last;
      }

      // skip columns which do not contain any edge points
      int cur;
      if( first < last )
      {
        cur = next[first];

        while( cur < last )
        {
          if( dist[cur] == MaxValue )
          {
            next[prev[cur]] = next[cur];
            prev[next[cur]] = prev[cur];
          }
          cur = next[cur];
        }
      }

      // Bad Point Test
      if( first < last )
      {
        cur = next[first];

        while( cur < last )
        {
          int pcur = prev[cur];
          int ncur = next[cur];

          float A = float(ncur - pcur);
          float B = float(dist[ncur] - dist[pcur]);
          float X = 0.5f * (ncur + pcur + B / A * (dist[ncur] + dist[pcur]));
          float D1 = (X - pcur) * (X - pcur) + dist[pcur] * dist[pcur];
          float D2 = (X - cur) * (X - cur) + dist[cur] * dist[cur];

          if( D1 <= D2 )
          {
            next[pcur] = ncur;
            prev[ncur] = pcur;
            if (pcur == first)
              cur = ncur;
            else
              cur = pcur;
          }
          else
             cur = ncur;
        }
      }
      //End of Bad Point Test

      std::vector<T2> res(N);
      std::vector<int> edgeX(N);
      std::vector<int> edgeY(N);

      cur = first;

      for( int i = 0; i < N; ++i )
      {
        T2 D = (cur - i) * (cur - i) + dist[cur] * dist[cur];
        int x = cur;
        int y = edgeYrow[cur];

        while( cur < last )
        {
          int ncur = next[cur];
          T2 D2 = (ncur - i) * (ncur - i) + dist[ncur] * dist[ncur];
          if( D < D2 )
            break;
          else
          {
            cur = next[cur];
            D = D2;
            x = ncur;
            y = edgeYrow[ncur];
          }
        }

        res[i] = D;
        edgeX[i] = x;
        edgeY[i] = y;
      }

      ::memcpy(dist, &(res[0]), N * sizeof(T2));
      ::memcpy(edgeXrow, &(edgeX[0]), N * sizeof(int));
      ::memcpy(edgeYrow, &(edgeY[0]), N * sizeof(int));
    }
  };

} // namespace Edt
} // namespace ImAlg
} // namespace Vtl


#endif // EDT_H