
#ifndef HARTLEY_H
#define HARTLEY_H

// system includings
#define _USE_MATH_DEFINES
#include <cmath>

// library includings
#include "imalg_tools.h"
#include "integral_transforms.h"

// own includings
#include "SaveImageArray.h"

template <typename T>
void SubtractAver(T *img,
                  int size)
{
  T aver = T(0);
  T *iter = img;

  for (int i = 0; i < size; i++, iter++)
  {
    aver += (*iter);
  }

  aver /= size;

  std::cout << std::endl << "aver = " << aver;

  for (int i = 0; i < size; i++, img++)
  {
    (*img) -= aver;
  }
}

//=============================================================================
template< typename T >
void MultGauss(T *img,
               int w,
               int h)
{
  T n = T(std::min(w, h));
  T sigma = n / T(8);
  n /= 2.0;
  
  for (int y = 0; y < h; y++)
  {
    for (int x = 0; x < w; x++)
    {
      T x_ = (T(x) - w/2) / sigma;
      T y_ = (T(y) - h/2) / sigma;

      (*img) *= T(/*1.0 / std::sqrt(2 * M_PI) / sigma * */
                  std::exp(-(x_ * x_ + y_ * y_) / 2));
      img++;
    }
  }
}

//=============================================================================
template< typename T >
int ExtentToSq(std::vector<T> &img,
                int w,
                int h)
{
  const int n = std::max(w, h);
  const int size = Vtl::NumAlg::IntTransf::Utils::CalcSmalestPow2(n);

  std::vector<T> newImg(size * size, 0);

  int startX = (size - w) / 2;
  int startY = (size - h) / 2;

  for (int y = startY; y < h + startY; y++)
  {
    for (int x = startX; x < w + startX; x++)
    {
      newImg[y * size + x] = img[(y - startY) * w + (x - startX)];
    }
  }

  img = newImg;

  return size;
}

//=============================================================================
template< typename T >
int HartleyTransform(std::vector<T> &img,
                     int w,
                     int h,
                     const char *fname)
{
  //SubtractAver(&(img[0]), w*h);

  //MultGauss(&(img[0]), w, h); 
  //SaveImageArray(&(img[0]), w, h, fname, "_MultGauss");

  const int size = ExtentToSq(img, w, h);
	Tests::SaveImageArray(&(img[0]), size, size, fname, "_ExtentToSq");
 
  int nn[] = {size, size};
	Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2, false>(&(img[0]), nn);
	Vtl::NumAlg::IntTransf::Hartley::fht2D_square(&(img[0]), size, size); 
	Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2,false>(&(img[0]), nn);

  Tests::SaveNormImageArray(&(img[0]), size, size, fname, "_Hartley");

  return size;
}

//=============================================================================
template< typename T >
void InvHartleyTransform(std::vector<T> &img,
                         int size,
                         const char *fname)
{
  int nn[] = {size, size};
	Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2, false>(&(img[0]), nn);
	Vtl::NumAlg::IntTransf::Hartley::ifht2D(&(img[0]), size, size); 
	Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2,false>(&(img[0]), nn);

  Tests::SaveNormImageArray(&(img[0]), size, size, fname, "_InvHartley");
}

#endif // HARTLEY_H