
#ifndef IMG_FROM_GRAD_H
#define IMG_FROM_GRAD_H

// own includings
#include "Hartley.h"

//=============================================================================
template<typename T>
class ReflectOrig
{
	T* v;
	
public:
	ReflectOrig(T in_v[]) 
		: v(in_v)	{}

  //---------------------------------------------------------------------------
	template<bool upper_half>
	void operator()(int plus_k,
                  int minus_k,
                  Loki::Int2Type<upper_half>)
	{
		Vtl::ImAlg::Tools::xchg(v[plus_k],v[minus_k]);
	}

  //--------------------------------------------------------------------------
	void operator()(int k,
                  Loki::Int2Type<false>)
	{
		v[k] = 0;
	}

  //---------------------------------------------------------------------------
	void operator()(int k,
                  Loki::Int2Type<true>)
	{
		v[k]=0;
	}
};

//=============================================================================
/*template<typename T>
class Hf
{
	T* _Hf;
  T* HLx;
  T* HLy;
  int sizeX;
  int sizeY;
	
public:
	ReflectOrig(T in_Hf[],
              T in_HLx[],
              T in_HLy[],
              int in_sizeX,
              int in_sizeY)
  {
    _Hf = in_Hf;
    HLx = in_HLx;
    HLy = in_HLy;

    sizeX = in_sizeX;
    sizeY = in_sizeY;
  }

  //---------------------------------------------------------------------------
	template<bool upper_half>
	void operator()(int plus_k,
                  int minus_k,
                  Loki::Int2Type<upper_half>)
	{
    int f = abs(plus_k % sizeX - sizeX / 2);
    int g = abs(plus_k / sizeX - sizeY / 2);

    double c = 2 * M_Pi * (f * f + g * g);

    _Hf(plus_k) = (f / HLx(plus_k) + g * HLy(plus_k)) / c;
    _Hf(minus_k) = -(f / HLx(minus_k) + g * HLy(minus_k)) / c;
	}

  //--------------------------------------------------------------------------
	void operator()(int k,
                  Loki::Int2Type<false>)
	{
		_Hf[k] = 0;
	}

  //---------------------------------------------------------------------------
	void operator()(int k,
                  Loki::Int2Type<true>)
	{
		_Hf[k] = 0;
	}
};*/

//=============================================================================
template< typename T >
void Hf(T *HLx,
        T *HLy,
        T *dst,
        int w,
        int h)
{
  T coef = T(2 * M_PI);
  int w2 = w / 2;
  int h2 = h / 2;

  for (int y = 0; y < h; y++)
  {
    for (int x = 0; x < w; x++)
    {
      if (x == w2 && y == h2)
      {
        continue;
      }
      else
      {
        int u = x - w2;
        int v = y - h2;

        T c = coef * T(u * u + v * v);

        *(dst + y * w + x) = (*(HLx + y * w + x) * u + *(HLy + y * w + x) * v) / c;
      }
    }
  }
}

//=============================================================================
template< typename T >
void ImgFromGrad(std::vector<T> &Lx,
                 std::vector<T> &Ly,
                 int w,
                 int h,
                 const char *fname)
{
  int size = HartleyTransform(Lx, w, h, fname);
  HartleyTransform(Ly, w, h, fname);

  int n[] = {size, size};

  ReflectOrig<T> reflLx(&(Lx[0]));
  Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<2>(n, reflLx);

  ReflectOrig<T> reflLy(&(Ly[0]));
  Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<2>(n, reflLy);

  /*std::vector<T> _Hf(size * size);
  Hf<T> reflLy(&(Ly[0]));
  Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<2>(n, reflLy);*/

  std::vector<T> _Hf(size * size);
  Hf(&(Lx[0]), &(Ly[0]), &(_Hf[0]), size, size);
	Tests::SaveNormImageArray(&(_Hf[0]), size, size, fname, "_Hf");

  InvHartleyTransform(_Hf, size, fname);
}

#endif // IMG_FROM_GRAD_H