#ifndef COMBINED_KAPPA_LINES_H
#define COMBINED_KAPPA_LINES_H

// library includings
#include "aimage.h"
#include "curvature/CurvatureFunc.h"

// own includings
#include "curvature/ImageContainer.h"
#include "curvature/Sigma.h"

namespace Vtl
{
namespace Curvature
{
	template <typename T>
	T SigmaToRadius(T s)
	{
	  return (s * s * s / 1.75);
	}

	//---------------------------------------------------------------------------
	template <typename T>
	T RadiusToSigma(T r)
	{
	  return T(1.75 * pow(double(fabs(r)), 1.0 / 3.0));
	}

	//---------------------------------------------------------------------------
	template <typename T, typename IterType, typename ImgType>
	T Average(T *meanDev,  //     
		        ImgType iter,
		        int c_x,
						int c_y,
						int w,
						int h,
						IterType gradX,
						IterType gradY,
						T a,
						T b)
	{
    T meanVal = T(0);
		int nonZeroCount = 0;

		T dirGradX = *(gradX + c_y * w + c_x);
		T dirGradY = *(gradY + c_y * w + c_x);
		T modGrad = sqrt(dirGradX * dirGradX + dirGradY * dirGradY);
    dirGradX /= modGrad;
		dirGradY /= modGrad;

    int prevX = -1;
		int prevY = -1;

		//std::cout << std::endl << "c_x = " << c_x << "   c_y = " << c_y
		//	<< "   a = " << a << "   b = " << b;
		for (T i = -a; i <= a; i += 0.5)
		{
		  for (T j = -b; j <= b; j += 0.5)
			{
			  int x = int(j * dirGradY + i * dirGradX + float(c_x) + 0.5);
        int y = int(- j * dirGradX + i * dirGradY + float(c_y) + 0.5);

				if (prevX == x && prevY == y)
				{
				  continue;
				}

				if (x < 0 || x >= w || y < 0 || y >= h)
				{
				  continue;
				}

				//std::cout << std::enld << "x = " << x << "  y = " << y;

				T temp = *(iter + y*w + x);
			
			  if (fabs(temp) > 0)
			  {
				  meanVal += temp;
          nonZeroCount++;
			  }
				
				prevX = x;
				prevY = y;
			}
		}
		
		/*int maxK = std::min(y + d, h - 1);
		int maxL = std::min(x + d, w - 1);

		for (int k = std::max(y - d, 0); k <= maxK; k++)
		  for (int l = std::max(x - d, 0); l <= maxL; l++)
		{
			T temp = *(iter + k*w + l);
			
			if (fabs(temp) > 0)
			{
				meanVal += temp;
        nonZeroCount++;
			}
	  }*/

		meanVal /= nonZeroCount;  // nonZeroCount is always >= 1

		*meanDev = T(0);

		prevX = -1;
		prevY = -1;

		for (T i = -a; i <= a; i += 0.5)
		{
		  for (T j = -b; j <= b; j += 0.5)
			{
			  int x = int(j * dirGradY + i * dirGradX + float(c_x) + 0.5);
        int y = int(- j * dirGradX + i * dirGradY + float(c_y) + 0.5);

				if (prevX == x && prevY == y)
				{
				  continue;
				}

				if (x < 0 || x >= w || y < 0 || y >= h)
				{
				  continue;
				}

				T temp = *(iter + y*w + x);
			
			  if (fabs(temp) > 0)
			  {
				  *meanDev += fabs(meanVal - temp);
			  }
				
				prevX = x;
				prevY = y;
			}
		}

		/*for (int k = std::max(y - d, 0); k <= maxK; k++)
		  for (int l = std::max(x - d, 0); l <= maxL; l++)
		{
			T temp = *(iter + k*w + l);
			
			if (fabs(temp) > 0)
			{
				*meanDev += fabs(meanVal - temp);
			}
	  }*/

		if (nonZeroCount > 1)
		{
		  *meanDev /= (nonZeroCount - 1);
		  *meanDev /= fabs(meanVal);
		}
		else
		{
		  *meanDev = T(0);
		}
		//T c = T(0.4) / T(j);
		//if (j > 0)
		//{
		//  value *= c;
		//}
		//value += *(iter + y*w + x) * (T(1) - c);
		//std::cout << std::endl << "  center = " << *(iter + c_y*w + c_x) <<
		//	"   value = " << meanVal << "  nonZeroCount = " << nonZeroCount <<
		//	"  meanDev = " << (*meanDev);

		return meanVal;
	}

	//---------------------------------------------------------------------------
  template <typename T, typename IterType, typename ImgType>
	T ProperRadius(int x,
								 int y,
								 int w,
								 int h,
								 IterType gradX,
								 IterType gradY,
								 const ImageContainer<ImgType> &kappaSet,
								 const Sigma<T> sigma)
	{
		T value;
		int N = kappaSet.size();
		T curSigma = sigma.Get(0);
		
		for (int i = 0; i < N; )
		{
			const ImgType *img = kappaSet[i];
			T meanDev;
			//std::cout << std::endl << "curSigma = " << curSigma;
			value = Average<T>(&meanDev, &(*img)[0], x, y, w, h, gradX, gradY,
				                 T(1)/*int(curSigma/4.0)*/, std::max(T(2), T(curSigma/2.0)));
			if (meanDev > 0.2)
			{
				//std::cout << "  too large dev";
				i++;
				curSigma = sigma.Next(curSigma);
				continue;
			}

			T radius = CurvatureScale<T>::value_to_radius(value);
			T reqSigma = RadiusToSigma(radius);
			
			//std::cout << std::endl << "prS = " << curSigma << "   i = " << i << "   value = " << value << "   r = " << radius << "   reqS = " << reqSigma;
			//if (4 * curSigma < radius)
			 if (reqSigma < sigma.Next(curSigma))
			{
					if (i > 0)
					{
						img = kappaSet[i - 1];
						T prevVal = Average<T>(&meanDev, &(*img)[0], x, y, w, h, gradX, gradY, 1, 1);

						//img = kappaSet[i + 1];
						//T nextVal = Average<T>(&meanDev, &(*img)[0], x, y, w, h, 1);

						//std::cout << std::endl << "  prevVal = " << prevVal << "   "  << 
						//	fabs(value - prevVal) / value << "  nextVal = " << nextVal << "  "
						//	<< fabs(value - nextVal) / value;

						if (/*fabs(value - nextVal) / value > 0.25 ||*/
								fabs(value - prevVal) / fabs(value) > 0.2)
						{
							//std::cout << std::endl << "  too large change between levels";
							i++;
						  curSigma = sigma.Next(curSigma);
							continue;
						}
					}
				break;
			}
			
			if (4 * curSigma > fabs(radius))
			{
			  break;
			}
			//std::cout << "   other sigma is requared";
			curSigma = sigma.Next(curSigma);/*reqSigma;*/
			//i = sigma.GetIndex(curSigma);
			i++;
		}
		return value;
	}

	//---------------------------------------------------------------------------
	template <typename IterType, typename T, typename ImgType>
	void CombinedKappaLines(IterType dst,
		                      IterType edgePoint,
													IterType gradX,
													IterType gradY,
										      int w,
										      int h,
										      T threshold,
													const ImageContainer<ImgType> &kappaSet,
											    const Sigma<T> sigma)
	{
		for (int y = 0; y < h; y++)
		{
			for (int x = 0; x < w; x++)
			{
				if ((*edgePoint) > threshold)
				{
					T val = ProperRadius(x, y, w, h, gradX, gradY, kappaSet, sigma);
					(*dst) = val;
				}
				else
				{
					(*dst) = T(0);
				}

				dst++;
				edgePoint++;
			}
		}
	}
	template <typename T,typename IterType>
	void SelectCurvatureRange(IterType dst,int w,int h,int radius_min=50,int radius_max=170)
	{
		T low_r= T(CurvatureScale<float>::radius_to_value(float(radius_min)));
		T high_r= T(CurvatureScale<float>::radius_to_value(float(radius_max)));
		for (int y = 0; y < h; y++)
		{
			for (int x = 0; x < w; x++)
			{
				T val = std::abs(*dst);
				if( (val < low_r) || (val > high_r) )
				{
				  (*dst) = 0;
				}
				dst++;
			}
		}
	}


}
}

#endif //COMBINED_KAPPA_LINES_H
