#ifndef HISTOGRAM_H
#define HISTOGRAM_H

namespace Vtl
{
///Histogramm is assumed stored as array or std::vector 
	namespace Histogramm
	{

		template<typename T> int TtoInt(T value) { return int(value); }

		template<typename T> 
		class TtoAbsIntBounded
		{
			const unsigned up;
		public:
			TtoAbsIntBounded(unsigned max_value)
				:up(max_value)
			{}

			unsigned operator()(T value)
			{ 
				unsigned res = std::min(unsigned(fabs(value)),up);
				/*if ( value < 0)
				{
				  std::cout << std::endl << "up = " << up << "  v = " << value
					  << "   fabs = " << unsigned(fabs(value)) << "  res = " << res;
				}*/
				return res;
			}
		};


		template<typename H>
		inline void Clear(H* hist,size_t N_hist)
		{
			std::fill(hist,hist+N_hist,H(0));
		}

		template<typename H>
		inline void Clear(std::vector<H>& hist)
		{
			hist.assign(hist.size(),H(0));
		}

		template<typename H,typename T,class F>
		inline void Add(H* hist,size_t N_hist,const T* data,int N_data,F TtoInt_=TtoInt)
		{
			for(int i=0; i<N_data ; ++i)
			{
				++hist[TtoInt_(data[i])];
			}
		}

		template<typename H,typename T,class F>
		inline void Add(H* hist,size_t N_hist,const T* image,int w,int h,int nls,F TtoInt_=TtoInt)
		{
			for( ; h > 0 ; --h, image+=nls)
				Add(hist,N_hist,image,w,TtoInt_);
		}

		template<typename H,typename T,class F>
		inline void Build(H* hist,size_t N_hist,const T* image,int w,int h,int nls,F TtoInt_=TtoInt)
		{
			Clear(hist,N_hist);
			Add(hist,N_hist,image,w,h,nls,TtoInt_);
		}

		template<typename H,typename T,class F>
		inline void Build(std::vector<H>& hist,const T* image,int w,int h,int nls,F TtoInt_=TtoInt)
		{
			Build(&(hist[0]),hist.size(),image,w,h,nls,TtoInt_);
		}

		template<typename H,typename T>
		inline void Draw(const H* hist,size_t N_hist,T* image,int w,int h,int nls)
		{
			H mh= *std::max_element(hist,hist+N_hist);
			
			assert(N_hist==size_t(w));
//			std::cout << "Draw : " << N_hist <<" " << w << " " << h << " " << nls << std::endl;
			for(size_t x=0; x<N_hist; ++x)
			{
				int len=int((log(hist[x]+1)/log(mh+1))*h);
				int y0=h-len;
				T* im=image+x;
//				std::cout << "len= " << len << "    y0= " << y0 << std::endl;
				int y=0;
				for( ; y<y0 ; ++y,im+=nls)
					*im=0;
				for( ; y<h ; ++y,im+=nls)
					*im=x;
			}
//			std::cout << "end_draw" << std::endl;
		}

		template<typename T>
		int split_max(const T* hist,int at_max,int at_end)
		{
			if((at_max-at_end)<3)
				return -1;
			T h_max=hist[at_max];
			++at_max;
			double alpha_old;
			int i_min=at_max;
			T h_min=hist[i_min];
			for( ;at_max<=at_end; ++at_max)
			{
				if(h_min > hist[at_max])
				{
					i_min=at_max;
					h_min=hist[i_min];
				}
				else
				{
					double alpha=h_max/(1.+h_min);
					if(alpha_old < alpha)
					{
					}
				}
			}

		}
		template<typename T>
		inline double _Alpha(T loc_min,T loc_max)
		{
			return double(loc_max)/(double(loc_min)+1.);
			//return double(loc_max+glob_max)/(double(loc_min)+1.);
			//return double(loc_max)/(double(loc_min)+1.)*double(loc_max)/(double(glob_max)+1.);
			//return double(loc_max)-double(loc_min);
		}

		class RevIndex
		{
			int i;
		public:
			RevIndex(int val) : i(val) {}
			operator int() const {return i;}
			RevIndex& operator=(int val) {i=val; return *this;}
			RevIndex& operator++() {--i;  return *this;} //in the other direction
			bool operator<=(int rvalue) const { return rvalue <= i;}  //in the other direction

		};
		template<typename T>
		inline double FindSecondMax(const T *data,int from,int to,int& l_min_ent,int& l_max_ent)
		{

			int l_min_ent2=from;
			T l_min=data[l_min_ent2];
			l_min_ent=l_min_ent2;
			l_max_ent=l_min_ent2;
			double alpha=_Alpha(l_min,l_min);
			if(from<to)
			{
				for(int x=from+1; x<= to ; ++x)
				{
					if(data[x]<l_min)
					{
						l_min_ent2=x;
						l_min=data[l_min_ent2];
					}
					else
					{
						double alpha2=_Alpha(l_min,data[x]);
						if(alpha2>alpha)
						{
							alpha=alpha2;
							l_max_ent=x;
							l_min_ent=l_min_ent2;
						}
					}
				}
			}
			else //(from>to)
			{
				for(int x=from-1; x >= to ; --x)
				{
					if(data[x]<l_min)
					{
						l_min_ent2=x;
						l_min=data[l_min_ent2];
					}
					else
					{
						double alpha2=_Alpha(l_min,data[x]);
						if(alpha2>alpha)
						{
							alpha=alpha2;
							l_max_ent=x;
							l_min_ent=l_min_ent2;
						}
					}
				}
			}
			return alpha;
		}
	}//namespace Histogramm
}//namespace Vtl


#endif  HISTOGRAM_H