// system includings
//#include <memory> 
#include <vector>
#include <iostream>

// library includings
#include "aimage.h"
#include "consolap.h"
#include "edt_meijster_2000.h"
#include "imalg_tools.h"

// own includings
#include "SaveImageArray.h"
#include "EdgeDetection/CannyEdgeDetector.h"
#include "EdtEdgeCoord.h"
#include "curvature/EdtEdgeBlur.h"


namespace Tests
{
  int TestEDT(int argc, char *argv[])
  {
    // Load input image
    if(argc < 2) 
    {
		  std::cout << "Argument - input image" << std::endl;
      return 0;
    }

    const char *fname = argv[1];

    std::auto_ptr<AImage> _img(::AImageLoad(0, true, fname, indicator));
		AImageMem8* img = dynamic_cast<AImageMem8*>(_img.get());
		if(!img)
		{
			return 0;
		}

		int w = img->Width();
		int h = img->Height();
		int size = w * h;
    int nls = img->NextLineShift();

    //         float
		std::vector<float> _src(size);
		float *src  = &(_src[0]);

		Vtl::Plf::F11::Convert<float> c_to_float;
		Vtl::ImAlg::Tools::copy_1_1((*img)[0], w, h, nls, src, w, c_to_float);

    const float sigma = 3.0f;
		const float nSigmasEnv = 3.0;
		const float histThStart = 10.0f;
		const float histThEnd = 0.0f;

		CannyEdgeDetector<float> canny(src, w, h);
		canny.Apply(sigma, nSigmasEnv, histThStart, histThEnd);

		Tests::SaveImageArray(&(canny.edges[0]), w, h, fname, "_Canny");

    std::vector<float> edt(size);
    std::vector<int> edgeX(size);
    std::vector<int> edgeY(size);
    
    Vtl::ImAlg::Edt::EdtEdgeCoord::edt(&(canny.edges[0]), w, h, nls, &(edt[0]),
              &(edgeX[0]), &(edgeY[0]), nls, Vtl::ImAlg::Edt::isEdge<float>());

    Tests::SaveNormImageArray(&(edt[0]), w, h, fname, "_EDT");

    /*int x = 249;
    int y = 243;
    std::cout << std::endl << "x = " << x << "  y = " << y << "  edgeY = "
       << edgeY[y * nls + x] << "  edgeX = " << edgeX[y * nls + x];
   
    x = 246;
    y = 237;
    std::cout << std::endl << "x = " << x << "  y = " << y << "  edgeY = "
       << edgeY[y * nls + x] << "  edgeX = " << edgeX[y * nls + x];

    x = 247;
   std::cout << std::endl << "x = " << x << "  y = " << y << "  edgeY = "
       << edgeY[y * nls + x] << "  edgeX = " << edgeX[y * nls + x];

   x = 248;
   std::cout << std::endl << "x = " << x << "  y = " << y << "  edgeY = "
       << edgeY[y * nls + x] << "  edgeX = " << edgeX[y * nls + x];

   x = 249;
   std::cout << std::endl << "x = " << x << "  y = " << y << "  edgeY = "
       << edgeY[y * nls + x] << "  edgeX = " << edgeX[y * nls + x];

    x = 246;
    y = 243;
    std::cout << std::endl << "x = " << x << "  y = " << y << "  edgeY = "
       << edgeY[y * nls + x] << "  edgeX = " << edgeX[y * nls + x];*/

    Tests::SaveNormImageArray(&(edgeX[0]), w, h, fname, "_edgeX");
    Tests::SaveNormImageArray(&(edgeY[0]), w, h, fname, "_edgeY");    

    float edgeLevel = 128.0f;
    std::vector<float> Lx(size);
    std::vector<float> Ly(size);
    EdtEdgeBlur(&(canny.edges[0]), &(Lx[0]), &(Ly[0]), &(edt[0]), &(edgeX[0]),
      &(edgeY[0]), w, h, sigma, edgeLevel);    

    Tests::SaveImageArray(&(canny.edges[0]), w, h, fname, "_CannyBlured");
    Tests::SaveImageArray(&(Lx[0]), w, h, fname, "_Lx");
    Tests::SaveImageArray(&(Ly[0]), w, h, fname, "_Ly");    
  }
}
