//#include <stdio.h>
#include <string>
#include <vector>
#include <iostream>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <sstream>
#include <memory> //auto_ptr()
#include <assert.h>

#include "aimage.h"
#include "integral_transforms.h"
#include "debug_test.h"
//#include "nummet.h"
#include "yfilesys.h"
#include "consolap.h"

#include "tests.h"
namespace Tests
{
	template<typename TT,typename II>
	class GaussFun
	{
		II N;
		TT sigma;
		TT c;
	public:
		GaussFun(II n,II n_sigmas) : N(n)
		{
			sigma=TT(N/2./TT(n_sigmas));
			c=0;
			for(II i=0; i<n; ++i)
				c+=fun_value(i);
			c=1/c;
		}
		TT Sigma() const { return sigma;}
		TT operator()(II i) const
		{
			return c*fun_value(i); 
		}
	private:
		TT fun_value(II i) const
		{
			TT x=TT(TT(i)-N/2)/sigma;
			//return TT(1./std::sqrt(2*M_PI)/sigma)*std::exp(-x*x/2); 
			return TT(std::exp(-x*x/2)); 
		}
	};


	bool test_Utils_smallest_log2_ge_size()
	{
		std::cout << "====== Test Vtl::NumAlg::IntTransf::Utils::smallest_log2_ge_size() function =======" << std::endl;
		bool is_ok=true;
		int size=1024;
		int log_size=Vtl::NumAlg::IntTransf::Utils::smallest_log2_ge_size(size);
		std::cout << "Vtl::NumAlg::IntTransf::Utils::smallest_log2_gt_size(" << size<< ")=" << log_size << std::endl;
		if(log_size!=10)
			is_ok=false;

		size=1023;
		log_size=Vtl::NumAlg::IntTransf::Utils::smallest_log2_ge_size(size);
		std::cout << "Vtl::NumAlg::IntTransf::Utils::smallest_log2_gt_size(" << size<< ")=" << log_size << std::endl;
		if(log_size!=10)
			is_ok=false;

		size=513;
		log_size=Vtl::NumAlg::IntTransf::Utils::smallest_log2_ge_size(size);
		std::cout << "Vtl::NumAlg::IntTransf::Utils::smallest_log2_gt_size(" << size<< ")=" << log_size << std::endl;
		if(log_size!=10)
			is_ok=false;
		size=512;
		log_size=Vtl::NumAlg::IntTransf::Utils::smallest_log2_ge_size(size);
		std::cout << "Vtl::NumAlg::IntTransf::Utils::smallest_log2_gt_size(" << size<< ")=" << log_size << std::endl;
		if(log_size!=9)
			is_ok=false;
		size=640;
		log_size=Vtl::NumAlg::IntTransf::Utils::CalcSmalestPow2(size);
		std::cout << "Vtl::NumAlg::IntTransf::Utils::CalcSmalestPow2(" << size<< ")=" << log_size << std::endl;
		if(log_size!=1024)
			is_ok=false;
		std::cout << (is_ok?"test OK":"test ERROR") << std::endl << "==============" << std::endl;
		return is_ok;
	}
	//-----------------------------
	bool test_Utils_rev()
	{
		std::cout << "================ Test bit reversing function ====================" << std::endl;
		unsigned size=8;
		unsigned log_size=3;
		unsigned step=2;
		bool is_ok=true;
		for(unsigned  i=0; i<size; ++i)
		{
			unsigned  rev=Vtl::NumAlg::IntTransf::Utils::rev(i, log_size);
			unsigned  rev2=Vtl::NumAlg::IntTransf::Utils::rev(i*step, log_size,step);
			std::cout << "Vtl::NumAlg::IntTransf::Utils::rev(" << i <<","<< log_size << ")=" << rev 
				<< " | i*step=" << i*step << " rev(i*step)=" << rev2 << " rev2/step= " << rev2/step << "==" << rev << std::endl;
			if(rev2/step != rev )
				is_ok=false;
			if(i != Vtl::NumAlg::IntTransf::Utils::rev(rev, log_size) )
				is_ok=false;
		}
		std::cout << (is_ok?"test OK":"test ERROR") << std::endl << "==============" << std::endl;
		return is_ok;
	}
	//-----------------------------
	bool test_Utils_bit_reversal()
	{
		std::cout << "======== Test reordering sparse REAL array in bitreverse order ============" << std::endl;
		bool is_ok=true;
		typedef float Real_t;
		typedef size_t Int_t;
		const bool is_complex=false;

		std::vector<Real_t> _data;
		for(Int_t n=1; n<32; n*=2)
		for(Int_t _step=1; _step<32; _step*=2)
		{
			std::cout << "size=" <<n<<" , step=" <<_step <<":"<<std::endl;
			Int_t step=(is_complex?2:1)*_step; //!!!!!!!!!!!1
			//const Int_t n=8;
			//const Int_t step=4;

			//allocate space and init
			const Int_t log_size=Vtl::NumAlg::IntTransf::Utils::smallest_log2_ge_size(n);
			Int_t data_sz=n*step;
			_data.resize(data_sz);
			Real_t *data=&(_data[0]);
			memset(data,0,data_sz*sizeof(Real_t));
			//Assign data for test 
			for(Int_t  i=0; i< n; ++i)
			{  // i  
				Int_t  j=i*step; //   
				//---------------------------------------
				data[j]=i+0.01f;
			}
			//call routine to test
			Vtl::NumAlg::IntTransf::Utils::bit_reversal(data,n,step); 

			//    
			for(Int_t  i=0; i< n; ++i)
			{  // i  
				Int_t  j=i*step; //   
				//---------------------------------------
				Int_t rev=Vtl::NumAlg::IntTransf::Utils::rev(i, log_size);
				std::cout << i << " , " << data[j] << " : rev(i)="<< rev << std::endl;
				if(abs(rev-data[j])>0.011)
					is_ok=false;
			}
			std::cout << "-----------" << std::endl;
		}
		std::cout << (is_ok?"test OK":"test ERROR") << std::endl << "==============" << std::endl;
		return is_ok;
	}
	//-----------------------------
	bool test_Utils_bit_reversal_complex_pair()
	{
		std::cout << "======== Test reordering sparse COMPLEX array in bitreverse order ============" << std::endl;
		bool is_ok=true;
		typedef float Real_t;
		typedef size_t Int_t;
		const bool is_complex=true;

		std::vector<Real_t> _data;
		for(Int_t n=1; n<32; n*=2)
		for(Int_t _step=1; _step<32; _step*=2)
		{
			std::cout << "size=" <<n<<" , step=" <<_step <<":"<<std::endl;
			Int_t step=(is_complex?2:1)*_step; //!!!!!!!!!!!1
			//const Int_t n=8;
			//const Int_t step=4;

			//allocate space and init
			const Int_t log_size=Vtl::NumAlg::IntTransf::Utils::smallest_log2_ge_size(n);
			Int_t data_sz=n*step;
			_data.resize(data_sz);
			Real_t *data=&(_data[0]);
			memset(data,0,data_sz*sizeof(Real_t));

			//Assign data for test 
			for(Int_t  i=0; i< n; ++i)
			{  // i  
				Int_t  j=i*step; //   
				//---------------------------------------
				data[j]=i+0.01f; //Re
				data[j+1]=i+0.02f;//Im
			}
			
			//call routine to test
			Vtl::NumAlg::IntTransf::Utils::bit_reversal_complex(data,n,step); 

			//    
			for(Int_t  i=0; i< n; ++i)
			{  // i  
				Int_t  j=i*step; //   
				//---------------------------------------
				Int_t rev=Vtl::NumAlg::IntTransf::Utils::rev(i, log_size);
				std::cout << i << " , (" << data[j] <<","<< data[j+1] << ") : rev(i)="<< rev << std::endl;
				if(abs(rev-data[j])>0.011 || abs(rev-data[j+1])>0.021)
					is_ok=false;
			}
			std::cout << "-----------" << std::endl;
		}
		std::cout << (is_ok?"test OK":"test ERROR") << std::endl << "==============" << std::endl;
		return is_ok;
	}
	//-----------------------------
	bool test_Utils_bit_reversal_complex_separate()
	{
		std::cout << "======== Test reordering sparse COMPLEX array (separate loc.) in bitreverse order ============" << std::endl;
		bool is_ok=true;
		typedef float Real_t;
		typedef size_t Int_t;
		const bool is_complex=true;

		std::vector<Real_t> _data;
		for(Int_t n=1; n<32; n*=2)
		for(Int_t _step=1; _step<32; _step*=2)
		{
			std::cout << "size=" <<n<<" , step=" <<_step <<":"<<std::endl;
			Int_t step=1*_step; //!!!!!!!!!!!1
			//const Int_t n=8;
			//const Int_t step=4;

			//allocate space and init
			const Int_t log_size=Vtl::NumAlg::IntTransf::Utils::smallest_log2_ge_size(n);
			Int_t data_sz=(is_complex?2:1)*n*step;
			_data.resize(data_sz);
			Real_t *re=&(_data[0]);
			Real_t *im=&(_data[data_sz/2]);
			memset(re,0,data_sz*sizeof(Real_t));

			//Assign data for test 
			for(Int_t  i=0; i< n; ++i)
			{  // i  
				Int_t  j=i*step; //   
				//---------------------------------------
				re[j]=i+0.01f; //Re
				im[j]=i+0.02f;//Im
			}
			
			//call routine to test
			Vtl::NumAlg::IntTransf::Utils::bit_reversal_complex(re,im,n,step); 

			//    
			for(Int_t  i=0; i< n; ++i)
			{  // i  
				Int_t  j=i*step; //   
				//---------------------------------------
				Int_t rev=Vtl::NumAlg::IntTransf::Utils::rev(i, log_size);
				std::cout << i << " , (" << re[j] <<","<< im[j] << ") : rev(i)="<< rev << std::endl;
				if(abs(rev-re[j])>0.011 || abs(rev-im[j])>0.021)
					is_ok=false;
			}
			std::cout << "-----------" << std::endl;
		}
		std::cout << (is_ok?"test OK":"test ERROR") << std::endl << "==============" << std::endl;
		return is_ok;
	}
	//-----------------------------
	//-----------------------------
	bool test_Utils_SwapArrayHalfs_real()
	{
		std::cout << "========= Test Utils::SwapArrayHalfs() for sparse REAL array =========" << std::endl;
		bool is_ok=true;
		typedef float Real_t;
		typedef size_t Int_t;
		const bool is_complex=false;

		std::vector<Real_t> _data;
		for(Int_t n=1; n<32; n*=2)
		for(Int_t _step=1; _step<32; _step*=2)
		{
			std::cout << "size=" <<n<<" , step=" <<_step <<":"<<std::endl;
			Int_t step=1*_step; //!!!!!!!!!!!1
			//const Int_t n=8;
			//const Int_t step=4;

			//allocate space and init
			const Int_t log_size=Vtl::NumAlg::IntTransf::Utils::smallest_log2_ge_size(n);
			Int_t data_sz=(is_complex?2:1)*n*step;
			_data.resize(data_sz);
			Real_t *data=&(_data[0]);
			memset(data,0,data_sz*sizeof(Real_t));

			//Assign data for test 
			for(Int_t  i=0; i< n; ++i)
			{  // i  
				Int_t  j=i*step; //   
				//---------------------------------------
				data[j]=i+0.01f; //Re
			}
			
			//call routine to test
			Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs<is_complex>(data,n,step);

			//    
			for(Int_t  i=0; i< n; ++i)
			{  // i  
				Int_t  j=i*step; //   
				//---------------------------------------
				int k;
				if(i<n/2)
					k=i+n/2;
				else
					k=i-n/2;
				std::cout << i << "  data[i]=" << data[j] <<"   =="<< (k+0.01) << std::endl;
				if( abs(data[j]-(k+0.01)) > 0.001 )
					is_ok=false;
			}
			std::cout << "-----------" << std::endl;
		}
		std::cout << (is_ok?"test OK":"test ERROR") << std::endl << "==============" << std::endl;
		return is_ok;
	}
	//-----------------------------
	bool test_Utils_SwapArrayHalfs_complex_pairs()
	{
		std::cout << "====== Test Utils::SwapArrayHalfs() for sparse COMPLEX array (loc. as pairs)======" << std::endl;
		bool is_ok=true;
		typedef float Real_t;
		typedef size_t Int_t;
		const bool is_complex=true;

		std::vector<Real_t> _data;
		for(Int_t n=1; n<32; n*=2)
		for(Int_t _step=1; _step<32; _step*=2)
		{
			std::cout << "size=" <<n<<" , step=" <<_step <<":"<<std::endl;
			Int_t step=(is_complex?2:1)*_step; //!!!!!!!!!!!2
			//const Int_t n=8;
			//const Int_t step=4;

			//allocate space and init
			const Int_t log_size=Vtl::NumAlg::IntTransf::Utils::smallest_log2_ge_size(n);
			Int_t data_sz=n*step;
			_data.resize(data_sz);
			Real_t *data=&(_data[0]);
			memset(data,0,data_sz*sizeof(Real_t));

			//Assign data for test 
			for(Int_t  i=0; i< n; ++i)
			{  // i  
				Int_t  j=i*step; //   
				//---------------------------------------
				data[j]=i+0.01f; //Re
				data[j+1]=i+0.02f; //Re
			}
			
			//call routine to test
			Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs<is_complex>(data,n,step);

			//    
			for(Int_t  i=0; i< n; ++i)
			{  // i  
				Int_t  j=i*step; //   
				//---------------------------------------
				int k;
				if(i<n/2)
					k=i+n/2;
				else
					k=i-n/2;
				std::cout << i << "  data[i]=(" << data[j] << "," << data[j+1] <<")"  << std::endl;
				if( abs(data[j]-(k+0.01)) > 0.001 || abs(data[j+1]-(k+0.02)) > 0.001 )
					is_ok=false;
			}
			std::cout << "-----------" << std::endl;
		}
		std::cout << (is_ok?"test OK":"test ERROR") << std::endl << "==============" << std::endl;
		return is_ok;
	}
	//-----------------------------
	bool test_Utils_SwapArrayHalfs_complex_separate()
	{
		std::cout << "====== Test Utils::SwapArrayHalfs() for sparse REAL array (separate loc.)======" << std::endl;
		bool is_ok=true;
		typedef float Real_t;
		typedef size_t Int_t;
		const bool is_complex=true;

		std::vector<Real_t> _data;
		for(Int_t n=1; n<32; n*=2)
		for(Int_t _step=1; _step<32; _step*=2)
		{
			std::cout << "size=" <<n<<" , step=" <<_step <<":"<<std::endl;
			Int_t step=1*_step; //!!!!!!!!!!!1
			//const Int_t n=8;
			//const Int_t step=4;

			//allocate space and init
			const Int_t log_size=Vtl::NumAlg::IntTransf::Utils::smallest_log2_ge_size(n);
			Int_t data_sz=(is_complex?2:1)*n*step;
			_data.resize(data_sz);
			Real_t *re=&(_data[0]);
			Real_t *im=&(_data[data_sz/2]);
			memset(re,0,data_sz*sizeof(Real_t));

			//Assign data for test 
			for(Int_t  i=0; i< n; ++i)
			{  // i  
				Int_t  j=i*step; //   
				//---------------------------------------
				re[j]=i+0.01f; //Re
				im[j]=i+0.02f;//Im
			}
			
			//call routine to test
			Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs(re,im,n,step);

			//    
			for(Int_t  i=0; i< n; ++i)
			{  // i  
				Int_t  j=i*step; //   
				//---------------------------------------
				int k;
				if(i<n/2)
					k=i+n/2;
				else
					k=i-n/2;
				std::cout << i << "  re[j],im[j]=(" << re[j] << "," << im[j] <<")"  << std::endl;
				if( abs(re[j]-(k+0.01)) > 0.001 || abs(im[j]-(k+0.02)) > 0.001 )
					is_ok=false;
			}
			std::cout << "-----------" << std::endl;
		}
		std::cout << (is_ok?"test OK":"test ERROR") << std::endl << "==============" << std::endl;
		return is_ok;
	}
	//-----------------------------
	//-----------------------------
	template<typename Real_t>
	bool test_Four1_forward_reverse_equality()
	{
		std::cout << "======== Test 1D FFT and IFFT make equality (complex at separate loc.) in bitreverse order ============" << std::endl;
		bool is_ok=true;
		//typedef float Real_t;
		typedef int Int_t;
		const bool is_complex=true;

		std::vector<Real_t> _data;
		for(Int_t n=2; n<64; n*=2)
		for(Int_t _step=1; _step<8; _step*=2)
		{
			std::cout << "size=" <<n<<" , step=" <<_step <<":" << std::endl;
			Int_t step=1*_step; //!!!!!!!!!!!1
			//const Int_t n=8;
			//const Int_t step=4;

			//allocate space and init
			Int_t data_sz=(is_complex?2:1)*n*step;
			_data.resize(data_sz);
			Real_t *re=&(_data[0]);
			Real_t *im=&(_data[data_sz/2]);
			memset(re,0,data_sz*sizeof(Real_t));

			//Assign data for test 
			GaussFun<Real_t,Int_t> fun(n,4);
			for(Int_t  i=0; i< n; ++i)
			{  // i  
				Int_t  j=i*step; //   
				//---------------------------------------
				re[j]=fun(i);
				im[j]=0;//Im
			}
			
			//call routine to test
			Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs(re,im,n,step);
			Vtl::NumAlg::IntTransf::Fourier::fft1D<false>(re,im,n,step); 
			Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs(re,im,n,step);
			Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs(re,im,n,step);
			Vtl::NumAlg::IntTransf::Fourier::fft1D<true>(re,im,n,step); 
			Vtl::NumAlg::IntTransf::Fourier::norm_ifft(re,n,step);
			Vtl::NumAlg::IntTransf::Fourier::norm_ifft(im,n,step);
			Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs(re,im,n,step);



			//    
			std::cout << "sigme=" << fun.Sigma() << std::endl;
			Real_t max_re_diff=0;
			Real_t max_im_diff=0;
			//---------------------------------------
			int c=(n/2)*step;
			Real_t amp_c=Real_t(std::sqrt(re[c]*re[c]+im[c]*im[c]));
			for(Int_t  i=0; i< n; ++i)
			{  // i  
				Int_t  j=i*step; //   
				//---------------------------------------
				Real_t amp=std::sqrt(re[j]*re[j]+im[j]*im[j]);
				Real_t phi=std::atan2(im[j],re[j]);
				std::cout << std::setprecision(7) <<std::setiosflags(std::ios::showpoint|std::ios::fixed);
				std::cout << std::setw(2) << i << ",(" << std::setw(10) << re[j] <<","<< std::setw(10) << im[j] 
				<< " (amp,phi)=("<< std::setw(10) << amp << ","<< std::setw(10) << phi  
					<< ") : gauss="<< std::setw(10) << fun(i) << std::endl;
				max_re_diff=std::max(max_re_diff,std::abs(re[j]-fun(i)));
				max_im_diff=std::max(max_im_diff,std::abs(im[j]-0));
			}
			// In next line 1 -means that function norm (integral, integral of modul is 1
			if( max_re_diff > 1*Vtl::NumAlg::RelerrEps<Real_t>() || max_im_diff > 1*Vtl::NumAlg::RelerrEps<Real_t>() )
				is_ok=false;
			std::cout << std::setprecision(9) << std::setiosflags(std::ios::showpoint|std::ios::scientific);
			std::cout << "max_re_diff=" << max_re_diff << "   max_im_diff=" << max_im_diff << std::endl;
			std::cout << std::setprecision(7) << std::resetiosflags(std::ios::scientific);
			std::cout << "-----------" << std::endl;
		}
		std::cout << (is_ok?"test OK":"test ERROR") << std::endl << "==============" << std::endl;
		return is_ok;
	}
	//-----------------------------
	//-----------------------------
	template<typename Real_t>
	bool test_Four1_step()
	{
		std::cout << "======== Test 1D FFT correct for all data steps  (complex at separate and seq. loc.) ============" << std::endl;
		bool is_ok=true;
		//typedef float Real_t;
		typedef int Int_t;
		//const bool is_complex=true;

		std::vector<Real_t> _data0;
		std::vector<Real_t> _data;
		for(Int_t n=2; n<64; n*=2)
		{
			std::cout << "size=" <<n<<" , step=" <<1 <<":" << std::endl;
			Int_t data_sz0=2*n*1; //2 as is_complex, step==1
			_data0.resize(data_sz0);
			Real_t *re0=&(_data0[0]);
			Real_t *im0=&(_data0[data_sz0/2]);
			memset(re0,0,data_sz0*sizeof(Real_t));
			//Assign data for test 
			GaussFun<Real_t,Int_t> fun(n,4);
			for(Int_t  i=0; i< n; ++i)
			{  // i  
				re0[i]=fun(i);
				im0[i]=0;//Im
			}
			Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs(re0,im0,n,1);
			Vtl::NumAlg::IntTransf::Fourier::fft1D<false>(re0,im0,n,1); 
			Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs(re0,im0,n,1);

			for(Int_t _step=2; _step<8; _step*=2)
			{
				std::cout << "size=" <<n<<" , step=" <<_step <<":" << std::endl;
				Int_t step=1*_step; //!!!!!!!!!!!1

				//allocate space and init
				Int_t data_sz=2*n*step; //2 as is_complex
				_data.resize(data_sz);
				Real_t *re=&(_data[0]);
				Real_t *im=&(_data[data_sz/2]);
				memset(re,0,data_sz*sizeof(Real_t));

				//Assign data for test 
				for(Int_t  i=0; i< n; ++i)
				{  // i  
					Int_t  j=i*step; //   
					//---------------------------------------
					re[j]=fun(i);
					im[j]=0;//Im
				}
				
				//call routine to test
				Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs(re,im,n,step);
				Vtl::NumAlg::IntTransf::Fourier::fft1D<false>(re,im,n,step); 
				Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs(re,im,n,step);


				//    
				std::cout << "sigme=" << fun.Sigma() << std::endl;
				Real_t max_re_diff=0;
				Real_t max_im_diff=0;
				//---------------------------------------
				for(Int_t  i=0; i< n; ++i)
				{  // i  
					Int_t  j=i*step; //   
					//---------------------------------------
					Real_t d_re=re[j]-re0[i];
					Real_t d_im=im[j]-im0[i];
					std::cout << std::setprecision(8) <<std::setiosflags(std::ios::showpoint|std::ios::fixed);
					std::cout << std::setw(2) << i << ",(" << std::setw(10) << re[j] <<","<< std::setw(10) << im[j] << ") "
						//<< "step=1:("<< std::setw(10) << re0[i] << ","<< std::setw(10) << im0[i] << ") " 
						<< ": delta="
						<< std::setprecision(2) << std::setiosflags(std::ios::showpoint|std::ios::scientific)
						<< d_re << "," << d_im << ")"
						<< std::endl;
					max_re_diff=std::max(max_re_diff,std::abs(d_re));
					max_im_diff=std::max(max_im_diff,std::abs(d_im));
				}
				// In next line 1 -means that function norm (integral, integral of modul is 1
				if( max_re_diff > 1*Vtl::NumAlg::RelerrEps<Real_t>() || max_im_diff > 1*Vtl::NumAlg::RelerrEps<Real_t>() )
					is_ok=false;
				std::cout << std::setprecision(9) << std::setiosflags(std::ios::showpoint|std::ios::scientific);
				std::cout << "max_re_diff=" << max_re_diff << "   max_im_diff=" << max_im_diff << std::endl;
				std::cout << std::setprecision(7) << std::resetiosflags(std::ios::scientific);
				std::cout << "-----------" << std::endl;
			}
		}
		std::cout << (is_ok?"test OK":"test ERROR") << std::endl << "==============" << std::endl;
		return is_ok;
	}
	//-----------------------------
	//-----------------------------
	bool test_Four1_mathlab()
	{
		std::cout << "======== Test 1D FFT  (complex at separate loc.) vs. matlab ============" << std::endl;
		bool is_ok=true;
		typedef float Real_t;
		typedef int Int_t;
		const bool is_complex=true;

		std::vector<Real_t> _data0;
		std::vector<Real_t> _data;
		for(Int_t n=2; n<64; n*=2)
		{
			std::fstream fs;
			std::ostringstream fn;
			fn << ".\\nummet\\mat\\gauss_fft_" << std::setw(2) << std::setfill('0') << n <<".dat" ;
			fs.open(fn.str().c_str(),std::ios_base::in);
			if(!fs)
			{
				std::ostringstream os;
				os<<"Can't open file "<< fn.str();
				throw os.str();
			}
			_data0.resize(2*n);
			Real_t *re0=&(_data0[0]);
			Real_t *im0=re0+n;
			for(Int_t i=0; i<n; ++i)
			{
				fs >> re0[i]; 
				fs >> im0[i]; 
			}
			fs.close();
			for(Int_t _step=1; _step<8; _step*=2)
			{
				std::cout << "size=" <<n<<" , step=" <<_step <<":" << std::endl;
				Int_t step=1*_step; //!!!!!!!!!!!1
				//const Int_t n=8;
				//const Int_t step=4;

				//allocate space and init
				const Int_t log_size=Vtl::NumAlg::IntTransf::Utils::smallest_log2_ge_size(n);
				Int_t data_sz=(is_complex?2:1)*n*step;
				_data.resize(data_sz);
				Real_t *re=&(_data[0]);
				Real_t *im=&(_data[data_sz/2]);
				memset(re,0,data_sz*sizeof(Real_t));

				//Assign data for test 
				Real_t sigma=Real_t(n/2./8.);
				Real_t sum=0;
				for(Int_t  i=0; i< n; ++i)
				{  // i  
					Int_t  j=i*step; //   
					//---------------------------------------
					Real_t x=Real_t(Real_t(i)-n/2)/sigma ;
					re[j]=Real_t(1./std::sqrt(2*M_PI)/sigma)*std::exp(-x*x/2); 
					im[j]=0;//Im
					sum+=re[j];
				}
				//Norm
				for(Int_t  i=0; i< n; ++i)
				{  // i  
					Int_t  j=i*step; //   
					//---------------------------------------
					re[j]=re[j]/sum; 
					im[j]=0;//Im
				}
				
				//call routine to test
				Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs(re,im,n,step);
				Vtl::NumAlg::IntTransf::Fourier::fft1D<false>(re,im,n,step); 
				Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs(re,im,n,step);


				//    
				//Real_t sigma_rev=Real_t(n/2./M_PI/sigma);//*1.07276f;
				Real_t max_re_diff=0;
				Real_t max_im_diff=0;
				//---------------------------------------
				for(Int_t  i=0; i< n; ++i)
				{  // i  
					Int_t  j=i*step; //   
					//---------------------------------------
					Real_t d_re=re[j]-re0[i];
					Real_t d_im=im[j]-im0[i];
					std::cout << std::setprecision(8) <<std::setiosflags(std::ios::showpoint|std::ios::fixed);
					std::cout << std::setw(2) << i << ",(" << std::setw(10) << re[j] <<","<< std::setw(10) << im[j] << ") "
						//<< "step=1:("<< std::setw(10) << re0[i] << ","<< std::setw(10) << im0[i] << ") " 
						<< ": delta="
						<< std::setprecision(2) << std::setiosflags(std::ios::showpoint|std::ios::scientific)
						<< d_re << "," << d_im << ")"
						<< std::endl;
					max_re_diff=std::max(max_re_diff,std::abs(d_re));
					max_im_diff=std::max(max_im_diff,std::abs(d_im));
				}
				// In next line 2 -means that function norm (integral, integral of modul is 1,
				// and 2=2*1 where 2 multiplier arises from RelerrEps() function returns value
				// 2 times smaller than that of value added to 1 is really increase 1
				if( max_re_diff > 2*Vtl::NumAlg::RelerrEps<Real_t>() || max_im_diff > 2*Vtl::NumAlg::RelerrEps<Real_t>() )
					is_ok=false;
				std::cout << std::setprecision(9) << std::setiosflags(std::ios::showpoint|std::ios::scientific);
				std::cout << "max_re_diff=" << max_re_diff << "   max_im_diff=" << max_im_diff << std::endl;
				std::cout << std::setprecision(7) << std::resetiosflags(std::ios::scientific);
				std::cout << "-----------" << std::endl;
			}
		}
		std::cout << (is_ok?"test OK":"test ERROR") << std::endl << "==============" << std::endl;
		return is_ok;
	}
	//-----------------------------
	//-----------------------------
	bool test_hartley_vs_fft()
	{
		std::cout << "======== Test Hartley vs 1D FFT  (complex at separate loc.) ============" << std::endl;
		bool is_ok=true;
		typedef float Real_t;
		typedef int Int_t;
		const bool is_complex=true;

		std::vector<Real_t> _data;
		std::vector<Real_t> _data_h;
		for(Int_t n=2; n<64; n*=2)
		for(Int_t _step=1; _step<32; _step*=2)
		{
			std::cout << "size=" <<n<<" , step=" <<_step <<":" << std::endl;
			Int_t step=1*_step; //!!!!!!!!!!!1
			//const Int_t n=8;
			//const Int_t step=4;

			//allocate space and init
			const Int_t log_size=Vtl::NumAlg::IntTransf::Utils::smallest_log2_ge_size(n);
			Int_t data_sz=2*n*step;//is_complex
			_data.resize(data_sz);
			_data_h.resize((data_sz/2)+2*n);//is real, 1 - for hartley, 2*n - fft_real and fft_imag
			Real_t *re=&(_data[0]);
			Real_t *im=&(_data[data_sz/2]);
			Real_t *hart=&(_data_h[0]);
			Real_t *hart_re=hart+(data_sz/2);
			Real_t *hart_im=hart_re+n;
			memset(re,0,data_sz*sizeof(Real_t));
			memset(hart,0,_data_h.size()*sizeof(Real_t));

			//Assign data for test 
			//Assign data for test 
			GaussFun<Real_t,Int_t> fun(n,8);
			for(Int_t  i=0; i< n; ++i)
			{  // i  
				Int_t  j=i*step; //   
				//---------------------------------------
				re[j]=fun(i);
				im[j]=0;//Im
				hart[j]=re[j];
			}
			
			
			//call routine to test
			Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs(re,im,n,step);
			Vtl::NumAlg::IntTransf::Fourier::fft1D<false>(re,im,n,step); 
			Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs(re,im,n,step);
			Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs<false>(hart,n,step);
			Vtl::NumAlg::IntTransf::Hartley::fht1D_(hart,n,step); 
			Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs<false>(hart,n,step);
			if(step==1)
			{
				int nn[]={n};
				Vtl::NumAlg::IntTransf::Hartley::GetReal<1>(hart,nn,hart_re); 
				Vtl::NumAlg::IntTransf::Hartley::GetImag<1>(hart,nn,hart_im); 
			}
			else
			{
				Vtl::NumAlg::IntTransf::Hartley::GetReal_1D(hart,n,hart_re,step,1); 
				Vtl::NumAlg::IntTransf::Hartley::GetImag_1D(hart,n,hart_im,step,1); 
			}


			//    
			std::cout << "sigma=" << fun.Sigma() << std::endl;
			Real_t max_re_diff=0;
			Real_t max_im_diff=0;
			//---------------------------------------
			int c=(n/2)*step;
			Real_t amp_c=Real_t(std::sqrt(re[c]*re[c]+im[c]*im[c]));
			for(Int_t  i=0; i< n; ++i)
			{  // i  
				Int_t  j=i*step; //   
				//---------------------------------------
				Real_t amp=std::sqrt(re[j]*re[j]+im[j]*im[j]);
				Real_t phi=std::atan2(im[j],re[j]);
				std::cout << std::setprecision(7) <<std::setiosflags(std::ios::showpoint|std::ios::fixed);
				std::cout << std::setw(2) << i << ",(" << std::setw(10) << re[j] <<","<< std::setw(10) << im[j] 
				<< " (hart_re,hart_im)=("<< std::setw(10) << hart_re[i] << ","<< std::setw(10) << hart_im[i]  << ") "
					<< std::endl;
				max_re_diff=std::max(max_re_diff,std::abs(re[j]-hart_re[i]));
				max_im_diff=std::max(max_im_diff,std::abs(im[j]-hart_im[i]));
			}
			// In next line 1 -means that function norm (integral, integral of modul is 1
			if( max_re_diff > 3*Vtl::NumAlg::RelerrEps<Real_t>() || max_im_diff > 3*Vtl::NumAlg::RelerrEps<Real_t>() )
				is_ok=false;
			std::cout << std::setprecision(9) << std::setiosflags(std::ios::showpoint|std::ios::scientific);
			std::cout << "max_re_diff=" << max_re_diff << "   max_im_diff=" << max_im_diff << std::endl;
			std::cout << std::setprecision(7) << std::resetiosflags(std::ios::scientific);
			std::cout << "-----------" << std::endl;
		}
		std::cout << (is_ok?"test OK":"test ERROR") << std::endl << "==============" << std::endl;
		return is_ok;
	}
	//-----------------------------
	bool test_speed()
	{
		std::cout << "Test FFT/Hartley speeed: " <<std::endl;
		std::vector<float> _data;
		size_t max_sz=4096;

		size_t data_sz=2*max_sz*max_sz;//is_complex
		_data.resize(data_sz+32);
		float *re=&(_data[0]);
		float *im=&(_data[data_sz/2+32]);

		for(size_t i=0; i<_data.size() ; ++i)
			re[i]=float(i);
		Vtl::Dbg::TimeElapsed t;
		int k=1;

		for(int n=max_sz; n>100; n/=2, k*=4)
		{
			size_t nn[]={n,n};

			t.start();
			for(int i=0; i<k ;++i)
			{
				Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2>(re,im,nn);
				Vtl::NumAlg::IntTransf::Fourier::fft2D(re,im,nn);
				Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2>(re,im,nn);
			}
			t.finish();
			std::cout << "FFT NxN , N=" << std::setw(6) << n <<"   time="<< std::setw(6) << t.Get()/k << "  sec." <<std::endl;

			t.start();
			for(int i=0; i<k ;++i)
			{
			Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2,false>(re,nn);
			Vtl::NumAlg::IntTransf::Hartley::fht2D_square(re,n,n); 
			Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2,false>(re,nn);
			}
			t.finish();
			std::cout << "Hartley NxN , N=" << std::setw(6) << n <<"   time="<< std::setw(6) << t.Get()/k  << "  sec." << std::endl;
	
		}
		std::cout << "Test speeed finished. " <<std::endl;
		return true;
	}
	//-----------------------------
	bool test_hartley_square()
	{
		std::cout << "======== Test 2D FHT  ============" << std::endl;
		bool is_ok=true;
		typedef float Real_t;
		typedef int Int_t;
		const bool is_complex=true;

		std::vector<Real_t> _data;
		for(Int_t n=2; n<16; n*=2)
		{
			Int_t data_sz=n*n;
			_data.resize(data_sz);
			Real_t *data=&(_data[0]);
			memset(data,0,data_sz*sizeof(Real_t));

			std::cout << "size = " <<n<<" x " << n <<":" << std::endl;

				//allocate space and init
//				const Int_t log_size=Vtl::NumAlg::IntTransf::Utils::smallest_log2_ge_size(n);

				//Assign data for test 
				Int_t step=1;
				Real_t sigma=Real_t(n/2./4.);
				Real_t sum=0;
				for(Int_t  y=0; y< n; ++y)
				for(Int_t  x=0; x< n; ++x)
				{  // i  
					Int_t  j=(y*n+x)*step; //   
					//---------------------------------------
					Real_t x_=Real_t(Real_t(x)-n/2)/sigma ;
					Real_t y_=Real_t(Real_t(y)-n/2)/sigma ;
					data[j]=Real_t(1./std::sqrt(2*M_PI)/sigma)*std::exp(-(x_*x_+y_*y_)/2); 
					sum+=data[j];
				}
				//Norm
				for(Int_t  i=0; i< n; ++i)
				{  // i  
					Int_t  j=i*step; //   
					//---------------------------------------
					data[j]=data[j]/sum; 
				}
				
				//call routine to test
				Int_t nn[]={n,n};
				Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2,false>(data,nn);
				//Vtl::NumAlg::IntTransf::CHartley<Real_t>::TransposeQuarters(data,n);
				Vtl::NumAlg::IntTransf::Hartley::fht2D_square(data,n,n); 
				//Vtl::NumAlg::IntTransf::CHartley<Real_t>::TransposeQuarters(data,n);
				Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2,false>(data,nn);
				
				//Vtl::NumAlg::IntTransf::CHartley<Real_t>::Do2D(data,n);


				//    
				//Real_t sigma_rev=Real_t(n/2./M_PI/sigma);//*1.07276f;
				//---------------------------------------
				for(Int_t  y=0; y< n; ++y)
				{
					for(Int_t  x=0; x< n; ++x)
					{  // i  
						Int_t j=y*n+x;
						std::cout << std::setprecision(4) <<std::setiosflags(std::ios::showpoint|std::ios::fixed);
						std::cout << std::setw(8) << data[j] <<" ";
//							<< std::setprecision(2) << std::setiosflags(std::ios::showpoint|std::ios::scientific)
					}
					std::cout << std::endl;
				}
				std::cout << "---" << std::endl;
				// In next line 2 -means that function norm (integral, integral of modul is 1,
				// and 2=2*1 where 2 multiplier arises from RelerrEps() function returns value
				// 2 times smaller than that of value added to 1 is really increase 1
//				if( max_re_diff > 2*NumAlg::RelerrEps<Real_t>() || max_im_diff > 2*NumAlg::RelerrEps<Real_t>() )
//					is_ok=false;
//				std::cout << std::setprecision(9) << std::setiosflags(std::ios::showpoint|std::ios::scientific);
//				std::cout << "max_re_diff=" << max_re_diff << "   max_im_diff=" << max_im_diff << std::endl;
//				std::cout << std::setprecision(7) << std::resetiosflags(std::ios::scientific);
				std::cout << "-----------" << std::endl;
		}
		std::cout << (is_ok?"test OK":"test ERROR") << std::endl << "==============" << std::endl;
		return is_ok;
	}



int test_Integral_Transforms(int argc, char* argv[])
{
//----------------
	std::cout << "============== test_Integral_Transforms() ==============" << std::endl;
	bool is_ok=true;
	is_ok=is_ok && test_Utils_smallest_log2_ge_size();
	is_ok=is_ok && test_Utils_rev();
	is_ok=is_ok && test_Utils_bit_reversal();
	is_ok=is_ok && test_Utils_bit_reversal_complex_pair();
	is_ok=is_ok && test_Utils_bit_reversal_complex_separate();
	is_ok=is_ok && test_Utils_SwapArrayHalfs_real();
	is_ok=is_ok && test_Utils_SwapArrayHalfs_complex_pairs();
	is_ok=is_ok && test_Utils_SwapArrayHalfs_complex_separate();
	is_ok=is_ok && test_Four1_forward_reverse_equality<float>();
	is_ok=is_ok && test_Four1_forward_reverse_equality<double>();
	is_ok=is_ok && test_Four1_step<float>();
	is_ok=is_ok && test_Four1_step<double>();
	is_ok=is_ok && test_Four1_mathlab();
	is_ok=is_ok && test_hartley_vs_fft();
	is_ok=is_ok && test_hartley_square();
	is_ok=is_ok && test_speed();


/*
	{
		double array_1[8]={0.1,1.1,2.1,3.1,4.1,5.1,6.1,7.1};
		size_t size= sizeof(array_1)/sizeof(array_1[0]);
		std::cout << "Original array_1[]" << std::endl;
		for(size_t i=0; i<size ; ++i)
			std::cout << "array_1[" << i <<"]="<< array_1[i] << std::endl;
		std::cout << "Vtl::NumAlg::IntTransf::Utils::bit_reversal(T data,I size)" << std::endl;
		Vtl::NumAlg::IntTransf::Utils::bit_reversal(array_1, size);
		for(size_t i=0; i<size ; ++i)
			std::cout << "array_1[" << i <<"]="<< array_1[i] << std::endl;
		std::cout << "Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs(T* data,I size)" << std::endl;
		Vtl::NumAlg::IntTransf::Utils::SwapArrayHalfs(array_1,size);
		for(size_t i=0; i<size ; ++i)
			std::cout << "array_1[" << i <<"]="<< array_1[i] << std::endl;
	}
*/


	{
		float data[]={1,2,3,4, 5,6,7,8, 9,10,11,12, 13,14,15,16};
		size_t data_sz=sizeof(data)/sizeof(data[0]);
		size_t n[]={4,4};
		const size_t ndim=sizeof(n)/sizeof(n[0]);
		Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<ndim,false>(data,n);
		//Vtl::NumAlg::IntTransf::CHartley<float>::TransposeQuarters(data,n[0]);

		std::cout << "SwapQuadrants:" << std::endl;
//		for(size_t i=0; i<data_sz  ; ++i)
//			std::cout << data[i] << std::endl;
		for(size_t y=0; y<n[1]  ; ++y)
		{
			for(size_t x=0; x<n[0]  ; ++x)
				std::cout << data[y*n[0]+x] <<" ";
			std::cout << std::endl;
		}
	}
	{
		float data[256];
		Vtl::NumAlg::IntTransf::Fourier::fft2D<float,size_t>(data,4,4);
		Vtl::NumAlg::IntTransf::Fourier::fft3D<float,size_t>(data,4,4,2);
	}
	{
		std::cout << "Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<1>() :" << std::endl;
		float data[8];
		const size_t data_sz=sizeof(data)/sizeof(data[0]);
		size_t n[]={data_sz};
		for(int i=0; i< data_sz; i++)
			data[i]=i+0.01f;
		Vtl::NumAlg::IntTransf::Utils::F::Test<float> pp(data);
		Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<1>(n,pp);
		for(size_t i=0; i<data_sz  ; ++i)
			std::cout << i << "  " <<  data[i] << " " << std::endl;
	}
	{
		std::cout << "Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<2>() :" << std::endl;
		const size_t w=8;
		const size_t h=4;
		size_t n[]={w,h};
		const size_t ndim=sizeof(n)/sizeof(n[0]);
		float data[w*h];
		const size_t data_sz=sizeof(data)/sizeof(data[0]);
		for(int i=0; i< data_sz; i++)
			data[i]=i+0.01f;
		Vtl::NumAlg::IntTransf::Utils::F::Test<float> pp(data);
		Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<2>(n,pp);
		int i=0;
		std::cout << std::setprecision(2);
		for(size_t i1=0; i1<n[1]  ; ++i1)
		{
			for(size_t i0=0; i0<n[0]  ; ++i0)
				std::cout << std::setw(5) << data[i++] << " ";
			std::cout << std::endl;
		}
	}
	{
		std::cout << "===" << std::endl <<"Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<3>() :" << std::endl;
		const size_t x=8;
		const size_t y=4;
		const size_t z=4;
		size_t n[]={x,y,z};
		const size_t ndim=sizeof(n)/sizeof(n[0]);
		float data[x*y*z];
		const size_t data_sz=sizeof(data)/sizeof(data[0]);
		for(int i=0; i< data_sz; i++)
			data[i]=i+0.01f;
		Vtl::NumAlg::IntTransf::Utils::F::Test<float> pp(data);
		Vtl::NumAlg::IntTransf::Utils::proc_pos_neg_freq<3>(n,pp);
		int i=0;
		std::cout << std::setprecision(2);
		for(size_t i2=0; i2<n[2]  ; ++i2)
		{
			for(size_t i1=0; i1<n[1]  ; ++i1)
			{
				for(size_t i0=0; i0<n[0]  ; ++i0)
					std::cout << std::setw(6) << data[i++] << " ";
				std::cout << std::endl;
			}
			std::cout << std::endl;
		}
	}

	//Vtl::NumAlg::IntTransf::Fourier<float,size_t>::four1(data,n,false);

	std::cout << "##########---" << (is_ok?"All tests OK":"Some tests ERROR") << std::endl << "==============" << std::endl;

	std::cout << std::setprecision(2) << std::setiosflags(std::ios::scientific);
	std::cout << "RelErr of float " << Vtl::NumAlg::RelerrEps<float>() << std::endl;
	std::cout << "RelErr of double " << Vtl::NumAlg::RelerrEps<double>() << std::endl;
	std::cout << "RelErr of long double " << Vtl::NumAlg::RelerrEps<long double>() << std::endl;

	return 0;
}

}//namespace Tests
