
#include <stdio.h>
#include <string>
#include <sstream>
#include <memory> //auto_ptr()
#include <iostream>
#include <iomanip>
#include <assert.h>

#include "aimage.h"
#include "consolap.h"
#include "img_registration.h"
#include "progrcnt.h" 

#include "tests.h"

namespace Tests
{
	typedef float Real_t;
	///For grayscale AImageMem...
	template<class Img>
	void rotate(Img* img,Img* out, float* mat,float dx,float dy)
	{
		int w=img->Width();
		int h=img->Height();
		int nls=img->NextLineShift();
		Vtl::ImAlg::Resamp::InterpBilinear<Img::Pixel> interp((*img)[0],w,h,nls,Img::Pixel(0));
		float x0=float(w/2);
		float y0=float(h/2);
		Vtl::ImAlg::Resamp::Affine(interp,x0,y0,x0-dx,y0-dy,mat,(*out)[0],w,h,out->NextLineShift());
		
	}
	///For color AImageMem...
	void rotate(AImageMemRgb8* img,AImageMemRgb8* out, float* mat,float dx,float dy)
	{
		for(int i=0;i<3;i++)
		rotate(img->GetPlane8(i),out->GetPlane8(i),mat,dx,dy);
	}

	///For grayscale AImageMem...
	template<class Img>
	int gray_registration(size_t sz,Img* img1,Img* img2,bool edge,const char* fname_out)
	{
		const int w=img1->Width();
		const int h=img1->Height();
		std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,w,h,img1->Bits()));
		Img* out=dynamic_cast<Img*>(_out.get());
		if (!out)
			return(1);

		Vtl::Plf::SrcAImage<Real_t,Img> src_1("img1",img1);

		Vtl::Plf::SrcAImage<Real_t,Img> src_2("img2",img2);


		std::list<Vtl::Plf::FilterElementary*> lst1;
		lst1.push_back(&src_1);
		std::list<Vtl::Plf::FilterElementary*> lst2;
		lst2.push_back(&src_2);


		//         
		const int bits=img1->Bits();
		const float maxPixelVal=float((1<<bits)-1);
	// Classic Canny
		float sigma=1.0f;
		float env_size_in_sigmas=2.5f;
		float gamma=2.2f;
		float threshold=20;
		//Vtl::Plf::F11::DoNothing<float> gam_cor;
		Vtl::Plf::Canny<float> canny_1("Canny",threshold,maxPixelVal,sigma,env_size_in_sigmas,gamma,Vtl::Plf::F11::DoNothing<float>());
		Vtl::Plf::Canny<float> canny_2("Canny",threshold,maxPixelVal,sigma,env_size_in_sigmas,gamma,Vtl::Plf::F11::DoNothing<float>());


		Vtl::Plf::OutPortNoPostProc<Real_t>* dst_1=0;
		Vtl::Plf::OutPortNoPostProc<Real_t>* dst_2=0;

		typedef Vtl::ImAlg::Registration::Hartley<Real_t,Vtl::ImAlg::Model::Similarity<Real_t> > ReddyChatterji;
		ReddyChatterji reddy;

		if (edge)
		{
			src_1.out.ConnectTo(&canny_1.in());
			dst_1=&canny_1.out();

			src_2.out.ConnectTo(&canny_2.in());
			dst_2=&canny_2.out();

			reddy.Init(sz, true);
		}
		else
		{
			dst_1=&src_1.out;
			dst_2=&src_2.out;

			reddy.Init(sz, false);
		}
		reddy.Insert<1>(dst_1,lst1.begin(),lst1.end());
		reddy.Insert<2>(dst_2,lst2.begin(),lst2.end());
		//float matrix_2x2[4];
		//float dx;
		//float dy;
		typename ReddyChatterji::Model model;
		reddy.Run(model,w,h);
		std::cout << model << std::endl;
		std::cout << model.getTransf(w,h) << std::endl;
		{
			const int nls=img2->NextLineShift();
			model.Interp((*img2)[0],w, h,nls,(*out)[0],out->NextLineShift(),Img::Pixel(0));
			::AImageSave(out,fname_out,indicator);
		}
		return(0);
	}

	int color_registration(size_t sz,AImageMemRgb8* img1,AImageMemRgb8* img2,bool edge,const char* fname_out)
	{
		AImageMem8 * r1=img1->GetPlane8(0);
		AImageMem8 * g1=img1->GetPlane8(1);
		AImageMem8 * b1=img1->GetPlane8(2);
		AImageMem8 * r2=img2->GetPlane8(0);
		AImageMem8 * g2=img2->GetPlane8(1);
		AImageMem8 * b2=img2->GetPlane8(2);
		const int w=img1->Width();
		const int h=img1->Height();

		std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_RGB,img1->Width(),img1->Height(),8));
		AImageMemRgb8* out=dynamic_cast<AImageMemRgb8*>(_out.get());
		if (!out)
			return(1);
		AImageMem8 * ro=out->GetPlane8(0);
		AImageMem8 * go=out->GetPlane8(1);
		AImageMem8 * bo=out->GetPlane8(2);


		Vtl::Plf::SrcAImage<Real_t,AImageMem8> src_r1("img1_r",r1);
		Vtl::Plf::SrcAImage<Real_t,AImageMem8> src_g1("img1_g",g1);
		Vtl::Plf::SrcAImage<Real_t,AImageMem8> src_b1("img1_b",b1);

		Vtl::Plf::SrcAImage<Real_t,AImageMem8> src_r2("img2_r",r2);
		Vtl::Plf::SrcAImage<Real_t,AImageMem8> src_g2("img2_g",g2);
		Vtl::Plf::SrcAImage<Real_t,AImageMem8> src_b2("img2_b",b2);





		std::list<Vtl::Plf::FilterElementary*> lst1;
		lst1.push_back(&src_r1);
		lst1.push_back(&src_g1);
		lst1.push_back(&src_b1);
		std::list<Vtl::Plf::FilterElementary*> lst2;
		lst2.push_back(&src_r2);
		lst2.push_back(&src_g2);
		lst2.push_back(&src_b2);


		//          
		Vtl::ImAlg::C::YCbCr::Y <float> Y1;
		Vtl::ImAlg::C::YCbCr::Y <float> Y2;
		Vtl::Plf::PixByPix3to1<float,Vtl::ImAlg::C::YCbCr::Y <float> >	rgb2gray_1("rgb2gray_1",Y1);
		Vtl::Plf::PixByPix3to1<float,Vtl::ImAlg::C::YCbCr::Y <float> >	rgb2gray_2("rgb2gray_2",Y2);

		//          
		const int bits=r1->Bits();	
		const int max_val=(1<<bits)-1;
		float threshold=0;
		Vtl::Plf::DiZenzo<float> dizenzo_1("DiZenzo",float(max_val),threshold,2.0,2.5f,4.2f);
		Vtl::Plf::DiZenzo<float> dizenzo_2("DiZenzo",float(max_val),threshold,2.0,2.5f,4.2f);



		Vtl::Plf::OutPortNoPostProc<Real_t>* dst_1=0;
		Vtl::Plf::OutPortNoPostProc<Real_t>* dst_2=0;

		typedef Vtl::ImAlg::Registration::Hartley<Real_t,Vtl::ImAlg::Model::Similarity<Real_t> > ReddyChatterji;
		ReddyChatterji reddy;
		if (edge)
		{
			src_r1.out.ConnectTo(&dizenzo_1.in_1);
			src_g1.out.ConnectTo(&dizenzo_1.in_2);
			src_b1.out.ConnectTo(&dizenzo_1.in_3);
			//dizenzo.out().ConnectTo(&dst.in);
			dst_1=&dizenzo_1.out();

			src_r2.out.ConnectTo(&dizenzo_2.in_1);
			src_g2.out.ConnectTo(&dizenzo_2.in_2);
			src_b2.out.ConnectTo(&dizenzo_2.in_3);
			//dizenzo.out().ConnectTo(&dst.in);
			dst_2=&dizenzo_2.out();


			reddy.Init(sz, true);
		}
		else
		{
			src_r1.out.ConnectTo(&rgb2gray_1.in_xx);
			src_g1.out.ConnectTo(&rgb2gray_1.in_xy);
			src_b1.out.ConnectTo(&rgb2gray_1.in_yy);
			dst_1=&rgb2gray_1.out;

			src_r2.out.ConnectTo(&rgb2gray_2.in_xx);
			src_g2.out.ConnectTo(&rgb2gray_2.in_xy);
			src_b2.out.ConnectTo(&rgb2gray_2.in_yy);
			dst_2=&rgb2gray_2.out;

			reddy.Init(sz, false);
		}
		reddy.Insert<1>(dst_1,lst1.begin(),lst1.end());
		reddy.Insert<2>(dst_2,lst2.begin(),lst2.end());
		//float matrix_2x2[4];
		//float dx;
		//float dy;
		ReddyChatterji::Model model;
		reddy.Run(model,w,h);


		std::cout << model << std::endl;
		std::cout << model.getTransf(w,h) << std::endl;
		{
			int w=img2->Width();
			int h=img2->Height();
			int nls=r2->NextLineShift();
			model.Interp((*r2)[0],w, h,nls,(*ro)[0],nls,AImageMem8::Pixel(0));
			    nls=g2->NextLineShift();
			model.Interp((*g2)[0],w, h,nls,(*go)[0],nls,AImageMem8::Pixel(0));
			    nls=b2->NextLineShift();
			model.Interp((*b2)[0],w, h,nls,(*bo)[0],nls,AImageMem8::Pixel(0));
			::AImageSave(out,fname_out,indicator);
		}
		return(0);
	}




	int test_color_registration(int argc,char *argv[])
	{
		std::cout << "==== color_registration() started ====" <<std::endl;
	
		//const char *img_fname_1="pic1.tif";
		//const char *img_fname_2="pic2.tif";

		const char *img_fname_1="budka1.bmp";
		const char *img_fname_2="budka2.bmp";

		//const char *img_fname_1=argv[1];
		//const char *img_fname_2=argv[2];

		std::cout << img_fname_1 <<std::endl;
		std::cout << img_fname_2 <<std::endl;
	
	//---------------------------------
		try
		{
			std::auto_ptr<AImage> _im1(::AImageLoad(0,true,img_fname_1,indicator));
			std::auto_ptr<AImage> _im2(::AImageLoad(0,true,img_fname_2,indicator));
			AImage* im1=_im1.get();
			AImage* im2=_im2.get();

			if(im1==0 || im2==0 )
				throw "Imaged not loaded";

			std::cout << typeid(im1).name() <<std::endl;
			std::cout << typeid(im2).name() <<std::endl;

			size_t sz=0;
			{
				size_t sz1=std::max(im1->Width(),im1->Height());
				size_t d1 =std::min(im1->Width(),im1->Height());
				std::cout << "sz2=" << sz1 << " d1=" <<d1 <<std::endl;
				if(sz1>2*d1)
				{
					throw "image 1 is too narraw!";
				}
				size_t sz2=std::max(im2->Width(),im2->Height());
				size_t d2 =std::min(im2->Width(),im2->Height());
				std::cout << "sz2=" << sz2 << " d2=" <<d2 <<std::endl;
				if(sz2>2*d2)
				{
					throw "image 2 is too narraw!";
				}
				if(2*std::min(sz1,sz2) < std::max(sz1,sz2))
				{
					throw "images more the 2 times different size!";
				}

				sz=Vtl::NumAlg::IntTransf::Utils::CalcSmalestPow2(std::max(sz1,sz2));
				std::cout << "New image size =" << sz <<std::endl;
			}


			if(im1->GetImageStore()!=im2->GetImageStore())
				throw "Imaged are of different storage type";
			if(im1->ImageClass()!=im2->ImageClass())
				throw "Imaged are of different photometric type";
			switch(im1->GetImageStore())
			{
			case AImage::IMGS_AImageMem8:
				{
					AImageMem8* img=dynamic_cast<AImageMem8*>(im1);
					if(img!=0)
					{ 
						switch(img->ImageClass())
						{
						case AImage::IMGC_GRAYSCALE:
							{ //  
								gray_registration(sz,img,dynamic_cast<AImageMem8*>(im2),false,"result_gray.tif");
								gray_registration(sz,img,dynamic_cast<AImageMem8*>(im2),true,"result_gray_via_edge.tif");
							}
							break;
						default:
							throw "Unsupported";
						};
					}
				}
				break;
			case AImage::IMGS_AImageMem16:
				{
					AImageMem16* img=dynamic_cast<AImageMem16*>(im1);
					if(img!=0)
					{ 
						if(AImage::IMGC_GRAYSCALE==img->ImageClass())
						{ //  
							gray_registration(sz,img,dynamic_cast<AImageMem16*>(im2),false,"result_gray.tif");
							gray_registration(sz,img,dynamic_cast<AImageMem16*>(im2),true,"result_gray_via_edge.tif");
						}
					}
				}
				break;

			case AImage::IMGS_AImageMemFloat:
				gray_registration(sz,dynamic_cast<AImageMemFloat*>(im1),dynamic_cast<AImageMemFloat*>(im2),false,"result_gray.tif");
				gray_registration(sz,dynamic_cast<AImageMemFloat*>(im1),dynamic_cast<AImageMemFloat*>(im2),true,"result_gray_via_edge.tif");
				break;

			case AImage::IMGS_AImageMemRgb8:
				{
					AImageMemRgb8* img=dynamic_cast<AImageMemRgb8*>(im1);
					if(img!=0)
					{
						color_registration(sz,dynamic_cast<AImageMemRgb8*>(im1),dynamic_cast<AImageMemRgb8*>(im2),true,"result_rgb_via_edge.bmp");
						color_registration(sz,dynamic_cast<AImageMemRgb8*>(im1),dynamic_cast<AImageMemRgb8*>(im2),false,"result_rgb_via_gray.bmp");
					}
				}
				break;
			default:
				throw "Unsupported";
			};

			std::cout << "==== color_registration() finished ====" <<std::endl;
			return(0);
		}
		catch(::AErrorStatusBase& msg)
		{
			printf("Exception: %s\n",msg.Msg());
		}
		catch(std::string msg)
		{
			printf("Exception: %s\n",msg.c_str());
		}
		catch(const char * msg)
		{
			printf("Exception: %s\n",msg);
		};
		return 1;
	}

}//namespace Tests

