#include <stdio.h>
#include <stdlib.h> //rand(),srand()
#include <time.h> //time()

#include <string>
#include <sstream>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <memory> //auto_ptr()

#include <list>

#include "plf_kernels_1d.h"
#include "plf_functors.h"
#include "imalg_tools.h"
#include "aimage.h"
#include "consolap.h"
#include "tests.h"



namespace Tests
{


	int test_fast_gauss(int argc, char* argv[])
	{
		//ConvKernel<float> ker; //    -       !
		{//Test Cyclic convolve
			std::cout << "========= Test Cyclic convolve ==========" <<std::endl;
			//-----Create Kernel
			//typedef Vtl::ImAlg::WndFun::Nuttall<float,0> FF;
			typedef Vtl::ImAlg::WndFun::Hann<float,0> FF;
			typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Corr,true>  K;
			//FF func(10);
			FF func(7);
			K ker(func);
			//-----Print Kernel
			std::cout << "Kernel is: " << func.Name()<<std::endl;
			for( Vtl::Plf::Kernel::FKern<float>::const_iterator it=ker.begin(); it!=ker.end(); ++it)
				std::cout << *it << "  ";
			std::cout << std::endl;

			//-----Create data
			const float data[]={1,2,3,4,5,6,7,8,9,10};
			const int sz=sizeof(data)/sizeof(data[0]);
			std::vector<float> out_c(sz);
			std::vector<float> out_ci(sz);
			std::vector<float> out_r(sz);
			std::vector<float> out_ri(sz);
			std::copy(data,data+sz,out_ci.begin());
			std::copy(data,data+sz,out_ri.begin());
			//-----Convolve each kind
			Vtl::ImAlg::Conv::CyclicConv<float,K> cc(ker);
			Vtl::ImAlg::Conv::CyclicConvInplace<float,K> cci(ker);
			Vtl::ImAlg::Conv::Conv<float,K> rc(ker);
			Vtl::ImAlg::Conv::ConvInplace<float,K> rci(ker);
			cc(data,data+sz,out_c.begin());
			rc(data,data+sz,out_r.begin());
			cci(out_ci.begin(),out_ci.end());
			rci(out_ri.begin(),out_ri.end());
			//-----Print convolution results
			bool ok=true;
			for(int i=0; i<sz; ++i)
			{
				if(out_ci[i]!=out_c[i])
				{
					ok=false;
					std::cout << "Cyclic conv inplace != CyclicConv at i="<< i << " out_ci[i]="<< out_ci[i]<< "  out_c="<< out_c[i] <<std::endl;
				}
				if(out_ri[i]!=out_r[i])
				{
					ok=false;
					std::cout << "Reflection conv inplace != ReflectConv at i="<< i << " out_ri[i]="<< out_ri[i]<< "  out_r="<< out_r[i] <<std::endl;
				}
			}
			std::cout << "Convolve comparisons is ===== "<< (ok?"OK!":"ERROR!") << " =====" <<std::endl;

			std::cout << "Results:"<<std::endl;
			int i=0;
			for( int i=0; i<sz ; ++i)
				std::cout << "i=" << i <<"\t in=" << data[i] <<"\t out_c[i]=" << out_c[i] << "\t out_ci[i]="<< out_ci[i] <<"\t out_r[i]=" << out_r[i] <<"\t out_ri[i]=" << out_r[i] << std::endl;
			std::cout << std::endl;
			
			std::list<float> lst;
			std::list<float> out_lst;
			for(int i=0; i<sz; ++i)
				lst.push_back(data[i]);
			out_lst.resize(lst.size());
			cc(lst.begin(),lst.end(),out_lst.begin());
			std::cout << "Vtl::ImAlg::Conv::CyclicConv<> work  ==== " << (std::equal(out_c.begin(),out_c.end( ),out_lst.begin()) ? "OK":"BAD")    << " ===== on data placed in list container" << std::endl;

			//rc.Run(lst.begin(),lst.end(),out1.begin());
			cci(lst.begin(),lst.end());
			std::cout << "Vtl::ImAlg::Conv::CyclicConvInplace<> work ==== " << (std::equal(out_c.begin(),out_c.end( ),lst.begin()) ? "OK":"BAD") << " ===== on data placed in list container" << std::endl;
			//rci.Run(lst.begin(),lst.end());
			
		}
		{//test fast Gauss -      -  .
			std::vector<float> data0;
			std::vector<float> data1;
			std::vector<float> data2;
			int len=200;
			data0.resize(len);
			data1.resize(len);
			data2.resize(len);
			
//			::srand( (unsigned)time( NULL ) );
//			for(int i=0; i<len;++i)
//			{
//				data0[i]=float( (rand()/double(RAND_MAX))* 200. - 100. ); 
//			}
			for(int i=0; i<len;++i)
			{
				data0[i]=float(((i/50)+1)*100); 
			}
//			for(int i=0; i<len;++i)
//			{
//				data0[i]=float(((i+20)%50==0)?17.54:0); 
//			}
			std::copy(data0.begin(),data0.end(),data1.begin());
			std::copy(data0.begin(),data0.end(),data2.begin());
			const float * d0=&(data0[0]);
			float * d1=&(data1[0]);
			float * d2=&(data2[0]);
			//double sigma=6.04f;
			double sigma=7.0f;
#define DERIV 0
			Vtl::ImAlg::Conv::FastGaussConvolve<float,DERIV> g0(sigma);
			g0(d1,d1+len);


			typedef Vtl::ImAlg::WndFun::Gauss<float,DERIV> FF;
#undef DERIV
			FF func(sigma,10);
			typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K;
			K ker(func);
			//Vtl::Plf::Kernel::ConvolveNoExt(d0,d0+len,d2,ker);
			Vtl::ImAlg::Conv::ConvInplace<float,K> conv(ker);
			//conv.Run(d0,d0+len,d2);
			conv(d2,d2+len);


			float max_err=0;
			float max_relerr=0;
			float max_val=0;
			int i;
			for(i=0; i<len-50; ++i)
			{
				//float dif=(d2[i]-d1[i])/std::max(std::abs(d2[i]),std::abs(d1[i]));
				float dif=(d2[i]-d1[i]);
				float max=std::max(std::abs(d2[i]),std::abs(d2[i]))+0.0000001f;
				float adif=std::abs(dif);
				max_err=std::max(max_err,adif);
				max_val=std::max(max_val,std::abs(d2[i]));
				max_relerr=std::max(max_relerr,adif/max);
				//if(std::abs(dif) > std::max(0.1f*std::abs(d2[i]),0.001f) )
				//	std::cout << "dif["<<i<<"]= " <<dif <<std::endl;
					std::cout << "i="<<i <<"   d0[i]= " <<d0[i]   <<"   d2[i]= " <<d2[i]   <<"   d1[i]= " <<d1[i]   <<"   dif[i]= " <<dif   <<std::endl;
			}
			float max_err2=max_err;
			float max_relerr2=max_relerr;
			float max_val2=max_val;
			for(; i<len; ++i)
			{
				//float dif=(d2[i]-d1[i])/std::max(std::abs(d2[i]),std::abs(d1[i]));
				float dif=(d2[i]-d1[i]);
				float max=std::max(std::abs(d2[i]),std::abs(d2[i]))+0.0000001f;
				float adif=std::abs(dif);
				max_err=std::max(max_err,adif);
				max_val=std::max(max_val,std::abs(d2[i]));
				max_relerr=std::max(max_relerr,adif/max);
				//if(std::abs(dif) > std::max(0.1f*std::abs(d2[i]),0.001f) )
				//	std::cout << "dif["<<i<<"]= " <<dif <<std::endl;
					std::cout << "i="<<i <<"   d0[i]= " <<d0[i]   <<"   d2[i]= " <<d2[i]   <<"   d1[i]= " <<d1[i]   <<"   dif[i]= " <<dif   <<std::endl;
			}
			std::cout << "max_err ="<<max_err <<"   max_relerr = " <<max_relerr  <<"   max_err/max_val  = " <<max_err/max_val    <<std::endl;
			std::cout << "max_err2="<<max_err2 <<"   max_relerr2= " <<max_relerr2  <<"   max_err2/max_val2= " <<max_err2/max_val2    <<std::endl;
			

		}
		{ //   -     
			std::auto_ptr<AImage> _im(::AImageLoad(0,true,argv[1],indicator));
			AImageMem8* img=dynamic_cast<AImageMem8*>(_im.get());
			if(img)
			{
				//    (   
				//,      AImage.
				int w=img->Width();
				int h=img->Height();
				int sz=w*h;
				std::vector<float> _data0(sz);
				std::vector<float> _data1(sz);
				std::vector<float> _data2(sz);
				std::vector<float> _data3(sz);
				float *data0=&(_data0[0]);
				float *data1=&(_data1[0]);
				float *data2=&(_data2[0]);
				float *data3=&(_data3[0]);

				Vtl::Plf::F11::Convert<float> c_to_float;
				Vtl::ImAlg::Tools::copy_1_1((*img)[0],w,h,img->NextLineShift(),data0,w,c_to_float);

				memcpy(data2,data0,sz*sizeof(float));
				memcpy(data1,data0,sz*sizeof(float));


				// 
				float sigma=3.5f;
				//      -  3  
				Vtl::ImAlg::Conv::FastGaussConvolve<float,0> g0(sigma);
				Vtl::ImAlg::Conv::FastGaussConvolve<float,1> g1(sigma);
				Vtl::ImAlg::Conv::FastGaussConvolve<float,2> g2(sigma);
				Vtl::ImAlg::Conv::FastGaussConvolve<float,3> g3(sigma);

				//-----------    X   Y   ---
				//  X :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(data1,w,h,w,g1,g0);
				//  Y :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(data2,w,h,w,g0,g1);
				//      8- .
				//   -  Mult()     ,   
				//    0-M,  M -    .
				//    -     128 (    ).
				//   :
				Vtl::Plf::F11::Lin<float> lin_1(g1.Deriv==0?g1.Mult(sigma):g1.Mult(sigma)/2.0f,g1.Deriv==0?0.0f:128.0f);
				//  :
				Vtl::Plf::F11::Convert<uint8_t> c_to_Byte;
				//   :
				//Vtl::Plf::F::Comb<float> lin(mult,bias);
				Vtl::Plf::F11::Comb<uint8_t,Vtl::Plf::F11::Lin<float>,Vtl::Plf::F11::Convert<uint8_t> > adj_1(lin_1,c_to_Byte);
				//    ,      :
				Vtl::ImAlg::Tools::copy_1_1(data1,w,h,w,(*img)[0],img->NextLineShift(),adj_1);
				::AImageSave(img,"fast_d1.tif");
				//    ,      :
				Vtl::ImAlg::Tools::copy_1_1(data2,w,h,w,(*img)[0],img->NextLineShift(),adj_1);
				::AImageSave(img,"fast_d2.tif");

				//-----------     X   Y      ---
				//  :
				Vtl::Plf::F21::SqrtOfSumSqr<float> sumsqr;
				//    :
				Vtl::ImAlg::Tools::copy_2_1(data2,w,h,w,data1,w,data2,w,sumsqr);

				//  ,       ,   :
				Vtl::Plf::F11::Lin<float> lin_1mod(g1.Mult(sigma),0);
				Vtl::Plf::F11::Comb<uint8_t,Vtl::Plf::F11::Lin<float>,Vtl::Plf::F11::Convert<uint8_t> > adj_1mod(lin_1mod,c_to_Byte);
				Vtl::ImAlg::Tools::copy_1_1(data2,w,h,w,(*img)[0],img->NextLineShift(),adj_1mod);
				::AImageSave(img,"fast_gmod.tif");

				//---------------------------------------------------
				//     1  2 
				//(        ):
				memcpy(data1,data0,sz*sizeof(float));
				memcpy(data2,data0,sz*sizeof(float));
				//-----------     X   Y   ---
				//  XX :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(data1,w,h,w,g2,g0);
				//  YY :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(data2,w,h,w,g0,g2);
				// , ..        Y				
				Vtl::Plf::F21::Sum<float> sum;
				Vtl::ImAlg::Tools::copy_2_1(data2,w,h,w,data1,w,data2,w,sum);

				//  ,      
				Vtl::Plf::F11::Lin<float> lin_2(g2.Deriv==0?g2.Mult(sigma):g2.Mult(sigma)/2.0f,g2.Deriv==0?0.0f:128.0f);
				Vtl::Plf::F11::Comb<uint8_t,Vtl::Plf::F11::Lin<float>,Vtl::Plf::F11::Convert<uint8_t> > adj_2(lin_2,c_to_Byte);
				Vtl::ImAlg::Tools::copy_1_1(data2,w,h,w,(*img)[0],img->NextLineShift(),adj_2);
				::AImageSave(img,"fast_lapl.tif");

				//     -   
				//       -  .
				//       
				//  data2   AImage,  data2  .
				Vtl::Plf::F11::Lin<float> lin_2mod(g2.Mult(sigma),0.0f);
				Vtl::Plf::F11::Abs<float> abs;
				Vtl::Plf::F11::Comb< float , Vtl::Plf::F11::Abs<float> , Vtl::Plf::F11::Lin<float> > lin_2mod_(abs,lin_2mod);
				Vtl::Plf::F11::Comb<uint8_t, Vtl::Plf::F11::Comb< float ,Vtl::Plf::F11::Abs<float> , Vtl::Plf::F11::Lin<float> > ,Vtl::Plf::F11::Convert<uint8_t> > adj_2mod(lin_2mod_,c_to_Byte);
				Vtl::ImAlg::Tools::copy_1_1(data2,w,h,w,(*img)[0],img->NextLineShift(),adj_2mod);
				::AImageSave(img,"fast_lapl_modul.tif");

				// ,     .
				::memset(data1,0,sz*sizeof(float));
				Vtl::ImAlg::Tools::findMax::SetPixel<float,1> functor(data1,w);
				Vtl::ImAlg::Tools::findMax::is_center_max_in3x3<1> icm;
				//Vtl::ImAlg::Tools::findMax::BoundsNo<1>(data2,w,h,w,icm,functor );
				Vtl::ImAlg::Tools::findMax::BoundsCyclic<1>(data2,w,h,w,icm,functor );
				Vtl::ImAlg::Tools::copy_1_1(data1,w,h,w,(*img)[0],img->NextLineShift(),adj_2mod);

				::AImageSave(img,"fast_lapl_modul_maxs.tif");

				//     1  2 
				//(     ):
				memcpy(data1,data0,sz*sizeof(float));
				memcpy(data2,data0,sz*sizeof(float));
				memcpy(data3,data0,sz*sizeof(float));
				//   :
				//  XX :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(data1,w,h,w,g2,g0);
				//  YY :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(data2,w,h,w,g0,g2);
				//  XY :
				Vtl::ImAlg::Conv::Conv2D::conv2DInplace(data3,w,h,w,g1,g1);
				//  
				Vtl::Plf::F31::HessianDetSqrt<float> detH; 
				//  data1,data3,data2     
				Vtl::ImAlg::Tools::copy_3_1(data1,w,h,w,data3,w,data2,w,data1,w,detH);

				//  ,       ,   :
				//Vtl::Plf::F::Lin<float> lin_3mod(2*g2.Mult(sigma),0);
				Vtl::Plf::F11::GammaCorMult<float> lin_3mod(255,2.2,1.0*g2.Mult(sigma));
				Vtl::Plf::F11::Comb<uint8_t,Vtl::Plf::F11::GammaCorMult<float>,Vtl::Plf::F11::Convert<uint8_t> > adj_3mod(lin_3mod,c_to_Byte);
				Vtl::ImAlg::Tools::copy_1_1(data1,w,h,w,(*img)[0],img->NextLineShift(),adj_3mod);
				::AImageSave(img,"fast_hess_det.tif");
				{
					std::cout << "Test that convolution with gaussian derivative via FastGaussConvolve results in true image derivatives estimation" << std::endl;
					Vtl::ImAlg::Conv::FastGaussConvolve<float,0> g0(sigma);
					Vtl::ImAlg::Conv::FastGaussConvolve<float,1> g1(sigma);
					static float d[100*100];
					for(int a=0; a<5; ++a)
					{
						for(int y=0; y<100; ++y)
						for(int x=0; x<100; ++x)
							d[100*y+x]=float(a*x);
						Vtl::ImAlg::Conv::Conv2D::conv2DInplace(d,100,100,100,g1,g0);
						std::cout << "sigma= " << sigma << "  a=" << a <<std::endl;
						std::cout << "d(10,10),(30,30),(70,50)=" << d[100*10+10] << " " << d[100*30+30] << " " << d[100*50+70] << std::endl;
					}
				}
			}
		}

		return 0;
	}
}//namespace Tests

