#include <iostream>
#include <iomanip>
#include <sstream>

#include <string>
#include <memory>
#include "aimage.h"
#include "consolap.h"

#include "img_registration.h"

#include "tests.h"


namespace Tests
{

	typedef float Real_t;

	template<typename T>
	std::pair<T,T> MinMaxImage(const T* p,int w, int h,int nls)
	{
		T min_v=p[0];
		T max_v=p[0];
		for(int y=0; y<h ; ++y,p+=nls)
		{
			for(int x=0; x<w; ++x)
				if(p[x]<min_v)
					min_v=p[x];
				else if(p[x]>max_v)
					max_v=p[x];
		}
		return std::pair<T,T>(min_v,max_v);
	}

	template<typename T>
	void SaveImage(std::string fname,const T* p,int w, int h,int nls,T add,T mul,bool truc_neg=false)
	{
		std::auto_ptr<AImage> _im(::AImageCreate(0,AImage::IMGC_GRAYSCALE,w,h,8));
		AImageMem8* im=dynamic_cast<AImageMem8*>(_im.get());
		for(int y=0; y<h; ++y)
		{
			AImageMem8::Pixel *out=(*im)[y];
			for(int x=0; x<w; ++x)
			{
				if(truc_neg)
				{
					T v=(p[x]+add)*mul;
					if(v<0) v=0;
					out[x]=AImageMem8::Pixel( v );
				}
				else
					out[x]=AImageMem8::Pixel( (p[x]+add)*mul );
			}
			p+=nls;
		}
		::AImageSave(im,fname.c_str(),0);
	}
	template<unsigned blk_n>
	void test_save(const Vtl::ImAlg::Registration::Hartley<Real_t,Vtl::ImAlg::Model::Similarity<Real_t> >& reddy,std::string fname_root,size_t w,size_t h,bool gray)
	{
		//test print
		std::pair<Real_t,Real_t> mmax1=MinMaxImage(reddy.storage().block<blk_n>(),w,h,w);
		std::cout << fname_root <<": " << mmax1.first << " to " << mmax1.second <<std::endl;
		Real_t add=0;
		Real_t mul=1;
		if(!gray)
		{
			add-=mmax1.first;
			mul=mmax1.second-mmax1.first;
			mul=Real_t(253)/mul;
			//add/=mul;
		}
		else
		{
			add=128;
			mul=std::max(std::abs(mmax1.first),std::abs(mmax1.second));
			mul=Real_t(127)/mul;
			add/=mul;
		}
		SaveImage(std::string("hreg_")+fname_root+".tif",reddy.storage().block<blk_n>(),w,h,w,add,mul,true);
	}
	void test_save_x(float *block,std::string fname_root,size_t w,size_t h,bool gray)
	{
		//test print
		std::pair<Real_t,Real_t> mmax1=MinMaxImage(block,w,h,w);
		std::cout << fname_root <<": " << mmax1.first << " to " << mmax1.second <<std::endl;
		Real_t add=0;
		Real_t mul=1;
		if(!gray)
		{
			add-=mmax1.first;
			mul=mmax1.second-mmax1.first;
			mul=Real_t(253)/mul;
			//add/=mul;
		}
		else
		{
			add=128;
			mul=std::max(std::abs(mmax1.first),std::abs(mmax1.second));
			mul=Real_t(127)/mul;
			add/=mul;
		}
		SaveImage(std::string("hreg_")+fname_root+".tif",block,w,h,w,add,mul,true);
	}


	template<class T>
	void image_registration_via_fft(T* img1,T* img2)
	{
		//-----------------      -------------------
		std::cout << typeid(img1).name() <<std::endl;
		std::cout << typeid(img2).name() <<std::endl;
		size_t sz1=std::max(img1->Width(),img1->Height());
		size_t d1 =std::min(img1->Width(),img1->Height());
		std::cout << "sz2=" << sz1 << " d1=" <<d1 <<std::endl;
		if(sz1>2*d1)
		{
			throw "image 1 is too narraw!";
		}
		size_t sz2=std::max(img2->Width(),img2->Height());
		size_t d2 =std::min(img2->Width(),img2->Height());
		std::cout << "sz2=" << sz2 << " d2=" <<d2 <<std::endl;
		if(sz2>2*d2)
		{
			throw "image 2 is too narraw!";
		}
		if(2*std::min(sz1,sz2) < std::max(sz1,sz2))
		{
			throw "images more the 2 times different size!";
		}
		//-----------------   -------------------------------
		size_t sz=Vtl::NumAlg::IntTransf::Utils::CalcSmalestPow2(std::max(sz1,sz2));
		//sz<<=1; //Zero padding to next 2
		std::cout << "New image size =" << sz <<std::endl;

		typedef Vtl::ImAlg::Model::SkewYOffset<Real_t> Model;
		
		Vtl::ImAlg::Registration::Hartley<Real_t,Model> reddy;
		
		reddy.Init(sz);
		{		
			Vtl::Plf::SrcAImage<Real_t,T> src("img1",img1);
			std::list<Vtl::Plf::FilterElementary*> lst;
			lst.push_back(&src);
			reddy.Insert<1>(&(src.out),lst.begin(),lst.end());
			//reddy.AfterInsert<1>();
		}
		{		
			Vtl::Plf::SrcAImage<Real_t,T> src("img2",img2);
			std::list<Vtl::Plf::FilterElementary*> lst;
			lst.push_back(&src);
			reddy.Insert<2>(&(src.out),lst.begin(),lst.end());
			//reddy.AfterInsert<2>();
		}

		Model model;
		int w=img2->Width();
		int h=img2->Height();
		reddy.Run(model,w,h);//,matrix_2x2,&dx,&dy);
		//registration finished!

		std::cout << model << std::endl;
		std::cout << model.getTransf(w,h) << std::endl;
		{
			int w=img2->Width();
			int h=img2->Height();
			int nls=img2->NextLineShift();
			std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,w,h,8));
			T *out=dynamic_cast<T*>(_out.get());
			if(!out)
				std::cout << "out=NULL" <<std::endl;
			else
			{
				model.Interp((*img2)[0],w, h,nls,(*out)[0],nls,T::Pixel(0));
				::AImageSave(out,"hreg_result.tif",indicator);
			}
		}
	}

	#define FRG_NUMBER 3  //0-6
	#define CHANS_PAIR 2  // 1& 1,2 
						  // 2& 1,3
						  // 3& 2,3

	int test_image_registration_via_fft(int argc,char *argv[])
	{
		std::cout << "==== test_image_registration_via_fft() started ====" <<std::endl;
	//	const char *img_fname_1=argv[1];
	//	const char *img_fname_2=argv[2];
	//---------------------------------
		std::cout << "Fragment # " <<  FRG_NUMBER <<std::endl;
		std::cout << "Channels pair is  ( " <<  (CHANS_PAIR==3?2:1) <<" , " << (CHANS_PAIR==1?2:3)<< ")" <<std::endl;
	#if FRG_NUMBER==0
			const char *img_fname_1="frg_skew_1.tif"; //"FONAR.TIF";
			const char *img_fname_2="frg_skew_2.tif"; //"FONAR_S.TIF";
	#elif FRG_NUMBER==1
		#if CHANS_PAIR == 1    // 1: 1&2 
			const char *img_fname_1="frg1_c1.tif";
			const char *img_fname_2="frg1_c2.tif";
		#elif CHANS_PAIR == 2  // 2: 1&3
			const char *img_fname_1="frg1_c1.tif";
			const char *img_fname_2="frg1_c3.tif";
		#elif CHANS_PAIR == 3  // 3: 2&3
			const char *img_fname_1="frg1_c2.tif";
			const char *img_fname_2="frg1_c3.tif";
		#else
			std::cout << "macros CHANS_PAIR - inappropriate value" << std::endl;
		#endif
	#elif FRG_NUMBER==2
		#if CHANS_PAIR == 1    // 1: 1&2 
			const char *img_fname_1="frg2_c1.tif";
			const char *img_fname_2="frg2_c2.tif";
		#elif CHANS_PAIR == 2  // 2: 1&3
			const char *img_fname_1="frg2_c1.tif";//pgm";
			const char *img_fname_2="frg2_c3.tif";//pgm";
		#elif CHANS_PAIR == 3  // 3: 2&3
			const char *img_fname_1="frg2_c2.tif";
			const char *img_fname_2="frg2_c3.tif";
		#else
			std::cout << "macros CHANS_PAIR - inappropriate value" << std::endl;
		#endif
	#elif FRG_NUMBER==3
		#if CHANS_PAIR == 1    // 1: 1&2 
			const char *img_fname_1="frg3_c1.tif";
			const char *img_fname_2="frg3_c2.tif";
		#elif CHANS_PAIR == 2  // 2: 1&3
			const char *img_fname_1="frg3_c1.tif";//pgm";
			const char *img_fname_2="frg3_c3.tif";//pgm";
		#elif CHANS_PAIR == 3  // 3: 2&3
			const char *img_fname_1="frg3_c2.tif";
			const char *img_fname_2="frg3_c3.tif";
		#else
			std::cout << "macros CHANS_PAIR - inappropriate value" << std::endl;
		#endif
	#elif FRG_NUMBER==4
		#if CHANS_PAIR == 1    // 1: 1&2 
			const char *img_fname_1="frg4_c1.tif";
			const char *img_fname_2="frg4_c2.tif";
		#elif CHANS_PAIR == 2  // 2: 1&3
			const char *img_fname_1="frg4_c1.tif";//pgm";
			const char *img_fname_2="frg4_c3.tif";//pgm";
		#elif CHANS_PAIR == 3  // 3: 2&3
			const char *img_fname_1="frg4_c2.tif";
			const char *img_fname_2="frg4_c3.tif";
		#else
			std::cout << "macros CHANS_PAIR - inappropriate value" << std::endl;
		#endif
	#elif FRG_NUMBER==5
		#if CHANS_PAIR == 1    // 1: 1&2 
			const char *img_fname_1="frg5_c1.tif";
			const char *img_fname_2="frg5_c2.tif";
		#elif CHANS_PAIR == 2  // 2: 1&3
			const char *img_fname_1="frg5_c1.tif";//pgm";
			const char *img_fname_2="frg5_c3.tif";//pgm";
		#elif CHANS_PAIR == 3  // 3: 2&3
			const char *img_fname_1="frg5_c2.tif";
			const char *img_fname_2="frg5_c3.tif";
		#else
			std::cout << "macros CHANS_PAIR - inappropriate value" << std::endl;
		#endif
	#elif FRG_NUMBER==6
		#if CHANS_PAIR == 1    // 1: 1&2 
			const char *img_fname_1="frg6_c1.tif";
			const char *img_fname_2="frg6_c2.tif";
		#elif CHANS_PAIR == 2  // 2: 1&3
			const char *img_fname_1="frg6_c1.tif";//pgm";
			const char *img_fname_2="frg6_c3.tif";//pgm";
		#elif CHANS_PAIR == 3  // 3: 2&3
			const char *img_fname_1="frg6_c2.tif";
			const char *img_fname_2="frg6_c3.tif";
		#else
			std::cout << "macros CHANS_PAIR - inappropriate value" << std::endl;
		#endif
	#else
		std::cout << "macros FRG_NUMBER - inappropriate value" << std::endl;
	#endif
	//---------------------------------
		std::auto_ptr<AImage> _im1(::AImageLoad(0,true,img_fname_1,indicator));
		std::auto_ptr<AImage> _im2(::AImageLoad(0,true,img_fname_2,indicator));
		AImage* im1=_im1.get();
		AImage* im2=_im2.get();
		if(im1==0 || im2==0 )
			throw "Imaged not loaded";
		if(im1->GetImageStore()!=im2->GetImageStore())
			throw "Imaged are of different storage type";
		if(im1->ImageClass()!=im2->ImageClass())
			throw "Imaged are of different photometric type";
		if(im1->ImageClass()!=AImage::IMGC_GRAYSCALE && im1->ImageClass()!=AImage::IMGC_FLOAT)
			throw "Can process only grayscale images";
		switch(im1->GetImageStore())
		{
		case AImage::IMGS_AImageMem8:
			image_registration_via_fft(dynamic_cast<AImageMem8*>(im1),dynamic_cast<AImageMem8*>(im2));
			break;
		case AImage::IMGS_AImageMem16:
			image_registration_via_fft(dynamic_cast<AImageMem16*>(im1),dynamic_cast<AImageMem16*>(im2));
			break;
		case AImage::IMGS_AImageMemFloat:
			image_registration_via_fft(dynamic_cast<AImageMemFloat*>(im1),dynamic_cast<AImageMemFloat*>(im2));
			break;
		default:
			throw "Can process images stored as ...";
		};
		std::cout << "==== test_image_registration_via_fft() finished ====" <<std::endl;
		return 0;
	}

}//namespace Tests
