#include <iostream>
#include <iomanip>
#include <sstream>

#include <string>
#include <memory>
#include "aimage.h"
#include "consolap.h"

namespace Tests
{
	template<typename T>
	void SaveImage(std::string fname,const T* p,int w, int h,int nls,T add,T mul,bool truc_neg=false);
}
#include "img_registration_oflow.h"

#include "tests.h"

namespace Tests
{

	typedef float Real_t;

	template<class T>
	struct test_arg
	{
		const T& ma_;
		test_arg(const T& ma) : ma_(ma) {}
		template<unsigned n>
		void f()
		{
			std::cout << "n=" << n << "  Dx<n> = " << ma_.Dx<n>(1.01f,2.02f) << "  Dv<n> = " << ma_.Dy<n>(1.01f,2.02f) << std::endl;
		}
	};
	template<unsigned N>
	struct ForEach
	{
		template<class Func>
		static void _Do(Func func,Loki::Int2Type<N> )
		{
		}
		template<unsigned n,class Func>
		static void _Do(Func func,Loki::Int2Type<n> )
		{
			func.f<n>();
			_Do(func,Loki::Int2Type<n+1>() );
		}
		template<class Func>
		static void Do(Func func )
		{
			_Do(func,Loki::Int2Type<0>() );
		}

	};

	template<class Img>
	void image_registration_via_oflow(Img* im1,Img* im2)
	{
		if(im1->Width()!=im2->Width() || im1->Height()!=im2->Height())
			throw "Images are different size";
		if(im1->Bits()!=im2->Bits() )
			throw "Images are different bits";

		//        , 
		//    .
		typedef Vtl::ImAlg::Model::Affine<Real_t> Model;
//		typedef Vtl::ImAlg::Model::Offset<Real_t> Model;
		Model model;
		model.setIdentity();

		//test model
		std::cout << "Model: N=" << model.N << " is " << (model.isLinear?"Linear":"NonLinear") << std::endl;
		test_arg<Model> ta(model);
		ForEach<model.N>::Do(ta);

		const size_t w=im1->Width();
		const size_t h=im1->Height();
		//       (  ),
		// ,     ,  float32,     
		//(  ),      ,
		//       .
		typedef Vtl::ImAlg::Registration::OFlowAllocStrategy<Real_t,1,false> AllocStrategy;
		AllocStrategy alloc;
//		alloc.Init(w,h,32,32);
		alloc.Init(w,h);
		std::cout << "Detality levels: " << alloc.Levels() << std::endl;

		// .     ,   ,
		//  ,     ( 88).
		// ,       ,
		//     ,   ...
		Vtl::ImAlg::Registration::OFlow<Model,AllocStrategy> of(model,alloc);
		of.Insert1((*im1)[0],w,h,im1->NextLineShift());
		of.Insert2((*im2)[0],w,h,im2->NextLineShift());
		of.AfterInsert();

		std::cout << "=============== Results: ==================" << std::endl;
		std::cout << model << std::endl;
		std::cout << "=============== Save transformation results direct and reverse: ==================" << std::endl;
		std::auto_ptr<AImage> _img_out(::AImageCreate(0,im1->ImageClass(),im1->Width(),im1->Height(),im1->Bits()));
		Img* img_out=dynamic_cast<Img*>(_img_out.get());
		model.Interp((*im2)[0],w, h,im2->NextLineShift(),(*img_out)[0],img_out->NextLineShift(),Img::Pixel(0));
		::AImageSave(img_out,"a_result.tif",indicator);
		model.Inverse();
		model.Interp((*im1)[0],w, h,im1->NextLineShift(),(*img_out)[0],img_out->NextLineShift(),Img::Pixel(0));
		::AImageSave(img_out,"a_result_rev.tif",indicator);
		
	}



	int test_image_registration_via_oflow(int argc,char *argv[])
	{
		std::cout << "==== test_image_registration_via_oflow() started ====" <<std::endl;
	//---------------------------------
		const char *img_fname_1="A_orig.tif"; //"FONAR.TIF";
		const char *img_fname_2="A_skew_20-40.tif"; //"FONAR_S.TIF";
//		const char *img_fname_2="a_left3_up5.tif"; //"FONAR_S.TIF";
	//---------------------------------
		std::auto_ptr<AImage> _im1(::AImageLoad(0,true,img_fname_1,indicator));
		std::auto_ptr<AImage> _im2(::AImageLoad(0,true,img_fname_2,indicator));
		AImage* im1=_im1.get();
		AImage* im2=_im2.get();
		if(im1==0 || im2==0 )
			throw "Imaged not loaded";
		if(im1->GetImageStore()!=im2->GetImageStore())
			throw "Imaged are of different storage type";
		if(im1->ImageClass()!=im2->ImageClass())
			throw "Imaged are of different photometric type";
		if(im1->ImageClass()!=AImage::IMGC_GRAYSCALE && im1->ImageClass()!=AImage::IMGC_FLOAT)
			throw "Can process only grayscale images";
		switch(im1->GetImageStore())
		{
		case AImage::IMGS_AImageMem8:
			image_registration_via_oflow(dynamic_cast<AImageMem8*>(im1),dynamic_cast<AImageMem8*>(im2));
			break;
		case AImage::IMGS_AImageMem16:
			image_registration_via_oflow(dynamic_cast<AImageMem16*>(im1),dynamic_cast<AImageMem16*>(im2));
			break;
		case AImage::IMGS_AImageMemFloat:
			image_registration_via_oflow(dynamic_cast<AImageMemFloat*>(im1),dynamic_cast<AImageMemFloat*>(im2));
			break;
		default:
			throw "Can process images stored as ...";
		};
		std::cout << "==== test_image_registration_via_fft() finished ====" <<std::endl;
		return 0;
	}

}//namespace Tests
