//#include <string>
//#include <sstream>
//#include <memory> 
//#include <list>
//#include <iostream>
//#include <fstream>
//#include <iomanip>

#include <time.h>
#include <assert.h>
#include <math.h>

#include "yfilesys.h"
#include "aimage.h"
#include "color_spaces.h"
#include "convolve.h"
#include "imalg_tools.h"

#include "plf_functors.h"
#include "plf_kernels_1d.h"

#include "tests.h"
#include "consolap.h"

// own includings
#include "curvature/CurvatureFunc.h"
#include "CurveDetection/MarkCenters.h"
#include "CurveDetection/ModelTypes.h"
#include "CurveDetection/RandomizedLineDetection.h"
#include "CurveDetection/HoughTransform.h"
#include "CurveDetection/PointsContainer.h"
#include "CurveDetection/Point.h"
#include "CurveDetection/CenterClusterization.h"
#include "CurveDetection/Sinogram.h"
#include "CurveDetection/ClustViaAccum.h"
#include "KMeans.h"
#include "curvature/BuildPalletes.h"
#include "EdgeDetection/GradientOperations.h"

//--!!!----------------------------------------------------------------------------
//#define CDI_OPERATE_AS_LINE_DETECTION
//#define CDI_OPERATE_AS_CIRCLE_DETECTION
//--!!!----------------------------------------------------------------------------
namespace Tests
{
	int test_line_curvature(int argc, char *argv[])
	{
		if(argc < 2) 
    {
			std::cout << "Command string:\n test.exe input_img" <<std::endl;
          return 1;
    }

    //Load image, auto_ptr guarantee this object destruction when leave block {}
		const char *fname = argv[1];
		std::auto_ptr<AImage> _img(::AImageLoad(0, true, fname, indicator));

		// The image is assumed to be grayscale (8 bits).
		// Otherwise the function returns error code 1
		AImageMem8* img = dynamic_cast<AImageMem8*>(_img.get());
		if(!img)
    {
			return 1;
    }

		const int NColors = 1 << img->Bits();

		// Image for saving curvature in pseudocolors
		std::auto_ptr<AImage> _img_pal(::AImageCreate(0,AImage::IMGC_PALETTE,img->Width(),img->Height(),img->Bits()));
		AImageMem8* img_pal = dynamic_cast<AImageMem8*>(_img_pal.get());
		build_smooth_palette_spectra_2(img_pal->Palette());

		// Container and image for histogramm saving
		std::vector<double> hist(256);
		std::auto_ptr<AImage> _hist_img(::AImageCreate(0,AImage::IMGC_PALETTE,256,256,8));
		AImageMem8* hist_img = dynamic_cast<AImageMem8*>(_hist_img.get());
		build_smooth_palette_spectra_2(hist_img->Palette());

		// Save image size
		const int w = img->Width();
		const int h = img->Height();
		const int size = w * h;

		//     (). 
		//   1-    ,   , ,  .
		std::vector<float> _data(size);
		std::vector<float> _data22(size);
		std::vector<float> combination(size);
		
		//        .
		float *data  = &(_data[0]);
		float *origData = &(_data22[0]);
		float *comb = &(combination[0]);
		
		//          float
		Vtl::Plf::F11::Convert<float> c_to_float;
		Vtl::ImAlg::Tools::copy_1_1((*img)[0],w,h,img->NextLineShift(),data,w,c_to_float);

		memcpy(origData, data, size * sizeof(float));

		//    ()
		const float g2_Mult=1;
		Vtl::Plf::F11::GammaCorMult<float> lin_3mod(255,2.2,1.0*g2_Mult);
		//     float    byte. ,   float  
		//     0-255 -       
//		Vtl::Plf::F11::Convert<uint8_t> c_to_Byte;
		Vtl::Plf::F11::Convert2ByteBounded c_to_Byte;
		//  ,      
//		Vtl::Plf::F11::Comb<uint8_t,Vtl::Plf::F11::GammaCorMult<float>,Vtl::Plf::F11::Convert<uint8_t> > adj_3mod(lin_3mod,c_to_Byte);
		Vtl::Plf::F11::Comb<uint8_t,Vtl::Plf::F11::GammaCorMult<float>,Vtl::Plf::F11::Convert2ByteBounded > adj_3mod(lin_3mod,c_to_Byte);


    //////////////////////////////////////////////////////////////////////////////
    // Multiply image to sinus
    int cx = w / 2;
    int cy = h / 2;

    for (int y = 0; y < h; y++)
    {
      for (int x = 0; x < w; x++)
      {
        float dx = float(x - cx);
        float dy = float(y - cy);

        float fi = atan2(dy, dx);
        float sinFi = sin(10 * fi);

        *(data + y * w + x) *= 0.1f * sinFi + 0.4f;
        //*(data + y * w + x) /= 2;
      }
    }

    //NormalizeImage(data, data, w, h);
    SaveImageArray(data, w, h, "Sinus", "Mult");


    //////////////////////////////////////////////////////////////////////////////




// Edge detection
		const float sigma_canny = 3.0f;
		const float nsigmas_env_canny = 3.5f;
		const float histStart = 15.0f;

		CannyEdgeDetector<float> canny(data, w, h);
		canny.Apply(sigma_canny, nsigmas_env_canny, histStart);

		float* Canny = &(canny.edges[0]);
		float* GradX = &(canny.Lx[0]);
		float* GradY = &(canny.Ly[0]);

		SaveImageArray(&(canny.edges[0]), w, h, fname, "_Canny");

    //-------------------------------------------------------------------------
   
		Vtl::Curvature::MultiscaleCurvatureEst(fname, w, h, data, Canny, comb,
		               GradX, GradY, hist, img_pal, hist_img, c_to_Byte, adj_3mod);

    //-------------------------------------------------------------------------
    /*float rad = 75.0f;
    Levelling(&(canny.edges[0]), size, rad);
    //SaveImageArray(&(canny.edges[0]), w, h, fname, "_GroundTruth");

    int edgePointsNumb = 0;
    int passedNumb = 0;
    float averErr = 0.0f;
    float errThresh = 0.15f * rad;

    for (int i = 0; i < size; ++i)
    {
      if (canny.edges[i] > 0.0f)
      {
        edgePointsNumb++;

        float r = Vtl::Curvature::CurvatureScale<float>::value_to_radius(comb[i]);
        float err = fabs(fabs(r) - canny.edges[i]);
        //std::cout << std::endl << "val = " << comb[i] << ",  r = " << r
        //  << ",  canny = " << canny.edges[i] << ",  err = " << err;

        averErr += err;

        if (err < errThresh)
        {
          passedNumb++;
        }
      }
    }

    averErr /= edgePointsNumb;

    std::cout << std::endl << std::endl << "edgePointsNumb = " << edgePointsNumb
      << ",   passedNumb = " << passedNumb << ",  averErr = " << averErr;
    std::cout << "";*/

    
		
#ifdef CDI_OPERATE_AS_LINE_DETECTION
		// Sinograms

		// Gradient module + nms
		std::vector<Vtl::CurveDetection::Point> points;
	  Vtl::CurveDetection::SelectPoints(points, Canny, w, h, 0.0f);
		Vtl::CurveDetection::HoughFuncForLine<float> line1(0.2f, 1.0f, w, h);
		Vtl::CurveDetection::HoughTransform(points, line1, origData, w, h, w, fname, "Canny"); 
		// Gradient module
		find_edges(data, w, h, Canny, sigma_canny, nsigmas_env_canny, false);
    SaveImage(img, Canny, w, h, w, adj_3mod, sigma_canny, fname, "GradMod");
    Vtl::CurveDetection::SelectPoints(points, Canny, w, h, 0.0f);
		Vtl::CurveDetection::HoughFuncForLine<float> line2(0.2f, 1.0f, w, h);
		Vtl::CurveDetection::HoughTransform(points, line2, origData, w, h, w, fname, "GradMod"); 
		// Curvature filtering
	  Vtl::Plf::F11::Abs<float> abs;
		Vtl::ImAlg::Tools::copy_1_1(comb, w, h, w, Canny, w, abs);
		Vtl::CurveDetection::SelectPoints(points, Canny, w, h, 0.0f);
		Vtl::CurveDetection::HoughFuncForLine<float> line3(0.2f, 1.0f, w, h);
		Vtl::CurveDetection::HoughTransform(points, line3, origData, w, h, w, fname, "Multiscale_ISO"); 

		// Detection
    /*Vtl::CurveDetection::RandPatternDetection<float, float *, float *,
		   Vtl::CurveDetection::Line <float> > (points, GradX, GradY, comb,
			                               //origData, w, h, w, 3.0f, 30, 0.05f, 0.85f); 
																		 origData, w, h, w, 2.0f, 30, 0.15f, 0.85f);*/
#endif	

#ifdef CDI_OPERATE_AS_CIRCLE_DETECTION
    //Vtl::CurveDetection::TwoStepsClusterization(comb, GradX, GradY, img,
    //                                img_pal, w, h, fname, c_to_Byte, adj_3mod);

    // CLUSTERIZATION OF EGDE POINTS OVER CIRCLE PARAMETERS

    // 1. SELECT EDGE POINTS
    std::vector< DatumR<float> > points;
    Vtl::CurveDetection::SelectPoints(points, comb, w, h);

    // 2. BUILD ACCUMULATOR
    Vtl::CurveDetection::RadiusMesh<float> rMesh(5.0f, w / 2.0f, 0.1f);
    int rNumb = rMesh.size(); // number of radius layers in the accumulator
   
    ImageContainer< std::vector<float> > circleAccum;
    for (int i = 0; i < rNumb; i++)
    {
      std::vector<float> _radLayer(size, 0);
      circleAccum.Add(_radLayer);
    }

    // 3. FILL ACCUMULATOR WITH CIRCLE CENTERS
    std::vector< std::vector< Vtl::CurveDetection::Circle > > cent(rNumb);
    float max = Vtl::CurveDetection::FillCircleAccum(circleAccum, cent, points,
                                           GradX, GradY, w, h, w, rMesh, 1.0f);
    //std::cout << std::endl << "max = " << max;

    // 4. BLUR EACH RADIUS LAYER
    for (int i = 0; i < rNumb; i++)
    {
      std::vector<float> *_radLayer = circleAccum[i];
      float *radL = &((*_radLayer)[0]);

      /*if (max > 0)
      {
        Vtl::CurveDetection::Normalize(radL, w, h, 2 * max);
      }*/

      /*std::ostringstream os;
			os <<	"Accum_" << i << "_" << 5.0 * pow(1.1, i);
      std::string outfname = os.str();
			Tests::SaveImage(img, radL, w, h, w, adj_3mod, 0.0f, fname,
                       outfname.c_str());*/

      std::cout << std::endl << i << "  " << rMesh(i) << "  " << cent[i].size();


      float sigmaCent = rMesh(i) / 9.0f;
      float nsigmas_env = 1.5f;
      typedef Vtl::ImAlg::WndFun::Gauss<float, 0, false> FF_; 
		  FF_ f_(sigmaCent, nsigmas_env);
		  typedef Vtl::Plf::Kernel::FKern<float, Vtl::Plf::Kernel::KT_Conv, true> K_;
		  K_ ker_(f_);
		  Vtl::ImAlg::Conv::ConvInplace<float, K_> g_(ker_);

      Vtl::ImAlg::Conv::Conv2D::conv2DInplace(radL, w, h, w, g_, g_);

      /*os <<	"_blur";
      std::string outfname = os.str();
			Tests::SaveImage(img, radL, w, h, w, adj_3mod, 0.0f, fname,
                       outfname.c_str());*/
    }

    // 5. DETECT MAXIMA IN ACCUMULATOR
    std::vector< std::vector<int> > maxima(rNumb);
    Vtl::CurveDetection::DetectMax(maxima, circleAccum, cent, w, h, w, 5);

    for (int i = 0; i < rNumb; i++)
    {
      std::vector<float> *_radLayer = circleAccum[i];
      float *radL = &((*_radLayer)[0]);

      /*std::ostringstream os;
			os <<	"Accum_max_" << i << "_" << rMesh(i);
      std::string outfname = os.str();
			Tests::SaveImage(img, radL, w, h, w, adj_3mod, 0.0f, fname,
                       outfname.c_str());*/
    }

    // 6. CLUSTERIZE EDGE POINTS
    std::vector< std::vector< Vtl::CurveDetection::Circle > > clust;
    std::vector< std::vector<float> > d;
    Vtl::CurveDetection::ClassifyViaAccum(clust, d, maxima, cent);

    Vtl::CurveDetection::DrawCentClusters(circleAccum, clust, cent, w, h, w);
    for (int i = 0; i < rNumb; i++)
    {
      std::vector<float> *_radLayer = circleAccum[i];
      float *radL = &((*_radLayer)[0]);

      /*std::ostringstream os;
			os <<	"_" << i << "_" << rMesh(i) << "_classes";
      std::string outfname = os.str();
			Tests::SaveImage(img_pal, radL, w, h, w, c_to_Byte, 0.0f, fname, outfname.c_str());*/
    }

    Vtl::CurveDetection::ClustCenters(clust, d, cent);
 
    Vtl::CurveDetection::DrawCentClusters(circleAccum, clust, cent, w, h, w);
    for (int i = 0; i < rNumb; i++)
    {
      std::vector<float> *_radLayer = circleAccum[i];
      float *radL = &((*_radLayer)[0]);

      /*std::ostringstream os;
			os <<	"_" << i << "_" << rMesh(i) << "_clusters";
      std::string outfname = os.str();
			Tests::SaveImage(img_pal, radL, w, h, w, c_to_Byte, 0.0f, fname, outfname.c_str());*/
    }
   
    std::vector< Vtl::CurveDetection::Circle >  pointsClust;
    Vtl::CurveDetection::ClustEdgePoints(pointsClust, clust, cent, 5);
    int cClustNumb = pointsClust.size();

    const int colorsNumb = 10;
    std::vector<float> color;

    if (colorsNumb >= cClustNumb)
    {
      color.resize(colorsNumb);
      color[0] = 1;
      color[1] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(15);
      color[2] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(35);
      color[3] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(40);
      color[4] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(47);
      color[5] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(195);
      color[6] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(160);
      color[7] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(200);
      color[8] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(400);
      color[9] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(1500);
    }
    else
    {
      color.resize(cClustNumb);
      float min = 25.0f;
      float step = 125.0f / cClustNumb;

      for (int i = 0; i < cClustNumb; i++)
      {
        color[i] = min + i * step;
      }
    }

    Vtl::CurveDetection::ThreshDraw(comb, w, h, 0.0f, 255.0f);

    for (int i = cClustNumb - 1; i >= 0; i--)
    {
      Vtl::CurveDetection::DrawPointsClust(pointsClust[i], points, comb, w, h, w, color[i]);
    }
    
    Tests::SaveImage(img_pal, comb, w, h, w, c_to_Byte, 0.0f, fname, "Clusters");

    // 7. RECLUSTERIZE SINGLE POINTS
    int neigSize = 10;

    for (int y = 0; y < h; y++)
    {
      for (int x = 0; x < w; x++)
      {
        if (float(*(comb + y * w + x)) != 0.0f)
        {
          int minX = std::max(0, x - neigSize);
          int maxX = std::min(w - 1, x + neigSize);
          int minY = std::max(0, y - neigSize);
          int maxY = std::min(h - 1, y + neigSize);

          std::vector<float> vals;
          std::vector<int> count;
          int maxCount = -1;
          float maxVal;
          int totalCount = 0;

          for (int n_x = minX; n_x <= maxX; n_x++)
          {
            for (int n_y = minY; n_y <= maxY; n_y++)
            {
              float val = *(comb + n_y * w + n_x);
              if (val != 0.0f)
              {
                totalCount++;

                if (val != 255.0f)
                {
                  int newVal = 1;
                  int incInd;

                  for (int i = 0; i < int(vals.size()); i++)
                  {
                    if (vals[i] == val)
                    {
                      count[i]++;
                      incInd = i;
                      newVal = 0;
                      break;
                    }
                  }

                  if (newVal)
                  {
                    vals.push_back(val);
                    count.push_back(1);
                    incInd = vals.size() - 1;
                  }

                  if (count[incInd] > maxCount)
                  {
                    maxCount = count[incInd];
                    maxVal = vals[incInd];
                  }
                }
              }
            }
          }

          if (totalCount != 0 && maxCount > 0)
          {
            if (float(maxCount) / float(totalCount) > 0.6)
            {
              if (float(*(comb + y * w + x)) != maxVal)
              {
                std::cout << std::endl << "x = " << x << "  y = " << y << "  val = " 
                  << float(*(comb + y * w + x)) << "  newVal = " << maxVal 
                  << "  perc = " << float(maxCount) / float(totalCount);
              }

              *(comb + y * w + x) = maxVal;
            }

            //std::cout << "  -> " << float(*(comb + y * w + x));
            std::cout << "";
          }
        }
      }
    }

    Tests::SaveImage(img_pal, comb, w, h, w, c_to_Byte, 0.0f, fname, "ClustersRecl");


 std::cout << "";
		
#endif
return 0;
	}
}
