#include <stdio.h>
#include <string>
#include <sstream>
#include <memory> //auto_ptr()
#include <assert.h>

#include "aimage.h"
#include "yfilesys.h"
#include "consolap.h"

#include "plf_filters.h"
#include "plf_filters_complex.h"
#include "plf_sys.h"
#include "tests.h"
#include <time.h>
#include "Sidorkina/plt_filter_image.h"
namespace Tests
{
/*
class Filter1to1
{
	AImage::Info inf;
public:
	Filter1to1()
	{
	}
	void Init(AImage *input)
	{
		switch(input->GetImageStore())
		{
		case AImage::IMGS_AImageMem8:
		case AImage::IMGS_AImageMem16:
		case AImage::IMGS_AImageMemFloat:
			break;
		case AImage::IMGS_AImageMemRgb8:
		case AImage::IMGS_AImageMemRgb16:
			break;
		default:
			;
		};
	}
	AImage::Info GetImageInfo() const {return inf;}
	void Run(AImage* output,ProgressCounter* indicator)
	{
	}
	protected:
};
*/
///

template<class Img>
int plf_test_Median(int r,Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<unsigned,Img> src("src",input);
	Vtl::Plf::DstAImage<unsigned,Img> dst("dst");

	Vtl::Plf::medianFilter<unsigned> medianf(r,input->Bits(),"medain");
	src.out.ConnectTo(&medianf.in);
	medianf.out.ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,input->Bits()));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);
	clock_t t1=clock();
	sys.Run(indicator,out_name.c_str());
	clock_t t2=clock();
	printf("classic time %d", int(t2-t1));
	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}

template<class Img>
int plf_test_Median2(int r,Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<unsigned,Img> src("src",input);
	Vtl::Plf::DstAImage<unsigned,Img> dst("dst");

	Vtl::Plf::medianFilter2<unsigned> medianf(r,input->Bits(),"medain");
	src.out.ConnectTo(&medianf.in);
	medianf.out.ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,input->Bits()));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);
	clock_t t1=clock();
	sys.Run(indicator,out_name.c_str());
	clock_t t2=clock();
	printf("new time %d", int(t2-t1));
	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}

///
template<class Img>
int plf_test_Resample(Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	Vtl::Plf::Resample<float> resamp("Resample",712,800);
	src.out.ConnectTo(&resamp.in());
	resamp.out().ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,input->Bits()));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}


///

template<class Img>
int plf_test_AverageGauss(Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
	float sigma=1.7f;
	float env_size_in_sigmas=2.5f;
	Vtl::Plf::AverageGauss<float> average("Average",sigma,env_size_in_sigmas);
	src.out.ConnectTo(&average.in());
	average.out().ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}
template<class Img>
int plf_test_Average(Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
//	const float maxPixelVal=float((1<<bits)-1);
	typedef Vtl::ImAlg::WndFun::Hann<float,0> FF;
	FF hann(9);
	Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv> kern(hann);
	Vtl::Plf::Average<float> average("AverageHann",&kern);
	src.out.ConnectTo(&average.in());
	average.out().ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}


template<class Img>
int plf_test_Sharpen(Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
	float sigma=3.0f;
	float env_size_in_sigmas=2.0f;
	float amplification=4.0f;
	typedef Vtl::ImAlg::WndFun::Gauss<float,0> G0;
	Vtl::Plf::Kernel::FKern<float> gauss(G0(sigma,env_size_in_sigmas));
//	Vtl::Plf::Kernel::Gauss<float,0,Vtl::Plf::Kernel::KT_Conv> gauss(sigma,env_size_in_sigmas);
	Vtl::Plf::F11::ToBounds<float> bounds(0,maxPixelVal);
	Vtl::Plf::Sharpen<float,Vtl::Plf::F11::ToBounds<float> > sharpen("Sharpen",&gauss,amplification,bounds);
	src.out.ConnectTo(&sharpen.in);
	sharpen.out().ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}

template<class Img>
int plf_test_MeanAndDisp(Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst_m("dst_mean");
	Vtl::Plf::DstAImage<float,Img> dst_d("dst_disp");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
	float sigma=3.0f;
	float env_size_in_sigmas=2.0f;
	Vtl::Plf::MeanAndDisp<float > mean_and_disp("MeanAndDisp",maxPixelVal,sigma,env_size_in_sigmas);
	src.out.ConnectTo(&mean_and_disp.in);
	mean_and_disp.out_mean().ConnectTo(&dst_m.in);
	mean_and_disp.out_disp().ConnectTo(&dst_d.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	//Vtl::Plf::FilterElementary* _dd[2]={&dst_m,&dst_d};
	std::list<Vtl::Plf::FilterElementary*> _dd;
	_dd.push_back(&dst_m);
	_dd.push_back(&dst_d);
	Vtl::Plf::FilterElementary* _ss[1]={&src};
	sys.Assign(&src,_dd.begin(),_dd.end());
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is_m=dst_m.in.getImageSize();
	std::auto_ptr<AImage> _out_m(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is_m.width,is_m.height,bits));
	Img *out_m=dynamic_cast<Img *>(_out_m.get());
	if(out_m==0)
		return 1;
	dst_m.setDstImage(out_m);
	Vtl::Plf::ImgSize is_d=dst_d.in.getImageSize();
	std::auto_ptr<AImage> _out_d(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is_d.width,is_d.height,bits));
	Img *out_d=dynamic_cast<Img *>(_out_d.get());
	if(out_d==0)
		return 1;
	dst_d.setDstImage(out_d);

	sys.Run(indicator,out_name.c_str());

	std::string fn_m=out_name+"_mean.pgm";
	::AImageSave(out_m,fn_m.c_str(),indicator);
	std::string fn_d=out_name+"_disp.pgm";
	::AImageSave(out_d,fn_d.c_str(),indicator);
	return 0;
}
template<class Img>
int plf_test_StatDiff(Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
	float sigma=5.0f;
	float env_size_in_sigmas=2.0f;
	float amplification=6.0f;
	float sigma_desired=maxPixelVal*85.0f/256.0f;
	float mean_desired=maxPixelVal/2;
	float alpha=0.8f;
	Vtl::Plf::F11::ToBounds<float> bounds(0,maxPixelVal);

	Vtl::Plf::StatDiff<float,Vtl::Plf::F11::ToBounds<float> > 
		statdiff("StatDiff",maxPixelVal,sigma,env_size_in_sigmas,amplification,sigma_desired,mean_desired,alpha,bounds);

	src.out.ConnectTo(&statdiff.in);
	statdiff.out().ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}

template<class Img>
int plf_test_DeNoise(Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
	float sigma=3.0f;
	float env_size_in_sigmas=1.0f;
	float in_k_sigmas=2.0f;
	Vtl::Plf::F11::ToBounds<float> bounds(0,maxPixelVal);

	Vtl::Plf::DeNoise<float,Vtl::Plf::F11::ToBounds<float> > 
		denoise("DeNoise",maxPixelVal,sigma,env_size_in_sigmas,in_k_sigmas,bounds);

	src.out.ConnectTo(&denoise.in);
	denoise.out().ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}

template<class Img>
int plf_test_HImage(Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
	int radius=7;
	float gamma=2.2f;
	Vtl::Plf::F11::GammaCorMult<float> gam_mult(maxPixelVal,gamma,1);
	Vtl::Plf::HImage<float,Vtl::Plf::F11::GammaCorMult<float> > himage("H-image",radius,gam_mult);
	src.out.ConnectTo(&himage.in());
	himage.out().ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}

template<class Img>
int plf_test_BiLateral(Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
	float epsilon=7.0f/255.0f*maxPixelVal;
	float sigma=2.0f;
	float env_sigmas=2.0f;
	Vtl::Plf::BiLateral<float> bilateral("BiLateral",epsilon,sigma,env_sigmas);
	src.out.ConnectTo(&bilateral.in());
	bilateral.out().ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}


template<class Img>
int plf_test_GradientModul(Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
// Classic Canny
	float sigma=1.0f;
	float env_size_in_sigmas=2.5f;
	float gamma=2.2f;
	Vtl::Plf::GradientModul<float> grad_modul("GradModul",maxPixelVal,sigma,env_size_in_sigmas,gamma);
	src.out.ConnectTo(&grad_modul.in);
	grad_modul.out().ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}

template<class Img>
int plf_test_Canny(Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
// Classic Canny
	float sigma=1.0f;
	float env_size_in_sigmas=2.5f;
	float gamma=2.2f;
	float threshold=20;
	//Vtl::Plf::F11::DoNothing<float> gam_cor;
	Vtl::Plf::Canny<float> canny("Canny",threshold,maxPixelVal,sigma,env_size_in_sigmas,gamma,Vtl::Plf::F11::DoNothing<float>());

	src.out.ConnectTo(&canny.in());
	canny.out().ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}

template<class Img>
int plf_test_HImageEdge(Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
	float threshold=10;
	int radius=7;
	float gamma=2.2f;
	Vtl::Plf::F11::GammaCor<float> gam_mult(maxPixelVal,gamma);
	//Vtl::Plf::F11::DoNothing<float> gam_cor;
	Vtl::Plf::HImageEdge<float,Vtl::Plf::F11::GammaCor<float> > h_edge("HImageEdge",radius,threshold,gam_mult);

	src.out.ConnectTo(&h_edge.in());
	h_edge.out().ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}

template<class Img>
int plf_test_Harris(Img* input,std::string out_name,float k)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
// Classic Canny
	float sigma=1.5f;
	float env_size_in_sigmas=2.5f;
	float gamma=2.2f;
	Vtl::Plf::F11::GammaCor<float> gam_cor(maxPixelVal,gamma);
	//Vtl::Plf::F11::DoNothing<float> gam_cor;
	Vtl::Plf::F31::Harris<float> func(sigma,k); 
	Vtl::Plf::Detect<float,Vtl::Plf::F31::Harris<float>,Vtl::Plf::F11::GammaCor<float> > det(out_name,sigma,env_size_in_sigmas,func,gam_cor);

	src.out.ConnectTo(&det.in());
	det.out().ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}



template<class Img>
int plf_test_HarrisLaplace(Img* input,std::string out_name,float k)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
// Classic Canny
	float sigma=1.5f;
	float env_size_in_sigmas=2.5f;
	float gamma=2.2f;
	Vtl::Plf::F11::GammaCor<float> gam_cor(maxPixelVal,gamma);
	//Vtl::Plf::F11::DoNothing<float> gam_cor;
	Vtl::Plf::F51::HarrisLaplace<float> func(sigma,k); 
	Vtl::Plf::HarrisLaplace<float,Vtl::Plf::F51::HarrisLaplace<float>,Vtl::Plf::F11::GammaCor<float> > det(out_name,sigma,env_size_in_sigmas,func,gam_cor);

	src.out.ConnectTo(&det.in);
	det.out().ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}

template<class Img>
int plf_test_Trace(Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
// Classic Canny
	float sigma=1.5f;
	float env_size_in_sigmas=2.5f;
	float gamma=2.2f;
	Vtl::Plf::F11::GammaCor<float> gam_cor(maxPixelVal,gamma);
	//Vtl::Plf::F11::DoNothing<float> gam_cor;
	Vtl::Plf::F31::Trace<float> func(sigma); 
	Vtl::Plf::Detect<float,Vtl::Plf::F31::Trace<float>,Vtl::Plf::F11::GammaCor<float> > det(out_name,sigma,env_size_in_sigmas,func,gam_cor);

	src.out.ConnectTo(&det.in());
	det.out().ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}
template<class Img>
int plf_test_MinEigenVal(Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
// Classic Canny
	float sigma=1.5f;
	float env_size_in_sigmas=2.5f;
	float gamma=2.2f;
	Vtl::Plf::F11::GammaCor<float> gam_cor(maxPixelVal,gamma);
	//Vtl::Plf::F11::DoNothing<float> gam_cor;
	Vtl::Plf::F31::MinEigenVal<float> func(sigma);
	Vtl::Plf::Detect<float,Vtl::Plf::F31::MinEigenVal<float>,Vtl::Plf::F11::GammaCor<float> > det(out_name,sigma,env_size_in_sigmas,func,gam_cor);

	src.out.ConnectTo(&det.in());
	det.out().ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}
template<class Img>
int plf_test_MaxEigenVal(Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
// Classic Canny
	float sigma=1.5f;
	float env_size_in_sigmas=2.5f;
	float gamma=2.2f;
	Vtl::Plf::F11::GammaCor<float> gam_cor(maxPixelVal,gamma);
	//Vtl::Plf::F11::DoNothing<float> gam_cor;
	Vtl::Plf::F31::MaxEigenVal<float> func(sigma);
	Vtl::Plf::Detect<float,Vtl::Plf::F31::MaxEigenVal<float>,Vtl::Plf::F11::GammaCor<float> > det(out_name,sigma,env_size_in_sigmas,func,gam_cor);

	src.out.ConnectTo(&det.in());
	det.out().ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}
template<class Img>
int plf_test_DiffEigenVal(Img* input,std::string out_name)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
// Classic Canny
	float sigma=1.5f;
	float env_size_in_sigmas=2.5f;
	float gamma=2.2f;
	Vtl::Plf::F11::GammaCor<float> gam_cor(maxPixelVal,gamma);
	//Vtl::Plf::F11::DoNothing<float> gam_cor;
	Vtl::Plf::F31::DiffEigenVal<float> func(sigma);
	Vtl::Plf::Detect<float,Vtl::Plf::F31::DiffEigenVal<float>,Vtl::Plf::F11::GammaCor<float> > det(out_name,sigma,env_size_in_sigmas,func,gam_cor);

	src.out.ConnectTo(&det.in());
	det.out().ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}
template<class Img>
int plf_test_HarrisNMS(Img* input,std::string out_name,float k)
{
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
// Classic Canny
	float sigma=1.5f;
	float env_size_in_sigmas=2.5f;
	float gamma=2.2f;
	Vtl::Plf::F11::GammaCor<float> gam_cor(maxPixelVal,gamma);
	//Vtl::Plf::F11::DoNothing<float> gam_cor;
	Vtl::Plf::F31::Harris<float> func(sigma,k); 
	Vtl::Plf::Detect<float,Vtl::Plf::F31::Harris<float>,Vtl::Plf::F11::GammaCor<float> > det("Harris",sigma,env_size_in_sigmas,func,gam_cor);
	Vtl::Plf::Fpie::HarrisNMS<float> harris_nms;
	Vtl::Plf::EnvProc1to1<float,Vtl::Plf::Fpie::HarrisNMS<float> > loc_max("loc_max",harris_nms);

	src.out.ConnectTo(&det.in());
	det.out().ConnectTo(&loc_max.in);
	loc_max.out.ConnectTo(&dst.in);

	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
	sys.Assign(&src,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}



template<class Img>
int plf_test_DifferentBrahch(Img* input,std::string out_name)
{
#define DIFFERENT_DELAY_BRANCHES

#ifdef DIFFERENT_DELAY_BRANCHES
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
#else
	Vtl::Plf::SrcAImage<float,Img> src("src",input);
	Vtl::Plf::SrcAImage<float,Img> src2("src2",input);
#endif
	Vtl::Plf::DstAImage<float,Img> dst("dst");

	const int bits=input->Bits();
	const float maxPixelVal=float((1<<bits)-1);
	float sigma=1.0f;
	float env_size_in_sigmas=2.5f;
	float gamma=2.2f;

	typedef Vtl::ImAlg::WndFun::Gauss<float,0> G0;
	typedef Vtl::ImAlg::WndFun::Gauss<float,1> G1;
	Vtl::Plf::Kernel::FKern<float> gauss(G0(sigma,env_size_in_sigmas));
	Vtl::Plf::Kernel::FKern<float> gauss_deriv(G1(sigma,env_size_in_sigmas));
	//Vtl::Plf::Kernel::Gauss<float,0,Vtl::Plf::Kernel::KT_Conv> gauss(sigma,env_size_in_sigmas); 
	//Vtl::Plf::Kernel::Gauss<float,1,Vtl::Plf::Kernel::KT_Conv> gauss_deriv(sigma,env_size_in_sigmas); 

	Vtl::Plf::ConvolveHorz<float> gx_h_conv("gx H-conv",&gauss_deriv);
	Vtl::Plf::ConvolveVert<float> gx_v_conv("gx V-conv",&gauss);
	Vtl::Plf::ConvolveHorz<float> gy_h_conv("gy H-conv",&gauss);
	Vtl::Plf::ConvolveVert<float> gy_v_conv("gy V-conv",&gauss_deriv);


	const Vtl::Plf::F21::SqrtOfSumSqr<float> ssqr;
	Vtl::Plf::F11::GammaCorMult<float> gam_mult(maxPixelVal,gamma,sqrt(2.0f*3.1415926f)*sigma);
	Vtl::Plf::PixByPix2to1<float,Vtl::Plf::F21::SqrtOfSumSqr<float>,Vtl::Plf::F11::GammaCorMult<float> > sum_sqr;
	sum_sqr.Init("SumSqr",ssqr,gam_mult);

	Vtl::Plf::Resample<float> resamp("Resample",712,800);

	Vtl::Plf::F11::Mult<float> m2(0.5);
	Vtl::Plf::Levels<float,Vtl::Plf::F11::Mult<float> > mult05_1("mult05_1",m2);
	Vtl::Plf::Levels<float,Vtl::Plf::F11::Mult<float> > mult05_2("mult05_2",m2);
	Vtl::Plf::PixByPix2to1<float,Vtl::Plf::F21::SqrtOfSumSqr<float> > sum_sqr2("SumSqr-2",ssqr);

	//------ create graph --------------------
	src.out.ConnectTo(&gx_h_conv.in);
	src.out.ConnectTo(&gy_h_conv.in);
	gx_h_conv.out.ConnectTo(&gx_v_conv.in);
	gy_h_conv.out.ConnectTo(&gy_v_conv.in);
	gx_h_conv.out.ConnectTo(&gx_v_conv.in);
	gy_h_conv.out.ConnectTo(&gy_v_conv.in);
	gx_v_conv.out.ConnectTo(&sum_sqr.in_1);
	gy_v_conv.out.ConnectTo(&sum_sqr.in_2);
	sum_sqr.out.ConnectTo(&mult05_1.in);

#ifdef DIFFERENT_DELAY_BRANCHES
	src.out.ConnectTo(&mult05_2.in);
#else
	src2.out.ConnectTo(&mult05_2.in);
#endif
	mult05_1.out.ConnectTo(&sum_sqr2.in_1);
	mult05_2.out.ConnectTo(&sum_sqr2.in_2);
	sum_sqr2.out.ConnectTo(&resamp.in());
	resamp.out().ConnectTo(&dst.in);
//------ create filters --------------------
	Vtl::Plf::FiltersSystem sys(out_name+".dot");
//------ Assing task------------------------
#ifdef DIFFERENT_DELAY_BRANCHES
	sys.Assign(&src,&dst);
#else
//	std::list<Vtl::Plf::FilterElementary*> _ss;
//	_ss.push_back(&src);
//	_ss.push_back(&src2);
//	sys.Assign(_ss.begin(),_ss.end(),&dst);
	Vtl::Plf::FilterElementary* _ss[2]={&src,&src2};
	sys.Assign(_ss,_ss+2,&dst);
#endif

//	std::list<Vtl::Plf::FilterBase*> _ss;
//	_ss.push_back(&src);
//	_ss.push_back(&src2);
//	sys.Assign(_ss.begin(),_ss.end(),&dst);

//	Vtl::Plf::FilterBase* _ss[2]={&src,&src2};
//	sys.Assign(_ss,_ss+2,&dst);
//------ Initialize destinations -----------
	Vtl::Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
	Img *out=dynamic_cast<Img *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);


	sys.Run(indicator,out_name.c_str());

	std::string fn=out_name+".pgm";
	::AImageSave(out,fn.c_str(),indicator);
	return 0;
}

int test_plf_Filter(int argc, char* argv[])
{
//------ load_images ---------------------
	std::auto_ptr<AImage> _in(::AImageLoad(0,true,argv[1],indicator));
	AImageMem8 *in=dynamic_cast<AImageMem8 *>(_in.get());
	if(in==0)
		return 1;
	const int bits=in->Bits();
	const float maxPixelVal=float((1<<bits)-1);



	clock_t t1=clock();
	plf_test_Median(5,in,"Median");
	clock_t t2=clock();
	plf_test_Median2(5,in,"Median2");
	clock_t t3=clock();
	printf("%d   %d", int(t2-t1), int(t3-t2) );

	plf_test_AverageGauss(in,"AverGauss");
	plf_test_Average(in,"AverWnd");
	plf_test_Canny(in,"Canny");
	plf_test_DifferentBrahch(in,"DiffBranch");
	plf_test_StatDiff(in,"StatDiff");

	plf_test_Resample(in,"Resample");
	plf_test_Sharpen(in,"Sharpen");
	plf_test_MeanAndDisp(in,"MeanAndDisp");
	plf_test_DeNoise(in,"DeNoise");
	plf_test_HImage(in,"H-image");
	plf_test_BiLateral(in,"BiLateral");
	plf_test_GradientModul(in,"ModGrad");
	plf_test_HImageEdge(in,"H-image-Edges");
	plf_test_Harris(in,"Harris(k=0.04)",0.04f);
	plf_test_Harris(in,"Harris(k=0.00)",0.0f);
	plf_test_Trace(in,"Trace");
	plf_test_MinEigenVal(in,"MinEigenVal");
	plf_test_MaxEigenVal(in,"MaxEigenVal");
	plf_test_DiffEigenVal(in,"DiffEigenVal");
	plf_test_HarrisNMS(in,"HarrisNMS(0.04)",0.04f);


	plf_test_HarrisLaplace(in,"HarrisLaplace(k=0.04)",0.04f);


	return 0;
}

}//namespace Tests
