#include <stdio.h>
#include <string>
#include <sstream>
#include <memory> //auto_ptr()
#include <assert.h>

#include "aimage.h"
#include "yfilesys.h"
#include "consolap.h"

#include "plf_filters.h"
#include "plf_filters_complex.h"
#include "plf_sys.h"
#include "tests.h"
#include <time.h>
#include "Sidorkina/plt_filter_image.h"


namespace Tests
{
	template<class Img>
	int plf_test_DiZenzo(Img* r,Img* g,Img* b,std::string out_name)
	{
		Vtl::Plf::SrcAImage<float,Img> src_r("src_r",r);
		Vtl::Plf::SrcAImage<float,Img> src_g("src_g",g);
		Vtl::Plf::SrcAImage<float,Img> src_b("src_b",b);
		Vtl::Plf::DstAImage<float,Img> dst("dst");

		const int bits=r->Bits();	
		const int max_val=(1<<bits)-1;
		float threshold=20;
		Vtl::Plf::DiZenzo<float> dizenzo("DiZenzo",float(max_val),threshold,1.5,2.0f,4.2f);


		src_r.out.ConnectTo(&dizenzo.in_1);
		src_g.out.ConnectTo(&dizenzo.in_2);
		src_b.out.ConnectTo(&dizenzo.in_3);
		dizenzo.out().ConnectTo(&dst.in);

		Vtl::Plf::FiltersSystem sys(out_name+".dot");
	//------ Assing task------------------------
		std::list<Vtl::Plf::SrcAImage<float,Img>* > src_list;
		src_list.push_back(&src_r);
		src_list.push_back(&src_g);
		src_list.push_back(&src_b);
		sys.Assign(src_list.begin(),src_list.end(),&dst);
	//------ Initialize destinations -----------
		Vtl::Plf::ImgSize is=dst.in.getImageSize();
		std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,is.width,is.height,bits));
		Img *out=dynamic_cast<Img *>(_out.get());
		if(out==0)
			return 1;
		dst.setDstImage(out);

		sys.Run(indicator,out_name.c_str());

		std::string fn=out_name+".pgm";
		::AImageSave(out,fn.c_str(),indicator);
		return 0;
	}

	int test_plf_Filter_rgb(int argc, char* argv[])
	{
	//------ load_images ---------------------
		std::auto_ptr<AImage> _in(::AImageLoad(0,true,argv[2],indicator));
		AImageMemRgb8 *in=dynamic_cast<AImageMemRgb8 *>(_in.get());
		if(in==0)
			return 1;
		AImageMem8 * r=in->GetPlane8(0);
		AImageMem8 * g=in->GetPlane8(1);
		AImageMem8 * b=in->GetPlane8(2);
		//const int bits=in->Bits();
		//const float maxPixelVal=float((1<<bits)-1);

		plf_test_DiZenzo(r,g,b,"DiZenzo");


		return 0;
	}

}//namespace Tests
