#include <string>
#include <sstream>
#include <memory> //auto_ptr()
#include <assert.h>

#include "aimage.h"
#include "yfilesys.h"
#include "consolap.h"

#include "plf_filters.h"
//#include "plf_filters_complex.h"
#include "plf_sys.h"
//#include <time.h>
#include "plt_filter_image.h"
#include "color_spaces.h" 
#include "yfilesys.h" 

#include "tests.h"
namespace Tests
{

template<typename T,class ColorSystem>
void check_invertibility()
{
	std::cout << "check_invertibility() : " << ColorSystem::name() << std::endl;
	bool ok=true;
	ColorSystem::toRGB<T> to; 
	ColorSystem::fromRGB<T> from;
	const T eps=Vtl::NumAlg::RelerrEps<T>();
	T r1;
	T g1;
	T b1;
	T c0;
	T c1;
	T c2;
	T eps_max=0;
	for(T r=0; r<255; ++r)
	for(T g=0; g<255; ++g)
	for(T b=0; b<255; ++b)
	{
		from(r,g,b,c0,c1,c2);
//		c0*=(1.f-1.0f*eps);
//		c1*=(1.f-1.0f*eps);;
//		c2*=(1.f-1.0f*eps);;
		to(c0,c1,c2,r1,g1,b1);

		T eps0 =std::max(r,r1)-std::min(r,r1);
		T eps1 =std::max(g,g1)-std::min(g,g1);
		T eps2 =std::max(b,b1)-std::min(b,b1);
		T em=Vtl::max(eps0,eps1,eps2);
		if(em > eps_max)
		{
			std::cout 
				<< "( "	<< r  <<" , "<< g  <<" , "<< b  <<" )    "
				<< "( "	<< r1 <<" , "<< g1 <<" , "<< b1 <<" )    "
				<< "( "	<< c0 <<" , "<< c1 <<" , "<< c2 <<" )    "
			<< std::endl;
		}
		eps_max=std::max(eps_max,em);
		if(em>=eps )
		{
/*			std::cout << "eps= "<<eps << std::endl;
			std::cout << "abs(r-r1)= "<<eps0<< std::endl;
			std::cout << "abs(g-g1)= "<<eps1 << std::endl;
			std::cout << "abs(b-b1)= "<<eps2 << std::endl;

			ok=false;
			std::cout << "Error: ( "<<r<<" , "<<g<<" , "<<b<<" )   -> ( "
			<< r1 <<" , "<< g1 <<" , "<< b1 <<" )"
			<<std::endl;
*/
//			continue;
			//goto finish;
		}
	}
//finish:
	std::cout << "eps_max=" << ((sizeof(eps_max)==1) ? unsigned(eps_max) : eps_max)  << "     eps=" << ((sizeof(eps)==1) ? unsigned(eps) : eps) <<std::endl;
	std::cout << (ok?"OK":"Fail") <<std::endl;
}



template<int n>
void ttt()
{
	const float MaxColor=255;
	std::cout << "====================" << std::endl << "sample n=" << n <<std::endl; 
	std::cout << "is_sample_cyclic_var= " << 
			Vtl::ImAlg::C::EElEll::is_sample_cyclic_var<n>()
	<<std::endl; 
	std::cout << "is_sample_sign_alternating= " << 
		Vtl::ImAlg::C::EElEll::is_sample_sign_alternating<n>()
	<<std::endl; 
	std::cout << "sample_max<n,float>= " << 
		Vtl::ImAlg::C::EElEll::sample_max<n,float>(MaxColor)
	<<std::endl; 
//	std::cout << "sample_max<n,int>= " << 
//		Vtl::ImAlg::C::EElEll::sample_max<n,int>()
//	<<std::endl; 
	std::cout << "sample_min<n,float>= " << 
		Vtl::ImAlg::C::EElEll::sample_min<n,float>(MaxColor)
	<<std::endl; 
	std::cout << "sample_min<n,double>= " << 
		Vtl::ImAlg::C::EElEll::sample_min<n,double>(MaxColor)
	<<std::endl; 
//	std::cout << "sample_max<n,int>= " << 
//		Vtl::ImAlg::C::EElEll::sample_min<n,int>()
//	<<std::endl; 
//	std::cout << "sample_max<n,unsigned char>= " << 
//		(unsigned)Vtl::ImAlg::C::EElEll::sample_min<n,unsigned char>()
//	<<std::endl; 
}

template<typename T,class Img_t,class CFunctor>
void SplitRGBto(const char* out_name,Img_t* r,Img_t* g,Img_t* b,Img_t* c0,Img_t* c1,Img_t* c2,CFunctor rgb2xxx,ProgressCounter* indicator)
{
	Vtl::Plf::SrcAImage<T,Img_t> src_r("src_r",r);
	Vtl::Plf::SrcAImage<T,Img_t> src_g("src_g",g);
	Vtl::Plf::SrcAImage<T,Img_t> src_b("src_b",b);
	typedef typename CFunctor::ColorSystem ColorSystem;
	std::string dst("dst_");

	Vtl::Plf::DstAImage<T,Img_t> dst_E(dst+ColorSystem::sample_name<0>());
	Vtl::Plf::DstAImage<T,Img_t> dst_Elam(dst+ColorSystem::sample_name<1>());
	Vtl::Plf::DstAImage<T,Img_t> dst_Elamlam(dst+ColorSystem::sample_name<2>());

	Vtl::Plf::PixByPix3to3<T,CFunctor> color_convert(out_name,rgb2xxx);

	src_r.out.ConnectTo(&color_convert.in_1);
	src_g.out.ConnectTo(&color_convert.in_2);
	src_b.out.ConnectTo(&color_convert.in_3);
	color_convert.out_1.ConnectTo(&dst_E.in);
	color_convert.out_2.ConnectTo(&dst_Elam.in);
	color_convert.out_3.ConnectTo(&dst_Elamlam.in);

	Vtl::Plf::FiltersSystem sys(std::string("SplitRGBto_")+out_name+".dot");
//------ Assing task------------------------
	std::list<Vtl::Plf::SrcAImage<T,Img_t>* > src_list;
	std::list<Vtl::Plf::DstAImage<T,Img_t>* > dst_list;
	src_list.push_back(&src_r);
	src_list.push_back(&src_g);
	src_list.push_back(&src_b);
	dst_list.push_back(&dst_E);
	dst_list.push_back(&dst_Elam);
	dst_list.push_back(&dst_Elamlam);
	sys.Assign(src_list.begin(),src_list.end(),dst_list.begin(),dst_list.end());
//------ Initialize destinations -----------
//	Vtl::Plf::ImgSize is=dst_E.in.getImageSize();
	dst_E.setDstImage(c0);
	dst_Elam.setDstImage(c1);
	dst_Elamlam.setDstImage(c2);

	sys.Run(indicator,out_name);
}

template<typename T,class Img_t,class CFunctor>
void SplitAndSaveRGBto(const char* out_name,Img_t* r,Img_t* g,Img_t* b,Img_t* c0,Img_t* c1,Img_t* c2,CFunctor rgb2xxx,ProgressCounter* indicator)
{
	std::string fname=YFileSys::remExt(out_name)+"_";
	std::string ext=".pgm";
	typedef typename CFunctor::ColorSystem ColorSystem;
	fname+=ColorSystem::name();
	fname+="_";
		
	SplitRGBto<float>(ColorSystem::name().c_str(),r,g,b,c0,c1,c2,rgb2xxx,indicator);

	std::string fn=fname+"0-"+ColorSystem::sample_name<0>()+ext;
	::AImageSave(c0,fn.c_str(),indicator);
	fn=fname+"1-"+ColorSystem::sample_name<1>()+ext;
	::AImageSave(c1,fn.c_str(),indicator);
	fn=fname+"2-"+ColorSystem::sample_name<2>()+ext;
	::AImageSave(c2,fn.c_str(),indicator);
}

int test_shadow_detection(int argc,char *argv[])
{
	//My temp tests start
	{
		ttt<0>();
		ttt<1>();
		ttt<2>();
		//check_invertibility<unsigned char,Vtl::ImAlg::C::HSV>();
		//check_invertibility<unsigned int,Vtl::ImAlg::C::HSV>();
		//check_invertibility<float,Vtl::ImAlg::C::HSV>();
		//check_invertibility<double,Vtl::ImAlg::C::HSV>();
	}
	//My temp tests finished
		std::string out_name=argv[2];
		std::auto_ptr<AImage> _in(::AImageLoad(0,true,argv[2],indicator));
		AImageMemRgb8 *in=dynamic_cast<AImageMemRgb8 *>(_in.get());
		if(in==0)
			return 1;
		std::auto_ptr<AImage> _out(::AImageCreate(0,in->ImageClass(),in->Width(),in->Height(),in->Bits()));
		AImageMemRgb8 *out=dynamic_cast<AImageMemRgb8 *>(_out.get());
		if(out==0)
			return 1;
		const int bits=in->Bits();
		AImageMem8 * r=in->GetPlane8(0);
		AImageMem8 * g=in->GetPlane8(1);
		AImageMem8 * b=in->GetPlane8(2);
		AImageMem8 * r_o=out->GetPlane8(0);
		AImageMem8 * g_o=out->GetPlane8(1);
		AImageMem8 * b_o=out->GetPlane8(2);

		using namespace Vtl::ImAlg::C;
		//typedef EElEll::fromRGB<float> ColorConv;
		{
			typedef XYZ::fromRGB<float> ColorConv;
			Normalizer<float,XYZ,nDefault> norm(255,0,128);
			ColorConv rgb2xxx(norm);
			SplitAndSaveRGBto<float>(out_name.c_str(),r,g,b,r_o,g_o,b_o,rgb2xxx,indicator);
		}
		{
			typedef YCbCr::fromRGB<float> ColorConv;
			Normalizer<float,YCbCr,nDefault> norm(255,0,128);
			ColorConv rgb2xxx(norm);
			SplitAndSaveRGBto<float>(out_name.c_str(),r,g,b,r_o,g_o,b_o,rgb2xxx,indicator);
		}
		{
//			typedef EElEll::fromRGB_positive<8,float> ColorConv;
			typedef EElEll::fromRGB<float> ColorConv;
			Normalizer<float,EElEll,nDefault> norm(255,0,128);
			ColorConv rgb2xxx(norm);
			SplitAndSaveRGBto<float>(out_name.c_str(),r,g,b,r_o,g_o,b_o,rgb2xxx,indicator);
		}
		{
			typedef YIQ::fromRGB<float> ColorConv;
			Normalizer<float,YIQ,nDefault> norm(255,0,128);
			ColorConv rgb2xxx(norm);
			SplitAndSaveRGBto<float>(out_name.c_str(),r,g,b,r_o,g_o,b_o,rgb2xxx,indicator);
		}
		{
			typedef YUV::fromRGB<float> ColorConv;
			Normalizer<float,YUV,nDefault> norm(255,0,128);
			ColorConv rgb2xxx(norm);
			SplitAndSaveRGBto<float>(out_name.c_str(),r,g,b,r_o,g_o,b_o,rgb2xxx,indicator);
		}
		{
			typedef HSV::fromRGB<float> ColorConv;
			ColorConv rgb2xxx;
			SplitAndSaveRGBto<float>(out_name.c_str(),r,g,b,r_o,g_o,b_o,rgb2xxx,indicator);

			HSV::toRGB<unsigned> to;
			unsigned r,g,b;
			to(0u,255u,255u,r,g,b);
			std::cout << "HSV:(0,255,255) -> ( " << r <<" , " << g << " , " << b << " )" << std::endl;
			to(128u,255u,255u,r,g,b);
			std::cout << "HSV:(128,255,255) -> ( " << r <<" , " << g << " , " << b << " )"<<  std::endl;
			to(255u,255u,255u,r,g,b);
			std::cout << "HSV:(255,255,255) -> ( " << r <<" , " << g << " , " << b << " )" << std::endl;
		}
		{
			typedef HSL::fromRGB<float> ColorConv;
			Normalizer<float,HSL,nDefault> norm(255,0,128);
			ColorConv rgb2xxx(norm);
			SplitAndSaveRGBto<float>(out_name.c_str(),r,g,b,r_o,g_o,b_o,rgb2xxx,indicator);
		}
		{
			typedef IV1V2::fromRGB<float> ColorConv;
			Normalizer<float,IV1V2,nDefault> norm(255,0,128);
			ColorConv rgb2xxx(norm);
			SplitAndSaveRGBto<float>(out_name.c_str(),r,g,b,r_o,g_o,b_o,rgb2xxx,indicator);
		}
		{
			typedef HSI<>::fromRGB<float> ColorConv;
			ColorConv rgb2xxx;
			SplitAndSaveRGBto<float>(out_name.c_str(),r,g,b,r_o,g_o,b_o,rgb2xxx,indicator);

			HSI<>::toRGB<float> to;
			float r,g,b;
			to(0,255,255,r,g,b);
			std::cout << "HSI:(0,255,255) -> ( " << r <<" , " << g << " , " << b << " )" << std::endl;
			to(128,255,255,r,g,b);
			std::cout << "HSI:(128,255,255) -> ( " << r <<" , " << g << " , " << b << " )" << std::endl;
			to(255,255,255,r,g,b);
			std::cout << "HSI:(255,255,255) -> ( " << r <<" , " << g << " , " << b << " )" << std::endl;
		}
		{
			typedef HSI<EElEll>::fromRGB<float> ColorConv;
			ColorConv rgb2xxx;
			SplitAndSaveRGBto<float>(out_name.c_str(),r,g,b,r_o,g_o,b_o,rgb2xxx,indicator);

			HSI<EElEll>::toRGB<float > to;
			float r,g,b;
			to(0,255,255,r,g,b);
			std::cout << "HSI<EElEll>:(0,255,255) -> ( " << r <<" , " << g << " , " << b << " )" << std::endl;
			to(128,255,255,r,g,b);
			std::cout << "HSI<EElEll>:(128,255,255) -> ( " << r <<" , " << g << " , " << b << " )" << std::endl;
			to(255,255,255,r,g,b);
			std::cout << "HSI<EElEll>:(255,255,255) -> ( " << r <<" , " << g << " , " << b << " )" << std::endl;
		}
		{
			typedef Lab::fromRGB<float> ColorConv;
			Normalizer<float,Lab,nDefault> norm(255,0,128);
			ColorConv rgb2xxx(norm);
			SplitAndSaveRGBto<float>(out_name.c_str(),r,g,b,r_o,g_o,b_o,rgb2xxx,indicator);
		}


		return 0;
}
}