#include <stdio.h>
#include <string>
#include <sstream>
#include <iostream>
#include <iomanip>
#include <memory> //auto_ptr()

#include "spec_functions.h"
#include "tests.h"


#define EXT_INFO
namespace Tests
{
	typedef double XandRes[2];
	static double  NoMult(double x)
	{
		return 1;
	}
	template<class F,class M> double chkFun(F fun,const XandRes *p,int len,M mult=NoMult)
	{
		double e=0;
		for(int i=0; i<len ; ++i)
		{
			double x=p[i][0];
			double r=fun(float(x));
			e=std::max(e,fabs( r*mult(x) - p[i][1] ) );
		}
		return e;
	}
	template<class F> double chkFun(F fun,const XandRes *p,int len)
	{
		return chkFun(fun,p,len,NoMult);
	}
	bool info(double e,double eps,const char *name)
	{
		if(e<eps)
		{
			std::cout << name << " is OK!" ;
			#ifdef EXT_INFO
			std::cout << "   eps="<<e;
			#endif
			std::cout << std::endl;
			return true;
		}
		else
		{
			std::cout << name << " is BAD!!!!!  Real error=" << e << "  Theor. error=" << eps << std::endl;
			return false;
		}
	}

	bool check_J0()
	{
		static const XandRes d[]=  //data from Abramovitz-Stegun
		{
			{ 0. ,+1.},
			{ 0.5,+0.938469807240813},
			{ 1.0,+0.765197686557967},
			{ 4.0,-0.397149809863847},
			{ 8.0,+0.171650807137554},
			{10.0,-0.245935764451348},
			{15.0,-0.014224472826781}
		};
		double e=chkFun(Vtl::NumAlg::Spf::bess::J<0>,d,sizeof(d)/sizeof(d[0]));
		return info(e,2.0e-8,"Vtl::NumAlg::Spf::bess::J<0>");
	}
	bool check_J1()
	{
		static const XandRes d[]=  //data from Abramovitz-Stegun
		{
			{ 0. ,+0.},
			{ 0.5,+0.2422684577},
			{ 1.0,+0.4400505857},
			{ 4.0,-0.0660433280},
			{ 8.0,+0.2346363469},
			{10.0,+0.0434727462},
			{15.0,+0.2051040386}
		};
		double e=chkFun(Vtl::NumAlg::Spf::bess::J<1>,d,sizeof(d)/sizeof(d[0]));
		return info(e,1.4e-8,"Vtl::NumAlg::Spf::bess::J<1>");
	}
	bool check_J2()
	{
		static const XandRes d[]=  //data from Abramovitz-Stegun
		{
			{ 0. ,+0.},
			{ 0.5,+0.0306040235},
			{ 1.0,+0.1149034849},
			{ 4.0,+0.3641281459},
			{ 8.0,-0.1129917204},
			{10.0,+0.2546303137},
			{15.0,+0.0415716780}
		};
		double e=chkFun(Vtl::NumAlg::Spf::bess::J<2>,d,sizeof(d)/sizeof(d[0]));
		return info(e,1.4e-8,"Vtl::NumAlg::Spf::bess::J<2>");
	}
	bool check_Y0()
	{
		static const XandRes d[]=  //data from Abramovitz-Stegun
		{
			{ 0.1,-1.5342386514},
			{ 0.5,-0.4445187335},
			{ 1.0,+0.0882569642},
			{ 4.0,-0.0169407393},
			{ 8.0,+0.2235214894},
			{10.0,+0.0556711673},
			{15.0,+0.2054642960}
		};
		double e=chkFun(Vtl::NumAlg::Spf::bess::Y<0>,d,sizeof(d)/sizeof(d[0]));
		return info(e,4.5e-8,"Vtl::NumAlg::Spf::bess::Y<0>");
	}
	bool check_Y1()
	{
		static const XandRes d[]=  //data from Abramovitz-Stegun
		{
			{ 0.1,-6.4589510947},
			{ 0.5,-1.4714723927},
			{ 1.0,-0.7812128213},
			{ 4.0,+0.3979257106},
			{ 8.0,-0.1580604617},
			{10.0,+0.2490154242},
			{15.0,+0.0210736280}
		};
		double e=chkFun(Vtl::NumAlg::Spf::bess::Y<1>,d,sizeof(d)/sizeof(d[0]));
		return info(e,9.9e-8,"Vtl::NumAlg::Spf::bess::Y<1>");
	}
	bool check_Y2()
	{
		static const XandRes d[]=  //data from Abramovitz-Stegun
		{
			{ 0.1,-127.64478324},
			{ 0.5,  -5.44137084},
			{ 1.0,  -1.65068261},
			{ 4.0,  +0.21590359},
			{ 8.0,  -0.26303660},
			{10.0,  -0.00586808},
			{15.0,  -0.20265448}
		};
		double e=chkFun(Vtl::NumAlg::Spf::bess::Y<2>,d,sizeof(d)/sizeof(d[0]));
		return info(e,2.2e-7,"Vtl::NumAlg::Spf::bess::Y<2>");
	}
//----------------
	static double  I_Mult(double x)
	{
		return exp(-x);
	}
	static double  I2_0to5_Mult(double x)
	{
		return 1./(x*x);
	}
	bool check_I0()
	{
		static const XandRes d[]=  //data from Abramovitz-Stegun
		{
			{ 0.0,1.0},
			{ 0.5,+0.6450352704},
			{ 1.0,+0.4657596076},
			{ 4.0,+0.2070018212},
			{ 8.0,+0.1434317818},
			{10.0,+0.1278333371},
			{15.0,+0.1038995314}
		};
		double e=chkFun(Vtl::NumAlg::Spf::bess::I<0>,d,sizeof(d)/sizeof(d[0]),I_Mult);
		return info(e,1.0e-7,"Vtl::NumAlg::Spf::bess::I<0>");
	}
	bool check_I1()
	{
		static const XandRes d[]=  //data from Abramovitz-Stegun
		{
			{ 0.0, 0.0},
			{ 0.5, 0.1564208032},
			{ 1.0, 0.2079104154},
			{ 4.0, 0.1787508395},
			{ 8.0, 0.1341424933},
			{10.0, 0.1212626814},
			{15.0, 0.1003741751}
		};
		double e=chkFun(Vtl::NumAlg::Spf::bess::I<1>,d,sizeof(d)/sizeof(d[0]),I_Mult);
		return info(e,8.0e-9,"Vtl::NumAlg::Spf::bess::I<1>");
	}
	bool check_I2()
	{
		static const XandRes d0[]=  //data from Abramovitz-Stegun
		{
			{ 0.0, 0.125},
			{ 0.5, 0.1276245967},
			{ 1.0, 0.1357476698},
			{ 4.0, 0.4013868359},
		};
		double e=chkFun(Vtl::NumAlg::Spf::bess::I<2>,d0,sizeof(d0)/sizeof(d0[0]),I2_0to5_Mult);

		static const XandRes d[]=  //data from Abramovitz-Stegun
		{
			{ 8.0, 0.109896158},
			{10.0, 0.103580801},
			{15.0, 0.090516308}
		};
		e=std::max(e,chkFun(Vtl::NumAlg::Spf::bess::I<2>,d,sizeof(d)/sizeof(d[0]),I_Mult));
		return info(e,3.9e-8,"Vtl::NumAlg::Spf::bess::I<2>");
	}
//----------------
	static double  K_Mult(double x)
	{
		return exp(x);
	}
	static double  K2_0to5_Mult(double x)
	{
		return x*x;
	}
	bool check_K0()
	{
		static const XandRes d[]=  //data from Abramovitz-Stegun
		{
			{ 0.1,2.6823261023},
			{ 0.5,1.5241093857},
			{ 1.0,1.14446307977},
			{ 4.0,0.6092976693},
			{ 8.0,0.4366230185},
			{10.0,0.3916319344},
			{15.0,0.3210023534}
		};
		double e=chkFun(Vtl::NumAlg::Spf::bess::K<0>,d,sizeof(d)/sizeof(d[0]),K_Mult);
		return info(e,1.8e-7,"Vtl::NumAlg::Spf::bess::K<0>");
	}
	bool check_K1()
	{
		static const XandRes d[]=  //data from Abramovitz-Stegun
		{
			{ 0.1,10.890182683},
			{ 0.5, 2.7310097082},
			{ 1.0, 1.6361534863},
			{ 4.0, 0.6815759452},
			{ 8.0, 0.4631490928},
			{10.0, 0.4107665704},
			{15.0, 0.3315348949}
		};
		double e=chkFun(Vtl::NumAlg::Spf::bess::K<1>,d,sizeof(d)/sizeof(d[0]),K_Mult);
		return info(e,1.6e-7,"Vtl::NumAlg::Spf::bess::K<1>");
	}
	bool check_K2()
	{
		static const XandRes d0[]=  //data from Abramovitz-Stegun
		{
			{ 0.1, 1.995039646},
			{ 0.5, 1.887545888},
			{ 1.0, 1.624838899},
			{ 4.0, 0.278422808},
		};
		double e=chkFun(Vtl::NumAlg::Spf::bess::K<2>,d0,sizeof(d0)/sizeof(d0[0]),K2_0to5_Mult);

		static const XandRes d[]=  //data from Abramovitz-Stegun
		{
			{ 8.0, 0.55241029},
			{10.0, 0.47378525},
			{15.0, 0.36520701}
		};
		e=std::max(e,chkFun(Vtl::NumAlg::Spf::bess::K<2>,d,sizeof(d)/sizeof(d[0]),K_Mult));
		return info(e,6.4e-8,"Vtl::NumAlg::Spf::bess::K<2>");
	}

	int test_spec_functions(int argc, char* argv[])
	{
		bool isOk=true;
		isOk &= check_J0();
		isOk &= check_J1();
		isOk &= check_J2();

		isOk &= check_Y0();
		isOk &= check_Y1();
		isOk &= check_Y2();

		isOk &= check_I0();
		isOk &= check_I1();
		isOk &= check_I2();

		isOk &= check_K0();
		isOk &= check_K1();
		isOk &= check_K2();

		{
			float v;
			// test for (x< 8.0 ) and (x>8.0)
			v=Vtl::NumAlg::Spf::bess::J<0>(1.0f);
			v=Vtl::NumAlg::Spf::bess::J<0>(9.0f);
			v=Vtl::NumAlg::Spf::bess::Y<0>(1.0f);
			v=Vtl::NumAlg::Spf::bess::Y<0>(9.0f);
			v=Vtl::NumAlg::Spf::bess::J<1>(1.0f);
			v=Vtl::NumAlg::Spf::bess::J<1>(9.0f);
			v=Vtl::NumAlg::Spf::bess::Y<1>(1.0f);
			v=Vtl::NumAlg::Spf::bess::Y<1>(9.0f);
			v=Vtl::NumAlg::Spf::bess::J<2>(1.0f);
			v=Vtl::NumAlg::Spf::bess::J<2>(9.0f);
			v=Vtl::NumAlg::Spf::bess::Y<2>(1.0f);
			v=Vtl::NumAlg::Spf::bess::Y<2>(9.0f);
			v=Vtl::NumAlg::Spf::bess::J<7>(1.0f);
			v=Vtl::NumAlg::Spf::bess::J<7>(9.0f);
			v=Vtl::NumAlg::Spf::bess::Y<7>(1.0f);
			v=Vtl::NumAlg::Spf::bess::Y<7>(9.0f);
			// test for (x< 3.75 ) and (x>3.75)
			v=Vtl::NumAlg::Spf::bess::I<0>(1.0f);
			v=Vtl::NumAlg::Spf::bess::I<0>(4.0f);
			v=Vtl::NumAlg::Spf::bess::I<1>(1.0f);
			v=Vtl::NumAlg::Spf::bess::I<1>(4.0f);
			v=Vtl::NumAlg::Spf::bess::I<2>(1.0f);
			v=Vtl::NumAlg::Spf::bess::I<2>(4.0f);
			v=Vtl::NumAlg::Spf::bess::I<7>(1.0f);
			v=Vtl::NumAlg::Spf::bess::I<7>(4.0f);
			// test for (x< 2.0 ) and (x>2.0)
			v=Vtl::NumAlg::Spf::bess::K<0>(1.0f);
			v=Vtl::NumAlg::Spf::bess::K<0>(3.0f);
			v=Vtl::NumAlg::Spf::bess::K<1>(1.0f);
			v=Vtl::NumAlg::Spf::bess::K<1>(3.0f);
			v=Vtl::NumAlg::Spf::bess::K<2>(1.0f);
			v=Vtl::NumAlg::Spf::bess::K<2>(3.0f);
			v=Vtl::NumAlg::Spf::bess::K<7>(1.0f);
			v=Vtl::NumAlg::Spf::bess::K<7>(3.0f);
		}
		return 0;
	}
}