/*! 
	@page dox_gen_worh_with_PLF How to ...   


         :\n
@li               (   ,            , x/width, y/height)\
@li       ,   :

@dot

digraph test123 {
   node [fontname=Helvetica, fontsize=12];
   edge [style="solid", color="black" ,arrowhead="open"];
   node0 [label="src \n(0)", shape=oval, color=black, fillcolor="#CBFCB4", fontcolor=black, style=filled];
   node1 [label="H-conv-d \n(1)", shape=record, color=black, fillcolor="#FCDDB4", fontcolor=black, style=filled];
   node2 [label="V-conv \n(2)", shape=record, color=black, fillcolor="#FCDDB4", fontcolor=black, style=filled];
   node3 [label="H-conv \n(3)", shape=record, color=black, fillcolor="#FCDDB4", fontcolor=black, style=filled];
   node4 [label="V-conv-d \n(4)", shape=record, color=black, fillcolor="#FCDDB4", fontcolor=black, style=filled];
   node5 [label="SumSqr \n(5)", shape=record, color=black, fillcolor="#FCDDB4", fontcolor=black, style=filled];
   node6 [label="gamma_mult \n(6)", shape=record, color=black, fillcolor="#FCDDB4", fontcolor=black, style=filled];
   node7 [label="dst \n(7)", shape=oval, color=black, fillcolor="#FCB4B4", fontcolor=black, style=filled];
   node0-> node3 [label="a" taillabel="0" headlabel="0"];
   node0-> node1 [label="a" taillabel="0" headlabel="0"];
   node1-> node2 [label="a" taillabel="0" headlabel="0"];
   node2-> node5 [label="a" taillabel="0" headlabel="0"];
   node3-> node4 [label="a" taillabel="0" headlabel="0"];
   node4-> node5 [label="a" taillabel="0" headlabel="1"];
   node5-> node6 [label="a" taillabel="0" headlabel="0"];
   node6-> node7 [label="a" taillabel="0" headlabel="0"];
}
@enddot

 :

  Plf::FiltersSystem;

     (plf_io_filter.h),     (plf_filters.h).
   Plf::FiltersSystem::Assign(...)     .     
    .      (  ) 
    getImageSize()            
  InPort<T>::setDstImage()   -  .
     Plf::FiltersSystem::Run().
    (/)    .
,   , -     :

@code

#include "plf_sys.h"
#include "plf_filters.h"
#include "plf_io_filter.h"

//    :       , 
//     ,    *.pgm

int test_plf_Filter(int argc, char* argv[])
{
//------ load_images ---------------------
	std::auto_ptr<AImage> _in(::AImageLoad(0,true,argv[1],indicator));
	AImageMem8 *in=dynamic_cast<AImageMem8 *>(_in.get());
	if(in==0)
		return 1;
	const int bits=in->Bits();
//------ create sources ------------------
	Plf::PlfSrc_AImageMem8<float> src("src",in);
	Plf::PlfDst_AImageMem8<float> dst("dst");
// Classic Canny
//------ create filters ------------------
	float sigma=1.0f;
	const Plf::ConvKernelGauss<float> gauss(sigma,2.5f); 
	const Plf::ConvKernelGauss1Deriv<float> gauss_deriv(sigma,2.5f); 
	Plf::ConvolveHorz<float> h_conv("H-conv",&gauss);
	Plf::ConvolveVert<float> v_conv("V-conv",&gauss);
	Plf::ConvolveHorz<float> h_conv_d("H-conv-d",&gauss_deriv);
	Plf::ConvolveVert<float> v_conv_d("V-conv-d",&gauss_deriv);
	const Plf::F::sqrtOfSumSqr<float> ssqr;
	Plf::FilterPixByPix2to1<float,Plf::F::sqrtOfSumSqr<float> > sum_sqr("SumSqr",ssqr);

	Plf::F::GammaCorMult<float> gam_mult(float((1<<bits)-1),2.2,sqrt(2.0f*3.1415926f)*sigma);
	Plf::FilterLevels<float,Plf::F::GammaCorMult<float> > gamma_m("gamma_mult",gam_mult);
//------ create graph --------------------
	src.out.ConnectTo(&h_conv.in);
	src.out.ConnectTo(&h_conv_d.in);
	h_conv.out.ConnectTo(&v_conv_d.in);
	h_conv_d.out.ConnectTo(&v_conv.in);
	v_conv.out.ConnectTo(&sum_sqr.in_1);
	v_conv_d.out.ConnectTo(&sum_sqr.in_2);
	sum_sqr.out.ConnectTo(&gamma_m.in);
	gamma_m.out.ConnectTo(&dst.in);

//--- Create system object
	Plf::FiltersSystem sys;
//------ Assing task------------------------
	sys.Assign(&src,&dst);

//------ Initialize destinations -----------
	Plf::ImgSize is=dst.in.getImageSize();
	std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::YImageClass::IMGC_GRAYSCALE
									,is.width,is.height,in->Bits()));
	AImageMem8 *out=dynamic_cast<AImageMem8 *>(_out.get());
	if(out==0)
		return 1;
	dst.setDstImage(out);

//----- Run filter
	sys.Run();
// Save data to file
	::AImageSave(out,argv[2],indicator);

	return 0;
}
@endcode

     ,     Plf::FilterElementary
  ,   Plf::FilterElementary::processLine() ,     
      -    
Plf::FilterElementary::set_OutPortsImagesSizeAndInPortsExtents(),  
                
 .     ,    (Plf::Extent) 
  \c ,      .

        (Inport<T>, OutPort<T>)        
  Plf::FilterElementary::AddPort().

 ,     /       
.      Plf::FilterComplex.  .   plf_filter_complex.h  



   .  :

\example test_plf_strip_buffer.cpp 
\example test_plf_io_ports.cpp 
\example test_plf_kernels.cpp 
\example test_plf_filters.cpp 
\example test_plf_filters_rgb.cpp 

*/




