#include <string>
#include <sstream>
//#include <stdio.h>
#include <assert.h>

#include "apipe.h"
#ifndef min
	template<class T> T min(T& a,T& b) {return (a<b)?a:b;}
#endif

size_t childPipe::envCalcBinBlockSize(const childPipe::Env& env)
{
		size_t len=0;
		childPipe::Env::const_iterator i;
		for(i=env.begin(); i!=env.end(); i++)
		{
			len+=(*i).first.size()+1;//append '='
			len+=(*i).second.size()+1;//append string terminated null

		}
		len+=1;//terminating block null
		return len;
}
void childPipe::envBuildBinBlock(const childPipe::Env& env,char *e,char delim,char delim_line)
{
	if(e)
	{
		childPipe::Env::const_iterator i;
		for(i=env.begin(); i!=env.end(); i++)
		{
			size_t len=(*i).first.size();
			size_t len2=(*i).second.size();
			if(len!=0 && len2!=0)
			{//do not transefe empty variables
				memcpy(e,(*i).first.c_str(),len);
				e+=len;
				*e++=delim;
				memcpy(e,(*i).second.c_str(),len2);
				e+=len2;
				*e++=delim_line;
			}
		}
		*e=char(0);
	}
}

void childPipe::envBuildFromBinBlock(childPipe::Env& env,const char *e)
{
		std::string var;
		std::string value;
		while(*e)
		{
			var=e;
			e+=var.size()+1;
			if(*e==0)
				break;
			value=e;
			e+=value.size()+1;
			env[var]=value;
		}
}

void childPipe::envFreeBinBlock(char* env_blk)
{
	if(env_blk)
		delete[] env_blk;
}
char * childPipe::envMakeBinBlock(childPipe::Env* env)
{
	if(!env || env->size()==0)
		return 0;
	else
	{
		size_t len=childPipe::envCalcBinBlockSize(*env);
		char * blk=new char[len];
		childPipe::envBuildBinBlock(*env,blk,0);
		return blk;
	}
}

void* childPipe::BuildEnvBlock(std::vector<char>& e,const childPipe::Env* env)
{
	if(!env || env->size()==0)
		return 0;
	else
	{
		size_t len=childPipe::envCalcBinBlockSize(*env);
		e.resize(len);
		childPipe::envBuildBinBlock(*env,&(e[0]),'=');
		return &e[0];
	}
}

void childPipe::nullify()
{
    ZeroMemory(&pi, sizeof(PROCESS_INFORMATION));
}

void childPipe::Init(std::string command,const Env *env,Flags flags )
{
	//F_Split         F_StdOut  F_StdErr!
	assert(!Is(flags,F_Split)||(Is(flags,F_Split)&&Is(flags,F_StdOut)&&Is(flags,F_StdErr)));
	the_same_stderr_stdout=!Is(flags,F_Split)&&Is(flags,F_StdOutErr);
    // Set up members of the PROCESS_INFORMATION structure.
    ZeroMemory(&pi, sizeof(PROCESS_INFORMATION));
    SECURITY_ATTRIBUTES sa; 
    sa.nLength = sizeof(SECURITY_ATTRIBUTES); 
    sa.bInheritHandle = TRUE; 
    sa.lpSecurityDescriptor = NULL; 

    // Create a pipe for the child process's STDOUT.
	HANDLE hStdout=INVALID_HANDLE_VALUE;
	HANDLE hChildStdoutWr=INVALID_HANDLE_VALUE;
	if(Is(flags,F_StdOut))
	{
		if( !CreatePipe(&hStdout, &hChildStdoutWr, &sa, 0) )
			throw std::string("Error ::CreatePipe()");
	}
	else
		hChildStdoutWr=CreateFile("NUL",GENERIC_WRITE,FILE_SHARE_WRITE,&sa,OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,NULL);
	stdout_.open(hStdout);

	HANDLE hStderr=INVALID_HANDLE_VALUE;
	HANDLE hChildStderrWr=INVALID_HANDLE_VALUE;
	if(Is(flags,F_StdErr) && !the_same_stderr_stdout)
	{
		if( !CreatePipe(&hStderr, &hChildStderrWr, &sa, 0) )
			throw std::string("Error ::CreatePipe()");
	}
	else
		hChildStderrWr=CreateFile("NUL",GENERIC_WRITE,FILE_SHARE_WRITE,&sa,OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,NULL);
	stderr_.open(hStderr);
    // Create a pipe for the child process's STDIN.
    HANDLE hChildStdinRd=INVALID_HANDLE_VALUE;
	HANDLE hStdin=INVALID_HANDLE_VALUE;
	if(Is(flags,F_StdIn))
	{
		if( !CreatePipe(&hChildStdinRd, &hStdin, &sa, 0) )
			throw std::string("Error ::CreatePipe()");
	}
	stdin_.open(hStdin);

    // Now create the child process.
    // Set up members of the STARTUPINFO structure.
    STARTUPINFO si;
     ZeroMemory(&si, sizeof(STARTUPINFO));
    si.cb = sizeof(STARTUPINFO);
    si.dwFlags = STARTF_USESTDHANDLES;
    si.hStdInput = hChildStdinRd;
    si.hStdOutput =hChildStdoutWr; 
	si.hStdError = the_same_stderr_stdout ? hChildStdoutWr : hChildStderrWr;

    // Create the child process.
    //  ,    -  environment,
    // ,   ...

	std::vector<char> env_block;
	LPSTR cmd=LPSTR(command.c_str());
	char *e=(char*)BuildEnvBlock(env_block,env);
	if( ! CreateProcess(NULL,
      cmd,       // command line. , const! Windows NT/2000/XP: The Unicode version of this function, CreateProcessW, will fail if this parameter is a const string. 
      NULL,          // process security attributes
      NULL,          // primary thread security attributes
      TRUE,          // handles are inherited
      CREATE_NO_WINDOW|HIGH_PRIORITY_CLASS|CREATE_NEW_PROCESS_GROUP,//0,             // creation flags
      e,          // use parent's environment
      NULL,          // use parent's current directory
      &si,           // STARTUPINFO pointer
      &pi)           // receives PROCESS_INFORMATION
      )
        throw "Create process failed";
    // close child-side handles.
    //     ,    EOF.
	if(hChildStdoutWr!=INVALID_HANDLE_VALUE)
		if( !CloseHandle(hChildStdoutWr) ) 
			throw std::string("Error ::CloseHandle()");
	if(hChildStderrWr!=INVALID_HANDLE_VALUE)
	    if( !CloseHandle(hChildStderrWr) ) 
			throw std::string("Error ::CloseHandle()");
	if(hChildStdinRd!=INVALID_HANDLE_VALUE)
	    if( !CloseHandle(hChildStdinRd) ) 
			throw std::string("Error ::CloseHandle()");
    if( !CloseHandle( pi.hThread ) ) 
			throw std::string("Error ::CloseHandle()");
    exitStatus = STILL_ACTIVE;              //   (!) .
}

void childPipe::Close()
{
    //       ,  :
    // -   
    // -   
    // -    ""
    //  ,  ,   ,
    //       ...
	stdin_.close();
    if( pi.hProcess != 0 )
        CloseHandle(pi.hProcess);
	stdout_.close();
	stderr_.close();
}



// :    .
//          STDIN/STDOUT.
//    ,  ,   .
bool childPipe::isRunning()
{
    if( exitStatus != STILL_ACTIVE )
        return false;
    assert(pi.hProcess != 0);
    if( !GetExitCodeProcess(pi.hProcess, &exitStatus) ) 
	{
        //      hProcess   ,
        //       .
        throw std::string("Error ::GetExitCodeProcess()");
    }
    if( exitStatus == STILL_ACTIVE )
        return true;

    //       ,    
    //   hProcess --  zombie
    if( !CloseHandle(pi.hProcess) )
        throw std::string("Error ::CloseHandle()");
    pi.hProcess = 0;

    //   ,   ,     .
	stdin_.close();
    //   hStdOut    ,    .
    return false;
}

void childPipe::wait()
{
    assert(pi.hProcess != 0);
    //  
    WaitForSingleObject(pi.hProcess, INFINITE);
    //  ,  .
    isRunning();
}
void childPipe::terminate()
{
    if( exitStatus != STILL_ACTIVE )
        return;
    if( pi.hProcess == 0)
		return;
	if(childPipe::isRunning())
	{
		if(TerminateProcess(pi.hProcess, // handle to the process
			UINT(-1)   // exit code for the process
			))
			return;
		else
		{
			DWORD err=GetLastError();
			//char buf[80];
			//sprintf(buf," error=%u\n",err);
			//::MessageBox(0,buf,"childPipe::terminate()",MB_OK);
			std::ostringstream os;
			os << " error=" << err << std::endl;
			::MessageBox(0,os.str().c_str(),"childPipe::terminate()",MB_OK);
		}
	}
}


unsigned ReadPipe::readNoWait(char *data,unsigned len)
{
	Sleep(1);
	if(handle_==INVALID_HANDLE_VALUE)
		return 0;
	DWORD bytes_avail=0;
	if(!PeekNamedPipe(handle_,0,0,0,&bytes_avail,0))
	{
		DWORD err=GetLastError();
		if(ERROR_BROKEN_PIPE==err)
		{
			if( !CloseHandle(handle_) ) 
				throw std::string("Error ::CloseHandle()");
			handle_ = INVALID_HANDLE_VALUE;
		}
		else
			throw "Error read from Pipe";
		return 0;
	}
	else if(bytes_avail==0)
		return 0;
//	return readWait(data,std::min(len,bytes_avail));
	return readWait(data,min(len,unsigned(bytes_avail)));
//	return readWait(data,std::min(len,unsigned(bytes_avail)));
}
unsigned ReadPipe::readWait(char *data,unsigned len)
{
	DWORD dwRead=0;
	if( !ReadFile( handle_, data, len, &dwRead,NULL) )
	{
		DWORD err=GetLastError();
		if(ERROR_BROKEN_PIPE==err)
		{
			if( !CloseHandle(handle_) ) 
				throw std::string("Error ::CloseHandle()");
			handle_ = INVALID_HANDLE_VALUE;
		}
		else
			throw "Error read from Pipe";
	}
	return dwRead;
}
void ReadPipe::close()
{
	if( !isEof() )
		if( !CloseHandle(handle_) ) 
			throw std::string("Error ::CloseHandle()");
	handle_ = INVALID_HANDLE_VALUE;
}

void WritePipe::close()
{
	if( !isEof() )
		if( !CloseHandle(handle_) ) 
			throw std::string("Error ::CloseHandle()");
	handle_ = INVALID_HANDLE_VALUE;
}
unsigned WritePipe::writeNoWait(const char *data,unsigned len)
{
	DWORD dwWritten=0;
    OVERLAPPED os;
    ZeroMemory(&os, sizeof(OVERLAPPED));
    if(!WriteFileEx(handle_ , data, len, &os, NULL))
	{
		DWORD err=GetLastError();
		if(ERROR_BROKEN_PIPE==err)
		{
			if( !CloseHandle(handle_) ) 
				throw std::string("Error ::CloseHandle()");
			handle_ = INVALID_HANDLE_VALUE;
		}
		else
			throw "Error write to Pipe";
	}
	else
		dwWritten=os.Offset;
	return dwWritten;
}
unsigned WritePipe::writeWait(const char *data,unsigned len)
{
	DWORD dwWritten=0;
    if(!WriteFile(handle_ , data, len, &dwWritten, NULL))
	{
		DWORD err=GetLastError();
		if(ERROR_BROKEN_PIPE==err)
		{
			if( !CloseHandle(handle_) ) 
				throw std::string("Error ::CloseHandle()");
			handle_ = INVALID_HANDLE_VALUE;
		}
		else
			throw "Error write to Pipe";
	}
	return dwWritten;
}

ReadString::ReadString(ReadPipe& rdpipe,unsigned max_string_len) 
	: rdpipe_(rdpipe),max_len(max_string_len) 
{
	buffer.resize(2*max_len+1);
	last_pos2=first_pos2=max_len+1;
	pos=0;
	null_detected=false;
	eof_detected=false;
	last_char='a';//no next line character
}

char* ReadString::gets(char *buf)
{
	if(IsAvail())
	{
//		strcpy(buf,&buffer[0]);
		::memcpy(buf,&buffer[0],buffer.size());
		null_detected=false;
		pos=0;
		return buf;
	}
	return 0;
}
bool ReadString::IsAvail()
{
	if(null_detected)
		return true;
	for(;;)
	{
		copy();
		if(null_detected)
			break;
		if(eof_detected)
			break;
		if(first_pos2==last_pos2)
		{
			if(rdpipe_.isEof())
			{
				if(pos)
				{
					buffer[pos]=0;
					null_detected=true;
				}
				eof_detected=true;
				rdpipe_.close();
				break;
			}
			else
			{
				first_pos2=last_pos2=max_len+1;
				last_pos2+=rdpipe_.readNoWait(&buffer[first_pos2],max_len);
			}
		}
	}
	return null_detected;
}
void ReadString::copy()
{
	while(first_pos2<last_pos2)
	{
		char c=buffer[first_pos2++];
		if(c==0 || c=='\n' || c=='\r' )
		{
			if(last_char==0)
				continue;
			else
			{
				last_char=0;
				buffer[pos++]=0;
				null_detected=true;
				break;
			}
		}
		else
			buffer[pos++]=c;
		last_char=c;
		if(pos==max_len)
		{
			buffer[pos]=0;
			null_detected=true;
			last_char=0;
			break;
		}
	}
}
