#ifndef _XML_INTERFACE_H
#define _XML_INTERFACE_H
//////////////////////////////////////////////////////////////////////////////////
//file #include "xml_interface.h"
//
//////////////////////////////////////////////////////////////////////////////////
#include <vector>
#include <string>
#include <functional> //less<>,equal_to<>, etc.
#include "datastring.h"
#include "xml_interface_imp.h"

/*! @brief   XML.
 *  !   ()    DLL,    
 *   PATH.      !!!     !
 *
 *         DLL   IBM-Xercesc :\n
 *  - icudt24l.dll 
 *  - icuuc24.dll 
 *  - xml4c_5_1_0.dll 
 *
 *  DLL \n
 *  - yxml_drv.dll 
 *
 *   XercesMessages_en_US.res ,         
 *  XERCESC_NLS_HOME,  \n
 * set XERCESC_NLS_HOME=d:\\dll \n
 *   Windows-XP    My Computer -> Advanced -> Environment Variables.
 *
 *     : \n
 * - msvcrt.dll
 * - msvcirt.dll\n
 *   Ms Visual Studio 6.0
 * 
 *        general/xml_interface_imp.cpp ,  
 *   yxml_drv.dll ,        !  
 * dll   Ms Visual Studio 6.0,    IBM-Xercesc 
 *    .         
 *  .
 *
 *  IBM-Xercesc  -  .  xml.zip .    
 *    .  Xercesc   ,  IBM  
 *     . 
 *  yxml_drv.dll  -  ,      
 *  IBM-Xercesc,      IBM-Xercesc    
 *        .   
 * XML -        windows-1251.
 *
 *************************************************************************************/
namespace YXML
{
/*! @addtogroup XMLandINI_DoxGroup   XML  INI    
 *  @{
 */
	class XMLdoc;
	class DNode;
	//////////////////////////////////////////////////////////////////////////////////
  /**
   * @class Xercesc
	 *   .     
	 *     XERCESC/ICU (xml4c)
	 */
	class Xercesc
	{
		static int reference_counter;
	public:
		Xercesc();
		virtual ~Xercesc();
	};
	//////////////////////////////////////////////////////////////////////////////////
	class XMLdoc
		:private Xercesc
	{
	private:
		YXML_DRV_DocHandle handle;
		mutable std::string filename;
		bool _autosave;
	public:
		XMLdoc();
		XMLdoc(const char *fname,const char* docname,bool autosave);
		~XMLdoc();
		static bool IsFileExist(const char *fname);
		std::string FName() const { return filename;}
		void Create(const char *fname,const char *docname);
		void Load(const char *fname,bool condenseWS=true);
		void Save(const char *fname,bool formatted=false) const;
		DNode Root();
	};
	////////////////////////////////////////////////////////////////////////////////
	/**
	 * @class DNode
	 *        .
	 *     ,   
	 *   .
	 *
	 */
	class DNode
	{
	private:
		YXML_DRV_NodeHandle node;
		class IsOR;
		class IsAND;
	public:
		enum Type
		{
			ELEMENT       =YXML_DRV_NODETYPE_ELEMENT,
			ATTRIBUTE     =YXML_DRV_NODETYPE_ATTRIBUTE,
			TEXT          =YXML_DRV_NODETYPE_TEXT,
			COMMENT       =YXML_DRV_NODETYPE_COMMENT,
			UNSUPPORTED   =YXML_DRV_NODETYPE_UNSUPPORTED
		};
		class AttrRef;
    /**
     * @class Is 
		 *     ,   
		 *   .         ,
		 *     @see DNode::IsCond,DNode::IsOR,DNode::IsAND
		 *     . 
		 *             
		 *     
		 * @see IsElem,IsName,IsNameEq,IsAttr,IsAttrIn
     */
		struct Is
		{
			virtual ~Is() {}
			virtual bool operator()(const DNode n) const=0;
			virtual Is* clone() const =0; //Must return new this object
			
			IsOR operator||(const Is& rhs) const;
			IsAND operator&&(const Is& rhs) const;
		};
		class IsCond
			: public Is
		{
			Is* cond;
			void clear() {if(cond) {delete cond;cond=0;}  }
		public:
			IsCond() : cond(0) {}
			IsCond(const Is& is) : cond(is.clone()) {}
			virtual ~IsCond() { clear(); }
			IsCond& operator=(const IsCond& is) { if(cond) clear(); cond=is.clone(); return *this;}
			virtual bool operator()(const DNode n) const { return (cond) ? (*cond)(n) : true;}
			virtual Is* clone() const { return new IsCond(*this); }
		};
	public:
		explicit DNode(YXML_DRV_NodeHandle h) : node(h) {}
		DNode(const DNode& n) : node(n.node) {}
		DNode& operator=(const DNode& n) { node=n.node; return *this;}
		~DNode() {}
		YXML_DRV_NodeHandle _GetYXML_NodeHandle() const {return node;}//>For special use ONLY
/** @defgroup SpecificXML  XML specific interface.
  *  @ingroup XMLandINI_DoxGroup
  *  ,  XML,    INI .
  *  @{
  */
		Type getType() const;    ///<              
		std::string getName() const;          ///<  ,      ELEMENT
		bool IsNode() const { return node!=0;} ///<     NULL.
		//Work only at ELEMENT node
    /** @defgroup InsertRemove        XML .
     *         XML .
     *  @{
     */
		DNode AppendChild(Type type,const DataString data);
		DNode InsertAfter(Type type,const DataString data);
		DNode InsertBefore(Type type,const DataString data);
		void Delete();
		void setValue(const DataString data);///<Work only at TEXT and COMMENT nodes
		DataString getValue() const; ///<Work only at TEXT and COMMENT nodes
	/** @} */ // end of InsertRemove
	/** @defgroup SearchTree XML tree search functions
     *     XML .
     *  @{
     */
		DNode Parent() const;
		DNode FirstChild() const;
		DNode LastChild() const;
		DNode Next() const;
		DNode Prev() const;
  //@{
		/**   ()  ,  ,   
		  * @param cond - ,   DNode::Is, 
			* @return   ,       . see IsNode
			*   @see DNode::IsCond,IsElem,IsName,IsNameEq,IsAttr,IsAttrIn,DNode::IsOR,DNode::IsAND
			*
			*    ,     DNode::Is.
			*       , 
			* @see FirstChild(std::string name) const
			* @see Next(std::string name) const
			* @see LastChild(std::string name) const
			* @see Prev(std::string name) const
			* @see FirstChild(std::string ,std::string attr,const T& ) const
			* @see Next(std::string ,std::string attr,const T& ) const
			* @see LastChild(std::string ,std::string attr,const T& ) const
			* @see Prev(std::string ,std::string attr,const T& ) const
		  */
		DNode FirstChild(const Is& cond) const;
		DNode Next(const Is& cond) const;
		DNode LastChild(const Is& cond) const;
		DNode Prev(const Is& cond) const;
   //@}
		DNode FirstChild(std::string name) const;
		DNode Next(std::string name) const;
		DNode LastChild(std::string name) const;
		DNode Prev(std::string name) const;
		template<typename T>
		DNode FirstChild(std::string name,std::string attr,const T& attr_val) const
		{ return FirstChild(IsNameEq(name)&&IsAttrEq(attr,attr_val));}
		template<typename T>
		DNode Next(std::string name,std::string attr,const T& attr_val) const
		{ return Next(IsNameEq(name)&&IsAttrEq(attr,attr_val));}
		template<typename T>
		DNode LastChild(std::string name,std::string attr,const T& attr_val) const
		{ return LastChild(IsNameEq(name)&&IsAttrEq(attr,attr_val));}
		template<typename T>
		DNode Prev(std::string name,std::string attr,const T& attr_val) const
		{ return Prev(IsNameEq(name)&&IsAttrEq(attr,attr_val));}
	/** @} */ // end of SearchTree
/** @} */ // end of SpecificXML
/** @addtogroup Common_XML_INI_DoxGroup INI and XML common interface
 *  @ingroup XMLandINI_DoxGroup
 *   /       -  ,
 *   -    . \n
 *  INI -       /,\n
 *  XML -    GetNode()        -  ,\n
 *   -    ,        -  
 *     ,     "INI and XML common interface" !!! 
 *     XML,      
 *    XML.
 *  @{
 */
	/** @defgroup BaseFunctionsXML_DoxGroup   XML
	  *      
		*   @see XMLAlternative_XML_INI_DoxGroup,    
		*   
	  * @{ 
		*/
		DNode GetNode(const std::string node_name); ///<     ,   ,  ,   ,    .
		void SetValue(const DataString data); ///<      ,  ELEMENT -       ,  TEXT  COMMENT -     .
		DataString GetValue() const;
		void SetAttr(const std::string attr_name,const DataString attr_value);///<     XML  INI
		DataString GetAttr(const std::string attr_name) const;///<     XML  INI
	/**@} */ 
	/** @defgroup XMLAlternative_XML_INI_DoxGroup XML: Alternative INI and XML common interface
		*  INI and XML common interface,    .
	  * @{ 
		*/
		DNode  operator()(const std::string node_name) {return GetNode(node_name);}
		AttrRef operator[](const std::string& attr);///<   ,      DataString      DataString 
		const AttrRef operator[](const std::string& attr) const;
		DNode& operator=(const DataString data) { SetValue(data); return *this; }
		operator DataString() const { return GetValue();}
	/**@} */ 
/** @} */ // end of Common_XML_INI
		friend std::ostream& operator<<(std::ostream& s,DNode& in) { return s<<in.operator DataString();}
	};
	//------------------Implementation, used indirectly via WiNode operator[] ------
	class DNode::AttrRef 
	{
		friend class DNode; 
		DNode n;
		std::string attrib;
	private:
		AttrRef();
		AttrRef(const AttrRef&);
		AttrRef operator=(const AttrRef&);
		AttrRef(const DNode& node,const std::string& attr) 
			:n(node),attrib(attr) {}
	public:
		operator DataString() const { return n.GetAttr(attrib.c_str());}
		void operator=(const DataString& s) { n.SetAttr(attrib.c_str(),s.c_str());} 
		friend std::ostream& operator<<(std::ostream& s,DNode::AttrRef& ir) { return s<<ir.operator DataString();}
	};
	//---------------------------------------------------------------------------------
	inline DNode::AttrRef DNode::operator[](const std::string& attr) { return DNode::AttrRef(*this,attr); }
	inline const DNode::AttrRef DNode::operator[](const std::string& attr) const { return DNode::AttrRef(*this,attr); }
	//------------- implementation of private classes used for logical search 
	/** 
	 * 	@class DNode::IsOR
	 *  ,  DNode::Is       
	 *     . 
	 * @see FirstChild(const Is& cond)  const
	 * @see Next(const Is& cond)  const
   * 
	 *   @see DNode::IsCond,IsElem,IsName,IsNameEq,IsAttr,IsAttrIn,DNode::IsOR,DNode::IsAND
   */
		class DNode::IsOR
			: public Is
		{
			Is* l;
			Is* r;
			void clear() {if(l) {delete l;l=0;} if(r) {delete r;r=0;} }
		public:
			IsOR(Is* lhs,Is* rhs) : l(lhs), r(rhs) {}
			IsOR(const IsOR& obj) : l(obj.l->clone()), r(obj.r->clone()) {}
			virtual ~IsOR() { clear(); }
			IsOR& operator=(const IsOR& obj) { clear(); l=obj.l->clone(); r=obj.r->clone(); return *this;}
			virtual bool operator()(const DNode n) const { return (*l)(n) || (*r)(n); }
			virtual Is* clone() const { return new IsOR(*this); }
		};
	//---------------------------------------------------------------------------------
	/** 
	 * 	@class DNode::IsAND
	 *  ,  DNode::Is       
	 *     . 
	 * @see FirstChild(const Is& cond)  const
	 * @see Next(const Is& cond)  const
   * 
	 *   @see DNode::IsCond,IsElem,IsName,IsNameEq,IsAttr,IsAttrIn,DNode::IsOR,DNode::IsAND
   */
		class DNode::IsAND
			: public Is
		{
			Is* l;
			Is* r;
			void clear() {if(l) {delete l;l=0;} if(r) {delete r;r=0;} }
		public:
			IsAND(Is* lhs,Is* rhs) : l(lhs), r(rhs) {}
			IsAND(const IsAND& obj) : l(obj.l->clone()), r(obj.r->clone()) {}
			virtual ~IsAND() { clear(); }
			IsAND& operator=(const IsAND& obj) { clear(); l=obj.l->clone(); r=obj.r->clone(); return *this;}
			virtual bool operator()(const DNode n) const { return (*l)(n) && (*r)(n); }
			virtual Is* clone() const { return new IsAND(*this); }
		};
	//---------------------------------------------------------------------------------
	inline DNode::IsOR DNode::Is::operator||(const Is& rhs) const { return IsOR(this->clone(),rhs.clone()); }
	inline DNode::IsAND DNode::Is::operator&&(const Is& rhs) const { return IsAND(this->clone(),rhs.clone()); }
	//----------classes for logical search tags in XML --------------------------------
	/** 
	 * 	@struct IsElem
	 *   ,        . 
	 * @see FirstChild(const Is& cond)  const
	 * @see Next(const Is& cond)  const
   * 
	 *   ,  ELEMENT -  true
   * 
	 *   @see DNode::IsCond,IsElem,IsName,IsNameEq,IsAttr,IsAttrIn,DNode::IsOR,DNode::IsAND
   */
		struct IsElem
			: public DNode::Is
		{
			virtual bool operator()(const DNode n) const { return (DNode::ELEMENT==n.getType());}
		private:
			virtual DNode::Is* clone() const { return new IsElem(*this); }
		};
	//---------------------------------------------------------------------------------
	/** 
	 * 	@struct IsName
	 *   ,        . 
	 * @see FirstChild(const Is& cond)  const
	 * @see Next(const Is& cond)  const
   * 
	 *     ELEMENT,      
	 * Cmp -  true
   *  
   * @param Cmp -     ,   STL,
   *    "<"functional">",   less<>,equal_to<>,  ..,  
	 *    .        
	 *      IsNameEq.   
	 * typedef ,             
	 *   Borland C++ Builder 6.0. Microsoft Visual C++ 7.0  Intel C++ 7.0  
	 *   .         
	 *     .
   * 
	 *   @see DNode::IsCond,IsElem,IsName,IsNameEq,IsAttr,IsAttrIn,DNode::IsOR,DNode::IsAND
   */
		template<class Cmp/*=std::equal_to<std::string>*/ >
		class IsName
			: public IsElem
		{
			std::string _name;
		public:
			IsName(std::string name) : _name(name) {}
			virtual bool operator()(const DNode n) const { return IsElem::operator()(n) && Cmp()(n.getName(),_name);}
			virtual DNode::Is* clone() const { return new IsName(*this); }
		};
	//---------------------------------------------------------------------------------
		typedef IsName<std::equal_to<std::string> > IsNameEq;///<,     ELEMENT?     .
	//---------------------------------------------------------------------------------
	/** 
	 * 	@struct IsAttr
	 *   ,        . 
	 * @see FirstChild(const Is& cond)  const
	 * @see Next(const Is& cond)  const
	 *     @param at @see IsAttr::IsAttr(),     @param T 
	 *    DataString      @param value c  
	 *   @param Cmp,    .    @param Cmp
	 *         STL.    @see IsAttrEq
	 */
	template<typename T,class Cmp> 
	class IsAttr
		: public DNode::Is
	{
		std::string a;
		T v;
	public:
		IsAttr(std::string at,T value) : a(at), v(value) {}
		virtual bool operator()(const DNode n) const 
		{ DataString ds=n.GetAttr(a); return (ds.empty())? false : Cmp()(ds,v);	}
		virtual DNode::Is* clone() const { return new IsAttr(*this); }
	};
	//---------------------------------------------------------------------------------
	typedef IsAttr<std::string,std::equal_to<std::string> > IsAttrEq;///< ,      .
	//---------------------------------------------------------------------------------
	/** 
	 * 	@struct IsAttrIn
	 *   ,        . 
	 * ,    @param at,    @param T 
	 *     @param value1  @param value2   . 
	 *      ,   , 
	 *  /,  ,     ..
	 */
	template<typename T> 
	class IsAttrIn
		: public DNode::Is
	{
		std::string a;
		T v1,v2;
	public:
		IsAttrIn(std::string at,T value1,T value2) : a(at), v1(value1), v2(value2) {}
		virtual bool operator()(const DNode n) const 
		{ DataString ds=n.GetAttr(a); return (ds.empty())? false : ( std::greater_equal<T>()(ds,v1) && std::less_equal<T>()(ds,v2) );	}
		virtual DNode::Is* clone() const { return new IsAttrIn(*this); }
	};

//Inline functions implementation
	inline DNode DNode::FirstChild(std::string name) const { return FirstChild(IsNameEq(name));}
	inline DNode DNode::Next(std::string name) const       { return Next(IsNameEq(name));}
	inline DNode DNode::LastChild(std::string name) const  { return LastChild(IsNameEq(name));}
	inline DNode DNode::Prev(std::string name) const       { return Prev(IsNameEq(name));}
/** @} */ // end of XMLandINI_DoxGroup
}
//////////////////////////////////////////////////////////////////////////////////
#endif //_XML_INTERFACE_H
