#include "yfilesys.h"

#ifdef WIN32
	#define YFILESYS_WINDOWS
	#include <io.h>
	#include "apipe.h"
#endif

#ifdef YFILESYS_WINDOWS
	#include <windows.h>
#else
	#include <sys\stat.h> //S_IWRITE, S_IREAD
	#include <sys\types.h> //struct stat
	#include <time.h>
#endif

//#include <stdio.h>//sprintf()

//////////////////////////////////////////////////////////////////////////////
// General, Windows/UNIX compatible
//////////////////////////////////////////////////////////////////////////////
std::string YFileSys::getEnv(const char* var_name)
{
#ifdef YFILESYS_WINDOWS
  
	DWORD sz=GetEnvironmentVariable(var_name,0,0);
	if(sz==0)
	{
		return std::string();
		//if(GetLastError()==ERROR_ENVVAR_NOT_FOUND)
	}
	std::vector<char> v(sz);
	char *s=&(v[0]);
	if(sz!=GetEnvironmentVariable(var_name,s,sz))
		throw std::string("Error in YFileSys::getEnv()");
	return std::string(s);
#else
		const char *e=getenv(var_name);
		return (e!=0)? std::string(e) : std::string();
#endif
}

bool YFileSys::isFullName( std::string path)
{//>Returns true if path contains absolute path
	if(path.size()>2)
		if(path[1]==':')
			return true;
	return false;
}
static std::string _dir_separator("/\\");

static bool _CreateDir(std::string path)
{
	size_t i=path.find_last_of(_dir_separator);
	if(i>=path.length())
	{
		std::ostringstream os;
		os << "can't create directory " << path;
		throw os.str();
		return false;
	}
	std::string s1=path.substr(0,i);
	if(!YFileSys::isDirExist(s1))
		_CreateDir(s1);
	if(0==CreateDirectory(path.c_str(),0))
	{
		std::ostringstream os;
		os << "can't create directory " << path;
		throw os.str();
		return false;
	}
	return true;
//	return (0==CreateDirectoryEx(0,path.c_str(),0));
/*
	_mkdir	CreateDirectory
 _rmdir                    RemoveDirectory
   _searchenv                SearchPath
*/
}
bool YFileSys::CreateDir(std::string path)
{
	try
	{
		if(!YFileSys::isFullName(path))
			throw std::string("you must specify full path");
		if(YFileSys::isDirExist(path))
			return false;
		std::string s=path;
		if(path[path.size()-1]=='\\' || path[path.size()-1]=='/')
			s=path.substr(0,path.size()-1);
		return _CreateDir(s);
	}
	catch(std::string msg)
	{
		std::ostringstream os;
		os << "Can't create directory " << path << " because " << msg;
		throw os.str();
	};
	return false;
}


std::string YFileSys::getExt(std::string path)
{
	std::string ext;
	std::string::size_type i=path.find_last_of("./\\");
	if(i<path.length())
		if(path[i]=='.')
			ext.assign(path,i,path.size());
	return ext;
}

std::string YFileSys::remExt(std::string path)
{
	std::string ext;
	std::string::size_type i=path.find_last_of("./\\");
	if(i<path.length())
	{
		if(path[i]=='.')
		{
			ext.assign(path,0,i);
			return ext;
		}
		else
			return path;
	}
	return path;
}

std::string YFileSys::getDir(std::string path)
{
	std::string dir;
	std::string::size_type i=path.find_last_of("/\\");
	if(i<path.length())
		if(path[i]=='/'||path[i]=='\\')
			dir.assign(path,0,i+1);
	return dir;
}

std::string YFileSys::getName(std::string path)
{
	std::string::size_type i=path.find_last_of("/\\",0);
	if(i<path.length())
	{
		if(path[i]=='/'||path[i]=='\\')
		{
			std::string name;
			name.assign(path,i+1,path.size());
			return name;
		}
	}
	return path;
}
int YFileSys::StrCmpI(const std::string& s1,const std::string& s2)
{
	return lstrcmpi(s1.c_str(),s2.c_str());
}
std::string YFileSys::changeExt(std::string path,std::string ext)
{
	//MessageBox(0,path.c_str(),"",MB_OK);
	if(!YFileSys::getExt(path).empty())
		path=YFileSys::remExt(path);
	if(ext[0]!='.')
		path+=".";
	path+=ext;
	return path;
}
std::string YFileSys::changeExtIf(std::string path,std::string extToChange,std::string ext)
{
	std::string e=extToChange;
	std::string e2=YFileSys::getExt(e);
	std::string e3;
	while(!e2.empty())
	{
		e3=YFileSys::getExt(path);
		if(0!=YFileSys::StrCmpI(YFileSys::getExt(path),e2))
			break;
		path=YFileSys::remExt(path);
		e=YFileSys::remExt(e);
		e2=YFileSys::getExt(e);
	}
	if(ext[0]!='.')
		path+=".";
	path+=ext;
	return path;
}

bool YFileSys::isExists( std::string path)
{
	if(path.size()>1 && ( path[path.size()-1]=='\\' || path[path.size()-1]=='/'))
		path.resize(path.size()-1);
#ifdef YFILESYS_WINDOWS
      return ::GetFileAttributesA( path.c_str() ) != 0xFFFFFFFF;
#else
      struct stat buffer;
      return ::stat( path.c_str(), &buffer) == 0;  
#endif
}
 
bool YFileSys::isFileExist(std::string path)
{//>Check if file exists
#ifdef YFILESYS_WINDOWS
     DWORD fa=::GetFileAttributesA( path.c_str() );
	 if( fa!= 0xFFFFFFFF)
		 if((fa&FILE_ATTRIBUTE_DIRECTORY)==0)
			 return true;
	 return false;
#else
	struct stat buffer;
	if(0==::stat(path.c_str(),&buffer))
	{
		if((buffer.st_mode & S_IFDIR )==0)
			return true;
	}
	return false;
#endif
}
bool YFileSys::isDirExist(std::string path)
{//>Check if directory exists
	if(path.size()>1 && ( path[path.size()-1]=='\\' || path[path.size()-1]=='/'))
		path.resize(path.size()-1);
#ifdef YFILESYS_WINDOWS
     DWORD fa=::GetFileAttributesA( path.c_str() );
	 if( fa!= 0xFFFFFFFF)
		 if((fa&FILE_ATTRIBUTE_DIRECTORY))
			 return true;
	 return false;
#else
	struct stat buffer;
	if(path.size()>2 && path[path.size()-1]==':')
		path+='/';
	if(0==::stat(path.c_str(),&buffer))
	{
		if( buffer.st_mode & S_IFDIR )
			return true;
	}
	return false;
#endif
}
bool YFileSys::isPathExist(std::string path)
{//>Remove from path filename and check if directory exists
	if(YFileSys::isDirExist(path))
		return true;
	path=YFileSys::getDir(path);
	return YFileSys::isDirExist(path);
}

size_t YFileSys::getFileSize(std::string path)
{
#ifdef YFILESYS_WINDOWS
     WIN32_FILE_ATTRIBUTE_DATA fad;      
	if ( !::GetFileAttributesExA( path.c_str(),	::GetFileExInfoStandard, &fad ) )
		return 0;
	return (fad.nFileSizeHigh)?0XFFFFFFFF:fad.nFileSizeLow; 
#else
	struct stat buffer;
	if(0==::stat(path.c_str(),&buffer))
		return buffer.st_size;
	return 0;
#endif
}
YDataTime YFileSys::getTime(std::string path)
{
#ifdef YFILESYS_WINDOWS
     WIN32_FILE_ATTRIBUTE_DATA fad;
      if ( !::GetFileAttributesExA( path.c_str(),
        ::GetFileExInfoStandard, &fad ) )
			return YDataTime();
	SYSTEMTIME st;
	FileTimeToSystemTime(&fad.ftLastWriteTime,&st);
	//SystemTimeToTzSpecificLocalTime(NULL, &st, &stLocal);
	return  YDataTime(st.wYear,st.wMonth,st.wDay,st.wHour,st.wMinute,st.wSecond+double(st.wMilliseconds)/1000);
#else
	struct stat buffer;
	if(0==::stat(path.c_str(),&buffer))
	{
		tm * t = gmtime(&buffer.st_mtime); 
		if(t)
			return  YDataTime(t->tm_year+1900,t->tm_mon+1,t->tm_mday,t->tm_hour,t->tm_min,t->tm_sec);
	}
	return YDataTime();
#endif
}


bool YFileSys::rm(std::string path)
{
	return 0==::remove(path.c_str());
}

std::string YFileSys::cwd()
{
   char Buffer[MAX_PATH];
   char *b=Buffer;

   DWORD dwRet = GetCurrentDirectory(MAX_PATH, Buffer);
   if(dwRet>MAX_PATH)
   {
	   std::vector<char> buf(dwRet);
	   b=&(buf[0]);
	   dwRet = GetCurrentDirectory(dwRet,b);
   }
   if(dwRet==0)
   {
	   std::ostringstream os;
	   os << "GetCurrentDirectory failed " << GetLastError() ;
      throw os.str();
   }
}
void YFileSys::cwd(std::string path)
{
   if( ! SetCurrentDirectory(path.c_str()) )
   {
	   std::ostringstream os;
	   os << "SetCurrentDirectory failed " << GetLastError() ;
      throw os.str();
   }
}
//////////////////////////////////////////////////////////////////////////////
// Windows specific
//////////////////////////////////////////////////////////////////////////////
#ifdef WIN32
std::string YFileSys::getDriveLetter(std::string path) 
{
	std::string dr; 
	if(path.size()>=2 && path[1]==':') 
		dr=path[0]; 
	else if((path.size()>=2 && path[0]=='"' && path[2]==':') )
		dr=path[1]; 
	return dr;
}
std::string YFileSys::remDrive(std::string path)
{
	if(path.size()>=2 && path[1]==':') 
		return path.substr(2);
	else if((path.size()>=2 && path[0]=='"' && path[2]==':') )
		return path.replace(1,2,"");
	return path;
}
std::list<std::string> YFileSys::FilesInDir(const char* dir,const char* mask)
{
	std::string s(dir);
	char c=dir[strlen(dir)-1];
	if(c!= '/' && c!='\\')
		s+='/';
	s+=mask;
	std::list<std::string> lst;
    struct _finddata_t c_file;
    long hFile;
	if( (hFile = _findfirst( s.c_str(), &c_file )) != -1L )
	{
		if(!(c_file.attrib & _A_SUBDIR))
			lst.push_back(c_file.name);
        while( _findnext( hFile, &c_file ) == 0 )
		{
			if(!(c_file.attrib & _A_SUBDIR))
				lst.push_back(c_file.name);
		}
       _findclose( hFile );
	}
	return lst;
}
#endif

//////////////////////////////////////////////////////////////////////////////
// Cygwin specific
//////////////////////////////////////////////////////////////////////////////
#ifdef WIN32
	static std::string _cyg_win_root;
	static std::string _cyg_win_bin;

	static void _initCygwinPathes()
	{
	HKEY hKey;
		LONG lSuccess = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
		"SOFTWARE\\Cygnus Solutions\\Cygwin\\mounts v2\\/",
			0, KEY_QUERY_VALUE,&hKey);
	if(lSuccess != ERROR_SUCCESS)
			throw "Error in GetCygwinRoot(): RegOpenKeyEx";
		unsigned char buf[512];
		DWORD sz=sizeof(buf);
		lSuccess=RegQueryValueEx(hKey,"native",NULL,NULL,buf,&sz);
		RegCloseKey(hKey);// handle to key to close
	if(lSuccess != ERROR_SUCCESS)
			throw "Error in GetCygwinRoot(): RegQueryValueEx";
		_cyg_win_root=std::string((char*)buf);
		_cyg_win_bin=_cyg_win_root+"\\bin\\";
	}

	std::string YFileSys::GetCygwinRoot()
	{
		if(_cyg_win_root.empty())
			_initCygwinPathes();
		return _cyg_win_root;
	}

	std::string YFileSys::GetCygwinBin()
	{
		if(_cyg_win_root.empty())
			_initCygwinPathes();
		return _cyg_win_bin;
	}

	std::string YFileSys::WinPathToPosix(std::string path,bool is_list)
	{
		if(_cyg_win_root.empty())
			_initCygwinPathes();

		char buf[512];
		//sprintf(buf,"%s%s %s",_cyg_win_bin.c_str(),"cygpath.exe -u -a",path.c_str());
		std::string cmd=_cyg_win_bin;
		if(is_list)
			cmd+="cygpath.exe -u -a -p ";
		else
			cmd+="cygpath.exe -u -a ";
		cmd+=path;
		childPipe cp(cmd.c_str(),0,childPipe::F_StdOutErr);
		ReadString rs(cp.Stdout(),sizeof(buf));
		int cnt=300;
		while(!rs.IsAvail() && cnt--) Sleep(10);
		rs.gets(buf);
		cp.wait();
		if(cp.isRunning())
			cp.terminate();
		return std::string(buf);
	}

	std::string YFileSys::PosixPathToWin(std::string path,bool is_list)
	{
		if(_cyg_win_root.empty())
			_initCygwinPathes();

		char buf[512];
		std::string cmd=_cyg_win_bin;
		if(is_list)
			cmd+="cygpath.exe -w -a -p ";
		else
			cmd+="cygpath.exe -w -a ";
		cmd+=path;
		//sprintf(buf,"%s%s %s",_cyg_win_bin.c_str(),"cygpath.exe -w -a",path.c_str());
		childPipe cp(cmd.c_str(),0,childPipe::F_StdOutErr);
		ReadString rs(cp.Stdout(),sizeof(buf));
		int cnt=300;
		while(!rs.IsAvail() && cnt--) Sleep(10);
		rs.gets(buf);
		cp.wait();
		if(cp.isRunning())
			cp.terminate();
		return std::string(buf);
	}
#endif //#ifdef WIN32
//////////////////////////////////////////////////////////////////////////////
// 
//////////////////////////////////////////////////////////////////////////////
