#ifndef _BMP_H
#define _BMP_H
////////////////////////////////////////////////////////////////
#include "yfile.h"
/// @brief Image File Formats, ONLY intrinsic used by library, DO NOT use directly!
namespace IFF
{

#pragma pack(push,1)
struct BMP_BitMapFileHeader
{
	uint16_t bfType;      // always = 0x4d42 ( 'BM')
	uint32_t bfSize;      // file size in bytes (disk space)
	uint16_t bfReserved1; // =0
	uint16_t bfReserved2; // =0
	uint32_t bfOffBits;   // offset for raster data beginning
};
#pragma pack(pop)
#pragma pack(push,1)
struct BMP_BitMapInfoHeader
{
	uint32_t biSize;          // this structure size,=40. Use for found RgbQuad data.
	uint32_t biWidth;
	uint32_t biHeight;
	uint16_t biPlanes;        // always = 1
	uint16_t biBitCount;      // 1,4,8 or 24.
	uint32_t biCompression;   // BI_RGB (no compression), BI_RLE8 or BI_RLE4.
	uint32_t biSizeImage;     // size of image in bytes, may be 0 for BI_RGB.
	uint32_t biXPelsPerMeter; // pixels per meter
	uint32_t biYPelsPerMeter;
	uint32_t biClrUsed;  // 0 - used all colors
	uint32_t biClrImportant; // n - first n colors (index) important, 0 -all.
};
#pragma pack(pop)

#ifndef BI_RGB
	#define BI_RGB  0
	#define BI_RLE8 1
	#define BI_RLE4 2
#endif
#ifndef RgbQuad
#pragma pack(push,1)
struct RgbQuad
{
	Byte b;
	Byte g;
	Byte r;
	Byte reserved;
};
#pragma pack(pop)
#endif
//file structure:
// BMP_BitMapFileHeader
// BMP_BitMapInfoHeader
// RgbQuad[0x1FF & (1<<biBitCount)]
// Byte ImadeData[];
// EOF

class BmpRead
: private YFileRead
{
public:
	enum Type
	{
		BT_BILEVEL,
		BT_INDEX,
		BT_RGB
	};
private:
	BMP_BitMapFileHeader  header;
	BMP_BitMapInfoHeader  info;
  uint32_t bytes_per_row;
	int pal_entries;
	RgbQuad *_pal;
	Type img_class;
	uint32_t base_offset;
public:
	BmpRead(const char *filename);
	virtual ~BmpRead();
	Type GetType() { return img_class; }
	int GetPaletteEntries() const { return pal_entries;}
	const RgbQuad *GetPalette() { return _pal;}
	int Width()  const {return info.biWidth;}
	int Height() const {return info.biHeight;}
	int Bits()   const {return info.biBitCount; }
	void read(int x, int y,Byte *data,int len);
//	void write(int x, int y,const Byte *data,int len);
};

class BmpWrite
: private YFileWrite
{
public:
	enum Type
	{
		BT_BILEVEL,
		BT_INDEX,
		BT_RGB
	};
private:
	BMP_BitMapFileHeader  header;
	BMP_BitMapInfoHeader  info;
  uint32_t bytes_per_row;
	int pal_entries;
	uint32_t pal_offset;
//	RgbQuad *_pal;
	Type img_class;
	uint32_t base_offset;
public:
	BmpWrite(const char *filename,int width,int height,IFF::BmpWrite::Type iclass,const RgbQuad *pal);
	virtual ~BmpWrite();
	Type GetType() { return img_class; }
	int GetPaletteEntries() const { return pal_entries;}
//	const RgbQuad *GetPalette() { return _pal;}
	int Width()  const {return info.biWidth;}
	int Height() const {return info.biHeight;}
	int Bits()   const {return info.biBitCount; }
//	void read(int x, int y,Byte *data,int len);
	void write(int x, int y,const Byte *data,int len);
};
} //namespace IFF
////////////////////////////////////////////////////////////////
#endif //_BMP_H
