#include <stdio.h>  //sprintf(),vsprintf()
#include <stdarg.h> //va_list, va_end
#include <string.h> //strlen(), strcmpi()
//#include "AImage.h"
#include "yfilesys.h"//
#include "YMem.h"//
#include "ImageMem.h"//#include "AImage.h"
#include "PBM_file.h"
#include "Tiff.h"
#include "bmp.h"
#include "Oz.h"
#include "IFloat.H"
#include "YFileSys.h"//

using namespace IFF;
/////////////////////////////////////////////////////////////////////////////////////
static void message_load_file(ProgressCounter* indicator,const char *fname,int height)
{
	if(indicator)
	{
		//::sprintf(msg,indicator->russian()?"  %s":"Load file %s",fname);
		std::ostringstream os;
		os << (indicator->russian()?"  ":"Load file ") << fname;
		indicator->start(os.str().c_str(),0,height-1);
//		indicator->start(msg,0,height-1);
	}
}
static void message_load_file_aborted(ProgressCounter* indicator,const char *fname)
{
	if(indicator)
	{
		std::ostringstream os;
		if(indicator->russian())
			os << "  " << fname << "  !";
		else
			os << "User aborted loading file " << fname << " !";
		//::sprintf(msg,indicator->russian()?"  %s  ":"User aborted loading file %s",fname);
		//indicator->error(msg);
		indicator->error(os.str().c_str());
	}
}
static void message_unknown_image_format(ProgressCounter* indicator,const char *fname)
{
	if(indicator)
	{
		std::ostringstream os;
		if(indicator->russian())
			os << "    " << fname << " .";
		else
			os << "Unknown image file format " << fname << " .";
		indicator->error(os.str().c_str());
	}
}
static void message_error_when_load_image_file(ProgressCounter* indicator,const char *fname)
{
	if(indicator)
	{
		std::ostringstream os;
		if(indicator->russian())
			os << "     " << fname;
		else		
			os << "Error load image file " << fname;
		indicator->error(os.str().c_str());
	}
}

/////////////////////////////////////////////////////////////////////////////////////
//////////// class ErrorStatusBase //////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
ErrorStatusBase::ErrorStatusBase(const char *fname,int nline,const char *_cls_this,const char *_cls,bool del_msg)
{
	file=fname;
	line=nline;
  cls_this=_cls_this;
	cls=_cls;
	msg_need_delete=del_msg;
	msg=0;
}
//virtual 
ErrorStatusBase::~ErrorStatusBase()
{
	if(msg_need_delete && msg)
		delete msg;
}
//virtual 
const char * ErrorStatusBase::Msg() const 
{ 
	return msg?msg:"Probably insufficient memory"; 
}  // return error message

AImage::Error::Error(const char *fname,int nline,const AImage *obj,const char *msgfmt,...)
:ErrorStatusBase(fname,nline,AImage::ThisClassName(),obj?obj->ClassName():"(null)",false)
{
		msg=message;
//		va_list argptr;
//		va_start(argptr, msgfmt);
//		::vsprintf(msg,msgfmt,argptr);
//		va_end(argptr);
}
/*
#define ERROR_NotCompatImages(im,iclass,width,height,bits) throw AImage::Error(__FILE__,__LINE__,im,\
        "Inadmissible destination AImage object, attempt load image %dx%dx%d %s into object %dx%dx%d %s .",\
        im->Width(),im->Height(),im->Bits(),im->ImageClassName(),\
        width,height,bits,AImage::ImageClassName(iclass) )

#define ERROR_CreateImage(im,iclass,width,height,bits) throw AImage::Error(__FILE__,__LINE__,im,\
        "Can't create image %dx%dx%d %s , probably insufficient memory.",\
        width,height,bits,AImage::ImageClassName(iclass) )
#define ERROR_SaveImageNotValid() throw AImage::Error(__FILE__,__LINE__,0,\
        "Can't save image because it is not valid.")

#define ERROR_CantSaveFormat(im,format,bits,iclass) throw AImage::Error(__FILE__,__LINE__,im,\
        "Can't save image %dx%dx%d %s in %s-format, because image is not %d-bit %s.",\
				im->Width(),im->Height(),im->Bits(),im->ImageClassName(),\
        format,bits,AImage::ImageClassName(iclass) )

#define ERROR_UnknownImageFileExtension(ext) throw AImage::Error(__FILE__,__LINE__,0,\
        "Unsupported image file extension %s.",ext)
*/

void ERROR_NotCompatImages(const AImage* im,AImage::YImageClass iclass,int width,int height,int bits) 
{
	std::ostringstream os;
	os << "In file image_io.cpp: "
	   << "Inadmissible destination AImage object, attempt load image "
	   << im->Width() << "x" << im->Height() << "x" << im->Bits() << "  " << im->ImageClassName() 
	   << " into object " 
	   << width << "x" << height << "x" << bits << "  " << AImage::ImageClassName(iclass)
	   << " .";
	throw os.str();
}
void ERROR_CreateImage(const AImage* im,AImage::YImageClass iclass,int width,int height,int bits) 
{
	std::ostringstream os;
	os << "In file image_io.cpp: "
	   << "Can't create image  "
	   << width << "x" << height << "x" << bits << "  " << AImage::ImageClassName(iclass)
	   << " , probably insufficient memory.";
	throw os.str();
}
void ERROR_SaveImageNotValid() 
{
	throw std::string("Can't save image because it is not valid.");
}
void ERROR_CantSaveFormat(const AImage* im,std::string format,int bits,AImage::YImageClass iclass) 
{
	std::ostringstream os;
	os << "In file image_io.cpp: "
	   << "Can't save image  "
	   << im->Width() << "x" << im->Height() << "x" << im->Bits() << "  " << im->ImageClassName() 
	   << " in " << format <<" - format, because image is not "
	   << bits << " -bits " << AImage::ImageClassName(iclass)
	   << " .";
	throw os.str();
}
void ERROR_UnknownImageFileExtension(std::string ext) 
{
	std::ostringstream os;
	os << "Unsupported image file extension " << ext <<" .";
	throw os.str();
}

/////////////////////////////////////////////////////////////////////////////////////
//////////// public functions to deal with AImage ///////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
static AImage *  LoadPbmFile(AImage *im,bool replace,const char *fname,ProgressCounter* indicator)
{
	PBMfile pbm;
	pbm.open(fname,YFile::UIO_RDONLY);
	//get image info from file
	const short width=pbm.Width();
	const short height=pbm.Length();
	//if(pbm.type==5)
	{
		const short bits=8;
		AImage::YImageClass iclass=AImage::IMGC_GRAYSCALE;
		//test possibility to load
		if(!::AImageIsCompatible(im,iclass,width,height,bits))
		{
			if(!replace)
				ERROR_NotCompatImages(im,iclass,width,height,bits);
		}
		//Create image
		im=::AImageCreate(im,iclass,width,height,bits);
		if(!im)
			ERROR_CreateImage(im,iclass,width,height,bits);
		AImageMem8 *img=dynamic_cast<AImageMem8 *>(im);
		if(!img)
			throw "Unexpected at LoadPbmFile(): CreateAImage() creates image of impropere type";
		//Load image
//		char msg[512];
		message_load_file(indicator,fname,height);
		Pixel8 *p=(*img)[0];
		int shift=img->NextLineShift();
		for(int y=0; y<height; y++)
		{
			if(indicator)
			{
				if((y&0x3F)==0) //inform only each 64-th string read
				{
					if(indicator->current(y))
					{
						message_load_file_aborted(indicator,fname);
						goto l_end;
					}
				}
			}
			pbm.read(0,y,p,width);
			p+=shift;
		}
		if(indicator)	
		{
			indicator->current(height-1);
			indicator->finish();
		}
	}
l_end:
	return im;
}

static AImage *  LoadPpmFile(AImage *im,bool replace,const char *fname,ProgressCounter* indicator)
{
	PBMfile pbm;
	pbm.open(fname,YFile::UIO_RDONLY);
	//get image info from file
	const short width=pbm.Width();
	const short height=pbm.Length();
	PtrDelete<Pixel8> __line;
	Pixel8 *line=__line.Make(width*3);
	if(!line)
		throw "LoadPpmFile(): insufficient memory";
	if(pbm.Type()!=PBMfile::PBM_RGB)
		throw "LoadPpmFile() : no PPM file";
	{
		const short bits=8;
		AImage::YImageClass iclass=AImage::IMGC_RGB;
		//test possibility to load
		if(!::AImageIsCompatible(im,iclass,width,height,bits))
		{
			if(!replace)
				ERROR_NotCompatImages(im,iclass,width,height,bits);
		}
		//Create image
		im=::AImageCreate(im,iclass,width,height,bits);
		if(!im)
			ERROR_CreateImage(im,iclass,width,height,bits);
		AImageMemRgb8 *img=dynamic_cast<AImageMemRgb8 *>(im);
		if(!img)
			throw "Unexpected at LoadPpmFile(): CreateAImage() creates image of impropere type";
		//Load image
		message_load_file(indicator,fname,height);
		//Pixel8 *p=(*img)[0];
		AImageMem8* Red=img->GetPlane8(0);
		AImageMem8* Green=img->GetPlane8(1);
		AImageMem8* Blue=img->GetPlane8(2);
		for(int y=0; y<height; y++)
		{
			if(indicator)
			{
				if((y&0x3F)==0) //inform only each 64-th string read
				{
					if(indicator->current(y))
					{
						message_load_file_aborted(indicator,fname);
						goto l_end;
					}
				}
			}
			pbm.read(0,y,line,3*width);
			{
				Pixel8 *r=(*Red)[y];
				Pixel8 *g=(*Green)[y];
				Pixel8 *b=(*Blue)[y];
				Pixel8 *d=line;
				int c=width;
				while(c--)
				{
					*r++=*d++;
					*g++=*d++;
					*b++=*d++;
				}
			}
		}
		if(indicator)	
		{
			indicator->current(height-1);
			indicator->finish();
		}
	}
l_end:
	return im;
}

static AImage * LoadOzFile(AImage *im,bool replace,const char *fname,ProgressCounter* indicator)
{
	OzFile oz;
	//get image info from file
	oz.open(fname,YFile::UIO_RDONLY);
	const short width=oz.Width();
	const short height=oz.Length();
	const short bits=oz.Bits();
	AImage::YImageClass iclass=AImage::IMGC_GRAYSCALE;
	//test possibility to load
	if(!::AImageIsCompatible(im,iclass,width,height,bits))
	{
		if(!replace)
			ERROR_NotCompatImages(im,iclass,width,height,bits);
	}
	//Create image
	im=::AImageCreate(im,iclass,width,height,bits);
	if(!im)
		ERROR_CreateImage(im,iclass,width,height,bits);
	AImageMem16  *img=dynamic_cast<AImageMem16 *>(im);
	if(!img)
		throw "Unexpected at LoadOzFile(): CreateAImage() creates image of impropere type";
	//Load image
	message_load_file(indicator,fname,height);
	for(int y=0; y<height; y++)
	{
		oz.read(0,y,(OzWord *)(*img)[y],width);
		if(indicator)
		{
			if(indicator->current(y))
			{
				message_load_file_aborted(indicator,fname);
				goto l_end;
			}
		}
	}
	if(indicator)	
	{
		indicator->current(height-1);
		indicator->finish();
	}
l_end:
	return im;
}
static AImage *  LoadTiffFile(AImage *im,bool replace,const char *fname,ProgressCounter* indicator)
{
	Tiff8Read tf(fname);
	//get image info from file
	const short width=tf.Width();
	const short height=tf.Length();
	const short bits=tf.Bits();
	const int ipal=tf.GetPaletteLength();
	AImage::YImageClass iclass;
	if(ipal)
		iclass=AImage::IMGC_PALETTE;
	else
		iclass=AImage::IMGC_GRAYSCALE;
	//test possibility to load
	if(!::AImageIsCompatible(im,iclass,width,height,bits))
	{
		if(!replace)
			ERROR_NotCompatImages(im,iclass,width,height,bits);
	}
	//Create image
	im=::AImageCreate(im,iclass,width,height,bits);
	if(!im)
		ERROR_CreateImage(im,iclass,width,height,bits);
	//Load image
	message_load_file(indicator,fname,height);
	{
		if(indicator)
		{
			if(indicator->current(height/2))
			{
				message_load_file_aborted(indicator,fname);
				goto l_end;
			}
		}
		if(tf.Bits()==8)
		{
			AImageMem8 *img=dynamic_cast<AImageMem8 *>(im);
			if(!img)
				throw "Unexpected at LoadTiffFile(): CreateAImage() creates image of improper type";
			tf.readImage((*img)[0]);
			//load palette
			if(ipal)
			{
				AImage::Rgb *pal=img->Palette();
				uint32_t map[256];
				tf.GetPalette(map);
				if(pal)
				{
					int ent=img->GetPaletteEntries();
					int i;
					for(i=0;i<ent; i++)
					{
						pal[i].r=Pixel8(i);
						pal[i].g=Pixel8(i);
						pal[i].b=Pixel8(i);
					}
					if( map)
					{
						ent=MIN(img->GetPaletteEntries(),ipal);
						for(i=0;i<ent; i++)
						{
							uint32_t m=map[i];
							pal[i].r=Byte(m&0xFF); m>>=8;
							pal[i].g=Byte(m&0xFF); m>>=8;
							pal[i].b=Byte(m&0xFF); m>>=8;
						}
					}
				}
			}
		}
		else if(tf.Bits()==16)
		{
			AImageMem16 *img=dynamic_cast<AImageMem16 *>(im);
			if(!img)
				throw "Unexpected at LoadTiffFile(): CreateAImage() creates image of impropere type";
			tf.readImage(Pixel8Ptr((*img)[0]));
		}
	}
	if(indicator)	
	{
		indicator->current(height-1);
		indicator->finish();
	}
l_end:
	return im;
}
static AImage *  LoadBmpFile(AImage *im,bool replace,const char *fname,ProgressCounter* indicator)
{

	BmpRead bm(fname);
	//get image info from file
	const short width=bm.Width();
	const short height=bm.Height();
	const short bits=bm.Bits();
	switch(bits)
	{
//	case 1:
//		break;
//	case 4:
//		break;
	case 8:
		{
			AImage::YImageClass iclass=AImage::IMGC_PALETTE;
			//test possibility to load
			if(!::AImageIsCompatible(im,iclass,width,height,bits))
			{
				if(!replace)
					ERROR_NotCompatImages(im,iclass,width,height,bits);
			}
			//Create image
			im=::AImageCreate(im,iclass,width,height,bits);
			if(!im)
				ERROR_CreateImage(im,iclass,width,height,bits);
			AImageMem8 *img=dynamic_cast<AImageMem8 *>(im);
			if(!img)
				"Unexpected at LoadBmpFile(): CreateAImage() creates image of impropere type";
			//Load image
			message_load_file(indicator,fname,height);
			for(int y=height-1; y>=0; y--)
			{
				bm.read(0,y,(*img)[y],width);
				if(indicator)
				{
					if(indicator->current(height-y-1))
					{
						message_load_file_aborted(indicator,fname);
						goto l_end_8;
					}
				}
			}
			if(indicator)	
			{
				indicator->current(0);
				indicator->finish();
			}
		l_end_8:
			//load palette
			AImage::Rgb *pal=img->Palette();
			const RgbQuad *map=bm.GetPalette();
			if(pal)
			{
				int ent=img->GetPaletteEntries();
				int i;
				for(i=0;i<ent; i++)
				{
					pal[i].r=Pixel8(i);
					pal[i].g=Pixel8(i);
					pal[i].b=Pixel8(i);
				}
				if( map)
				{
					ent=MIN(img->GetPaletteEntries(),bm.GetPaletteEntries());
					for(i=0;i<ent; i++)
					{
						pal[i].r=map[i].r;
						pal[i].g=map[i].g;
						pal[i].b=map[i].b;
					}
				}
			}
		}
		break;
	case 24:
		{
			AImage::YImageClass iclass=AImage::IMGC_RGB;
			//test possibility to load
			if(!::AImageIsCompatible(im,iclass,width,height,8))
			{
				if(!replace)
					ERROR_NotCompatImages(im,iclass,width,height,bits);
			}
			//alloc temporary space
			PtrDelete<Byte> _buf;
			Byte *buf=_buf.Make(bm.Width()*3);
			if(!buf)
				throw "LoadBmpFile() : insufficient memory when load RGB image";
			//Create image
			im=::AImageCreate(im,iclass,width,height,8);
			if(!im)
				ERROR_CreateImage(im,iclass,width,height,8);
			AImageMemRgb8 *img=dynamic_cast<AImageMemRgb8 *>(im);
			if(!img)
				"Unexpected at LoadBmpFile(): CreateAImage() creates image of impropere type";
			//Load image
			message_load_file(indicator,fname,height);
			for(int y=height-1; y>=0; y--)
			{
				bm.read(0,y,buf,3*width);
				for(int plane=0; plane<3; plane++)
				{
					AImageMem8* im=img->GetPlane8(plane);
					Byte *dst=(*im)[y];
					Byte *src=buf+(2-plane);
					int w=width;
					while(w--)
					{
						*dst++=*src;
						src+=3;
					}
				}
				if(indicator)
				{
					if(indicator->current(height-y-1))
					{
						message_load_file_aborted(indicator,fname);
						goto l_end_24;
					}
				}
			}
l_end_24:
			if(indicator)	
			{
				indicator->current(0);
				indicator->finish();
			}
		}
		break;
	default:
		throw "currently BMP supports only 8 and 24 bits per pixel";
	};
	return im;
}
static AImage * LoadFloatFile(AImage *im,bool replace,const char *fname,ProgressCounter* indicator)
{
	FloatImageFile ff;
	//get image info from file
	ff.open(fname,YFile::UIO_RDONLY);
	const short width=ff.Width();
	const short height=ff.Height();
	const short bits=8*sizeof(PixelFloat);
	AImage::YImageClass iclass=AImage::IMGC_FLOAT;
	//test possibility to load
	if(!::AImageIsCompatible(im,iclass,width,height,bits))
	{
		if(!replace)
			ERROR_NotCompatImages(im,iclass,width,height,bits);
	}
	//Create image
	im=::AImageCreate(im,iclass,width,height,bits);
	if(!im)
		ERROR_CreateImage(im,iclass,width,height,bits);
	AImageMemFloat *img=dynamic_cast<AImageMemFloat *>(im);
	if(!img)
		throw "Unexpected at LoadFloatFile(): CreateAImage() creates image of impropere type";
	//Load image
	message_load_file(indicator,fname,height);
	for(int y=0; y<height; y++)
	{
		ff.read(0,y,(FloatWord *)(*img)[y],width);
		if(indicator)
		{
			if(indicator->current(y))
			{
				message_load_file_aborted(indicator,fname);
				goto l_end;
			}
		}
	}
	if(indicator)
	{
		indicator->current(height-1);
		indicator->finish();
	}
l_end:
	return im;
}


static void SavePbmFile(const AImage * img,const char *fname,ProgressCounter* indicator,const char *banner)
{
			const AImageMem8 *im=dynamic_cast<const AImageMem8 *>(img);
			if(!im->IsOk()) ERROR_SaveImageNotValid();
			if(!im || im->ImageClass()!=AImage::IMGC_GRAYSCALE) 
				ERROR_CantSaveFormat(im,"PGM",8,AImage::IMGC_GRAYSCALE);

			PBMfile pbm;
			//		im->Load(lpszPathName);
			const int width=im->Width();
			const int height=im->Height();
			pbm.open(fname,YFile::UIO_CREATE_OVERWRITE,width,height,im->Bits());
			{
				if(indicator)
				{
					indicator->start(banner,0,height-1);
				}
				Pixel8 *p=(*im)[0];
				int shift=im->NextLineShift();
				for(int y=0; y<height; y++)
				{
					pbm.write(0,y,p,width);
					p+=shift;
					if(indicator)
					{
						if((y&0x3F)==0)
						{
							indicator->current(y);
						}
					}
				}
				if(indicator)
				{
					indicator->current(height-1);
					indicator->finish();
				}
			}
}
static void SavePpmFile(const AImage * img,const char *fname,ProgressCounter* indicator,const char *banner)
{
			if(!img) ERROR_SaveImageNotValid();
			const AImageMemRgb8 *_im=dynamic_cast<const AImageMemRgb8 *>(img);
			if(!_im ) 
				ERROR_CantSaveFormat(_im,"PPM",8,AImage::IMGC_RGB);
			if(!_im->IsOk()) ERROR_SaveImageNotValid();
			if( _im->ImageClass()!=AImage::IMGC_RGB) 
				ERROR_CantSaveFormat(_im,"PPM",8,AImage::IMGC_RGB);

			PBMfile pbm;
			//		im->Load(lpszPathName);
			const int width=_im->Width();
			const int height=_im->Height();
			PtrDelete<Byte> __line(new Byte[width*3],width*3);
			Byte *line=__line.Get();
			if(!line)
				throw "SavePpmFile(): inssufficient memoty";
			pbm.open(fname,YFile::UIO_CREATE_OVERWRITE,width,height,_im->Bits(),PBMfile::PBM_RGB);
			{
				if(indicator)
				{
					indicator->start(banner,0,height-1);
				}
				int shift[3];
				Pixel8 *p[3];
				{
					for(int i=0; i<3; i++)
					{
						AImageMem8 *im=_im->GetPlane8(i);
						p[i]=(*im)[0];
						shift[i]=im->NextLineShift();
					}
				}
				for(int y=0; y<height; y++)
				{
					Byte *l=line;
					Pixel8 *p0=p[0];
					Pixel8 *p1=p[1];
					Pixel8 *p2=p[2];
					for(int x=0; x<width; x++)
					{
						*l++=*p0++;
						*l++=*p1++;
						*l++=*p2++;
					}
					pbm.write(0,y,line,width*3);
					{
						for(int i=0; i<3; i++)
						{
							p[i]+=shift[i];
						}
					}
					if(indicator)
					{
						if((y&0x3F)==0)
						{
							indicator->current(y);
						}
					}
				}
				if(indicator)
				{
					indicator->current(height-1);
					indicator->finish();
				}
			}
}

static void SaveOzFile(const AImage * img,const char *fname,ProgressCounter* indicator,const char *banner)
{
			const AImageMem16 *im=dynamic_cast<const AImageMem16 *>(img);
			if(!im->IsOk()) ERROR_SaveImageNotValid();
			if(!im) 
				ERROR_CantSaveFormat(img,"OZ",16,AImage::IMGC_GRAYSCALE);

			OzFile oz;
			//		im->Load(lpszPathName);
			const int width=im->Width();
			const int height=im->Height();
			oz.open(fname,YFile::UIO_CREATE_OVERWRITE,width,height,im->Bits());
			{
				if(indicator)
				{
					indicator->start(banner,0,height-1);
				}
				for(int y=0; y<height; y++)
				{
					oz.write(0,y,(OzWord *)(*im)[y],width);
					if(indicator)
					{
						indicator->current(y);
					}
				}
				if(indicator)
				{
					indicator->current(height-1);
					indicator->finish();
				}
			}
}
static void  SaveTiffFile(const AImage * img,const char *fname,ProgressCounter* indicator,const char *banner)
{
			const int ent=256;
			uint32_t pal_space[ent];
			uint32_t *paletta=0;
			if(!img || !img->IsOk()) ERROR_SaveImageNotValid();
			if(img->ImageClass()==AImage::IMGC_GRAYSCALE) 
				;
			else if(img->ImageClass()==AImage::IMGC_PALETTE 
				&& img->GetPaletteEntries()>0 
				&& img->Palette())
			{
				const AImageMem8 *im=dynamic_cast<const AImageMem8 *>(img);
				if(!im || !im->IsOk()) ERROR_SaveImageNotValid();
				const int ent2=MIN(im->GetPaletteEntries(),ent);
				paletta=pal_space;
				if(paletta)
				{
					AImage::Rgb *pal=im->Palette();
					for( int i=0; i<ent2; i++)
					{
						uint32_t p=0;
						p|=pal[i].r;
						p<<=8;
						p|=pal[i].g;
						p<<=8;
						p|=pal[i].b;
						//p<<=8;
						paletta[i]=p;
					}
				}
			}
			else
				ERROR_CantSaveFormat(img,"TIFF",8,AImage::IMGC_GRAYSCALE);
			const int width=img->Width();
			const int height=img->Height();
			const int bits=img->Bits();
			if(indicator)
			{
				indicator->start(banner,0,height-1);
			}
			Tiff8Write tf(fname,width,height,bits,paletta);
			{
				if(indicator)
				{
					int y=height/2;
					indicator->current(y);
				}
				if(bits<=8)
				{
					const AImageMem8 *im=dynamic_cast<const AImageMem8 *>(img);
					if(!im)
						ERROR_CantSaveFormat(img,"TIFF",8,AImage::IMGC_GRAYSCALE);
					tf.writeImage((*im)[0]);
				}
				else if(bits<=16)
				{
					const AImageMem16 *im=dynamic_cast<const AImageMem16 *>(img);
					if(!im)
						ERROR_CantSaveFormat(img,"TIFF",8,AImage::IMGC_GRAYSCALE);
					tf.writeImage((Byte *)((*im)[0]));
				}
			}
			if(indicator)
			{
				indicator->current(height-1);
				indicator->finish();
			}
}
static void  SaveBmpFile(const AImage * img,const char *fname,ProgressCounter* indicator,const char *banner)
{
	if(!img || !img->IsOk()) ERROR_SaveImageNotValid();
	switch(img->GetImageStore())
	{
	case AImage::IMGS_AImageMem8:
		{
			const AImageMem8 *im=dynamic_cast<const AImageMem8 *>(img);
			if(!im || im->ImageClass()!=AImage::IMGC_PALETTE) 
				ERROR_CantSaveFormat(im,"BMP",8,im->ImageClass());
			const int width=im->Width();
			const int height=im->Height();
			const int colors=256;
			RgbQuad pal[colors];
			{
				AImage::Rgb *p=im->Palette();
				int ent=MIN(im->GetPaletteEntries(),colors);
				int i;
				for(i=0; i<ent; i++)
				{
					pal[i].r=p[i].r;
					pal[i].g=p[i].g;
					pal[i].b=p[i].b;
					pal[i].reserved=0;
				}
				for(   ; i<colors; i++)
				{
					pal[i].r=Byte(i);
					pal[i].g=Byte(i);
					pal[i].b=Byte(i);
					pal[i].reserved=0;
				}
			}
			BmpWrite bm(fname,width,height,BmpWrite::BT_INDEX,pal);
			{
				if(indicator)
				{
					indicator->start(banner,0,height-1);
				}
				for(int y=0; y<height; y++)
				{
					bm.write(0,y,(*im)[y],width);
					if(indicator)
					{
						indicator->current(y);
					}
				}
				if(indicator)
				{
					indicator->current(height-1);
					indicator->finish();
				}
			}
		}
		break;
	case AImage::IMGS_AImageMemRgb8:
		{
			const AImageMemRgb8 *im=dynamic_cast<const AImageMemRgb8 *>(img);
			if(!im || im->ImageClass()!=AImage::IMGC_RGB) 
				ERROR_CantSaveFormat(im,"BMP",8,im->ImageClass());
			const int width=im->Width();
			const int height=im->Height();
			BmpWrite bm(fname,width,height,BmpWrite::BT_RGB,0);
			{
				PtrDelete<Byte> _buf;
				Byte * buf=_buf.Make(width*3+4);
				if(indicator)
				{
					indicator->start(banner,0,height-1);
				}
				for(int y=0; y<height; y++)
				{
					for(int plane=0; plane < 3; plane++)
					{
						Byte* dst=buf+(2-plane);
						const Byte* src=(*(im->GetPlane8(plane)))[y];
						int w=width;
						while(w--)
						{
							*dst=*src++;
							dst+=3;
						}
					}
					bm.write(0,y,buf,width*3);
					if(indicator)
					{
						indicator->current(y);
					}
				}
				if(indicator)
				{
					indicator->current(height-1);
					indicator->finish();
				}
			}
		}
		break;
	//case AImage::IMGS_AImageMemRgb16:
	//case AImage::IMGS_AImageMem16:
	//case AImage::IMGS_AImageMemFloat:
	default:
			ERROR_CantSaveFormat(img,"BMP",img->Bits(),img->ImageClass());
	};
}
static void  SaveFloatFile(const AImage * img,const char *fname,ProgressCounter* indicator,const char *banner)
{
			const AImageMemFloat *im=dynamic_cast<const AImageMemFloat *>(img);
			if(!im->IsOk()) ERROR_SaveImageNotValid();
			if(!im) 
				ERROR_CantSaveFormat(img,"FIM",8*sizeof(PixelFloat),AImage::IMGC_FLOAT);
			//		im->Load(lpszPathName);
			const int width=im->Width();
			const int height=im->Height();
			FloatImageFile ff;
			ff.open(fname,YFile::UIO_CREATE_OVERWRITE,width,height);
			if(indicator)
			{
				indicator->start(banner,0,height-1);
			}
			for(int y=0; y<height; y++)
			{
				ff.write(0,y,(*im)[y],width);
				if(indicator)
				{
					indicator->current(y);
				}
			}
			if(indicator)
			{
				indicator->current(height-1);
				indicator->finish();
			}
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
///// Integration ////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
AImage * AImageLoad(AImage *img,bool replace,const char *fname,ProgressCounter* indicator)
{
	static struct 
	{
		const char *ext;
		AImage * (*fun)(AImage *im,bool replace,const char *fname,ProgressCounter* indicator);
	} __ImageFileFormatLoad[]=
	{
		{".oz" ,LoadOzFile  },
		{".tif",LoadTiffFile},
		{".bmp",LoadBmpFile },
		{".fim",LoadFloatFile},
		{".pgm",LoadPbmFile},
		{".ppm",LoadPpmFile}
	};
	const int __ImageFileFormatLoadNumber=sizeof(__ImageFileFormatLoad)/sizeof(__ImageFileFormatLoad[0]);

	//const char *ext=YFileName::GetExt(fname);
	std::string ext=YFileSys::getExt(fname);
	for(int i=0; i<__ImageFileFormatLoadNumber; i++)
	{
		if(0==YFileSys::StrCmpI(ext,__ImageFileFormatLoad[i].ext))
		{
			img=__ImageFileFormatLoad[i].fun(img,replace,fname,indicator);
			goto finish;
		}
	}
	message_unknown_image_format(indicator,fname);
finish:
	if(img)
	{
		if(!img->IsOk())
		{
			delete img;
			img=0;
			message_error_when_load_image_file(indicator,fname);
		}
	}
	return img;
}
void AImageSave(const AImage * img,const char *fname,ProgressCounter* indicator)
{
	static struct 
	{
		const char *ext;

		void (*fun)(const AImage *im,const char *fname,ProgressCounter* indicator,const char *banner);
	} __ImageFileFormatSave[]=
	{
		{".oz" ,SaveOzFile  },
		{".tif",SaveTiffFile},
		{".bmp",SaveBmpFile },
		{".fim",SaveFloatFile},
		{".pgm",SavePbmFile},
		{".ppm",SavePpmFile}
	};
	const int __ImageFileFormatSaveNumber=sizeof(__ImageFileFormatSave)/sizeof(__ImageFileFormatSave[0]);

	std::string ext=YFileSys::getExt(fname);
	for(int i=0; i<__ImageFileFormatSaveNumber; i++)
	{
		if(0==YFileSys::StrCmpI(ext,__ImageFileFormatSave[i].ext))
		{
			try
			{
				//msg[0]=0;
				std::ostringstream os;
				os << "Save file " << fname;
				std::string msg=os.str();
				__ImageFileFormatSave[i].fun(img,fname,indicator,msg.c_str());
			}
			catch(...)
			{
				message_error_when_load_image_file(indicator,fname);
				if(indicator)
				{
					std::ostringstream os;
					os << "Error when save file " << fname;
					std::string msg=os.str();
					indicator->error(msg.c_str());
				}
				throw;
			};
			return;
		}
	}
	if(indicator)
	{
		message_unknown_image_format(indicator,fname);
		ERROR_UnknownImageFileExtension(YFileSys::getExt(fname).c_str());
	}
}

