#ifndef _IMAGEMEM_H
#define _IMAGEMEM_H
//////////////////////////////////////////////////////
#include "aimage.h"
//---------------------------------------------------------------------------------------
class ImageMem8
: public AImageMem8
{
	short wid;
  short len;
  short bits;
  int next_line_shift;
  Pixel8 *data;
	int pal_entries;
	AImage::Rgb * pal;
	YImageClass img_class;
  void nullify();
public:
  static const char *ThisClassName();
  virtual const char *ClassName() const { return ThisClassName(); }
	ImageMem8();
	virtual ~ImageMem8();

	virtual bool IsOk() const { return (data!=0); }
  virtual bool Create(short width,short height,short bitsPerPix=8,bool hasPalette=false);
  virtual void Clear();
	virtual int Width()  const  { return wid; }
	virtual int Height() const  { return len;}
  virtual int Bits()   const  { return bits;}
  virtual int NextLineShift() const { return next_line_shift; }
  virtual Pixel8* operator[](int y)const {return data+y*next_line_shift; }//@@need test bounds and NULL
	virtual YImageClass  ImageClass() const;
	virtual int GetPaletteEntries() const {return pal_entries;}
	virtual AImage::Rgb *Palette() const { return pal;}
	virtual void read(int x,int y,Pixel16 *data,int len) const;
	virtual void ChangeBits(int _bits) { if(_bits>=1 && _bits<=8) bits=short(_bits);}
};
//---------------------------------------------------------------------------------------
class ImageMem16
: public AImageMem16
{
private:
	short wid;
  short len;
  short bits;
  Pixel16 *data;
  void nullify();
	ImageMem16(const ImageMem16& im);
public:
  static const char *ThisClassName();
  virtual const char *ClassName() const { return ThisClassName(); }
	ImageMem16() { nullify(); }
	virtual ~ImageMem16() { Clear(); }

	virtual bool IsOk() const { return (data!=0); }
  virtual bool Create(short width,short height,short bitsPerPix,bool hasPalette=false);
  virtual void Clear();
	virtual int Width()  const  { return wid; }
	virtual int Height() const  { return len; }
  virtual int Bits()   const  { return bits;}
  virtual int NextLineShift() const { return wid; }
  virtual Pixel16* operator[](int y)const {return (data+long(wid)*y); }//@@need test bounds and NULL
	virtual YImageClass  ImageClass() const;
	virtual int GetPaletteEntries() const {return 0;}
	virtual AImage::Rgb *Palette() const  {return 0;}
	virtual void read(int x,int y,Pixel16 *data,int len) const;
	virtual void ChangeBits(int _bits) { if(_bits>=1 && _bits<=16) bits=short(_bits);}
};
//---------------------------------------------------------------------------------------
class ImageMemFloat
: public AImageMemFloat
{
public:
	short wid;
	short len;
	short bits;
	PixelFloat *data;
	void nullify();
public:
	static const char *ThisClassName();
	virtual const char *ClassName() const { return ThisClassName(); }
	ImageMemFloat();
	virtual ~ImageMemFloat();

	virtual bool IsOk() const { return (data!=0); }
	virtual bool Create(short width,short height,short bitsPerPix=8*sizeof(PixelFloat),bool hasPalette=false);
	virtual void Clear();
	virtual int Width()  const  { return wid; }
	virtual int Height() const  { return len;}
	virtual int Bits()   const  { return bits;}
	virtual int NextLineShift() const { return wid; }
	virtual PixelFloat* operator[](int y)const {return data+y*wid; }//need test bounds and NULL
	virtual YImageClass  ImageClass() const { return AImage::IMGC_FLOAT; }
  virtual void SetTo(PixelFloat color);
	virtual void read(int x,int y,Pixel16 *data,int len) const;
};
//---------------------------------------------------------------------------------------
class ImageMemRgb8
: public AImageMemRgb8
{
	short wid;
  short len;
  short bits;
  AImageMem8 * data8[3];
private:
  void nullify();
	ImageMemRgb8(const ImageMemRgb8& im);
public:
  static const char *ThisClassName();
  virtual const char *ClassName() const { return ThisClassName(); }
	ImageMemRgb8() {nullify();}
  virtual ~ImageMemRgb8() {Clear();}
	virtual bool IsOk() const;
  virtual bool Create(short width,short height,short bitsPerPix=8,bool hasPalette=false);
  virtual void Clear();
	virtual YImageClass  ImageClass() const;
	virtual int Width()  const  { return wid; }
	virtual int Height() const  { return len; }
  virtual int Bits()   const  { return bits;}
  virtual int NextLineShift() const { return 0; }
  virtual AImageMem8  * GetPlane8(int plane) const;
	virtual void read(int x,int y,Pixel16 *data,int len) const;
};
//////////////////////////////////////////////////////
#endif //_IMAGEMEM_H
