#include <string>
#include <sstream>
//#include <stdio.h> //sprint();
//#include <string.h>//strlen(),strcat();
#include "PBM_file.h"
using namespace IFF;

PBMfile::TYPE PBMfile::Type() 
const
{
	switch(type)
	{
	//case 4:
		//return PBM_BILEVEL;
	case 5:
		return PBM_GRAYSCALE;
	case 6:
		return PBM_RGB;
	};
	return PBM_UNKNOWN;
}

void PBMfile::nullify()
{
	width=0;
	height=0;
	bits=0;
	type=-1;
	data_offset=0;
}
PBMfile::PBMfile()
{
	nullify();
}
//virtual 
PBMfile::~PBMfile()
{
	close();
}
void PBMfile::open(const char *fname,OpenMode access,
			int  Width,int  Height,int  bitsPerPixel,PBMfile::TYPE _type)
{
	try
	{
		if(access==YFile::UIO_CREATE || access==YFile::UIO_CREATE_OVERWRITE)
		{
			if(Width<=0 || Height<=0 )
				throw "PBMfile::open() : Invalid image size parameter";
			if(bitsPerPixel<=0 || bitsPerPixel!=8)
				throw "PBMfile::open() : Invalid image depth parameter";
			switch(_type)
			{
			case PBM_GRAYSCALE:
				type=5;
				break;
			case PBM_RGB:
				type=6;
				break;
			default:
				type=-1;
			};
			if(type<0)
				throw "PBMfile::open() : Invalid image type parameter";
		}
		YFile::_open(fname,access);
		if(!YFile::IsOk())
			throw "file open error";

		switch(this->GetMode())
		{
		case UIO_CREATE:
		case UIO_CREATE_OVERWRITE:
				width=Width;
				height=Height;
				bits=bitsPerPixel;
				writeHeader();
				break;
		case UIO_RDONLY:
		case UIO_MODIFY:
				readHeader();
				break;
		default:
			throw "unknown file open mode";
		};
	}
	catch(...)
	{
		close();
		throw;
	};
}
void PBMfile::writeHeader()
{
/*
	const char nl[2]={0x0A,0};
	char buf[40];
	char head[512];
	::sprintf(head,"P%d",type);
	::strcat(head,nl);
	::sprintf(buf,"%d %d",width,height);
	::strcat(head,buf);
	::strcat(head,nl);
	if(type>4)
	{
		::sprintf(buf,"%d",255);//1<<bits);
		::strcat(head,buf);
		::strcat(head,nl);
	}
	data_offset=unsigned(::strlen(head));
	_write(0,head,data_offset);
*/
	char nl=0x0A;
	char space=32;
	std::ostringstream os;
	os << "P" << type <<nl;
	os << width << space << height << nl;
	if(type>4)
	{
		os << 255 << nl;//1<<bits);
	}
	std::string head=os.str();
	data_offset=head.length();
	_write(0,head.c_str(),data_offset);

}
static char *gets_from_buf(char * const buf,int bufsz,char *curpos)
{
	char *plast=(&buf[bufsz-1]);
	while(curpos<plast)
	{
		if(*curpos==0x0A)
		{
			*curpos++=0;
			return curpos;
		}
		curpos++;
	}
return 0;
}
void PBMfile::readHeader()
{
	char head[512];
	char *curpos=head;
	_read(0,head,sizeof(head));
	if(head[0]!='P')
		throw "No Portable Bitmap File";
//verify type
	if(head[1]=='5')
	{
		type=5;
	}
	else if(head[1]=='6')
	{
		type=6;
	}
	else
	{
		throw "Unsupported Portable Bitmap File type";
	}
//read width and height
	do
	{
		curpos=::gets_from_buf(head,sizeof(head),curpos);
	} while( curpos && curpos[0]=='#');
	if(!curpos)
		throw "No Portable Bitmap File";
/*
	if(2!=::sscanf(curpos,"%d%d",&width,&height))
		throw "No Portable Bitmap File";
*/
	{
		width =0;
		height =0;
		std::istringstream is(curpos);
		is >> width; 
		is >> height;
		if(width==0 || height ==0 )
			throw "No Portable Bitmap File";
	}

//read bits depth
	bits=1;
	if(type>4)
	{
		do
		{
			curpos=::gets_from_buf(head,sizeof(head),curpos);
		} while( curpos && curpos[0]=='#');
		if(!curpos)
			throw "No Portable Bitmap File";
/*
		if(1!=::sscanf(curpos,"%d",&bits))
			throw "No Portable Bitmap File";
*/
		{
			bits =0;
			std::istringstream is(curpos);
			is >> bits; 
			if(bits==0 )
				throw "No Portable Bitmap File";
		}
		curpos=::gets_from_buf(head,sizeof(head),curpos);
		if(bits!=255 || curpos==0)
			throw "No Portable Bitmap File";
		bits=8;
	}
	data_offset=unsigned(curpos-head);
}
void PBMfile::close()
{
	YFile::_close();
	nullify();
}
//--------------------------------------------------
void PBMfile::read(int x,int y,Byte *line,int len)
{
	switch(type)
	{
	case 5:
		YFile::_read(data_offset+x+y*width,line,len);
		break;
	case 6:
		YFile::_read(data_offset+3*(x+y*width),line,len);
		break;
	};
}
void PBMfile::write(int x,int y,const Byte *line,int len)
{
	switch(type)
	{
	case 5:
		YFile::_write(data_offset+x+y*width,line,len);
		break;
	case 6:
		YFile::_write(data_offset+3*(x+y*width),line,len);
		break;
	};
}
