#ifndef _TIFF_H
#define _TIFF_H
//////////////////////////////////////////////////////////////////////
#include "yfile.h"
//////////////////////////////////////////////////////////////////////
namespace IFF
{
#define LSB					0x4949
#define MSB					0x4d4d
#define NEW_SUBFILE_TYPE_TAG			254
#define SUBFILE_TYPE_TAG			255
#define IMAGE_WIDTH_TAG				256
#define IMAGE_LENGTH_TAG			257
#define BITS_PER_SAMPLE_TAG			258
#define TAG_COMPRESSION				259
#define PHOTOMETRIC_INTERPRETATION_TAG		262
#define STRIPS_OFFSET_TAG 	    		273
#define SAMPLES_PER_PIXEL_TAG   		277
#define ROWS_PER_STRIP_TAG 	    		278
#define STRIPS_BYTE_COUNTS_TAG  		279
#define TAG_PLANAR_CONFIGURATION		284
#define PALETTE_TAG 		    		320
//////////////////////////////////////////////////////////////////////
#define TTYP_BYTE			    	  1
#define TTYP_ASCII  		     		  2
#define TTYP_SHORT  		    	  	  3
#define TTYP_LONG			    	  4
#define TTYP_RATIONAL   	    		  5
//////////////////////////////////////////////////////////////////////
#pragma pack(push,1)
struct TIFF_IFDentry 
{
	uint16_t tag;
	uint16_t type;
	uint32_t length;
	uint32_t value_or_offset;
};
#pragma pack(pop)
//////////////////////////////////////////////////////////////////////
#pragma pack(push,1)
struct TIFF_IFD 
{
	uint16_t entry_count;
	struct TIFF_IFDentry *Entry0;
	uint32_t next_IFD_offset;
};
#pragma pack(pop)
//////////////////////////////////////////////////////////////////////
#pragma pack(push,1)
struct TIFF_Header 
{
	uint16_t byte_order;
	uint16_t version;
	uint32_t IFD0_offset;
};
#pragma pack(pop)
//////////////////////////////////////////////////////////////////////
#pragma pack(push,1)
struct TIFF_ImageParms 
{
	uint32_t StripLengthOffset;
	uint32_t StripOffset;
	uint16_t StripsPerImage; //strips per image
	uint16_t width;
	uint16_t length;
};
#pragma pack(pop)
//////////////////////////////////////////////////////////////////////
#pragma pack(push,1)
struct TiffTag 
{
	uint16_t tag;
	uint16_t type;
	uint32_t length;
	uint32_t value_or_offset;
};
#pragma pack(pop)
//////////////////////////////////////////////////////////////////////


class TiffWriteIFD
: public YFileWrite
{
	uint16_t entries;
	TiffTag ifd[40];
	//------------------
	uint16_t width;
	uint16_t length;
	uint16_t bits;
	uint16_t spi; //strips per image
	uint16_t rowsPS; // RowsPerStrip
	//------------------
protected:
	uint32_t str_offset;
	uint32_t str_length;
	uint32_t pal_offset;
	uint16_t pal_length;
private:
	void writeIFD(void);
	uint16_t fillIFD(void);
public:
	TiffWriteIFD(const char* filename,int wid,int len,int bits_ps,uint32_t *palette=0);
	virtual ~TiffWriteIFD(void) {;}
	int Width(void) { return width; }
	int Length(void) { return length; }
	int StripsPerImage(void) { return spi; }
	int RowsPerStrip(void)   { return rowsPS; }
	int PutPalette(uint32_t *palette=0);
};
//////////////////////////////////////////////////////////////////////

class Tiff8Write
: public TiffWriteIFD
{
private:
	Tiff8Write(const Tiff8Write&);
public:
	Tiff8Write(const char* filename,int wid,int len,int bits_ps,uint32_t *paletta);
	virtual ~Tiff8Write(void) {;}
	int writeStrip(int nStrip,Byte* buf); //returns number bytes readed
	int writeImage(Byte* buf);
};

//////////////////////////////////////////////////////////////////////

class TiffReadIFD
: public YFileRead
{
	uint16_t entries;
	TiffTag ifd[40];
	//------------------
	uint16_t width;
	uint16_t length;
	uint16_t bits;
	uint16_t spi; //strips per image
	uint16_t rowsPS; // RowsPerStrip
	//------------------
protected:
	uint32_t str_offset;
	uint32_t str_length;
	uint32_t pal_offset;
	uint16_t pal_length;
private:
	void readIFD();
public:
	TiffReadIFD(const char* filename);
	virtual ~TiffReadIFD(void) {;}
	int Width(void) { return width; }
	int Length(void) { return length; }
	int Bits(void) { return bits; }
	int StripsPerImage(void) { return spi; }
	int RowsPerStrip(void)   { return rowsPS; }
	int GetPalette(uint32_t *palette);
	int GetPaletteLength() const { return pal_length;}
};
//////////////////////////////////////////////////////////////////////

class Tiff8Read
: public TiffReadIFD
{
	uint32_t *buf_alloc;
	uint32_t *StripLength;
	uint32_t *StripOffset;
private:
	Tiff8Read(const Tiff8Read&);
public:
	Tiff8Read(const char* filename);
	virtual ~Tiff8Read(void);
	int readStrip(int nStrip,Byte* buf); //returns number bytes readed
	int readImage(Byte* buf);
};
} //namespace IFF
//////////////////////////////////////////////////////////////////////
#endif //_TIFF_H
