#ifndef _VTL_IMAGE_H
#define _VTL_IMAGE_H
///////////////////////////////////////////////////////////////////////////////////////////
#include "loki_typeManip.h"
#include "loki_Typelist.h"
#include "loki_static_check.h" 
///////////////////////////////////////////////////////////////////////////////////////////
#include "alexandresku.h"
#include "vtl_meta_base.h"
#include "vtl_image_types.h"
///////////////////////////////////////////////////////////////////////////////////////////

namespace Vtl
{
namespace Img
{
	typedef int offset_type;

	namespace Impl
	{
	namespace Layout
	{
		template<typename T>
		static int align(int sz,int align_to)
		{
			return (sz*sizeof(T)+align_to-1)/sizeof(T);
		}
		template<typename T,int N,int Interleave,int stride_0>
		offset_type build_strides(offset_type* strides,const int* sizes,int align_to,int nbands)
		{
			int prev=(0==Interleave)?nbands:1;
			if(stride_0==0)
			{
				*strides++=prev;
			}
			for(int i=1; i<N ; ++i)
			{
				prev=align<T>(*sizes++,align_to) * prev * ((i==Interleave)?nbands:1);
				*strides++=prev;
			}
			return prev;
		}
	}//namespace Layout
	}//namespace Impl

	/// T_Ndim -  , 2D, 3D  ..  -   .
	/// T_Nbands -  ( ) .  .   ,     RUNTIME.
	/// T_Interleave
	///
	template<typename T,int T_Ndim,int T_Nbands=1,int T_Interleave=0>
	struct Layout;
	///Specialization Compile time known Nbands
	template<typename T,int T_Ndim,int T_Nbands,int T_Interleave>
	struct Layout
	{
		typedef T value_type;
		enum {N=T_Ndim};
		enum {Interleave=(T_Nbands==1)?0:T_Interleave};
		enum {NbandsIsConst=T_Nbands};
		enum {Nbands=T_Nbands};
		enum {stride_0=(Interleave==0)?Nbands:1};
		enum {const_band_stride=(Interleave==0 && Nbands>0)?Nbands:0};
		///  ,         1   i  (x,y,z,...)    .
		enum {strides_array_len=(stride_0)?(N-1):N};
		offset_type strides[strides_array_len];
	public:
		//Access to stride in i dimension. You should optimize the access by cheking if this stride is compile time constant!
		template<int i> int Stride() const {return Stride(Loki::Int2Type<stride_0!=0>(),Loki::Int2Type<i>());}
		template<int i> int Stride(Loki::Int2Type<false> ,Loki::Int2Type<i> )  const { return strides[i];} 
		template<int i> static int Stride(Loki::Int2Type<true> ,Loki::Int2Type<0> )  { return  stride_0;} 
		template<int i> int Stride(Loki::Int2Type<true> ,Loki::Int2Type<i> ) const { return strides[i-1];} 
		template<int i> static bool StrideIsEnum() { return (i==0 && stride_0!=0);} //!< Compile time constant!
		//Adjust layout to some external image
		Layout(offset_type* known_strides)
		{
			Vtl::Meta::copy<strides_array_len>(strides,known_strides);
		}
		// Create default layout for set of parameters assigned
		Layout(const int* sizes,int align_to,int nbands=Nbands)
		{
			nbands=(Nbands)?Nbands:nbands;
			offset_type sz=Impl::Layout::build_strides<value_type,N,Interleave,stride_0>(strides,sizes,align_to,nbands);
			std::cout << "MemBlockSize=" << sz <<std::endl;
		}
	};
	///Specialization RunTime known Nbands
	template<typename T,int T_Ndim,int T_Interleave>
	struct Layout<T,T_Ndim,0,T_Interleave>
	{
		typedef T value_type;
		enum {N=T_Ndim};
		enum {Interleave=T_Interleave};
		enum {NbandsIsConst=0};
		const int Nbands;
		enum {stride_0=(Interleave==0)?0:1};
		enum {const_band_stride=0};
		///  ,         1   i  (x,y,z,...)    .
		enum {strides_array_len=(stride_0)?(N-1):N};
		offset_type strides[strides_array_len];
	public:
		//Access to stride in i dimension. You should optimize the access by cheking if this stride is compile time constant!
		template<int i> int Stride() const {return Stride(Loki::Int2Type<stride_0!=0>(),Loki::Int2Type<i>());}
		template<int i> int Stride(Loki::Int2Type<false> ,Loki::Int2Type<i> )  const { return strides[i];} 
		template<int i> static int Stride(Loki::Int2Type<true> ,Loki::Int2Type<0> )  { return  stride_0;} 
		template<int i> int Stride(Loki::Int2Type<true> ,Loki::Int2Type<i> ) const { return strides[i-1];} 

		template<int i> static bool StrideIsEnum() { return (i==0 && stride_0!=0);} //!< Compile time constant!
		//Adjust layout to some external image
		Layout(int nbands,offset_type* known_strides)
			: Nbands(nbands)
		{
			Vtl::Img::Meta::copy<strides_array_len>(strides,known_strides);
		}
		// Create default layout for set of parameters assigned
		Layout(const int* sizes,int align_to,int nbands)
			: Nbands(nbands)
		{
			nbands=(Nbands)?Nbands:nbands;
			offset_type sz=Impl::Layout::build_strides<value_type,N,Interleave,stride_0>(strides,sizes,align_to,nbands);
			std::cout << "MemBlockSize=" << sz <<std::endl;
		}
	};

	template<typename T,int T_Ndim,int T_Nbands,int T_Layout,int T_Align> 
	class ImageNDref
	{
		typedef T value_type;
	public:
		virtual ~ImageNDref() {}
		ImageNDref()
		{

		}

	};
}//namespace Img
}//namespace Vtl
///////////////////////////////////////////////////////////////////////////////////////////
#endif //_VTL_IMAGE_H
