#ifndef _VTL_IMAGE_TYPES_H
#define _VTL_IMAGE_TYPES_H
///////////////////////////////////////////////////////////////////////////////////////////
#include <vector>
#include "alexandresku.h"
///////////////////////////////////////////////////////////////////////////////////////////
namespace Vtl
{
	namespace Img
	{
//------------- General image info --------------------
		/// image classes --------------------
		enum ImageClass
		{
			Mask,
			Palette,
			Grayscale,
			//Complex,
			TrueColor,
			Multichan
		};
		/// Image layout in memory/file (type of ordering)
		enum Interleave
		{
			SEP=-1,  //!< bands at separated location, NOT COMPATIBLE with ENVI
			BIP=0, //!< band interleaved by pixel 
			BIL=1, //!< band interleaved by line 
			BSQ=2  //!< band sequential
		};
		///Pixel sample data types (fully ENVI-compatible, as in IMG/HDR file)
		///Do NOT change numerical values for format support work correctly!
		enum SampleType
		{
			ST_Unknown=255, //!< any non standart value, ERROR, Not supported by ENVI   
			ST_u8   =1,  //!<     u   1 =    8 bit byte; 
			ST_s8   =254,//!<     s   1 =    8 bit signed byte; NOT ENVI COMPATIBLE! Extension due to TIFF! 
			ST_s16  =2,  //!<     s   2 =   16-bit signed integer; 
			ST_s32  =3,  //!<     s   3 =   32-bit signed long integer; 
			ST_f32  =4,  //!<     f   4 =   32-bit floating point; 
			ST_f64  =5,  //!<     f   5 =   64-bit double precision floating point; 
			ST_c32f =6,  //!<     f   6 = 2x32-bit complex, real-imaginary pair of double precision; ??SINGLE PRECISION?? 
			ST_c64f =9 , //!<     f   9 = 2x64-bit double precision complex, real-imaginary pair of double precision; 
			ST_u16  =12, //!<     u  12 =   16-bit unsigned integer; 
			ST_u32  =13, //!<     u  13 =   32-bit unsigned integer; 
			ST_s64  =14, //!<     s  14 =   64-bit integer; 
			ST_u64  =15  //!<     u  15 =   64-bit unsigned integer.
		};
		template<typename T>	struct complex_type;
		//		{
		//			T real;
		//			T imag;
		//		};
//-------------- Complex type
		template<>	struct complex_type<float>
		{
			float real;
			float imag;
		};
		template<>	struct complex_type<double>
		{
			double real;
			double imag;
		};
		template<int data_type>	struct DataType;
		namespace Impl
		{
			template<class P,SampleType data_type,bool _signed,bool _float,bool _struct>	
			struct DataType
			{
				typedef P sample_type;
				enum { type_id=data_type };
				enum { sample_size=sizeof(P) };
				enum { is_signed=_signed };
				enum { is_float=_float   };
				enum { is_struct=_struct };
			};
		}
		template<>	struct DataType<ST_u8>   : Impl::DataType<uint8_t ,ST_u8 ,false,false,false> {};
		template<>	struct DataType<ST_s8>   : Impl::DataType< int8_t ,ST_s8 ,true ,false,false> {};
		template<>	struct DataType<ST_s16>  : Impl::DataType< int16_t,ST_s16,true ,false,false> {};
		template<>	struct DataType<ST_s32>  : Impl::DataType< int32_t,ST_s32,true ,false,false> {};
		template<>	struct DataType<ST_f32>  : Impl::DataType<  float ,ST_f32,true ,true,false> {};
		template<>	struct DataType<ST_f64>  : Impl::DataType< double ,ST_f64,true ,true,false> {};
		template<>	struct DataType<ST_c32f> : Impl::DataType<complex_type<float>,ST_c32f,true,true,true> {};
		template<>	struct DataType<ST_c64f> : Impl::DataType<complex_type<double>,ST_c64f,true,true,true> {};
		template<>	struct DataType<ST_u16>  : Impl::DataType<uint16_t,ST_u16,false,false,false> {};
		template<>	struct DataType<ST_u32>  : Impl::DataType<uint32_t,ST_u32,false,false,false> {};
		template<>	struct DataType<ST_s64>  : Impl::DataType< int64_t,ST_s64,true,false,false> {};
		template<>	struct DataType<ST_u64>  : Impl::DataType<uint64_t,ST_u64,false,false,false> {};
		/*!     (Sample of Pixel).    
         *    DataType<int TypeId>       
		 *    .    DataTypeId<typename SampleType>.
		 *           DataTypeId<typename SampleType>,
		 *    -    DataType<int TypeId>.
		 *          
		 * DataTypeList2TypeList<>          .
		 ************************************************************************************/
		typedef Vtl::Types::TL::NoDuplicates<TYPELIST_12(DataType<ST_u8>,DataType<ST_s8>,DataType<ST_s16>,DataType<ST_s32>
		   ,DataType<ST_f32>,DataType<ST_f64>,DataType<ST_c32f>,DataType<ST_c64f>
		   ,DataType<ST_u16>,DataType<ST_u32>,DataType<ST_s64>,DataType<ST_u64>) >::Result EnviSample_DataTypesList;


		//!       (see SampleType) .
		template<typename T> struct DataTypeId
		{
		private:
			template<class T1,class T2> struct Comp
			{
				enum { value=Vtl::Types::isSameType<typename T1::sample_type,T2>::value};
			};
			typedef typename Vtl::Types::TL::FindIf<Vtl::Img::EnviSample_DataTypesList,T,Comp>::Result xx_lst;
			typedef typename Vtl::Types::TL::TypeAt<xx_lst,0>::Result typ;
		public:
			enum { value=typ::type_id };
		};

		//!       DataType<>.
		template<typename T> struct T2DataType
		{
			typedef DataType<DataTypeId<T>::value> Result;
		};
		//!    DataType<T>    .
		template<class TList> struct DataTypeList2TypeList;
		template<class T1,class T2> struct DataTypeList2TypeList<Types::Typelist<T1,T2> >
		{
			typedef Types::Typelist<typename T1::sample_type,typename DataTypeList2TypeList<T2>::Result > Result;
		};
		template<class T> struct DataTypeList2TypeList
		{
			typedef T Result;
		};
		typedef DataTypeList2TypeList<EnviSample_DataTypesList>::Result EnviSampleTypesList;


		///    
#pragma pack(push,1)
		struct ImageInfo
		{
			uint32_t width;
			uint32_t height;
			uint32_t planes;
			uint32_t bitsPerSample;
			ImageClass iclass;
			SampleType sample_type_id;
			Interleave interleave;
		};
#pragma pack(pop)
#pragma pack(push,1)
		template<typename T> struct Rgb
		{
			T r;
			T g;
			T b;
		};
#pragma pack(pop)

//------------ image layout in memory/file --------------
#pragma pack(push,1)
		struct SampleAccess
		{
			void* sample_ptr;
			uint32_t reserv;
			uint32_t sample_step;
			uint32_t sample_nextLineShift;
		};
#pragma pack(pop)
#pragma pack(push,1)
		struct PaletteAccess
		{
			void* palette_ptr;
			uint32_t reserv;
			uint32_t palette_type;
			uint32_t palette_entries;
		};
#pragma pack(pop)

#define	IMAGE_DATA_LAYOUT_VERSION 100
#pragma pack(push,1)
		struct ImageDataLayout
		{
		//- stuct definition
			uint32_t ptr_sizeof;
			uint32_t struct_sz;
			uint32_t version;
			uint32_t reserv;
		//- general image container definition
			ImageInfo info;
		//- data access
			PaletteAccess pal;
			SampleAccess sample[1]; //sample[planes]
		};
#pragma pack(pop)
		class ImageDataLayoutHolder
		{
			std::vector<uint8_t> buf;
			ImageDataLayout* idl;
		public:
			ImageDataLayoutHolder(const ImageInfo& inf)
			{
				int sz=sizeof(ImageDataLayout)+(inf.planes-1)*sizeof(SampleAccess);
				buf.resize(sz);
				idl=(ImageDataLayout*)(&(buf[0]));
				idl->ptr_sizeof=sizeof(void*);
				idl->struct_sz=sz;
				idl->version=IMAGE_DATA_LAYOUT_VERSION;
				idl->reserv=0;
				idl->info=inf;
				if(inf.iclass==Img::Palette)
				{
				}
				else
				{
					idl->pal.palette_ptr=0;
					idl->pal.reserv=0;
					idl->pal.palette_type=Img::T2DataType<uint8_t>::Result.type_id;
					idl->pal.palette_entries=0;
				}
				for(uint32_t i=0; i<inf.planes; ++i)
				{
					idl->sample[i].sample_ptr=0;
					idl->sample[i].reserv=0;
					idl->sample[i].sample_step=1;
					idl->sample[i].sample_nextLineShift=inf.width;
				}
			}
			ImageDataLayout* Layout() {return idl; }
			void set_Palette()
			{
			}
			template <typename T> void set_BIP_data(T *data,int samples_per_row)
			{

				idl->info.sample_type=Img::T2DataType<T>::Result;
				for(int i=0; i<idl->info.planes; i++)
				{
					idl->sample[i].reserv=0;
					idl->sample[i].sample_ptr=data+i;
					idl->sample[i].sample_nextLineShift=samples_per_row;
					idl->sample[i].sample_step=idl->info.planes;
				}
			}
			template <typename T> void set_BIL_data(T *data,int samples_per_row)
			{

				idl->info.sample_type_id=Img::SampleType(Img::DataTypeId<T>::value);
				for(size_t i=0; i<idl->info.planes; i++)
				{
					idl->sample[i].reserv=0;
					idl->sample[i].sample_ptr=data+i*samples_per_row;
					idl->sample[i].sample_nextLineShift=samples_per_row*idl->info.planes;
					idl->sample[i].sample_step=1;
				}
			}
			template <typename T> void set_BSQ_data(T *data,int samples_per_row,int next_band_shift=0)
			{

				idl->info.sample_type=Img::T2DataType<T>::Result;
				if(next_band_shift==0)
					next_band_shift=samples_per_row*idl->info.height;
				for(int i=0; i<idl->info.planes; i++)
				{
					idl->sample[i].reserv=0;
					idl->sample[i].sample_ptr=data+i*next_band_shift;
					idl->sample[i].sample_nextLineShift=samples_per_row;
					idl->sample[i].sample_step=1;
				}
			}
			template <typename T> void set_SEP_data(int plane,T *data,int samples_per_row)
			{
				assign( (0<=plane) && (plane <idl->info.planes) );
				idl->info.sample_type=Img::T2DataType<T>::Result;
				int i=plane;
				idl->sample[i].reserv=0;
				idl->sample[i].sample_ptr=data;
				idl->sample[i].sample_nextLineShift=samples_per_row;
				idl->sample[i].sample_step=1;
			}
		};

		/*!
		 *
		 *
		 *********************************************************************************/
		template<typename T> class ImageStorageStrategyNew
		{
		public:
			T* Create(int size)
			{
				return new T[size];
			}
			void Destroy(T* data)
			{
				delete[] data;
			}
		};
		/*!
		 *
		 *
		 *********************************************************************************/
		template<typename P,int planes=1,int interleave=BIL,class AllocStrategy=ImageStorageStrategyNew<P> >
		class ImageHolder
		{
			std::auto_ptr<ImageDataLayout> idl;
		public:
			virtual ~ImageHolder()
			{
			}
			ImageHolder(int width,int height,int bits,ImageClass iclass)
			{
			}
			ImageHolder()
			{
			}
		};
		/*!
		 *
		 *
		 *********************************************************************************/
		template<typename pel, unsigned stride>
		class pic_plane 
		{
		public:
			pic_plane(pel* p) : ptr(p) {}
			operator pel*()          { return ptr; }
			void operator += (int i) { ptr += i*stride; }
			void operator -= (int i) { ptr -= i*stride; }
			void operator ++ ()      { ptr += stride;   }
			void operator -- ()      { ptr -= stride;   }
			pic_plane plane(int p) const { return pic_plane<pel,stride>(ptr+p); }
		//    friend pic_plane operator+(const pic_plane a, uint32_t s)       { return pic_plane(a.ptr+s*a.stride); }
		//    friend bool operator == (const pic_plane a, const pic_plane b)  { return a.ptr == b.ptr; }
		//    friend bool operator != (const pic_plane a, const pic_plane b)  { return a.ptr != b.ptr; }
		private:
			pel *ptr;
		};

		/*!
		 *
		 *
		 *********************************************************************************/
		template<typename P,int planes,int interleave >
		class ImageBase
		{
		protected:
			const ImageDataLayout* idl;
			//ImageDataLayout* idl;
			Vtl::Img::Interleave check_img_interleave(ImageDataLayout* idl_)
			{
/*
		struct ImageDataLayout
		{
		//- stuct definition
			uint32_t ptr_sizeof;
			uint32_t struct_sz;
			uint32_t version;
			uint32_t reserv;
		//- general image container definition
			ImageInfo info;
		//- data access
			PaletteAccess pal;
			SampleAccess sample[1]; //sample[planes]
		};
*/

					

				SampleAccess& sa=idl_->sample[0];
				sa.sample_ptr;
				return Vtl::Img::SEP;
			}
			void check_img_info_propere(const ImageInfo& info)
			{
				if(DataTypeId<P>::value != info.sample_type_id)
				{
					std::ostringstream os;

					throw os.str();
				}
				if(planes != info.planes)
				{
					std::ostringstream os;

					throw os.str();
				}
			}
			ImageDataLayout* check_idl_propere(ImageDataLayout* idl_)
			{
				assert(idl_!=0);
				check_img_info_propere(idl_->info);
				if(planes!=1)
				{ //  planes==1 -   interleave    
					check_img_interleave(idl_);
				}
				return idl_;
			}
		public:
			virtual ~ImageBase() {}
			ImageBase(ImageDataLayout* idl_)
				: idl(check_idl_propere(idl_))
			{
			}
			int Width() const {return idl->info.width;}
			int Height() const {return idl->info.height;}
			int Planes() const {return idl->info.planes;}
			int Bits() const {return idl->info.bitsPerSample;}
			ImageClass iClass() const {return idl->info.iclass;}
		};
		// ,   BSQ  BIL,
		//   "planes=1"   "interleave"
		template<typename P,int planes=1,int interleave=BSQ >
		class Image
			: public ImageBase<P,planes,interleave>
		{
			P* raw_data;
			int nls;
			//ImageDataLayout* idl;
			void check_img_info_propere()
			{
				if(DataTypeId<P>::value != idl_->info.sample_type)
				{
					std::ostringstream os;

					throw os.str();
				}
			}
		public:
			virtual ~Image() {}
			Image(ImageDataLayout* idl_)
				: ImageBase<P,planes,interleave>(idl_)
			{
				raw_data=(P*)(idl->sample[0].sample_ptr);
				nls=idl->sample[0].sample_nextLineShift;
			}
			P* operator[](int y) {return raw_data+nls*y;}
			Image<P,1,BSQ> GetPlane(int plane) 
			{
			}
		};
		//    BIP
		template<typename P,int planes >
		class Image<P,planes,BIP >
			: public ImageBase<P,planes,BIP>
		{
		public:
			virtual ~Image() {}
			Image(ImageDataLayout* idl_)
				: ImageBase(idl_)
			{
			}
		};
		//    SEP
		template<typename P,int planes >
		class Image<P,planes,SEP >
			: public ImageBase<P,planes,SEP>
		{
		public:
			virtual ~Image() {}
			Image(ImageDataLayout* idl_)
				: ImageBase(idl_)
			{
			}
		};
	};
};//namespace Vtl
///////////////////////////////////////////////////////////////////////////////////////////
#endif //_VTL_IMAGE_TYPES_H
