#include<math.h>
#include "ytypes.h"
#include "ymem.h"
#include "aimage.h"
#include "progrcnt.h"
#include "imresamp.h"
#include "ygeometr.h"
////////////////////////////////////////////////////////////////////////
#define	FILL_STR_FRG(fill,len,ptr,val) 	\
        {	                              \
          int ll=(len);                 \
          if((fill) && (ll>0))          \
          {                             \
             while(ll--)                \
							 *(ptr)++=(val);	        \
          }                             \
					else	                        \
					{                             \
						(ptr)+=ll;                  \
          }                             \
        }
////////////////////////////////////////////////////////////////////////
// ImageResampleT template
//////////////////////////////////////////////////////////////////////////////////
template<class ImageType, class PixelType>
class ImageResampleT
{
private:
	ImageResample::ImageResampleParams& prms;
private:
	int reflect( const ImageType* img, const int x, const int y, const int w, const int h);

	inline int bilinear( PixelType* pix, double dx, double dy, int nls);
	inline int bilinearReflect( const ImageType* from, double dx, double dy, int x, int y, int w, int h);
	inline int bicubical( PixelType* pix, double dx, double dy, int nls);
	inline int bicubicalReflect( const ImageType* from, double dx, double dy, int x, int y, int w, int h);

	bool imageResizeNearest( ImageType* to);
	bool imageResizeBilinear( ImageType* to);
	bool imageResizeBicubical( ImageType* to);
	bool imageResizeAverage( ImageType* to);
	bool imageResizeHistMax( ImageType* to);
	bool imageRotateNearest( ImageType* to);
	bool imageRotateBilinear( ImageType* to);
	bool imageRotateBicubical( ImageType* to);
public:
	ImageResampleT( ImageResample::ImageResampleParams& _prms) : prms(_prms) {}
	virtual ~ImageResampleT() {}
	static bool init( ImageResample::ImageResampleParams& prms);
	bool run( ImageType* to);
};
//////////////////////////////////////////////////////////////////////////////////
class RotateEpsBounds
{
	YPointInt tro, tlo, bro, blo, tri, tli, bri, bli;
public:
	RotateEpsBounds( const int eps, int w, int h,int shx,int shy,double radians);
	void calculateLengths( int y, int w, int* lengths);
};
//////////////////////////////////////////////////////////////////////////////////
RotateEpsBounds::RotateEpsBounds( const int eps, int w, int h,int shx,int shy,double radians)
{
	const double sn  = ::sin(radians);
	const double csn = ::cos(radians);
	blo = YPointInt(int( (-eps)*csn -  (-eps)*sn + 0.5) + shx, int( (-eps)*sn +  (-eps)*csn + 0.5) + shy);
	tlo = YPointInt(int( (-eps)*csn - (h+eps)*sn + 0.5) + shx, int( (-eps)*sn + (h+eps)*csn + 0.5) + shy);
	bro = YPointInt(int((w+eps)*csn -  (-eps)*sn + 0.5) + shx, int((w+eps)*sn +  (-eps)*csn + 0.5) + shy);
	tro = YPointInt(int((w+eps)*csn - (h+eps)*sn + 0.5) + shx, int((w+eps)*sn + (h+eps)*csn + 0.5) + shy);
	bli = YPointInt(int(  (eps)*csn -   (eps)*sn + 0.5) + shx, int(  (eps)*sn +   (eps)*csn + 0.5) + shy);
	tli = YPointInt(int(  (eps)*csn - (h-eps)*sn + 0.5) + shx, int(  (eps)*sn + (h-eps)*csn + 0.5) + shy);
	bri = YPointInt(int((w-eps)*csn -   (eps)*sn + 0.5) + shx, int((w-eps)*sn +   (eps)*csn + 0.5) + shy);
	tri = YPointInt(int((w-eps)*csn - (h-eps)*sn + 0.5) + shx, int((w-eps)*sn + (h-eps)*csn + 0.5) + shy);
}
//--------------------------------
void RotateEpsBounds::calculateLengths( int y, int w, int* lengths)
{
	int xo1 = w;
	int	xo2 = 0;
	int x_tmp;
	if( y <= MAX( tlo.y, tro.y) && y >= MIN( tlo.y, tro.y) )
	{
		x_tmp = LineFun::getIntersectionPointBres4X( tlo, tro, y); 
		if( x_tmp < xo1 )
			xo1 = x_tmp;

		if( (tlo.y-tro.y) == 0 ) x_tmp = LineFun::getIntersectionPointBres4X( tro, tlo, y);

		if( x_tmp > xo2 )
			xo2 = x_tmp;
	}
	if( y <= MAX( tro.y, bro.y) && y >= MIN( tro.y, bro.y) )
	{
		x_tmp = LineFun::getIntersectionPointBres4X( tro, bro, y); 
		if( x_tmp < xo1 )
			xo1 = x_tmp;

		if( (tro.y-bro.y) == 0 ) x_tmp = LineFun::getIntersectionPointBres4X( bro, tro, y); 

		if( x_tmp > xo2 )
			xo2 = x_tmp;
	}
	if( y <= MAX( bro.y, blo.y) && y >= MIN( bro.y, blo.y) )
	{
		x_tmp = LineFun::getIntersectionPointBres4X( bro, blo, y); 
		if( x_tmp < xo1 )
			xo1 = x_tmp;

		if( (bro.y-blo.y) == 0 ) x_tmp = LineFun::getIntersectionPointBres4X( blo, bro, y); 

		if( x_tmp > xo2 )
			xo2 = x_tmp;
	}
	if( y <= MAX( blo.y, tlo.y) && y >= MIN( blo.y, tlo.y) )
	{
		x_tmp = LineFun::getIntersectionPointBres4X( blo, tlo, y); 
		if( x_tmp < xo1 )
			xo1 = x_tmp;

		if( (tlo.y-blo.y) == 0 ) x_tmp = LineFun::getIntersectionPointBres4X( tlo, blo, y); 

		if( x_tmp > xo2 )
			xo2 = x_tmp;
	}

	int xi1 = w;
	int xi2 = 0;
	if( y <= MAX( tli.y, tri.y) && y >= MIN( tli.y, tri.y) )
	{
		x_tmp = LineFun::getIntersectionPointBres4X( tli, tri, y); 
		if( x_tmp < xi1 )
			xi1 = x_tmp;

		if( (tli.y-tri.y) == 0 ) x_tmp = LineFun::getIntersectionPointBres4X( tri, tli, y);

		if( x_tmp > xi2 )
			xi2 = x_tmp;
	}
	if( y <= MAX( tri.y, bri.y) && y >= MIN( tri.y, bri.y) )
	{
		x_tmp = LineFun::getIntersectionPointBres4X( tri, bri, y); 
		if( x_tmp < xi1 )
			xi1 = x_tmp;

		if( (tri.y-bri.y) == 0 ) x_tmp = LineFun::getIntersectionPointBres4X( bri, tri, y); 

		if( x_tmp > xi2 )
			xi2 = x_tmp;
	}
	if( y <= MAX( bri.y, bli.y) && y >= MIN( bri.y, bli.y) )
	{
		x_tmp = LineFun::getIntersectionPointBres4X( bri, bli, y); 
		if( x_tmp < xi1 )
			xi1 = x_tmp;

		if( (bri.y-bli.y) == 0 ) x_tmp = LineFun::getIntersectionPointBres4X( bli, bri, y); 

		if( x_tmp > xi2 )
			xi2 = x_tmp;
	}
	if( y <= MAX( bli.y, tli.y) && y >= MIN( bli.y, tli.y) )
	{
		x_tmp = LineFun::getIntersectionPointBres4X( bli, tli, y); 
		if( x_tmp < xi1 )
			xi1 = x_tmp;

		if( (tli.y-bli.y) == 0 ) x_tmp = LineFun::getIntersectionPointBres4X( tli, bli, y); 

		if( x_tmp > xi2 )
			xi2 = x_tmp;
	}

	const bool skip = ( xi1 == w && xi2 == 0 ) ? true : false;
	lengths[0] = MAX(xo1,0);
	if( skip )
	{
		lengths[1] = MIN(w,xo2) - MAX(xo1,0);
		lengths[2] = 0;
		lengths[3] = 0;
	}
	else 
	{
		lengths[1] = xi1 - MAX(xo1,0);
		lengths[2] = xi2 - xi1;
		lengths[3] = MIN(w,xo2) - xi2;
	}
	lengths[4] = w - MIN(xo2,w);

	return;
}
//////////////////////////////////////////////////////////////////////////////////
//
// ImageResampleT functions
//
//////////////////////////////////////////////////////////////////////////////////
template<class ImageType, class PixelType>
bool ImageResampleT<ImageType, PixelType>::init( ImageResample::ImageResampleParams& prms)
{
	switch(prms.mtype)
	{
	case ImageResample::mtResize:
		{
			if( prms.new_width <= 0
			 || prms.new_height <= 0
			 || prms.imeth == ImageResample::imUnknown )
				return false;
			prms.to_width = prms.new_width;
			prms.to_height = prms.new_height;
		}
		break;
	case ImageResample::mtRotate:
		{
			if( prms.imeth == ImageResample::imUnknown
			 || unsigned int(prms.fill_int) > ((1U<<prms.from->Bits()) - 1) )
				return false;
			const int h = prms.from->Height();
			const int w = prms.from->Width();

			const double sn  = ::sin(prms.radians);
			const double csn = ::cos(prms.radians);
			YPointInt bl(int( (-2)*csn -  (-2)*sn + 0.5), int( (-2)*sn +  (-2)*csn + 0.5));
			YPointInt tl(int( (-2)*csn - (h+2)*sn + 0.5), int( (-2)*sn + (h+2)*csn + 0.5));
			YPointInt br(int((w+2)*csn -  (-2)*sn + 0.5), int((w+2)*sn +  (-2)*csn + 0.5));
			YPointInt tr(int((w+2)*csn - (h+2)*sn + 0.5), int((w+2)*sn + (h+2)*csn + 0.5));
			//sn  = ::sin(-prms.radians);
			//csn = ::cos(-prms.radians);
			const int maxy = MAX(MAX(tl.y,bl.y),MAX(br.y,tr.y));
			const int miny = MIN4(tl.y,bl.y,br.y,tr.y);
			const int maxx = MAX4(tl.x,bl.x,br.x,tr.x);
			const int minx = MIN4(tl.x,bl.x,br.x,tr.x);
			prms.shiftx = ((minx<0)? ABS(minx) : 0)+1;
			prms.shifty = ((miny<0)? ABS(miny) : 0)+1;
			prms.to_width = maxx-minx;
			prms.to_height = maxy-miny;
		}
		break;
	default:
		throw "ImageResampleT<ImageType, PixelType>::init() : Invalid operation - neither resize nor rotate";
	};

	return true;
}

template<class ImageType, class PixelType>
bool ImageResampleT<ImageType, PixelType>::run( ImageType* to)
{
	bool ok;
	if( prms.mtype == ImageResample::mtRotate )
		switch(prms.imeth)
		{
		case ImageResample::imNearest:
			ok = imageRotateNearest( to);
			break;
		case ImageResample::imBilinear:
			ok = imageRotateBilinear( to);
			break;
		case ImageResample::imBicubical:
			ok = imageRotateBicubical( to);
			break;
		}
	if( prms.mtype == ImageResample::mtResize )
	{
//		if( prms.to_height == prms.from->Height() && prms.to_width == prms.from->Width() )
//			to = dynamic_cast<const ImageType*>(prms.from); /*FIXME*/
//		else
			switch(prms.imeth)
			{
			case ImageResample::imNearest:
				ok = imageResizeNearest( to);
				break;
			case ImageResample::imBilinear:
				ok = imageResizeBilinear( to);
				break;
			case ImageResample::imBicubical:
				ok = imageResizeBicubical( to);
				break;
			case ImageResample::imAverage:
				ok = imageResizeAverage( to);
				break;
			case ImageResample::imHistMax:
				ok = imageResizeHistMax( to);
				break;
			default:
				;
			};
	}

	if(!ok)
		return false; // Interrpupted by user

	AImage::CopyPalette(to,prms.from);
	return true;
}


template<class ImageType, class PixelType> 
bool ImageResampleT<ImageType, PixelType>::imageRotateNearest( ImageType* to)
{
	const int h = prms.from->Height();
	const int w = prms.from->Width();
	const double  sn = ::sin(-prms.radians);
	const double csn = ::cos(-prms.radians);
	const int eps = 4;

	RotateEpsBounds Bounds(eps,w,h,prms.shiftx,prms.shifty,prms.radians);

	PixelType* pix = &(*to)[0][0];

	const int _fill_int = prms.fill_int;
	const bool _fill = prms.fill;

	const int to_h = prms.to_height;
	const int to_w = prms.to_width;
	const int shx = prms.shiftx;
	const int shy = prms.shifty;

	const ImageType* from = dynamic_cast<const ImageType*>(prms.from);

	ProgressCounter* indicator = prms.indicator;
	if( indicator )
	{
			indicator->start(indicator->russian()?"  ( )":"Rotating image (nearest neighbour)", 0, to_h - 1);
	}

	for( int y = 0; y < to_h; y++)
	{
		if(indicator && (y%40 == 0) && indicator->current(y))
		{
			indicator->finish();
			return false;
		}
		int lengths[5];
		int i;
		int x = 0;
		int xf = 0;
		Bounds.calculateLengths( y, to_w, lengths);
		// #### fill ####
		FILL_STR_FRG(_fill,lengths[0],pix,_fill_int);
		x += lengths[0];
		xf = x + lengths[1];
		
		// #### interpolate with bounds check ####
		if( lengths[1])
		{
			for( i = x; i < xf; i++)
			{
				const int dshx = i - shx;
				const int dshy = y - shy;
				const int rxt = int(dshx*csn - dshy*sn  + 0.5);
				const int ryt = int(dshx*sn  + dshy*csn + 0.5);
				if( rxt < 0 || rxt >= w || ryt < 0 || ryt >= h )
				{
					if(_fill) 
						*pix++ = _fill_int;
				}
				else
				{
					*pix++ = (*from)[ryt][rxt];
				}
			}
		}
		x = xf;
		xf += lengths[2];
		// #### interpolate without bounds check ####
		if( lengths[2] )
		{
			for( i = x; i < xf; i++)
			{
				const int dshx = i - shx;
				const int dshy = y - shy;
				const int rxt = int(dshx*csn - dshy*sn  + 0.5);
				const int ryt = int(dshx* sn + dshy*csn + 0.5);
				*pix++ = (*from)[ryt][rxt];
			}
		}
		x = xf;
		xf += lengths[3];
		// #### interpolate with bounds check ####
		if( lengths[3] )
		{
			for( i = x; i < xf; i++)
			{
				const int dshx = i - shx;
				const int dshy = y - shy;
				const int rxt = int(dshx*csn - dshy*sn  + 0.5);
				const int ryt = int(dshx*sn  + dshy*csn + 0.5);
				if( rxt < 0 || rxt >= w || ryt < 0 || ryt >= h )
				{
					if(_fill) 
						*pix++ = _fill_int;
				}
				else
				{
					*pix++ = (*from)[ryt][rxt];
				}
			}
		}
		// #### fill ####
		FILL_STR_FRG(_fill,lengths[4],pix,_fill_int);
	}
	if( indicator && indicator->InProgress() )
		indicator->finish();
	return true;
}

template<class ImageType, class PixelType> 
bool ImageResampleT<ImageType, PixelType>::imageRotateBilinear( ImageType* to)
{
	const int h = prms.from->Height();
	const int w = prms.from->Width();
	const int eps = 6;
	RotateEpsBounds Bounds(eps,w,h,prms.shiftx,prms.shifty,prms.radians);
	const double sn = ::sin(-prms.radians);
	const double csn = ::cos(-prms.radians);

	PixelType* pix = &(*to)[0][0];

	const int boundw = w - 1;
	const int boundh = h - 1;

	const int MAX_PIXEL = (1<<prms.to_bits) - 1;
	const int _fill_int = prms.fill_int;
	const bool _fill = prms.fill;

	const int to_h = prms.to_height;
	const int to_w = prms.to_width;
	const int shx = prms.shiftx;
	const int shy = prms.shifty;

	const ImageType* from = dynamic_cast<const ImageType*>(prms.from);
	const int nls = from->NextLineShift();

	ProgressCounter* indicator = prms.indicator;
	if( indicator )
	{
		indicator->start(indicator->russian()?"  ( )":"Rotating image (bilinear)", 0, to_h - 1);
	}

	for( int y = 0; y < to_h; y++)
	{
		if(indicator && (y%40 == 0) && indicator->current(y))
		{
			indicator->finish();
			return false;
		}
		int lengths[5];
		int i;
		int x = 0;
		int xf = 0;
		Bounds.calculateLengths( y, to_w, lengths);
		// #### fill ####
		FILL_STR_FRG(_fill,lengths[0],pix,_fill_int);
		x += lengths[0];
		xf = x + lengths[1];
		
		// #### interpolate with bounds check ####
		if( lengths[1])
		{
			for( i = x; i < xf; i++)
			{
				const int dshx = i - shx;
				const int dshy = y - shy;
				const double xt = dshx*csn - dshy*sn  + 0.5;
				const double yt = dshx *sn + dshy*csn + 0.5;
				const int rxt = int(xt);
				const int ryt = int(yt);
				if( rxt < 0 || rxt >= w || ryt < 0 || ryt >= h )
				{
					if(_fill) 
						*pix++ = _fill_int;
				}
				else
				{
					const int _x = int(xt+0.5);
					const int _y = int(yt+0.5);
					const double dx = xt - (_x - 0.5);
					const double dy = yt - (_y - 0.5);
					if( rxt <= 1 || rxt >= boundw || ryt <= 1 || ryt >= boundh )
					{ // bilinear approximation on boundaries
						const int ires = bilinearReflect( from, dx, dy, _x, _y, w, h);
						*pix++ = IN_BOUNDS(ires,0,MAX_PIXEL);
					}
					else
					{
						PixelType* _pix = &(*from)[_y-1][_x-1];
						const int ires = bilinear( _pix, dx, dy, nls);
						*pix++ = IN_BOUNDS(ires,0,MAX_PIXEL);
					}
				}
			}
		}
		x = xf;
		xf += lengths[2];
		// #### interpolate without bounds check ####
		if( lengths[2] )
		{
			for( i = x; i < xf; i++)
			{
				const int dshx = i - shx;
				const int dshy = y - shy;
				const double xt = dshx*csn - dshy*sn  + 0.5;
				const double yt = dshx*sn  + dshy*csn + 0.5;
				const int _x = int(xt+0.5);
				const int _y = int(yt+0.5);
				const double dx = xt - (_x - 0.5);
				const double dy = yt - (_y - 0.5);
				PixelType* _pix = &(*from)[_y-1][_x-1];
				const int ires = bilinear( _pix, dx, dy, nls);
				*pix++ = IN_BOUNDS(ires,0,MAX_PIXEL);
			}
		}
		x = xf;
		xf += lengths[3];
		// #### interpolate with bounds check ####
		if( lengths[3] )
		{
			for( i = x; i < xf; i++)
			{
				const int dshx = i - shx;
				const int dshy = y - shy;
				const double xt = dshx*csn - dshy*sn  + 0.5;
				const double yt = dshx*sn  + dshy*csn + 0.5;
				const int rxt = int(xt);
				const int ryt = int(yt);
				if( rxt < 0 || rxt >= w || ryt < 0 || ryt >= h )
				{
					if(_fill) 
						*pix++ = _fill_int;
				}
				else
				{
					const int _x = int(xt+0.5);
					const int _y = int(yt+0.5);
					const double dx = xt - (_x - 0.5);
					const double dy = yt - (_y - 0.5);
					if( rxt <= 1 || rxt >= boundw || ryt <= 1 || ryt >= boundh )
					{ // bilinear approximation on boundaries
						const int ires = bilinearReflect( from, dx, dy, _x, _y, w, h);
						*pix++ = IN_BOUNDS(ires,0,MAX_PIXEL);
					}
					else
					{
						PixelType* _pix = &(*from)[_y-1][_x-1];
						const int ires = bilinear( _pix, dx, dy, nls);
						*pix++ = IN_BOUNDS(ires,0,MAX_PIXEL);
					}
				}
			}
		}
		// #### fill ####
		FILL_STR_FRG(_fill,lengths[4],pix,_fill_int);
	}
	if( indicator && indicator->InProgress() )
		indicator->finish();
	return true;
}

template<class ImageType, class PixelType> 
bool ImageResampleT<ImageType, PixelType>::imageRotateBicubical( ImageType* to)
{
	const int h = prms.from->Height();
	const int w = prms.from->Width();
	const int eps = 5;
	RotateEpsBounds Bounds(eps,w,h,prms.shiftx,prms.shifty,prms.radians);
	const double sn = ::sin(-prms.radians);
	const double csn = ::cos(-prms.radians);

	PixelType* pix = &(*to)[0][0];

	const int boundw = w - 2;
	const int boundh = h - 2;

	const int MAX_PIXEL = (1<<prms.to_bits) - 1;
	const int _fill_int = prms.fill_int;
	const bool _fill = prms.fill;

	const int to_h = prms.to_height;
	const int to_w = prms.to_width;
	const int shx = prms.shiftx;
	const int shy = prms.shifty;

	const ImageType* from = dynamic_cast<const ImageType*>(prms.from);
	const int nls = from->NextLineShift();

	ProgressCounter* indicator = prms.indicator;
	if( indicator )
	{
		indicator->start(indicator->russian()?"  ( )":"Rotating image (bicubical)", 0, to_h - 1);
	}

	for( int y = 0; y < to_h; y++)
	{
		if(indicator && (y%40 == 0) && indicator->current(y))
		{
			indicator->finish();
			return false;
		}
		int lengths[5];
		int i;
		int x = 0;
		int xf = 0;
		Bounds.calculateLengths( y, to_w, lengths);
		// #### fill ####
		FILL_STR_FRG(_fill,lengths[0],pix,_fill_int);
		x += lengths[0];
		xf = x + lengths[1];
		// #### interpolate with bounds check ####
		if( lengths[1])
		{
			for( i = x; i < xf; i++)
			{
				const int dshx = i - shx;
				const int dshy = y - shy;
				const double xt = dshx*csn - dshy*sn  + 0.5;
				const double yt = dshx*sn  + dshy*csn + 0.5;
				const int rxt = int(xt);
				const int ryt = int(yt);
				if( rxt < 0 || rxt >= w || ryt < 0 || ryt >= h )
				{
					if(_fill) 
						*pix++ = _fill_int;
				}
				else
				{
					const int _x = int(xt+0.5);
					const int _y = int(yt+0.5);
					const double dx = xt - (_x - 1.5);
					const double dy = yt - (_y - 1.5);
					if( rxt <= 2 || rxt >= boundw || ryt <= 2 || ryt >= boundh )
					{ // bicubical on boundaries
						const int ires = bicubicalReflect( from, dx, dy, _x, _y, w, h);
						*pix++ = IN_BOUNDS( ires, 0, MAX_PIXEL);
					}
					else
					{ // bicubical inside
						PixelType* _pix = &(*from)[_y-2][_x-2];
						const int ires = bicubical( _pix, dx, dy, nls);
						*pix++ = IN_BOUNDS( ires, 0, MAX_PIXEL);
					}
				}
			}
		}
		x = xf;
		xf += lengths[2];
		// #### interpolate without bounds check ####
		if( lengths[2] )
		{
			for( i = x; i < xf; i++)
			{
				const int dshx = i - shx;
				const int dshy = y - shy;
				const double xt = dshx*csn - dshy*sn  + 0.5;
				const double yt = dshx*sn  + dshy*csn + 0.5;
				const int _x = int(xt+0.5);
				const int _y = int(yt+0.5);
				const double dx = xt - (_x - 1.5);
				const double dy = yt - (_y - 1.5);
				PixelType* _pix = &(*from)[_y-2][_x-2];
				const int ires = bicubical( _pix, dx, dy, nls);
				*pix++ = IN_BOUNDS( ires, 0, MAX_PIXEL);
			}
		}
		x = xf;
		xf += lengths[3];
		// #### interpolate with bounds check ####
		if( lengths[3] )
		{
			for( i = x; i < xf; i++)
			{
				const int dshx = i - shx;
				const int dshy = y - shy;
				const double xt = dshx*csn - dshy*sn  + 0.5;
				const double yt = dshx*sn  + dshy*csn + 0.5;
				const int rxt = int(xt);
				const int ryt = int(yt);
				if( rxt < 0 || rxt >= w || ryt < 0 || ryt >= h )
				{
					if(_fill) 
						*pix++ = _fill_int;
				}
				else
				{
					const int _x = int(xt+0.5);
					const int _y = int(yt+0.5);
					const double dx = xt - (_x - 1.5);
					const double dy = yt - (_y - 1.5);
					if( rxt <= 2 || rxt >= boundw || ryt <= 2 || ryt >= boundh )
					{ // bicubical on boundaries
						const int ires = bicubicalReflect( from, dx, dy, _x, _y, w, h);
						*pix++ = IN_BOUNDS( ires, 0, MAX_PIXEL);
					}
					else
					{ // bicubical inside
						PixelType* _pix = &(*from)[_y-2][_x-2];
						const int ires = bicubical( _pix, dx, dy, nls);
						*pix++ = IN_BOUNDS( ires, 0, MAX_PIXEL);
					}
				}
			}
		}
		// #### fill ####
		FILL_STR_FRG(_fill,lengths[4],pix,_fill_int);
	}
	if( indicator && indicator->InProgress() )
		indicator->finish();
	return true;
}

template<class ImageType, class PixelType>
bool ImageResampleT<ImageType, PixelType>::imageResizeNearest( ImageType* to)
{
	const int h = prms.from->Height();
	const int w = prms.from->Width();

	PixelType* pix = &(*to)[0][0];

	const int to_h = prms.to_height;
	const int to_w = prms.to_width;

	const ImageType* from = dynamic_cast<const ImageType*>(prms.from);

	ProgressCounter* indicator = prms.indicator;
	if( indicator )
	{
		indicator->start(indicator->russian()?"   ( )":"Resize image (nearest neighbour)", 0, to_h - 1);
	}
	for( int y = 0; y < to_h; y++)
	{
		if(indicator && (y%40 == 0) && indicator->current(y))
		{
			indicator->finish();
			return false;
		}
		for(int x = 0; x < to_w; x++)
		{//@@Yurin?????  double ????
			const double xt = double(w)*double(x)/double(to_w);
			const double yt = double(h)*double(y)/double(to_h);
			*pix++ = (*from)[int(yt)][int(xt)];
		}
	}
	if( indicator && indicator->InProgress() )
		indicator->finish();
	return true;
}

template<class ImageType, class PixelType>
bool ImageResampleT<ImageType, PixelType>::imageResizeBilinear( ImageType* to)
{
	const int h = prms.from->Height();
	const int w = prms.from->Width();

	PixelType* pix = &(*to)[0][0];

	const int boundw = w - 1;
	const int boundh = h - 1;

	const int MAX_PIXEL = (1<<prms.to_bits) - 1;

	const int to_h = prms.to_height;
	const int to_w = prms.to_width;

	const ImageType* from = dynamic_cast<const ImageType*>(prms.from);
	const int nls = from->NextLineShift();
//@@YURIN ???          ???
	int epsX = int(2*to_w/w +0.5);
	if( epsX == 0 ) epsX = 1;
	int epsY = int(2*to_h/h +0.5);
	if( epsY == 0 ) epsY = 1;
	
	ProgressCounter* indicator = prms.indicator;
	if( indicator )
	{
		indicator->start(indicator->russian()?"   ( )":"Resize image (bilinear)", 0, to_h - 1);
	}

	for( int y = 0; y < to_h; y++)
	{
		if(indicator && (y%40 == 0) && indicator->current(y))
		{
			indicator->finish();
			return false;
		}
		int lengths[3];
		int i;
		int x = 0;
		if( y <= epsY || y >= (to_h - epsY) || to_w <= 2*epsX )
		{
			lengths[0] = to_w;
			lengths[1] = 0;
			lengths[2] = 0;
		}
		else
		{
			lengths[0] = epsX;
			lengths[1] = to_w - 2*epsX;
			lengths[2] = epsX;
		}

		int xf = lengths[0];
		
		// #### interpolate with bounds check ####
		if( lengths[0])
		{
			for( i = x; i < xf; i++)
			{
				const double xt = double(w)*double(i)/double(to_w);
				const double yt = double(h)*double(y)/double(to_h);
				const int rxt = int(xt);
				const int ryt = int(yt);
				const int _x = int(xt+0.5);
				const int _y = int(yt+0.5);
				const double dx = xt - (_x - 0.5);
				const double dy = yt - (_y - 0.5);
				if( rxt <= 1 || rxt >= boundw || ryt <= 1 || ryt >= boundh )
				{ // bilinear approximation on boundaries
					const int ires = bilinearReflect( from, dx, dy, _x, _y, w, h);
					*pix++ = IN_BOUNDS(ires,0,MAX_PIXEL);
				}
				else
				{
					PixelType* _pix = &(*from)[_y-1][_x-1];
					const int ires = bilinear( _pix, dx, dy, nls);
					*pix++ = IN_BOUNDS(ires,0,MAX_PIXEL);
				}
			}
		}
		x = xf;
		xf += lengths[1];
		// #### interpolate without bounds check ####
		if( lengths[1] )
		{
			for( i = x; i < xf; i++)
			{
				const double xt = double(w)*double(i)/double(to_w);
				const double yt = double(h)*double(y)/double(to_h);
				const int _x = int(xt+0.5);
				const int _y = int(yt+0.5);
				const double dx = xt - (_x - 0.5);
				const double dy = yt - (_y - 0.5);
				PixelType* _pix = &(*from)[_y-1][_x-1];
				const int ires = bilinear( _pix, dx, dy, nls);
				*pix++ = IN_BOUNDS(ires,0,MAX_PIXEL);
			}
		}
		x = xf;
		xf += lengths[2];
		// #### interpolate with bounds check ####
		if( lengths[2] )
		{
			for( i = x; i < xf; i++)
			{
				const double xt = double(w)*double(i)/double(to_w);
				const double yt = double(h)*double(y)/double(to_h);
				const int rxt = int(xt);
				const int ryt = int(yt);
				const int _x = int(xt+0.5);
				const int _y = int(yt+0.5);
				const double dx = xt - (_x - 0.5);
				const double dy = yt - (_y - 0.5);
				if( rxt <= 1 || rxt >= boundw || ryt <= 1 || ryt >= boundh )
				{ // bilinear approximation on boundaries
					const int ires = bilinearReflect( from, dx, dy, _x, _y, w, h);
					*pix++ = IN_BOUNDS(ires,0,MAX_PIXEL);
				}
				else
				{
					PixelType* _pix = &(*from)[_y-1][_x-1];
					const int ires = bilinear( _pix, dx, dy, nls);
					*pix++ = IN_BOUNDS(ires,0,MAX_PIXEL);
				}
			}
		}
	}
	if( indicator && indicator->InProgress() )
		indicator->finish();
	return true;
}

template<class ImageType, class PixelType>
bool ImageResampleT<ImageType, PixelType>::imageResizeBicubical( ImageType* to)
{
	const int h = prms.from->Height();
	const int w = prms.from->Width();

	PixelType* pix = &(*to)[0][0];

	const int boundw = w - 2;
	const int boundh = h - 2;

	const int MAX_PIXEL = (1<<prms.to_bits) - 1;

	const int to_h = prms.to_height;
	const int to_w = prms.to_width;

//@@YURIN ???          ???
	int epsX = int(4*to_w/w + 0.5);
	if( epsX == 0 ) epsX = 1;
	int epsY = int(4*to_h/h + 0.5);
	if( epsY == 0 ) epsY = 1;

	const ImageType* from = dynamic_cast<const ImageType*>(prms.from);
	const int nls = from->NextLineShift();

	ProgressCounter* indicator = prms.indicator;
	if( indicator )
	{
		indicator->start(indicator->russian()?"   ( )":"Resize image (bicubical)", 0, to_h - 1);
	}
	for( int y = 0; y < to_h; y++)
	{
		if(indicator && (y%40 == 0) && indicator->current(y))
		{
			indicator->finish();
			return false;
		}
		int lengths[5];
		int i;
		int x = 0;
		if( y <= epsY || y >= (to_h - epsY) || to_w <= 2*epsX )
		{
			lengths[0] = to_w;
			lengths[1] = 0;
			lengths[2] = 0;
		}
		else
		{
			lengths[0] = epsX;
			lengths[1] = to_w - 2*epsX;
			lengths[2] = epsX;
		}
		int xf = lengths[0];
		// #### interpolate with bounds check ####
		if( lengths[0])
		{
			for( i = x; i < xf; i++)
			{
				const double xt = double(w)*double(i)/double(to_w);
				const double yt = double(h)*double(y)/double(to_h);
				const int rxt = int(xt);
				const int ryt = int(yt);
				const int _x = int(xt+0.5);
				const int _y = int(yt+0.5);
				const double dx = xt - (_x - 1.5);
				const double dy = yt - (_y - 1.5);
				if( rxt <= 2 || rxt >= boundw || ryt <= 2 || ryt >= boundh )
				{ // bicubical on boundaries
					const int ires = bicubicalReflect( from, dx, dy, _x, _y, w, h);
					*pix++ = IN_BOUNDS( ires, 0, MAX_PIXEL);
				}
				else
				{ // bicubical inside
					PixelType* _pix = &(*from)[_y-2][_x-2];
					const int ires = bicubical( _pix, dx, dy, nls);
					*pix++ = IN_BOUNDS( ires, 0, MAX_PIXEL);
				}
			}
		}
		x = xf;
		xf += lengths[1];
		// #### interpolate without bounds check ####
		if( lengths[1] )
		{
			for( i = x; i < xf; i++)
			{
				const double xt = double(w)*double(i)/double(to_w);
				const double yt = double(h)*double(y)/double(to_h);
				const int _x = int(xt+0.5);
				const int _y = int(yt+0.5);
				const double dx = xt - (_x - 1.5);
				const double dy = yt - (_y - 1.5);
				PixelType* _pix = &(*from)[_y-2][_x-2];
				const int ires = bicubical( _pix, dx, dy, nls);
				*pix++ = IN_BOUNDS( ires, 0, MAX_PIXEL);
			}
		}
		x = xf;
		xf += lengths[2];
		// #### interpolate with bounds check ####
		if( lengths[2] )
		{
			for( i = x; i < xf; i++)
			{
				const double xt = double(w)*double(i)/double(to_w);
				const double yt = double(h)*double(y)/double(to_h);
				const int rxt = int(xt);
				const int ryt = int(yt);
				const int _x = int(xt+0.5);
				const int _y = int(yt+0.5);
				const double dx = xt - (_x - 1.5);
				const double dy = yt - (_y - 1.5);
				if( rxt <= 2 || rxt >= boundw || ryt <= 2 || ryt >= boundh )
				{ // bicubical on boundaries
					const int ires = bicubicalReflect( from, dx, dy, _x, _y, w, h);
					*pix++ = IN_BOUNDS( ires, 0, MAX_PIXEL);
				}
				else
				{ // bicubical inside
					PixelType* _pix = &(*from)[_y-2][_x-2];
					const int ires = bicubical( _pix, dx, dy, nls);
					*pix++ = IN_BOUNDS( ires, 0, MAX_PIXEL);
				}
			}
		}
	}
	if( indicator && indicator->InProgress() )
		indicator->finish();
	return true;
}

template<class ImageType, class PixelType>
bool ImageResampleT<ImageType, PixelType>::imageResizeAverage( ImageType* to)
{
	const int h = prms.from->Height();
	const int w = prms.from->Width();
	PixelType* pix = &(*to)[0][0];
	const int to_h = prms.to_height;
	const int to_w = prms.to_width;

	const int nx = MAX(1,int(double(w)/double(to_w) + 0.5));
	const int ny = MAX(1,int(double(h)/double(to_h) + 0.5));

	const ImageType* from = dynamic_cast<const ImageType*>(prms.from);
	const int nls = from->NextLineShift();

	ProgressCounter* indicator = prms.indicator;
	if( indicator )
	{
		indicator->start(indicator->russian()?"   ()":"Resize image (averaging)", 0, to_h - 1);
	}

	for(int y = 0; y < to_h; y++)
	{
		if(indicator && (y%40 == 0) && indicator->current(y))
		{
			indicator->finish();
			return false;
		}
		for(int x = 0; x < to_w; x++)
		{
			const double xt = double(w)*double(x)/double(to_w);
			const double yt = double(h)*double(y)/double(to_h);
			const int _y = int(yt);
			const int _x = int(xt);

			PixelType* _pix = &(*from)[_y][_x];

			int sum = 0;
			for( int i = 0; i < ny; i++)
			{
				for( int j = 0; j < nx; j++)
				{
					sum += *_pix;
					_pix ++;
				}
				_pix += nls - nx;
			}

			double res = double(sum)/double(nx*ny);

			*pix++ = int(res+0.5);
		}
	}
	if( indicator && indicator->InProgress() )
		indicator->finish();
	return true;
}

template<class ImageType, class PixelType>
bool ImageResampleT<ImageType, PixelType>::imageResizeHistMax( ImageType* to)
{
	const int h = prms.from->Height();
	const int w = prms.from->Width();
//@@Yurin!!!!!!!!!!!!!!!!!!!! stack variable too large !!! 
// Fixed:
	const int max_int = (1<<prms.to_bits);
	int* counts = new int[max_int];
//@@Yurin!!!!!!!!!!!!!!!!!!!! may not allocated - new returns NULL or throw exception
	PtrDelete<int> pd_counts( counts, max_int);
// ---
	PixelType* pix = &(*to)[0][0];

	const int to_h = prms.to_height;
	const int to_w = prms.to_width;

	const int nx = MAX(1,int((2*double(w)+double(to_w))/(2*double(to_w))));
	const int ny = MAX(1,int((2*double(h)+double(to_h))/(2*double(to_h))));

	const ImageType* from = dynamic_cast<const ImageType*>(prms.from);
	const int nls = from->NextLineShift();

	ProgressCounter* indicator = prms.indicator;
	if( indicator )
	{
		indicator->start(indicator->russian()?"   ( )":"Resize image (histogram maximum)", 0, to_h - 1);
	}

	for(int y = 0; y < to_h; y++)
	{
		if(indicator && (y%40 == 0) && indicator->current(y))
		{
			indicator->finish();
			return false;
		}
		for(int x = 0; x < to_w; x++)
		{
			const double xt = double(w)*double(x)/double(to_w);
			const double yt = double(h)*double(y)/double(to_h);
			const int _x = int(xt+1);
			const int _y = int(yt);

			PixelType* initpix = &(*from)[_y][_x];
			PixelType* _pix = initpix;

			int i;
			int j;
			for( i = 0; i < ny; i++)
			{
				for( j = 0; j < nx; j++)
				{
					counts[*_pix] = 0;
					_pix++;
				}
				_pix += nls - nx;
			}

			_pix = initpix;
			for( i = 0; i < ny; i++)
			{
				for( j = 0; j < nx; j++)
				{
					counts[*_pix]++;
					_pix++;
				}
				_pix += nls - nx;
			}

			int max = 0;
			int maxval = 0;
			_pix = initpix;

			for( i = 0; i < ny; i++)
			{
				for( j = 0; j < nx; j++)
				{
					if(counts[*_pix] > max)
					{
						max = counts[*_pix];
						maxval = *_pix;
						_pix++;
					}
				}
				_pix += nls - nx;
			}

			*pix++ = maxval;
		}
	}
	if( indicator && indicator->InProgress() )
		indicator->finish();
	return true;
}

template<class ImageType, class PixelType>
int ImageResampleT<ImageType, PixelType>::reflect( const ImageType* img, const int x, const int y, const int w, const int h)
{
	// borders

	if( x < 0 && y >= 0 && y < h )
		return (*img)[y][-x-1];
	if( x >= 0 && x < w && y >= h )
		return (*img)[h-(y-h+1)][x];
	if( x >= w && y >= 0 && y < h )
		return (*img)[y][w-(x-w+1)];
	if( x >= 0 && x < w && y < 0 )
		return (*img)[-y-1][x];

	// corners

	if( x >= w && y >= h )
		return (*img)[h-(y-h+1)][w-(x-w+1)];
	if( x < 0 && y >= h )
		return (*img)[h-(y-h+1)][-x-1];
	if( x < 0 && y < 0 )
		return (*img)[-y-1][-x-1];
	if( x >= w && y < 0 )
		return (*img)[-y-1][w-(x-w+1)];

	return (*img)[y][x];
}

// inline interpolation functions

template<class ImageType, class PixelType>
inline int ImageResampleT<ImageType, PixelType>::bilinear( PixelType* pix, double dx, double dy, int nls)
{
	const double P[] = { 1 - dx, dx };
	const double L[] = { 1 - dy, dy };
	double res = 0;
	for( int i = 0; i < 2; i++)
	{
		double rtmp = 0;
		for( int j = 0; j < 2; j++)
		{
			 rtmp += *pix * P[j];
			pix++;
		}
		res += rtmp * L[i];
		pix += nls - 2;
	}

	return int(res+0.5);
}

template<class ImageType, class PixelType>
inline int ImageResampleT<ImageType, PixelType>::bilinearReflect( const ImageType* from, double dx, double dy, int x, int y, int w, int h)
{
	const double P[] = { 1 - dx, dx };
	const double L[] = { 1 - dy, dy };
	double res = 0;
	for( int i = 0; i < 2; i++)
	{
		double rtmp = 0;
		for( int j = 0; j < 2; j++)
		{
//			rtmp += rtmp += reflect( from, x-1+j, y-1+i, w, h) * P[j];
			rtmp += reflect( from, x-1+j, y-1+i, w, h) * P[j];
		}
		res += rtmp * L[i];
	}

	return int(res+0.5);
//	return 127;
}

template<class ImageType, class PixelType>
inline int ImageResampleT<ImageType, PixelType>::bicubical( PixelType* pix, double dx, double dy, int nls)
{
	double P[4];
	P[0] = ((-1.0*dx +  6.0)*dx - 11.0)*dx + 6.0;
	P[1] = (( 3.0*dx - 15.0)*dx + 18.0)*dx;
	P[2] = ((-3.0*dx + 12.0)*dx -  9.0)*dx;
	P[3] = ((     dx -  3.0)*dx +  2.0)*dx;

	double L[4];
	L[0] = ((-1.0*dy +  6.0)*dy - 11.0)*dy + 6.0;
	L[1] = (( 3.0*dy - 15.0)*dy + 18.0)*dy;
	L[2] = ((-3.0*dy + 12.0)*dy -  9.0)*dy;
	L[3] = ((     dy -  3.0)*dy +  2.0)*dy;
	double res = 0;
	int i = 0;
	for(;;)
	{
		double rtmp = 0;
		for( int j = 0; j < 4; j++)
		{
			 rtmp += *pix * P[j];
			pix++;
		}
		res += rtmp * L[i];
		i++;
		if(!(i < 4))
			break;
		pix += nls - 4;
	}

	return int(res/36.0+0.5);
}

template<class ImageType, class PixelType>
inline int ImageResampleT<ImageType, PixelType>::bicubicalReflect( const ImageType* from, double dx, double dy, int x, int y, int w, int h)
{
	double P[4];
	P[0] = ((-1.0*dx +  6.0)*dx - 11.0)*dx + 6.0;
	P[1] = (( 3.0*dx - 15.0)*dx + 18.0)*dx;
	P[2] = ((-3.0*dx + 12.0)*dx -  9.0)*dx;
	P[3] = ((     dx -  3.0)*dx +  2.0)*dx;

	double L[4];
	L[0] = ((-1.0*dy +  6.0)*dy - 11.0)*dy + 6.0;
	L[1] = (( 3.0*dy - 15.0)*dy + 18.0)*dy;
	L[2] = ((-3.0*dy + 12.0)*dy -  9.0)*dy;
	L[3] = ((     dy -  3.0)*dy +  2.0)*dy;
	double res = 0;
	for( int i = 0; i < 4; i++)
	{
		double rtmp = 0;
		for( int j = 0; j < 4; j++)
		{
			 rtmp += reflect( from, x-2+j, y-2+i, w, h) * P[j];
		}
		res += rtmp * L[i];
	}

	return int(res/36.0+0.5);
}

// ImageResample functions
void ImageResample::nullifyPrms() 
{ 
	prms.from = 0; 
	prms.indicator = 0; 
	prms.imeth = imUnknown; 
	prms.is_valid = false; 
}

bool ImageResample::initResize( const AImage* img, InterpolationMethod meth, int new_width, int new_height)
{
	if(!img || !img->IsOk())
	{
		return false;
	}
	if(new_width<=0 && new_height<=0)
	{
		return false;
	}
	else if(new_width<=0 )
	{
			new_width = (new_height*img->Width()+img->Height()/2)/img->Height();
	}
	else if(new_height<=0)
	{
			new_height = (new_width*img->Height()+img->Width()/2)/img->Width();
	}
	ImageResampleParams _prms;
	_prms.from = img;
	_prms.imeth = meth;
	_prms.new_width = new_width;
	_prms.new_height = new_height;
	_prms.mtype = ImageResample::mtResize;
	_prms.is_valid = false;
	return init(_prms);
}
bool ImageResample::initRotate( const AImage* img, InterpolationMethod meth, double angle, bool radians, bool _fill, int _fill_int)
{
	if(!img || !img->IsOk())
	{
		return false;
	}
	ImageResampleParams _prms;
	_prms.from = img;
	_prms.imeth = meth;
	_prms.radians = radians ? angle : ((-2)*::atan2(-1.,0.)*angle/180);
	_prms.fill = _fill;
	_prms.fill_int = _fill_int;
	_prms.mtype = ImageResample::mtRotate;
	_prms.is_valid = false;
	return init(_prms);
}


bool ImageResample::init( ImageResampleParams _prms)
{
	prms = _prms;

	if( prms.imeth == imAuto )
	{
		if( prms.mtype == mtRotate )
		{
			if( prms.from->ImageClass() == AImage::IMGC_GRAYSCALE )
				prms.imeth = imBicubical;
			if( prms.from->ImageClass() == AImage::IMGC_PALETTE )
				prms.imeth = imNearest;
		}
		if( prms.mtype == mtResize )
		{
			if( prms.new_width*prms.new_height >= prms.from->Width()*prms.from->Height() )
			{
				if( prms.from->ImageClass() == AImage::IMGC_GRAYSCALE )
					prms.imeth = imBicubical;
				if( prms.from->ImageClass() == AImage::IMGC_PALETTE )
					prms.imeth = imNearest;
			}
			else
			{
				if( prms.from->ImageClass() == AImage::IMGC_GRAYSCALE )
					prms.imeth = imAverage;
				if( prms.from->ImageClass() == AImage::IMGC_PALETTE )
					prms.imeth = imHistMax;
			}
		}
	}

	prms.to_class = prms.from->ImageClass();
	if( prms.to_class != AImage::IMGC_GRAYSCALE && prms.to_class != AImage::IMGC_PALETTE ) return 0;
	if( prms.to_class == AImage::IMGC_PALETTE )
	{
		if( prms.imeth != ImageResample::imNearest && prms.imeth != ImageResample::imHistMax ) return 0;
	}
	prms.to_bits = prms.from->Bits();

	prms.is_valid = ImageResampleT<AImageMem8,Pixel8>::init( prms);

	return prms.is_valid;
}


bool ImageResample::Run( AImage* to, ProgressCounter* ind)
{
	if(    !prms.is_valid 
		|| to->Width() != prms.to_width
		|| to->Height() != prms.to_height
		|| to->Bits() != prms.to_bits
		|| to->ImageClass() != prms.to_class )
		throw "Error: ImageResample::Run(): Wrong output image.";

	prms.indicator = ind;

	const AImageMem8* from8 = dynamic_cast<const AImageMem8*>(prms.from);
	if(from8)
	{
		ImageResampleT<AImageMem8,Pixel8> mod8(prms);
		return mod8.run( dynamic_cast<AImageMem8*>(to));
	}
	else
	{
		ImageResampleT<AImageMem16,Pixel16> mod16(prms);
		return mod16.run( dynamic_cast<AImageMem16*>(to));
	}
}
