
#include "aimage.h"
#include "imfilter.h" 

////////////////////////////////////////////////////////////////////////////////////
///////////Duplicate/////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
template<class T,class P> void tpl_AImageFilter_Duplicate(const T* in,T* out,P max_color,ProgressCounter *ind,const char *name)
{
	int width=in->Width();
	int height=in->Height();
	if(ind) ind->start(name,0,height);
	for(int y=0; y<height; y++)
	{
		P* o=(*out)[y];
		P* i=(*in)[y];
		int w=width;
		while(w--)
			*o++=*i++;
		if(ind)
		{
			if(ind->current(y))
			{
				ind->finish();
				throw "User abort";
			}
		}
	}
	if(ind) ind->finish();
}
void AImageFilter_Duplicate(const AImage *in,AImage *out,ProgressCounter *ind)
{
	const char *name=(ind && ind->russian() )?"":"Duplicate";
	const AImageMem8 *in8=dynamic_cast<const AImageMem8  *>(in);
	AImageMem8 *out8=dynamic_cast<AImageMem8 *>(out);
	Pixel8 MAX_COLOR=Pixel8((1<<in->Bits())-1);
	if(in8 && out8)
	{
		tpl_AImageFilter_Duplicate(in8,out8,MAX_COLOR,ind,name);
	}
	else
	{
		const AImageMem16 *in16=dynamic_cast<const AImageMem16  *>(in);
		AImageMem16 *out16=dynamic_cast<AImageMem16 *>(out);
		Pixel16 MAX_COLOR=Pixel16((1<<in->Bits())-1);
		if(in16 && out16)
		{
			tpl_AImageFilter_Duplicate(in16,out16,MAX_COLOR,ind,name);
		}
		else
		{
			throw "Unsupported image type for Duplicate";
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////
///////////Negative/////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
template<class T,class P> void tpl_AImageFilter_Negative(const T* in,T* out,P max_color,ProgressCounter *ind,const char *name)
{
	int width=in->Width();
	int height=in->Height();
	if(ind) ind->start(name,0,height);
	for(int y=0; y<height; y++)
	{
		P* o=(*out)[y];
		P* i=(*in)[y];
		int w=width;
		while(w--)
			*o++=max_color - (*i++);
		if(ind)
		{
			if(ind->current(y))
			{
				ind->finish();
				throw "User abort";
			}
		}
	}
	if(ind) ind->finish();
}
void AImageFilter_Negative(const AImage *in,AImage *out,ProgressCounter *ind)
{
	const char *name=(ind && ind->russian() )?"":"Negative";
	const AImageMem8 *in8=dynamic_cast<const AImageMem8  *>(in);
	AImageMem8 *out8=dynamic_cast<AImageMem8 *>(out);
	Pixel8 MAX_COLOR=Pixel8((1<<in->Bits())-1);
	if(in8 && out8)
	{
		tpl_AImageFilter_Negative(in8,out8,MAX_COLOR,ind,name);
	}
	else
	{
		const AImageMem16 *in16=dynamic_cast<const AImageMem16  *>(in);
		AImageMem16 *out16=dynamic_cast<AImageMem16 *>(out);
		Pixel16 MAX_COLOR=Pixel16((1<<in->Bits())-1);
		if(in16 && out16)
		{
			tpl_AImageFilter_Negative(in16,out16,MAX_COLOR,ind,name);
		}
		else
		{
			throw "Unsupported image type for Negative";
		}
	}
}

