#ifndef _NUM_SVD_H
#define _NUM_SVD_H
////////////////////////////////////////////////////////////////////////////////////////////////
namespace Vtl
{
namespace NumAlg
{
namespace Lin
{ 

	template<typename T> inline T sing_fortran(T x1,T x2) { return (x2 >= 0) ? std::abs(x1) : -std::abs(x1); }
	class svd
	{
		///      
		template<typename T,class MatrixU,class VectorW,class VectorRV1 >
		static void hausholder(T& G,T& ANORM,VectorW& W,MatrixU& UU,VectorRV1& RV1)
		{
			const int M=UU.size1();
			const int N=UU.size2();
			G=0.;
			T SCALE=0.;
			ANORM=0.;
		//

			for(int i=0; i < N ; ++i )
			{
				int L=i+1;
				RV1[i]=SCALE*G;
				G=0.;
				T S=0.;
				SCALE=0.;
				if( i < M )
				{
					for(int k=i; k < M; ++k)
					{
						SCALE=SCALE+std::abs(UU(k,i));
					}
					if( SCALE != 0. )
					{
						for(int k=i; k < M; ++k)
						{
							UU(k,i)=UU(k,i)/SCALE;
							T t=UU(k,i);
							S=S+t*t;
						}
						T F=UU(i,i);
						G=-sing_fortran(std::sqrt(S),F);
						T H=F*G-S;
						UU(i,i)=F-G;
						if( i != (N-1) )
						{
							for(int j=L; j < N; ++j )
							{
								S=0.;
								for(int k=i; k < M; ++k)
								{
									S=S+UU(k,i)*UU(k,j);
								}
								T F=S/H;
								for(int k=i; k < M; ++k)
								{
									UU(k,j)=UU(k,j)+F*UU(k,i);
								}
							}
						}
						for(int k=i; k < M; ++k)
						{
							UU(k,i)=SCALE*UU(k,i);
						}
					}
				}
				W[i]=SCALE*G;
				G=0.;
				S=0.;
				SCALE=0.;
				if( !(i >= M || i == (N-1)) ) 
				{
					for(int k=L; k < N; ++k)
					{
						SCALE=SCALE+std::abs(UU(i,k));
					}
					if( SCALE != 0. )
					{
						for(int k=L; k < N; ++k)
						{
							UU(i,k)=UU(i,k)/SCALE;
							T t=UU(i,k);
							S=S+t*t;
						}
						T F=UU(i,L);
						G=-sing_fortran(std::sqrt(S),F);
						T H=F*G-S;
						UU(i,L)=F-G;
						for( int k=L; k < N; ++k)
						{
							RV1[k]=UU(i,k)/H;
						}
						if( i != (M-1) )
						{
							for( int j=L; j < M; ++j )
							{
								S=0.;
								for( int k=L; k < N; ++k)
								{
									S=S+UU(j,k)*UU(i,k);
								}
								for( int k=L; k < N; ++k)
								{
									UU(j,k)=UU(j,k)+S*RV1[k];
								}
							}
						}
						for( int k=L; k < N; ++k )
						{
							UU(i,k)=SCALE*UU(i,k);
						}
					}
				}
				ANORM=std::max(ANORM,std::abs(W[i])+std::abs(RV1[i]));
			}
		}
	public:
		/*!
		*      double A(NM,N),W(N),U(NM,N),V(NM,N),RV1(N)
		* //MY      double A(M,N),W(N),U(NM,N),V(N,N),RV1(N)
		*
		*           
		*                              T
		*                         A=USV
		*        A   M*N.
		*          -
		*         QR-.
		*
		*    :
		* """""""""""""""""
		*     M -    A  U.
		*
		*     N -    A  U   V.
		*
		*     NM =< max{ M,N }   -     -
		*       ,      
		*      .
		*
		*     A -    ,   -
		*        .
		*
		*     MATU  =.TRUE.      U  .
		*     MATU  =.FALSE.    .
		*
		*     MATV  =.TRUE.      V  .
		*     MATV  =.FALSE.    .
		*     !!!Yurin change:   MATU  =.FALSE.  
		*                        U=NULL
		*     !!!Yurin change:   MATV  =.FALSE.  
		*                        V=NULL
		*
		*    :
		*   """"""""""""""""
		*     A   (      U  V).
		*
		*     W  N ()   A (-
		*         S).   .    
		*         ,     IERR+1 -:- IERR + N 
		*           .
		*
		*     U   U (  )  -
		*       ,    MATU    .TRUE.  
		*         U    . U -
		*          A.     ,  -
		*        U ,        , 
		*          .
		*
		*     V   U (  )  ,
		*           MATV    .TRUE.
		*           V   . V  
		*          A  U  .  -
		*          ,   V ,  
		*          ,    .
		*
		*     IERR=0 -     .
		*     IERR=K - K-       30
		*           .
		*     IERR - !!!!!
		*
		*     RV1-    .
		*
		*int SVD(int NM,int M,int N,double **A,double *W,bool MATU,double **U,bool MATV,double **V,double *RV1)
		*********************************************************************************/
		template<class MatrixU,class MatrixV,class VectorW,class VectorRV1 >
		static int decomp(MatrixU& UU,VectorW& W,MatrixV& VV,VectorRV1& RV1,bool calcU=true,bool calcV=true)
		{
			typedef typename MatrixU::value_type T;
		//	int NM;
			const int M=UU.size1();
			const int N=UU.size2();
			if( UU.size2()!=VV.size1() || UU.size2()!=VV.size2())
				throw "svd: Matrix V must be square with N equal to number of columns of matrix A!";
			if( UU.size2()!=W.size() )
				throw "svd: Vector W must have size equal to number of columns of matrix A!";
		//------<      >----
			T G=0.;
			T ANORM=0.;
			hausholder(G,ANORM,W,UU,RV1);
		//---------<    >--
			if( calcV ) 
			{//  I=N   -1  1  
				int L;
				for( int ii=0; ii < N; ++ii )
				{
					int i=N-1-ii;
					if( i != (N-1) ) 
					{
						if( G != 0.)
						{
							for( int j=L; j < N; ++j)
								VV(j,i)=(UU(i,j)/UU(i,L))/G; //<       
							for( int j=L; j < N; j ++)
							{
								T S=0.;
								for( int k=L; k < N; ++k )
									S += UU(i,k)*VV(k,j);
								for( int k=L; k < N; ++k )
									VV(k,j) += S*VV(k,i);
							}
						}
						for( int j=L; j < N; ++j)
						{
							VV(i,j)=0.;
							VV(j,i)=0.;
						}
					}
					VV(i,i)=1.;
					G=RV1[i];
					L=i;
				}
			}
		//--------<    >--
			if( calcU ) 
			{ // I=min{ N,M }   -1  1  
				const int MN=std::min(M,N);
		//
				for( int ii=0; ii < MN; ++ii )
				{
					int i=MN-1-ii;
					int L=i+1;
					G=W[i];
					if( i != (N-1) )
					{
						for( int j=L; j < N; ++j )
							 UU(i,j)=0.;
					}
					if( G == 0. ) 
					{
						for( int j=i; j < M; ++j)
							UU(j,i)=0.;
					}
					else
					{
						if( i != (MN-1) )
						{
							for( int j=L; j < N; ++j)
							{
								T S=0.;
								for( int k=L; k < M; ++k )
									S += UU(k,i)*UU(k,j);
					//--------------<       >---
								T F=(S/UU(i,i))/G;
								for( int k=i; k < M; ++k )
									UU(k,j) += F*UU(k,i);
							}
						}
						for( int j=i; j < M; ++j)
							UU(j,i)=UU(j,i)/G;
					}
					UU(i,i) += 1.0f;
				}
			}
		//------<    >---
		//--------<  K=N   -1  1  - >---

				int ITS;
				for(int kk=0; kk <  N ; ++kk)
				{
					const int K1=N-1-kk;
					int K=K1+1;
					ITS=0;
		//--------<   .  L=K
		//            -1  1  >----
		L520:

					int L;
					int L1;
					for(int LL=0; LL < K; ++LL )
					{
						L=K-1-LL;
						L1=L-1;
						if( std::abs(RV1[L])+ANORM == ANORM ) goto L565;
						// RV1(0)   ,       . 
						if( std::abs(W[L1])+ANORM == ANORM ) break;
					}
		//  L ,  1,  RV1(L-1)  0.
					T C=0.;
					T S=1.;
					for(int i=L; i < K; ++i )
					{
						T F=S*RV1[i];
						RV1[i]=C*RV1[i];
						if( std::abs(F)+ANORM == ANORM ) goto L565;
						G=W[i];
						T H=std::sqrt(F*F+G*G);
						W[i]=H;
						C=G/H;
						S=-F/H;
						if( calcU ) 
						{
							for(int j=0; j < M ; ++j )
							{
								 T Y=UU(j,L1);
								 T Z=UU(j,i);
								 UU(j,L1) = Y*C+Z*S;
								 UU(j,i) = -Y*S+Z*C;
							}
						}
					}
		//-------<     >------
		L565:      
					T Z=W[K-1];
					if( (L+1) != K )
					{ //        2.
						if( ITS == 30 ) 
						{
							//     -  30  
							//     . 
							return K;//(IERR)
						}
						ITS=ITS+1;
						T X=W[L];
						T Y=W[K1-1];
						G=RV1[K1-1];
						T H=RV1[K-1];
						T F=((Y-Z)*(Y+Z)+(G-H)*(G+H))/(2*H*Y);
						G=std::sqrt(F*F+1);
						F=((X-Z)*(X+Z)+H*(Y/(F+sing_fortran(G,F))-H))/X;
			//-------<  QR- >---
						T C=1.;
						T S=1.;
						for(int i1=L; i1 < K1; ++i1 )
						{
							int i=i1+1;
							G=RV1[i];
							Y=W[i];
							H=S*G;
							G=C*G;
							Z=std::sqrt(F*F+H*H);
							RV1[i1]=Z;
							C=F/Z;
							S=H/Z;
							F=X*C+G*S;
							G=-X*S+G*C;
							H=Y*S;
							Y=Y*C;
							if( calcV ) 
							{
								for( int j=0; j < N ; ++j )
								{
									X=VV(j,i1);
									Z=VV(j,i);
									VV(j,i1)=X*C+Z*S;
									VV(j,i)=-X*S+Z*C;
								}
							}
							Z=std::sqrt(F*F+H*H);
							W[i1]=Z;
							//---<    ,  Z   >-- 
							if( Z != 0. ) 
							{
								C=F/Z;
								S=H/Z;
							}
							F =  C*G+S*Y;
							X = -S*G+C*Y;
							if( calcU ) 
							{
								for( int j=0; j < M; ++j )
								{
									Y=UU(j,i1);
									Z=UU(j,i);
									UU(j,i1) =  Y*C+Z*S;
									UU(j,i)  = -Y*S+Z*C;
								}
							}
						}
			//
						RV1[L]=0.;
						RV1[K-1]=F;
						W[K-1]=X;
						goto L520;
					}      
		//-------<  >----
					if( Z < 0. )
					{// W[K-1]   
						W[K-1]=-Z;
						if( calcV ) 
						{
							for( int j=0; j < N ; ++j )
								VV(j,K-1)=-VV(j,K-1);
						}
					}
				}
			return 0;//(IERR=0)
		}
		/// \brief     SVD 
		template<class MatrixU,class MatrixV,class VectorW>
		static void comp(MatrixU& A,MatrixU& UU,VectorW& W,MatrixV& VV)
		{
			A=UU;
			prod_diag(A,W);
			A=prod(A,trans(VV));
		}
		/*
		 * \brief   .
		 *
		 *   -   UU, S, VV. 
		 *
		 *   -   Ainv.
		 *
		 *    W  V
		 *
		 * Ainv  UU  .
		 *
		 *********************************************************************************/
		template<class MatrixAinv,class MatrixU,class VectorS,class MatrixV>
		static void inverse(MatrixAinv& Ainv,MatrixU& UU,VectorS& W,MatrixV& VV,double relerr)
		{
			inverse_diag_matrix_positive(W,relerr);
			prod_diag(VV,W);
			Ainv=prod(VV,trans(UU));
		}
		/*
		 * \brief   .
		 *
		 *   - A.   . 
		 *
		 *   -   Ainv.
		 *
		 *    A
		 * Ainv  A   .
		 *
		 *********************************************************************************/
		template<typename T,size_t M,size_t N>
		static void inverse(ublas::c_matrix<T,N,M>& Ainv,ublas::c_matrix<T,M,N>& A,double relerr)
		{
			ublas::c_matrix<T,N,N> V;
			ublas::c_vector<T,N> s;
			ublas::c_vector<T,N> rv1;
			decomp(A,s,V,rv1);
			inverse(Ainv,A,s,V,relerr);
		}
		/*
		 * \brief   .
		 *
		 *   - A.   . 
		 *
		 *   -   Ainv.
		 *
		 *     A
		 * Ainv  A  .
		 *
		 *********************************************************************************/
		template<typename T,size_t M,size_t N>
		static void inverse_save(ublas::c_matrix<T,N,M>& Ainv,ublas::c_matrix<T,M,N> A,double relerr)
		{
			inverse(Ainv,A,relerr);
		}
		/*
		 * \brief   .
		 *
		 *   - A.   . 
		 *
		 *  -  .
		 *
		 *    A
		 *
		 *********************************************************************************/
		template<typename T,size_t M,size_t N>
		static ublas::c_matrix<T,N,M> inverse(ublas::c_matrix<T,M,N>& A,double relerr)
		{
			ublas::c_matrix<T,N,M> Ainv;
			inverse(Ainv,A,relerr);
			return Ainv;
		}
		/*
		 * \brief   .
		 *
		 *   - A.   . 
		 *
		 *  -  .
		 *
		 *     A
		 *
		 *********************************************************************************/
		template<typename T,size_t M,size_t N>
		static ublas::c_matrix<T,N,M> inverse_save(ublas::c_matrix<T,M,N> A,double relerr)
		{
			return inverse(A,relerr);
		}
		template<class VectorRes,class VectorB,class MatrixU,class VectorS,class MatrixV>
		static void solve(VectorRes& result,VectorB& b,MatrixU& U,VectorS& S,MatrixV& V,double relerr)
		{
			inverse_diag_matrix(S,relerr);
			result=prod(trans(U),b);
			result=element_prod(S,result);
			result=prod(V,result);
		}
		template<typename T,size_t M,size_t N>
		static void solve(ublas::c_vector<T,N>& x,ublas::c_matrix<T,M,N>& A,ublas::c_vector<T,M>& b,double relerr)
		{
			ublas::c_matrix<T,N,N> V;
			ublas::c_vector<T,N> s;
			decomp(A,s,V,x);
			solve(x,b,A,s,V,relerr);
		}
		template<typename T,size_t M,size_t N>
		static void solve_save(ublas::c_vector<T,N>& x,ublas::c_matrix<T,M,N> A,ublas::c_vector<T,M> b,double relerr)
		{
			solve(x,A,b,relerr);
		}
		template<typename T,size_t M,size_t N>
		static ublas::c_vector<T,N> solve(ublas::c_matrix<T,M,N>& A,ublas::c_vector<T,M>& b,double relerr)
		{
			ublas::c_vector<T,N> x; 
			solve(x,A,b,relerr);
			return x;
		}
		template<typename T,size_t M,size_t N>
		static ublas::c_vector<T,N> solve_save(ublas::c_matrix<T,M,N> A,ublas::c_vector<T,M> b,double relerr)
		{
			return solve(A,b,relerr);
		}

		/// Reorder SVD so that singular values in decrement order
		template<typename T,size_t N> 
		static void sort(Vtl::NumAlg::Lin::ublas::c_matrix<T,N,N>& U
					,Vtl::NumAlg::Lin::ublas::c_vector<T,N>& W
					,Vtl::NumAlg::Lin::ublas::c_matrix<T,N,N>& V)
		{
			using namespace Vtl::NumAlg::Lin::ublas;
			//using namespace Vtl::NumAlg::Implicit;
			using namespace Vtl::NumAlg;
			
			typedef typename c_matrix<T, N, N> Matrix;
			typedef typename c_vector<T,N> Vector;
			typedef typename Matrix::value_type value_type;

			Implicit::Sort<Vector,Implicit::Pred::gt<Vector> > s(W);
			Matrix C;
			C=U;
			Implicit::sort_columns(U,C,s);
			C=V;
			Implicit::sort_columns(V,C,s);
			Vector Ct(W);
			Implicit::sort(W,Ct,s);
		}
	};
////////////////////////////////////////////////////////////////////////////////////////////////
}//namespace Lin
}//namespace NumAlg
}//namespace Vtl
////////////////////////////////////////////////////////////////////////////////////////////////
#endif //_NUM_SVD_H
