#ifndef __NUMERIC0_H__
#define __NUMERIC0_H__
////////////////////////////////////////////////////////////////////////////
// Template realization numeric analysis functions:
//!!!      Next functions suppose, that types T and TT may be
//!!!      reduced to double without lost in accuracy.
//
//		template <class T> T nmRang(int k,T * a,int n);
//		template <class T> T nmMedian(T * a,int n);
//		template <class T,class TT> void nmLinearLeastSquares(int n,const T *x,const TT *y,double &slope,double& intercept);
//    template <class T> double nmMeanAndDisp(int n,T* array,double *dispsqr);
//    template <class T> int nmFindBadByStudent(int n,T* array,double* tau2,double* mean,double* disp2);
//    template <class T> double nmFilteredMeanByStudent(int n,T* array,double* tau2,int* n_rest);
//    template <class T,class TT> int nmFindBadByStudent(int n,T* x,TT* y,double* tau2);
//
//
////////////////////////////////////////////////////////////////////////////
template <class T> T nmRang(int k,T * a,int n)
{
	int l=0;
	int r=n-1;
	while(l<r)
	{
		T x=a[k];
		int i=l;
		int j=r;
		do
		{
			while(a[i]<x) i++;
			while(x<a[j]) j--;
			if(i<=j)
			{
				T w=a[i];
				a[i]=a[j];
				a[j]=w;
				i++;
				j--;
			}
		} while(i<=j);
		if(j<k) l=i;
		if(k<i) r=j;
	}
	return a[k];
}
template <class T> T nmMedian(T * a,int n)
{
	return nmRang(n>>1,a,n);//array index is from 0 to n-1, k-is mean array index, so no 1 add need.
}

template <class T>
void nmMedian1D(T* data,int n,int half_wnd_size,T* buf,bool reflect_edjes) //buf[2*(half_wnd_size*2+1)]
{
	const int len=half_wnd_size*2+1;
	T* round_buf=buf;
	T* tmp=buf+len;
	//start init
	round_buf[half_wnd_size+1]=data[0];
	if(reflect_edjes)
	{
		for(int i=1; i<=half_wnd_size; i++)
		{
			round_buf[half_wnd_size+1-i]=round_buf[half_wnd_size+1+i]=data[i];
		}
	}
	else
	{
		for(int i=1; i<=half_wnd_size; i++)
		{
			round_buf[half_wnd_size+1+i]=data[i];
			round_buf[half_wnd_size+1-i]=data[0];
		}
	}
	//main loop
	int pos=0;
	for(int j=0; j<n; j++)
	{
		int l=j+half_wnd_size;
		if(l>=n)
		{ //near right (top array) edge
			if(reflect_edjes)
			{
				int pos2=pos;
				if(++pos2>=len) pos2=0;
				//l=n-1-(l-(n-1)); this data in array already overwrited, but it present in round_buf, find it
				round_buf[pos]=round_buf[pos2]; //override sample left window by entered sample
			}
			else
			{
				round_buf[pos]=data[n-1]; //override sample left window by entered sample
			}
		}
		else
		{
			round_buf[pos]=data[l]; //override sample left window by entered sample
		}
		if(++pos>=len) pos=0;
		for(int i=0; i<len; i++)
			tmp[i]=round_buf[i];
		data[j]=::nmMedian(tmp,len);
	}
}

template <class T,class TT> void nmLinearLeastSquares(int n,const T *x,const TT *y,double &slope,double& intercept)
{
	double m_x=0;
	double m_y=0;
	double m_xy=0;
	double m_x2=0;
	//  
	for(int i=0; i<n; i++)
	{
		double xx=x[i];
		double yy=y[i];
		m_x += xx;
		m_y += yy;
		m_xy+= xx*yy;
		m_x2+= xx*xx;
	}
	// slope & intercept  x[]    [0,1]
	double div=m_x2-m_x*m_x/n;
	slope=(m_xy-m_x*m_y/n);
	slope=( div > ABS(slope)/1.0e40 )? slope/div : 0; //to prevent divide by zero and overflow
	intercept=(m_y-slope*m_x)/n;
}
template <class T> double nmMeanAndDisp(int n,const T* array,double *dispsqr)
{
	double x=0;
	double s=0;
	for(int i=0; i<n ; i++)
	{
		T v=array[i];
		x+=v;
		s+=v*v;
	}
	x/=n;
	if(dispsqr)
	{
		s=s/n-x*x;
		*dispsqr=s;
	}
	return x;
}
template <class T> int nmFindBadByStudent(int n,const T* array,const double* tau2,double* mean,double* disp2)
{// :
 //     -  , -1    
 //     - ,   'mean' 
 //     - ,   'disp2'  
 //  :
 // n>3
 // T array[n];
 // double tau2[n+1];
	double s;
	double x=::nmMeanAndDisp(n,array,&s);
	int k=0;
	double d=0;
	for(int i=0; i<n; i++)
	{
		double dd=array[i]-x;
		dd=dd*dd;
		if(dd>d)
		{
			k=i;
			d=dd;
		}
	}
	if(mean)
		*mean=x;
	if(disp2)
		*disp2=s;
	if(d==0 || d<s*tau2[n])  //  s==0   d==0       
		return -1;
	return k;
}
template <class T> double nmFilteredMeanByStudent(int n,T* array,const double* tau2,int* n_rest)
{//   ,   ,
 //      
 //    n_rest>=3 (  )
 // int array[n];
 // double tau2[n+1];
	double mean=0;;
	do
	{
		int k=::nmFindBadByStudent(n,array,tau2,&mean,0);
		if(k<0 )  //  s==0   d==0       
		{
			n--;
			break;
		}
		T v=array[k];
		array[k]=array[n-1];
		array[n-1]=v;
		n--;
	}	while(n>=3);
	if(n_rest)
		*n_rest=n+1;
	return mean;
}

template <class T,class TT> int nmFindBadByStudent(int n,const T* x,const TT* y,const double* tau2)
{//    , -1   
 // double x[n],y[n];
 // double tau2[n+1];
	double a;
	double b;
	::nmLinearLeastSquares(n,x,y,a,b);
	double d=0;
	double s=0;
	int k=0;
	{
		for(int i=0; i<n ; i++)
		{
			double dd=y[i]-(a*x[i]+b);
			dd=dd*dd;
			s+=dd;
			if(dd>d)
			{
				k=i;
				d=dd;
			}
		}
		s/=n;
	}
	if(d==0 || d<s*tau2[n]) //  s==0   d==0       
		return -1;
	return k;
}
//>---------------------------------------------------------------------------------
template<class T> T nmMaxArrayElem(int n,const T* array)
{
	if(n<1 || !array)
		throw "Bad Param";
	T el=*array++;
	n--;
	while(n--)
	{
		if(el<*array) 
			el=*array;
		array++;
	}
	return el;
}
////////////////////////////////////////////////////////////////////////////
#endif //__NUMERIC0_H__
