#ifndef _PLF_FILTER_COMPLEX_H
#define _PLF_FILTER_COMPLEX_H
/////////////////////////////////////////////////////////////////////////////
#include "plf_filters.h"
#include "color_spaces.h"

namespace Vtl
{
// @brief     plf_filters.h,   
namespace Plf 
{
/** @addtogroup ComplexFiltersDoxGroup  Complex filters
 *  @ingroup PipeLineFiltersDoxGroup  
 *
 *  @brief  ()    
 *     (Pipe Line Filtering)    
 *   .
 *
 *  @{
 */
	///////////////////////////////////////////////////////////////////////////
	// class SeparableConv2D
	/// 
	/// @brief    .   
	///   PLF   ,   :
	/// 
	/// 
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class SeparableConv2D
		: public FilterComplex
	{
		Vtl::Plf::ConvolveHorz<T> h_conv;
		Vtl::Plf::ConvolveVert<T,P> v_conv;
	private:
		SeparableConv2D(const SeparableConv2D&);
		SeparableConv2D& operator=(const SeparableConv2D&);
	public:
		InPort<T> & in() {return v_conv.in; }
		OutPort<T,P> & out() {return h_conv.out; }
		SeparableConv2D()
			: FilterComplex(1,1)
		{
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		SeparableConv2D(std::string name,const Kernel::Base<T,Kernel::KT_Conv> *kernel_x,const Kernel::Base<T,Kernel::KT_Conv> *kernel_y,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterComplex(1,1)
		{
			Init(name,kernel_x,kernel_y,post_proc);
		}
		void Init(std::string name,const Kernel::Base<T,Kernel::KT_Conv> *kernel_x,const Kernel::Base<T,Kernel::KT_Conv> *kernel_y,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			out().setPostProc(post_proc);
			v_conv.Init(Name()+"::V-conv",kernel_y);
			h_conv.Init(Name()+"::H-conv",kernel_x);
		//------ create graph --------------------
			v_conv.out.ConnectTo(&h_conv.in);
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		virtual ~SeparableConv2D()
		{
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class Average
	/// 
	/// @brief     c    X   Y. 
	///     (Mean, Average) ,  
	///   (@see class AverageGauss).    
	/// ,      \f$ sigma\f$     , 
	///   ,  :\n
	/// @li    
	/// @li          
	///                
	///      ( ,    ).
	///     
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class Average
		: public FilterComplex
	{
		SeparableConv2D<T,P> conv2D;
	private:
		Average(const Average&);
		Average& operator=(const Average&);
	public:
		InPort<T> & in() {return conv2D.in(); }
		OutPort<T,P> & out() {return conv2D.out(); }
		Average()
			: FilterComplex(1,1)
		{
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		Average(std::string name,const Kernel::Base<T,Kernel::KT_Conv> *kernel_1D,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterComplex(1,1)
		{
			Init(name,kernel_1D,post_proc);
		}
		void Init(std::string name,const Kernel::Base<T,Kernel::KT_Conv> *kernel_1D,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			out().setPostProc(post_proc);
			conv2D.Init(Name(),kernel_1D,kernel_1D,post_proc);
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		virtual ~Average()
		{
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class Sharpen
	/// 
	/// @brief  <I(x,y)>+amp*(I(x,y-<I(x,y)>). 
	///       , 
	///    .      
	///   ( .. ).
	///  Sharpen,    \ref Lit_Pratt
	/// \f$ G(x,y)=\langle I(x,y)\rangle+ k \cdot \lbrack I(x,y)- \langle I(x,y)\rangle \rbrack \f$ ,
	///  \f$ k \f$ -  .  \f$ k=1 \f$    , 
	///  \f$ k=0 \f$ -,  \f$ 0 < k < 1\f$ -    
	/// ,  \f$ 1 < k \f$ -    .
	///
	/// \dotfile Sharpen.dot "  Vtl::Plf::Sharpen -   " 
	/// 
	///     \image html o_gray.jpg 
	///   \image html Sharpen.jpg 
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class Sharpen
		: public FilterComplex
	{
		Average<T> mean;
		PixByPix2to1<T,F21::Sharpen<T>,P > sharp;
	private:
		Sharpen(const Sharpen&);
		Sharpen& operator=(const Sharpen&);
	public:
		InPortComb<T> in; 
		OutPort<T,P> & out() {return sharp.out; }
		Sharpen()
			: FilterComplex(1,1)
		{
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		Sharpen(std::string name,const Kernel::Base<T,Kernel::KT_Conv> *kernel_1D,float amp=4.0,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterComplex(1,1)
		{
			Init(name,kernel_1D,amp,post_proc);
		}
		void Init(std::string name,const Kernel::Base<T,Kernel::KT_Conv> *kernel_1D,float amp=4.0,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			out().setPostProc(post_proc);
			mean.Init(Name()+"::Mean",kernel_1D);
			F21::Sharpen<T> s(amp);
			sharp.Init(Name()+"::Sharpen",s,post_proc);
			mean.out().ConnectTo(&sharp.in_2);
			in.Add(&mean.in());
			in.Add(&sharp.in_1);
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		virtual ~Sharpen()
		{
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class AverageGauss
	/// 
	/// @brief   (Mean, Average)    
	///  (@see class Average).
	/// 
	/// \dotfile AverGauss.dot "  Vtl::Plf::Average -     " 
	/// 
	///     \image html o_gray.jpg 
	///   \image html AverGauss.jpg 
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class AverageGauss
		: public FilterComplex
	{
		typedef Vtl::ImAlg::WndFun::Gauss<T,0> G0;
		Vtl::Plf::Kernel::FKern<T> gauss;
		Average<T,P> aver;
	private:
		AverageGauss(const AverageGauss&);
		AverageGauss& operator=(const AverageGauss&);
	public:
		InPort<T> & in() {return aver.in(); }
		OutPort<T,P> & out() {return aver.out(); }
		AverageGauss()
			: FilterComplex(1,1)
		{
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		AverageGauss(std::string name,float sigma=1.0f,float env_sigmas=2.5f,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterComplex(1,1)
		{
			Init(name,sigma,env_sigmas,post_proc);
		}
		void Init(std::string name,float sigma=1.0f,float env_sigmas=2.5f,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			out().setPostProc(post_proc);
			//gauss.Init(sigma,env_sigmas); 
			G0 g0(sigma,env_sigmas);
			gauss.Init(g0);

			aver.Init(Name(),&gauss,post_proc);
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		virtual ~AverageGauss()
		{
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class HImage
	/// 
	/// @brief   H-image \ref Lit_HImg, \see Vtl::Plf::F::Himg.
	/// 
	/// \dotfile H-image.dot  "  Vtl::Plf::HImage" 
	/// 
	///     \image html o_gray.jpg 
	///   \image html H-image.jpg 
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class HImage
		: public FilterComplex
	{

		EnvProc1to1<T,Fpie::Himg<T>,P > h;
	private:
		HImage(const HImage&);
		HImage& operator=(const HImage&);
	public:
		InPort<T> & in() {return h.in; }
		///   !  .
		OutPort<T,P> & out() {return h.out; }
		HImage()
			: FilterComplex(1,1)
		{
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		HImage(std::string name,int radius,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterComplex(1,1)
		{
			Init(name,radius,post_proc);
		}
		/// output_maxval==float((1<<bits)-1)
		void Init(std::string name,int radius,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			Fpie::Himg<T> hi(radius);
			h.Init(name,hi,post_proc);
		//------ create graph --------------------
			//not to do - one filter only 
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		virtual ~HImage()
		{
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class BiLateral
	/// 
	/// @brief BiLateral   \ref Lit_BiLateral, \see Vtl::Plf::F::BiLateral.
	/// 
	/// \dotfile BiLateral.dot   "  Vtl::Plf::BiLateral   (   , edge preserving)" 
	/// 
	///     \image html o_gray.jpg 
	///   \image html BiLateral.jpg 
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class BiLateral
		: public FilterComplex
	{

		EnvProc1to1<T,Fpie::BiLateralSmooth<T>,P > bi;
	private:
		BiLateral(const BiLateral&);
		BiLateral& operator=(const BiLateral&);
	public:
		InPort<T> & in() {return bi.in; }
		///   !  .
		OutPort<T,P> & out() {return bi.out; }
		BiLateral()
			: FilterComplex(1,1)
		{
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		BiLateral(std::string name,float epsilon,float sigma=2.0f,float env_sigmas=2.0f,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterComplex(1,1)
		{
			Init(name,epsilon,sigma,env_sigmas);
		}
		/// output_maxval==float((1<<bits)-1)
		void Init(std::string name,float epsilon,float sigma=2.0f,float env_sigmas=2.0f,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			Fpie::BiLateralSmooth<T> b(epsilon,sigma,env_sigmas);
			bi.Init(name,b,post_proc);
		//------ create graph --------------------
			//not to do - one filter only 
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		virtual ~BiLateral()
		{
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class GradientModul
	/// 
	/// @brief         
	///  .   .
	/// 
	/// \dotfile ModGrad.dot "  GradientModul    ( )" 
	/// 
	///     \image html o_gray.jpg 
	///   \image html ModGrad.jpg 
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class GradientModul
		: public FilterComplex
	{
		T sigma; 
		typedef typename Vtl::Plf::F11::Comb<T,Vtl::Plf::F11::GammaCorMult<T>,P> PostProc;

		typedef Vtl::ImAlg::WndFun::Gauss<T,0> G0;
		typedef Vtl::ImAlg::WndFun::Gauss<T,1> G1;
		Vtl::Plf::Kernel::FKern<T> gauss;
		Vtl::Plf::Kernel::FKern<T> gauss_deriv;

//		Vtl::Plf::Kernel::Gauss<T,0,Kernel::KT_Conv> gauss; 
//		Vtl::Plf::Kernel::Gauss<T,1,Kernel::KT_Conv> gauss_deriv; 
		Vtl::Plf::SeparableConv2D<T> gx;
		Vtl::Plf::SeparableConv2D<T> gy;
		Vtl::Plf::PixByPix2to1<T,Vtl::Plf::F21::SqrtOfSumSqr<T>,PostProc> sum_sqr;
	private:
		GradientModul(const GradientModul&);
		GradientModul& operator=(const GradientModul&);
	public:
		InPortComb<T> in;
		///   !  .
		OutPort<T,PostProc > & out() {return sum_sqr.out; }
		///    !   X (X-  ).
		OutPort<T,P> & out_gx() {return gx.out(); }
		///    !   Y (Y-  ).
		OutPort<T,P> & out_gy() {return gy.out(); }
		GradientModul()
			: FilterComplex(1,1)
		{
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		/// output_maxval==float((1<<bits)-1)
		GradientModul(std::string name,T output_maxval,float sigma_=1.0f,float env_sigmas=2.5f,float gamma_correction=2.2f,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterComplex(1,1)
		{
			Init(name,output_maxval,sigma_,env_sigmas,gamma_correction,post_proc);
		}
		/// output_maxval==float((1<<bits)-1)
		void Init(std::string name,T output_maxval,float sigma_=1.0f,float env_sigmas=2.5f,float gamma_correction=2.2f,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			sigma=sigma_;
			G0 g0(sigma,env_sigmas);
			G1 g1(sigma,env_sigmas);
			gauss.Init(g0); 
			gauss_deriv.Init(g1); 
//			gauss.Init(sigma,env_sigmas); 
//			gauss_deriv.Init(sigma,env_sigmas); 
			gx.Init(Name()+"::gx",&gauss_deriv,&gauss);
			gy.Init(Name()+"::gy",&gauss,&gauss_deriv);
			const Vtl::Plf::F21::SqrtOfSumSqr<T> ssqr;
			Vtl::Plf::F11::GammaCorMult<T> gam_mult(output_maxval,gamma_correction,sqrt(2.0f*3.1415926f)*sigma);
			Vtl::Plf::F11::Comb<T,Vtl::Plf::F11::GammaCorMult<T>,P> gam_mult_pproc(gam_mult,post_proc);
			sum_sqr.Init(Name()+"::SumSqr",ssqr,gam_mult_pproc);
		//------ create graph --------------------
			out_gx().ConnectTo(&sum_sqr.in_1);
			out_gy().ConnectTo(&sum_sqr.in_2);
			
			in.Add(&gx.in());
			in.Add(&gy.in());
//			AddPort(in,0);
//			AddPort(out(),0);
		}
		virtual ~GradientModul()
		{
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class MeanAndDisp
	/// 
	/// @brief       (   ).
	/// 
	/// \dotfile MeanAndDisp.dot  "  MeanAndDisp       " 
	/// 
	///     \image html o_gray.jpg 
	///   : \image html MeanAndDisp_mean.jpg 
	///   (  ): \image html MeanAndDisp_disp.jpg  
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class MeanAndDisp
		: public FilterComplex
	{
		Vtl::Plf::Levels<T,Vtl::Plf::F11::Sqr<T> > sqr;
		typedef Vtl::ImAlg::WndFun::Gauss<T,0> G0;
		Vtl::Plf::Kernel::FKern<T> gauss;

		Vtl::Plf::Average<T> mean;
		Vtl::Plf::Average<T> sqr_mean;
		Vtl::Plf::PixByPix2to1<T,Vtl::Plf::F21::Disp<T> > disp;
	private:
		MeanAndDisp(const MeanAndDisp&);
		MeanAndDisp& operator=(const MeanAndDisp&);
	public:
		InPortComb<T> in;
		///   X (X-  ).
		OutPort<T,P> & out_mean() {return mean.out(); }
		///   Y (Y-  ).
		OutPort<T,P> & out_disp() {return disp.out; }
		MeanAndDisp()
			: FilterComplex(1,1)
		{
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		/// output_maxval==float((1<<bits)-1)
		MeanAndDisp(std::string name,T output_maxval,float sigma=1.0f,float env_sigmas=2.5f,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterComplex(1,1)
		{
			Init(name,output_maxval,sigma,env_sigmas,post_proc);
		}
		/// output_maxval==float((1<<bits)-1)
		void Init(std::string name,T output_maxval,float sigma=1.0f,float env_sigmas=2.5f,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			//sigma=sigma_;
			G0 g0(sigma,env_sigmas);
			gauss.Init(g0);
//			gauss.Init(sigma,env_sigmas); 
			Vtl::Plf::F11::Sqr<T> s; 
			Vtl::Plf::F21::Disp<T> d;

			sqr.Init(Name()+"::sqr",s);
			mean.Init(Name()+"::mean(v)",&gauss);
			sqr_mean.Init(Name()+"::mean(v^2)",&gauss);
			disp.Init(Name()+"::(\\<v^2\\>-\\<v\\>^2)",d);
		//------ create graph --------------------
			sqr.out.ConnectTo(&sqr_mean.in());
			mean.out().ConnectTo(&disp.in_1);
			sqr_mean.out().ConnectTo(&disp.in_2);
			in.Add(&sqr.in);
			in.Add(&mean.in());
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		virtual ~MeanAndDisp()
		{
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class StatDiff
	/// 
	/// @brief    .  
	///         \see MeanAndDisp.
	///     :
	///	\f$ G(x,y)=\lbrack I(x,y)- \langle I(x,y)\rangle \rbrack \cdot \lbrack \frac{A \sigma_d}{A\sigma(x,y)+\sigma_d}\rbrack + \lbrack \alpha m_d + (1-\alpha)\langle I(x,y)\rangle \rbrack\f$ ,\n
	/// \n
	/// - \f$ G(x,y) \f$  - -
	/// - \f$ I(x,y) \f$  -  
	/// - \f$ \langle I(x,y)\rangle \f$  -  ,  () 
	///        (x,y).
	/// - \f$ \sigma(x,y)= \sqrt(\langle I(x,y)^2\rangle - \langle I(x,y)\rangle ^2 )\f$  - 
	///        .
	/// - \f$ m_d \f$    \f$ \sigma_d \f$  -     
	///    .
	/// - \f$ A \f$  -   ,    
	///           
	///    \f$ \sigma(x,y) \f$ .
	/// - \f$ \alpha \f$  -          
	///     .\n
	///
	///        Sharpen,  
	///   \f$ G(x,y)=\langle I(x,y)\rangle+ k \cdot \lbrack I(x,y)- \langle I(x,y)\rangle \rbrack \f$ ,
	///    \f$ k \f$ ( \f$ k=1 \f$    ,  \f$ k=0 \f$ -, 
	///  \f$ 0 < k < 1\f$ -    ,  \f$ 1 < k \f$ -    )
	///     \f$ k=k(x,y)=\frac{A \sigma_d}{A\sigma(x,y)+\sigma_d} \f$  
	///    ,     -  
	///   (   ).    
	///  \f$ G(x,y)=\langle I(x,y)\rangle \f$ -   - 
	///   \f$ \lbrack \alpha m_d + (1-\alpha)\langle I(x,y)\rangle \rbrack\  \f$ , 
	///   \f$ \alpha=1 \f$     \f$ m_d \f$, ..   ,
	///    \f$ \alpha=0 \f$       Sharpen.
	///
	/// \dotfile StatDiff.dot "  Vtl::Plf::StatDiff -   " 
	/// 
	///     \image html o_gray.jpg 
	///   \image html StatDiff.jpg 
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class StatDiff
		: public FilterComplex
	{
//		Vtl::Plf::Levels<T,Vtl::Plf::F::Sqr<T> > sqr;
//		Vtl::Plf::ConvKernelGauss<T> gauss; 
		Vtl::Plf::MeanAndDisp<T> mean_and_disp;
		PixByPix3to1<T,Vtl::Plf::F31::StatDiff<T>,P > stat;
	private:
		StatDiff(const StatDiff&);
		StatDiff& operator=(const StatDiff&);
	public:
		InPortComb<T> in;
		OutPort<T,P> & out() {return stat.out; }
		StatDiff()
			: FilterComplex(1,1)
		{
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		/// output_maxval==float((1<<bits)-1)
		StatDiff(std::string name,T output_maxval,float sigma=5.0f,float env_sigmas=2.0f,float amplification=6.0f,float sigma_desired=85.0f,float mean_desired=128.0f,float alpha=0.8f,P post_proc=Vtl::Plf::F11::DoNothing<T>() )
			: FilterComplex(1,1)
		{
			Init(name,output_maxval,sigma,env_sigmas,amplification,sigma_desired,mean_desired,alpha,post_proc);
		}
		/// output_maxval==float((1<<bits)-1)
		void Init(std::string name,T output_maxval,float sigma=5.0f,float env_sigmas=2.0f,float amplification=6.0f,float sigma_desired=85.0f,float mean_desired=128.0f,float alpha=0.8f,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			//out().setPostProc(post_proc);

			mean_and_disp.Init(Name()+"::MeanAndDisp",output_maxval,sigma,env_sigmas);
			Vtl::Plf::F31::StatDiff<T> s(amplification,sigma_desired,mean_desired,alpha); 
			stat.Init(Name()+"::StatDiff",s,post_proc);
		//------ create graph --------------------
			mean_and_disp.out_mean().ConnectTo(&stat.in_xy);
			mean_and_disp.out_disp().ConnectTo(&stat.in_yy);
			in.Add(&stat.in_xx);
			in.Add(&mean_and_disp.in);
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		virtual ~StatDiff()
		{
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class DeNoise
	/// 
	/// @brief    .  
	///         \see MeanAndDisp.
	/// 
	/// \dotfile DeNoise.dot   "  DeNoise -   " 
	/// 
	///     \image html o_gray.jpg 
	///   \image html DeNoise.jpg 
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class DeNoise
		: public FilterComplex
	{
//		Vtl::Plf::Levels<T,Vtl::Plf::F::Sqr<T> > sqr;
//		Vtl::Plf::ConvKernelGauss<T> gauss; 
		Vtl::Plf::MeanAndDisp<T> mean_and_disp;
		PixByPix3to1<T,Vtl::Plf::F31::DeNoise<T>,P > denoise;
	private:
		DeNoise(const DeNoise&);
		DeNoise& operator=(const DeNoise&);
	public:
		InPortComb<T> in;
		OutPort<T,P> & out() {return denoise.out; }
		DeNoise()
			: FilterComplex(1,1)
		{
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		/// output_maxval==float((1<<bits)-1)
		DeNoise(std::string name,T output_maxval,float sigma=5.0f,float env_sigmas=2.0f,float in_k_sigmas=2.0f,P post_proc=Vtl::Plf::F11::DoNothing<T>() )
			: FilterComplex(1,1)
		{
			Init(name,output_maxval,sigma,env_sigmas,in_k_sigmas,post_proc);
		}
		/// output_maxval==float((1<<bits)-1)
		void Init(std::string name,T output_maxval,float sigma=5.0f,float env_sigmas=2.0f,float in_k_sigmas=2.0f,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			//out().setPostProc(post_proc);

			mean_and_disp.Init(Name()+"::MeanAndDisp",output_maxval,sigma,env_sigmas);
			Vtl::Plf::F31::DeNoise<T> s(in_k_sigmas); 
			denoise.Init(Name()+"::DeNoise",s,post_proc);
		//------ create graph --------------------
			mean_and_disp.out_mean().ConnectTo(&denoise.in_xy);
			mean_and_disp.out_disp().ConnectTo(&denoise.in_yy);
			in.Add(&denoise.in_xx);
			in.Add(&mean_and_disp.in);
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		virtual ~DeNoise()
		{
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class Canny
	/// 
	/// @brief   .
	/// 
	/// \dotfile Canny.dot    "  Canny -   ()" 
	/// 
	///     \image html o_gray.jpg 
	///   \image html Canny.jpg 
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class Canny
		: public FilterComplex
	{
		float sigma; 
		typedef typename Vtl::Plf::F11::Comb<T,Vtl::Plf::F11::GammaCorMult<T>,P> PostProc;

		Vtl::Plf::GradientModul<T> mgrad; 
		Vtl::Plf::EnvProc3to1<T,Fpie::CannyNMS<T>,P> nms; 
	private:
		Canny(const Canny&);
		Canny& operator=(const Canny&);
	public:
		InPortComb<T>& in() {return mgrad.in;}
		OutPort<T,P>& out() {return nms.out;}
		Canny()
			: FilterComplex(1,1)
		{
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		/// output_maxval==float((1<<bits)-1)
		Canny(std::string name,T threshold,T output_maxval,float sigma=1.0f,float env_sigmas=2.5f,float gamma_correction=2.2f,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterComplex(1,1)
		{
			Init(name,threshold,output_maxval,sigma,env_sigmas,gamma_correction,post_proc);
		}
		/// output_maxval==float((1<<bits)-1)
		void Init(std::string name,T threshold,T output_maxval,float sigma=1.0f,float env_sigmas=2.5f,float gamma_correction=2.2f,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			mgrad.Init(Name()+"::MogGrad",output_maxval,sigma,env_sigmas,gamma_correction);
			Vtl::Plf::Fpie::CannyNMS<T> nms_proc(threshold);
			nms.Init(Name()+"::NMS",nms_proc,post_proc);
		//------ create graph --------------------
			mgrad.out().ConnectTo(&nms.in_val);
			mgrad.out_gx().ConnectTo(&nms.in_gx);
			mgrad.out_gy().ConnectTo(&nms.in_gy);
//			AddPort(in,0);
//			AddPort(out(),0);
		}
		virtual ~Canny()
		{
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class DiZenzo
	/// 
	/// 
	/// @brief   DiZenzo.   ,     
	///        ,    ,    
	///           . \ref Lit_DiZenzo 
	/// 
	/// \dotfile DiZenzo.dot    "  DiZenzo -   ()   " 
	/// 
	///     \image html o_color.jpg 
	///   \image html DiZenzo.jpg 
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class DiZenzo
		: public FilterComplex
	{
		float sigma; 
//		typedef typename Vtl::Plf::F::Comb<T,Vtl::Plf::F::GammaCorMult<T>,P> PostProc;

		typedef Vtl::ImAlg::WndFun::Gauss<T,0> G0;
		typedef Vtl::ImAlg::WndFun::Gauss<T,1> G1;
		Vtl::Plf::Kernel::FKern<T> gauss;
		Vtl::Plf::Kernel::FKern<T> gauss_deriv;
		Vtl::Plf::SeparableConv2D<T> gx_1;
		Vtl::Plf::SeparableConv2D<T> gy_1;
		Vtl::Plf::SeparableConv2D<T> gx_2;
		Vtl::Plf::SeparableConv2D<T> gy_2;
		Vtl::Plf::SeparableConv2D<T> gx_3;
		Vtl::Plf::SeparableConv2D<T> gy_3;

		Vtl::Plf::PixByPix2to3<T,Vtl::Plf::F23::CorrMatrix<T,T&> > mat_1;
		Vtl::Plf::PixByPix2to3<T,Vtl::Plf::F23::CorrMatrix<T,T&> > mat_2;
		Vtl::Plf::PixByPix2to3<T,Vtl::Plf::F23::CorrMatrix<T,T&> > mat_3;

		Vtl::Plf::PixByPix3to1<T,Vtl::Plf::F31::Sum3<T> > sum_1;
		Vtl::Plf::PixByPix3to1<T,Vtl::Plf::F31::Sum3<T> > sum_2;
		Vtl::Plf::PixByPix3to1<T,Vtl::Plf::F31::Sum3<T> > sum_3;

		PixByPix3to3<T, Vtl::Plf::F33::DiZenzo<T> ,Vtl::Plf::F11::GammaCorMult<T> > color_modul_eigvec;

		Vtl::Plf::EnvProc3to1<T,Fpie::CannyNMS<T>,P> nms;
	
	private:
		DiZenzo(const DiZenzo&);
		DiZenzo& operator=(const DiZenzo&);
	public:
		InPortComb<T> in_1;
		InPortComb<T> in_2;
		InPortComb<T> in_3;
		///   !  .
		OutPort<T > & out() {return nms.out; }
		DiZenzo()
			: FilterComplex(1,1)
		{
		}
		DiZenzo(std::string name,T output_maxval,T threshold=20,float sigma_=1.0f,float env_sigmas=2.5f,float gamma_correction=2.2f,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterComplex(1,1)
		{
			Init(name,output_maxval,threshold,sigma_,env_sigmas,gamma_correction,post_proc);
		}

		void Init(std::string name,T output_maxval,float threshold=20,float sigma_=1.0f,float env_sigmas=2.5f,float gamma_correction=2.2f,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			sigma=sigma_;
			G0 g0(sigma,env_sigmas);
			gauss.Init(g0);
			G1 g1(sigma,env_sigmas);
			gauss_deriv.Init(g1);
			//gauss.Init(sigma,env_sigmas); 
			//gauss_deriv.Init(sigma,env_sigmas); 
			gx_1.Init(Name()+"::gx",&gauss_deriv,&gauss);
			gy_1.Init(Name()+"::gy",&gauss,&gauss_deriv);
			gx_2.Init(Name()+"::gx",&gauss_deriv,&gauss);
			gy_2.Init(Name()+"::gy",&gauss,&gauss_deriv);
			gx_3.Init(Name()+"::gx",&gauss_deriv,&gauss);
			gy_3.Init(Name()+"::gy",&gauss,&gauss_deriv);
			const Vtl::Plf::F23::CorrMatrix<T,T&> cov;
			mat_1.Init("1-cov",cov);
			mat_2.Init("2-cov",cov);
			mat_3.Init("3-cov",cov);
			
			const Vtl::Plf::F31::Sum3<T> s3;
			sum_1.Init("sum gx^2",s3);
			sum_2.Init("sum gx*gy",s3);
			sum_3.Init("sum gy^2",s3);

			Vtl::Plf::F11::GammaCorMult<T> gam_mult(output_maxval,gamma_correction,sqrt(2.0f*M_PI)*sigma);

			const Vtl::Plf::F33::DiZenzo<T> dz;
			color_modul_eigvec.Init("color_modul_eigvec",dz,gam_mult);

			Vtl::Plf::Fpie::CannyNMS<T> canny(threshold); 
			nms.Init("NMS",canny,post_proc);

		//------ create graph --------------------

			gx_1.out().ConnectTo(&mat_1.in_1);
			gy_1.out().ConnectTo(&mat_1.in_2);
			gx_2.out().ConnectTo(&mat_2.in_1);
			gy_2.out().ConnectTo(&mat_2.in_2);
			gx_3.out().ConnectTo(&mat_3.in_1);
			gy_3.out().ConnectTo(&mat_3.in_2);

			mat_1.out_11.ConnectTo(&sum_1.in_xx);
			mat_2.out_11.ConnectTo(&sum_1.in_xy);
			mat_3.out_11.ConnectTo(&sum_1.in_yy);

			mat_1.out_12.ConnectTo(&sum_2.in_xx);
			mat_2.out_12.ConnectTo(&sum_2.in_xy);
			mat_3.out_12.ConnectTo(&sum_2.in_yy);

			mat_1.out_22.ConnectTo(&sum_3.in_xx);
			mat_2.out_22.ConnectTo(&sum_3.in_xy);
			mat_3.out_22.ConnectTo(&sum_3.in_yy);
			
			sum_1.out.ConnectTo(&color_modul_eigvec.in_1);
			sum_2.out.ConnectTo(&color_modul_eigvec.in_2);
			sum_3.out.ConnectTo(&color_modul_eigvec.in_3);

			color_modul_eigvec.out_1.ConnectTo(&nms.in_val);
			color_modul_eigvec.out_2.ConnectTo(&nms.in_gx);
			color_modul_eigvec.out_3.ConnectTo(&nms.in_gy);

			in_1.Add(&gx_1.in());
			in_1.Add(&gy_1.in());
			in_2.Add(&gx_2.in());
			in_2.Add(&gy_2.in());
			in_3.Add(&gx_3.in());
			in_3.Add(&gy_3.in());
		}
		virtual ~DiZenzo()
		{
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class HImageEdge
	/// 
	/// \brief       H-image \ref Lit_HImg, \\see Vtl::Plf::F::Himg, Vtl::Plf::HImage.
	/// 
	/// \dotfile H-image-Edges.dot  "  HImageEdge -     HImage"
	/// \ref Lit_HImg
	/// 
	///     \image html o_gray.jpg 
	///   \image html H-image-Edges.jpg 
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class HImageEdge
		: public FilterComplex
	{
		Vtl::Plf::EnvProc1to2<T,Fpie::Himg2<T> > himage;
		Vtl::Plf::PixByPix2to1<T,Vtl::Plf::F21::SqrtOfSumSqr<T> > sum_sqr;
		Vtl::Plf::EnvProc3to1<T,Fpie::CannyNMS<T>,P> nms; 
	private:
		HImageEdge(const HImageEdge&);
		HImageEdge& operator=(const HImageEdge&);
	public:
		InPort<T>& in() {return himage.in;}
		OutPort<T,P>& out() {return nms.out;}
		HImageEdge()
			: FilterComplex(1,1)
		{
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		/// output_maxval==float((1<<bits)-1)
		HImageEdge(std::string name,int radius,T threshold,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterComplex(1,1)
		{
			Init(name,radius,threshold,post_proc);
		}
		/// output_maxval==float((1<<bits)-1)
		void Init(std::string name,int radius,T threshold,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			const Vtl::Plf::Fpie::Himg2<T> hi(radius);
			himage.Init("H-Img-pseudograd",hi);
			const Vtl::Plf::F21::SqrtOfSumSqr<T> ssqr;
			sum_sqr.Init(Name()+"::SumSqr",ssqr);
			Vtl::Plf::Fpie::CannyNMS<T> nms_proc(threshold);
			nms.Init(Name()+"::NMS",nms_proc,post_proc);
		//------ create graph --------------------
			himage.out_x.ConnectTo(&sum_sqr.in_1);
			himage.out_y.ConnectTo(&sum_sqr.in_2);
			sum_sqr.out.ConnectTo(&nms.in_val);
			himage.out_x.ConnectTo(&nms.in_gx);
			himage.out_y.ConnectTo(&nms.in_gy);
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		virtual ~HImageEdge()
		{
		}
	};

	///////////////////////////////////////////////////////////////////////////
	// class GradientCorr
	/// 
	/// @brief     : \f$ g_x^2,\ g_x g_y,\  g_y^2 \f$.
	/// 
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=F::CorrMatrix<T,T&> >
	class GradientCorr
		: public FilterComplex
	{
		float sigma; 
		typedef Vtl::ImAlg::WndFun::Gauss<T,0> G0;
		typedef Vtl::ImAlg::WndFun::Gauss<T,1> G1;
		Vtl::Plf::Kernel::FKern<T> gauss;
		Vtl::Plf::Kernel::FKern<T> gauss_deriv;
		Vtl::Plf::SeparableConv2D<T> gx;
		Vtl::Plf::SeparableConv2D<T> gy;
		Vtl::Plf::PixByPix2to3<T,Vtl::Plf::F23::CorrMatrix<T,T&> > corr;
	private:
		GradientCorr(const GradientCorr&);
		GradientCorr& operator=(const GradientCorr&);
	public:
		InPortComb<T> in;
		///   !  .
		OutPort<T > & out_xx() {return corr.out_11; }
		///    !   X (X-  ).
		OutPort<T > & out_xy() {return corr.out_12; }
		///    !   Y (Y-  ).
		OutPort<T > & out_yy() {return corr.out_22; }
		/// output_maxval==float((1<<bits)-1)
		GradientCorr()
			: FilterComplex(1,1)
		{
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		GradientCorr(std::string name,float sigma_=1.0f,float env_sigmas=2.5f,P func=F23::CorrMatrix<T,T&>())
			: FilterComplex(1,1)
		{
			Init(name,sigma_,env_sigmas,func);
		}
		void Init(std::string name,float sigma_=1.0f,float env_sigmas=2.5f,P func=F23::CorrMatrix<T,T&>())
		{
			setName(name);
			sigma=sigma_;
			G0 g0(sigma,env_sigmas);
			G1 g1(sigma,env_sigmas);
			gauss.Init(g0); 
			gauss_deriv.Init(g1); 
			//gauss.Init(sigma,env_sigmas); 
			//gauss_deriv.Init(sigma,env_sigmas); 
			gx.Init(Name()+"::gx",&gauss_deriv,&gauss);
			gy.Init(Name()+"::gy",&gauss,&gauss_deriv);

			corr.Init(Name()+"::2to3",func );
		//------ create graph --------------------
			gx.out().ConnectTo(&corr.in_1);
			gy.out().ConnectTo(&corr.in_2);
			
			in.Add(&gx.in());
			in.Add(&gy.in());
//			AddPort(in,0);
//			AddPort(out(),0);
		}
		virtual ~GradientCorr()
		{
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class Detect
	/// 
	/// @brief  ,   ..     
	///       :
	/// \li class Vtl::Plf::F::Harris -   ( ).
	/// \li class Vtl::Plf::F::MaxEigenVal -  (DiZenzo).
	/// \li class Vtl::Plf::F::MinEigenVal -   ( , -).
	/// \li class Vtl::Plf::F::DiffEigenVal -  (DiZenzo)    ( ).
	///  ..
	/// 
	/// \dotfile "HarrisNMS(0.04).dot"  "  Detect -  " 
	///
	///     \image html o_gray.jpg 
	///    (corners, blobs):\n
	///     Vtl::Plf::F::Harris  k=0.0 \image html "Harris(k=0.00).jpg"
	///     Vtl::Plf::F::Harris  k=0.04 \image html "Harris(k=0.04).jpg" 
	///     Vtl::Plf::F::MinEigenVal \image html MinEigenVal.jpg 
	///    (edges detectors):\n
	///     Vtl::Plf::F::MaxEigenVal \image html MaxEigenVal.jpg 
	///     Vtl::Plf::Vtl::Plf::F::DiffEigenVal \image html DiffEigenVal.jpg 
	///     Vtl::Plf::Vtl::Plf::F::Trace \image html Trace.jpg 
	///
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class Func=Vtl::Plf::F31::Harris<T>,class P=Vtl::Plf::F11::DoNothing<T> >
	class Detect
		: public FilterComplex
	{
		Vtl::Plf::GradientCorr<T,F23::CorrMatrix<T,T&> > corr;
		Vtl::Plf::Average<T> avg_xx;
		Vtl::Plf::Average<T> avg_xy;
		Vtl::Plf::Average<T> avg_yy;
		typedef Vtl::ImAlg::WndFun::Gauss<T,0> G0;
		Vtl::Plf::Kernel::FKern<T> gauss;
		Vtl::Plf::PixByPix3to1<T,Func,P > det;
	private:
		Detect(const Detect&);
		Detect& operator=(const Detect&);
	public:
		InPortComb<T>&  in() { return corr.in; }
		OutPort<T,P> & out() {return det.out; }
		Detect()
			: FilterComplex(1,1)
		{
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		Detect(std::string name,float sigma_=1.0f,float env_sigmas=2.5f,Func functor=Vtl::Plf::F31::Harris<T>(),P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterComplex(1,1)
		{
			Init(name,sigma_,env_sigmas,functor,post_proc);
		}
		void Init(std::string name,float sigma=1.0f,float env_sigmas=2.5f,Func functor=Vtl::Plf::F31::Harris<T>(),P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			F23::CorrMatrix<T,T&> corr_mat;
			corr.Init(Name()+"::corr",sigma,env_sigmas,corr_mat);
			float coef=1.7f;
			G0 g0(coef*sigma,env_sigmas);
			gauss.Init(g0); 
//			gauss.Init(coef*sigma,env_sigmas); 
			avg_xx.Init(Name()+"::avg_xx",&gauss);
			avg_xy.Init(Name()+"::avg_xy",&gauss);
			avg_yy.Init(Name()+"::avg_yy",&gauss);
			det.Init(Name()+"::functor",functor,post_proc);
		//------ create graph --------------------
			corr.out_xx().ConnectTo(&avg_xx.in());
			corr.out_xy().ConnectTo(&avg_xy.in());
			corr.out_yy().ConnectTo(&avg_yy.in());
			avg_xx.out().ConnectTo(&det.in_xx);
			avg_xy.out().ConnectTo(&det.in_xy);
			avg_yy.out().ConnectTo(&det.in_yy);
			
//			AddPort(in,0);
//			AddPort(out(),0);
		}
		virtual ~Detect()
		{
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class Detect
	/// 
	/// @brief  
	///
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class Func=Vtl::Plf::F31::Harris<T>,class P=Vtl::Plf::F11::DoNothing<T> >
	class HarrisLaplace
		: public FilterComplex
	{
		Vtl::Plf::GradientCorr<T,F23::CorrMatrix<T,T&> > corr;
		Vtl::Plf::Average<T> avg_xx;
		Vtl::Plf::Average<T> avg_xy;
		Vtl::Plf::Average<T> avg_yy;
		typedef Vtl::ImAlg::WndFun::Gauss<T,0> G0;
		Vtl::Plf::Kernel::FKern<T> gauss;
		Vtl::Plf::PixByPix5to1<T,Func,P > det;
	private:
		HarrisLaplace(const HarrisLaplace&);
		HarrisLaplace& operator=(const HarrisLaplace&);
	public:
		InPortComb<T> in;
//		InPortComb<T>&  in() { return corr.in; }
		OutPort<T,P> & out() {return det.out; }
		HarrisLaplace()
			: FilterComplex(1,1)
		{
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		HarrisLaplace(std::string name,float sigma_=1.0f,float env_sigmas=2.5f,Func functor=Vtl::Plf::F31::Harris<T>(),P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterComplex(1,1)
		{
			Init(name,sigma_,env_sigmas,functor,post_proc);
		}
		void Init(std::string name,float sigma=1.0f,float env_sigmas=2.5f,Func functor=Vtl::Plf::F31::Harris<T>(),P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			F23::CorrMatrix<T,T&> corr_mat;
			corr.Init(Name()+"::corr",sigma,env_sigmas,corr_mat);
			float coef=1.7f;
			G0 g0(coef*sigma,env_sigmas);
			gauss.Init(g0); 

			avg_xx.Init(Name()+"::avg_xx",&gauss);
			avg_xy.Init(Name()+"::avg_xy",&gauss);
			avg_yy.Init(Name()+"::avg_yy",&gauss);
			det.Init(Name()+"::functor",functor,post_proc);
		//------ create graph --------------------
			corr.out_xx().ConnectTo(&avg_xx.in());
			corr.out_xy().ConnectTo(&avg_xy.in());
			corr.out_yy().ConnectTo(&avg_yy.in());
			
			avg_xx.out().ConnectTo(&det.in_1);
			avg_xy.out().ConnectTo(&det.in_2);
			avg_yy.out().ConnectTo(&det.in_3);
			//in.out().ConnectTo(&det.in_4);
			//avg_yy.out().ConnectTo(&det.in_5);
			
			
			in.Add(&corr.in);
			in.Add(&det.in_4);
			in.Add(&det.in_5);
//			AddPort(in,0);
//			AddPort(out(),0);
		}
		virtual ~HarrisLaplace()
		{
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class Resample
	/// 
	/// @brief   (resampling)  ( ).
	/// 
	/// \dotfile Resample.dot  "  Resample -  " 
	/// 
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class Resample
		: public FilterComplex
	{
		Vtl::Plf::FilterResampleVert<float> v_resamp;
		Vtl::Plf::FilterResampleHorz<float,P> h_resamp;
	private:
		Resample(const Resample&);
		Resample& operator=(const Resample&);
	public:
		InPort<T> & in() {return v_resamp.in; }
		OutPort<T,P> & out() {return h_resamp.out; }
		Resample()
			: FilterComplex(1,1)
		{
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		Resample(std::string name,int new_width,int new_height,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterComplex(1,1)
		{
			Init(name,new_width,new_height,post_proc);
		}
		void Init(std::string name,int new_width,int new_height,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			out().setPostProc(post_proc);

			v_resamp.Init(Name()+"::V-resample",new_height); 
			h_resamp.Init(Name()+"::H-resample",new_width); 
		//------ create graph --------------------
			v_resamp.out.ConnectTo(&h_resamp.in);
//			AddPort(in(),0);
//			AddPort(out(),0);
		}
		virtual ~Resample()
		{
		}
	};
/** @} */ // end of ComplexFiltersDoxGroup  
} //namespace Plf 
} //namespace Vtl
/////////////////////////////////////////////////////////////////////////////
#endif //_PLF_FILTER_COMPLEX_H
