#ifndef _PLF_ITERATORS_H
#define _PLF_ITERATORS_H
/////////////////////////////////////////////////////////////////////////////
#include <algorithm>
namespace Vtl
{
// @brief     plf_filters.h,   
namespace Plf 
{
/** @addtogroup IteratorsForElementaryFiltersDoxGroup  Iterators and Utilities for Elementary Filters
 *  @ingroup PipeLineFiltersDoxGroup  
 *
 *  @brief        
 *     (Pipe Line Filtering) 
 *
 *  @{
 */
	///////////////////////////////////////////////////////////////////////////
	/// class ColumnIteratorFast
	/// @brief      ImageRef. 
	/// 
	///      ,    .
	///   
	///       ,
	///          .
	///////////////////////////////////////////////////////////////////////////
	template<typename T>
	class ColumnIteratorFast
	{
		const T* ptr;
		int stride;
	private:
		ColumnIteratorFast() {}
	public:
		ColumnIteratorFast(const T* pointer,int step) : ptr(pointer), stride(step) {}
		ColumnIteratorFast(const ColumnIteratorFast& it) : ptr(it.ptr), stride(it.stride) {}

		const T& operator*()  { return *ptr;}
		ColumnIteratorFast& operator++()  {ptr+=stride; return *this;}
//		ColumnIteratorFast operator++(int)  {ColumnIteratorFast it(*this); ptr+=stride; return it;}
		ColumnIteratorFast& operator--()  {ptr-=stride; return *this;}
//		ColumnIteratorFast operator--(int)  {ColumnIteratorFast it(*this); ptr-=stride; return it;}
	};
	///////////////////////////////////////////////////////////////////////////
	/// class LineIterator
	/// @brief      ImageRef. 
	/// 
	///   
	///       ,
	///          .
	///      ,    
	///    class ColumnIteratorFast.
	///////////////////////////////////////////////////////////////////////////
	template<typename T>
	class LineIterator
	{
		const T* ptr;
		const int nls;
//		const int h;
	public:
		LineIterator(ImageRef<T>& img,bool end=false) 
			:ptr(img[-img.ext().before]),nls(img.nls())//,h(img.h()); 
		{
			if(end)
				ptr+=img.w();
		}
		ColumnIteratorFast<T> operator*() const { return ColumnIteratorFast<T>(ptr,nls);}
		LineIterator& operator++()  {++ptr; return *this;}
//		ColumnIteratorFast operator++(int)  {ColumnIteratorFast it(*this); ptr+=stride; return it;}
		LineIterator& operator--()  {--ptr; return *this;}
//		ColumnIteratorFast operator--(int)  {ColumnIteratorFast it(*this); ptr-=stride; return it;}
		bool operator!=(const LineIterator& li) const { return ptr!=li.ptr;}
	};
/** @} */ // end of IteratorsForElementaryFiltersDoxGroup
} //namespace Plf
} //namespace Vtl
#endif //_PLF_ITERATORS_H
