#ifndef _PLF_STRIP_BUFFER_H
#define _PLF_STRIP_BUFFER_H
/////////////////////////////////////////////////////////////////////////////
#include <algorithm>
#include "plf_strip_buffer_support.h"
namespace Vtl
{
// @brief     plf_filters.h,   
namespace Plf 
{
/** @addtogroup PipeLineFiltersDoxGroup  Pipe Line Filters. 
 *  @{
 */
/** @} */ // end of PipeLineFiltersDoxGroup  
/** @addtogroup PortsAndBuffersDoxGroup  Ports and buffers 
 *  @ingroup PipeLineFiltersDoxGroup 
 * 
 *  @brief ,      ,  
 *           .  :    
 *         ,      (    
 *           )    .  
 *         ,    .
 *
 *  @{
 */
//#define DEBUG_SHOW_ALL_BUFFER
	///////////////////////////////////////////////////////////////////////////
	/// class StripBuffer
	/// 
	/// @brief   .     ,
	///      ,    
	///  .
	///
	///  .\n
	/// @li @ref StripBufferInitDoxGroup
	/// @li @ref StripBufferPositionInfoDoxGroup  
	/// @li @ref StripBufferReadDataDoxGroup  
	/// @li @ref StripBufferInsertDoxGroup  
	/// 
	///////////////////////////////////////////////////////////////////////////
	template<typename T>
	class StripBuffer
	{
		std::vector<T> vec; //>      ( )  ().
		Extent ex; //>
		ImgSize ims;//>   ,    .
		int full_width;
		///   (   !)     'y'. 
		///   0  (ims.height-1).      
		/// , ..    void InsertNextLineFinished()   
		///    , ..   void PopTail().
		int y_wait;
		ImageRef<T> ir;
#ifdef DEBUG_SHOW_ALL_BUFFER
		ImageRef<T> __ir;
#endif
	private:
		T* ptrLineStart(int y) { return (&(vec[0]))+full_width*(y+ex.forward);}
		T* ptrLine0(int y) { return ptrLineStart(y)+ex.left;}
	public:
		StripBuffer() {}
		virtual ~StripBuffer() {}
/** @defgroup StripBufferInitDoxGroup  StripBuffer Initialization functions
 *  @ingroup PortsAndBuffersDoxGroup  
 * @brief    . 
 *
 *  :\n
 *
 * @li 1)    -  void init_1(const ImgSize& sz).     
 *      , ..   (OutPortNoPostProc),     
 *     (InPort)     .     
 *      ,       ,  
 *    ,       ,   1.7  , 
 *          640480  ..     
 *       , ..   ,  
 *      OutPortNoPostProc    .
 *
 * @li 2)  void init_2(const Extent& ext):     -Extent - ( 
 *        ).  ,      
 *     ,      Extent    
 *       InPort,     OutPortNoPostProc,   . OutPortNoPostProc, 
 *           InPort,     
 *      Extent   ,     
 *           ,  
 *     .
 *
 * @li 3)  void init_3(int increase_delay_by):  ,  .
 *              Y  
 *     Y   Extent     (  ),  
 *        @ref StripBufferPositionInfoDoxGroup.   
 *           :   
 *       -  ,      
 *         .     :  
 *              ,  
 *       .       
 *      ,       ,   Extents 
 *         ,        
 *     . ,     ,    ,      
 *    ,       ,       
 *      .      ,        
 *       ,     ,   ""  ,
 *        .              
 *        . ,   Extent.forward  ,   ,
 *     .  ,    ,     , 
 *           ,         
 *      .        .   
 *             Extent      
 *     Extent ,    .
 *
 * @li 4)  void init_4():       
 *      .  ,      .
 * 
 *  @{
 */
		void init_1(const ImgSize& sz)
		{
			assert(sz.width>0 && sz.height>0);
			ims=sz;
		}
		void init_2(const Extent& ext)
		{
			ext.checkCompatible(ims);
			ex=ext;
			if(ex.forward<ex.before)
				ex.forward=ex.before; //   !
			full_width=ex.left+ims.width+ex.right;
		}
		void init_3(int increase_delay_by)
		{
			Extent ext(ex);
			ext.forward+=increase_delay_by;
			ext.checkCompatible(ims);
			ex=ext;
		}

		void init_4()
		{
			vec.resize(full_width*(ex.forward+1+ex.forward));
			//y_wait=0; //      'y_wait'.
			//y_wait=-ex.before; //    'y_wait'  .
			Rewind();
			ir=ImageRef<T>(ptrLine0(0) ,ims.width,full_width,ex);
#ifdef DEBUG_SHOW_ALL_BUFFER
			__ir=ImageRef<T>(ptrLine0(0) ,ims.width,full_width,Extent(ex.left,ex.right,ex.forward,ex.forward));
#endif
		}
		void Rewind()
		{
			y_wait=0;
			ptr_beg_ins=0;
		}
/** @} */ // end of StripBufferInitDoxGroup
/** @defgroup StripBufferPositionInfoDoxGroup  StripBuffer current position / info
 *  @ingroup PortsAndBuffersDoxGroup  
 * 
 * @brief     .     , 
 *  , 
 *      ; \b   .
 *
 *        ,  
 *  Extent,    ,      
 *  Y_wait(),     Y() == Y_wait()-delay().
 *     ,     delay().
 *          
 *   Y_last()  Y_wait()-1
 * 
 *  @{
 */
		/// @brief   ,    ,
		///      ,   Extent. 
		ImgSize ImageSize() const {return ims;}
		/// @brief  .
		///  Extent  -      Y()  
		///  ImageSize(),        Y()-ext().before 
		///  Y()+ext().forward .       
		/// ext().left  ext().right    , 
		///         ext().before 
		///  ext().forward  . 
		Extent ext() const {return ex;}
		/// @brief      'y'        
		///  'y'    ext().
		int delay() const {return ex.forward; } 
		/// @brief      'y'        
		///  'y'    'extent'.   ,    
		///       .
		Strip delay(Extent extent) const 
		{
			return Strip(Y_last()+extent.before,Y_wait()-extent.forward); 
		} 
		/// @brief   ,    (    Extent).
		int Y() const {return Y_wait()-delay()-1; }
		/// @brief  ,     
		///      .     
		///    Y_last()   Y_wait()-1 .
		int Y_wait() const {return y_wait;}
		/// @brief        , 
		///     (      ), 
		///     ,   .
		int Y_last() const { return y_wait-(ex.forward+1+ex.before);}
		/// @brief      'y'        
		///  'y'    'extent'.   ,    
		///       .
		Strip getCurStrip() const 
		{ 
			return Strip(Y_last(),Y_wait());  
		}

/** @} */ // end of StripBufferPositionInfoDoxGroup  
/** @defgroup StripBufferReadDataDoxGroup  Image data reading from StripBuffer
 *  @ingroup PortsAndBuffersDoxGroup  
 *   
 * @brief    ()     .
 * 
 *    <tt>template<typename T> class ImageRef</tt>,   
 *     \b  ,  . 
 *       ,
 *     y=0    .    
 *      -ext().before  +ext().forward. =0    
 *    ,       w(). 
 *  \attention          
 *       .      
 *   ,      (    
 *  ).           
 *             .\n
 *      . \n
 * \li <tt> template<typename T> class ColumnIteratorFast; </tt>
 * \li <tt> template<typename T> class LineIterator; </tt>\n
 *       \n
 * \li <tt> template<typename T> class ConvolveHorz; </tt>
 * \li <tt> template<typename T> class ConvolveVert; </tt>
 *  @{
 */
#ifdef DEBUG_SHOW_ALL_BUFFER
		const ImageRef<T>& getImg() const {return __ir; }
#else
		const ImageRef<T>& getImg() const {return ir; }
#endif
		ImageRef<T> getImg(const Extent& ex1,const Strip& strip) const 
		{
			assert(getCurStrip().isContains(strip));
			int y_cent=strip.last()+ex1.before;
			int y_offset=y_cent-Y();
			return ImageRef<T>(getImg(),ex1,y_offset); 
		}
/** @} */ // end of StripBufferReadDataDoxGroup  
/** @defgroup StripBufferInsertDoxGroup  Writing into StripBuffer
 *  @ingroup PortsAndBuffersDoxGroup  
 *   
 * @brief       .
 * 
 *  @{
 */
		typedef T* Iter_Line;
		void InsertNextLineStart()
		{
//			assert(is_WaitData());
			if( y_wait <= ex.forward )
			{
				if( y_wait <= ex.before )
				{
					ptr_beg_ins=ptrLine0(y_wait);
					// : 
					// 1)    img_width 
					// 2)     
					// 3)  ()  y_wait   -y_wait.
				}
				else
				{
					ptr_beg_ins=ptrLine0(y_wait);
					// : 
					// 1)    img_width 
					// 2)     
					//    !!!
				}
			}
			else
			{
				if(ex.forward>0)
					memmove(ptrLineStart(-ex.before),ptrLineStart(-ex.before+1),sizeof(T)*full_width*(ex.before+ex.forward));
				ptr_beg_ins=ptrLine0(ex.forward);
				// : 
				// 1)    img_width 
				// 2)     
			}
		}
		void InsertNextLineFinished()
		{ // call next filters that data ready!
//			assert(is_WaitData() && (ptr_beg_ins!=0) );
			//  -        /   
			if(ex.left>0)
				_hFlipLeft(begin_line_insert(),ex.left);
			if(ex.right>0)
				_hFlipRight(end_line_insert(),ex.right);
			//
			if( y_wait <= ex.before ) 
			{ // 3)  ()  y_wait   -y_wait.
				if(y_wait!=0)
					_mirrorLineRelCenter(y_wait,-y_wait);
			}
			ptr_beg_ins=0;
			y_wait++;
		}
		void PopTail()
		{
//			assert(!is_WaitData());
//			assert(!is_EndData());
			int s=(y_wait-ims.height+1);
			int d=(ex.forward>ex.before)? s:ex.before;
			s=2*s;
			if(d>ex.forward || s >(2*ex.forward+1) )
			{
				std::ostringstream os;
				os << "StripBuffer::PopTail(): before= "<< ex.before << " forward=" <<ex.forward << " d = " << d << " s=" << s;
				throw os.str();
			}
			memmove(ptrLineStart(-d),ptrLineStart(-d+1),sizeof(T)*full_width*(d+ex.forward));
//			memmove(ptrLineStart(-ex.forward),ptrLineStart(-ex.forward+1),sizeof(T)*full_width*(ex.forward+ex.forward));
			memcpy(ptrLineStart(ex.forward),ptrLineStart(ex.forward-s),sizeof(T)*full_width);
			++y_wait;
		}
		Iter_Line begin_line_insert() { return ptr_beg_ins; }
		Iter_Line   end_line_insert() { return ptr_beg_ins+ims.width; }
	private:
		///    InsertNextLineStart(),
		///     begin_line_insert()  end_line_insert() ,
		///  InsertNextLineFinished()      -  
		///   .
		T* ptr_beg_ins;
		void _hFlipLeft(Iter_Line it,int len)
		{
				Iter_Line r=it;
				while(len--)
					*--it=*++r;
		}
		void _hFlipRight(Iter_Line it,int len)
		{
				Iter_Line r=it--;
				while(len--)
					*r++=*--it;
		}
		void _mirrorLineRelCenter(int src_y,int dst_y)
		{
			if(src_y==dst_y)
				return;
			T* s=ptrLineStart(src_y);
			T* d=ptrLineStart(dst_y);
			::memcpy(d,s,sizeof(T)*full_width);
		}
/** @} */ // end of StripBufferInsertDoxGroup
	};
/** @} */ // end of PortsAndBuffersDoxGroup    . 
} //namespace Plf
} //namespace Vtl
#endif //_PLF_STRIP_BUFFER_H
